//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 逆アセンブラ
//

#include "disasm.h"
#include "debugger_private.h"
#include "m68030disasm.h"
#include "m88100disasm.h"

//
// 逆アセンブラ(1行分)
//

// デストラクタ
DisasmLine::~DisasmLine()
{
}


//
// 逆アセンブルする人
//

// コンストラクタ
Disasm::Disasm()
	: inherited("Disasm")
{
	// 初期値
	saddr.SetSuper(true);
	saddr.SetData(true);
	SetAccessMode(MemoryMode::Physical);
}

// コンストラクタ (初期化込み)
Disasm::Disasm(DebuggerMD *md_)
	: Disasm()	// 移譲
{
	InitMD(md_);
}

// デストラクタ
Disasm::~Disasm()
{
}

// md を初期化(代入)する。
void
Disasm::InitMD(DebuggerMD *md_)
{
	md = md_;

	if (md->arch == DebuggerMD::Arch::M680x0) {
		line.reset(new m680x0disasm());
	} else {
		line.reset(new m88100disasm());
	}
}

// saddr を設定する。
void
Disasm::SetAddr(const saddr_t& saddr_)
{
	saddr = saddr_;
}

// アドレスを設定する。
void
Disasm::SetAddr(uint32 laddr_)
{
	saddr.SetAddr(laddr_);
}

// User/Super を設定する。
void
Disasm::SetSuper(bool super_)
{
	saddr.SetSuper(super_);
}

// Data/Inst を設定する。
void
Disasm::SetData(bool data_)
{
	saddr.SetData(data_);
}

// アクセスモードを設定する。
void
Disasm::SetAccessMode(MemoryMode access_mode_)
{
	access_mode = access_mode_;
}

// MMU ルックアップモードを設定する。
void
Disasm::SetLookupMode(MMULookupMode lookup_mode_)
{
	lookup_mode = lookup_mode_;
}

// アドレスをバイト単位で加算(減算)する。
void
Disasm::Offset(int bytes)
{
	saddr += (uint32)bytes;
}

// コンソールに出力。
// 出力後 saddr は更新されている。
void
Disasm::Print(FILE *cons, int row)
{
	DebuggerMemoryStream mem(md, saddr, access_mode, lookup_mode);

	for (int y = 0; y < row; y++) {
		std::string addrbuf;
		bool accessible;

		accessible = mem.FormatAddr(addrbuf);
		if (accessible == false) {
			fprintf(cons, "%s\n", addrbuf.c_str());
			mem.Offset(md->inst_bytes);
			continue;
		}

		line->Exec(&mem);
		fprintf(cons, "%-18s %s%s\n",
			addrbuf.c_str(),
			line->dump.c_str(), line->text.c_str());
	}

	// アドレスを更新
	saddr = mem.laddr;
}


//
// 逆アセンブラモニタ
//

// コンストラクタ
DisasmMonitor::DisasmMonitor(int idx_)
	: inherited()
{
	monitor.func = ToMonitorCallback(&DisasmMonitor::MonitorUpdate);
	monitor.SetSize(80, 16);
	monitor.Regist(ID_MONITOR_DISASM(idx_));
}

// デストラクタ
DisasmMonitor::~DisasmMonitor()
{
}

// モニタ更新
void
DisasmMonitor::MonitorUpdate(Monitor *, TextScreen& screen)
{
	DebuggerMemoryStream mem(md, saddr, access_mode, lookup_mode);

	screen.Clear();

	int row = screen.GetRow();

	if (nextaddr.size() != row) {
		nextaddr.resize(row);
	}

	for (int y = 0; y < row; y++) {
		std::string addrbuf;
		std::string mnemonic;
		std::string mnembuf;
		bool accessible;
		TA attr;

		// カーソル行をボールドに
		if (__predict_false((uint32)mem.laddr == md->GetPC())) {
			attr = TA::Em;
		} else {
			attr = TA::Normal;
		}

		accessible = mem.FormatAddr(addrbuf);
		screen.Puts(0, y, attr, addrbuf.c_str());
		if (__predict_false(accessible == false)) {
			mem.Offset(md->inst_bytes);
			continue;
		}

		line->Exec(&mem);
		screen.Print(19, y, attr, "%s%s",
			line->dump.c_str(), line->text.c_str());

		// 変換できなければここで終了
		if (__predict_false(line->bin.size() == 0)) {
			break;
		}

		// 次の命令開始アドレスを覚えておく。
		// (n行下へ移動する時のため)
		nextaddr[y] = mem.laddr.GetAddr();
	}
}

// 現在位置(1行目)より n 行下の命令までのオフセットを返す。
int
DisasmMonitor::GetOffset(int n) const
{
	int i;

	// 引数 n を nextaddr[] のインデックス i に変換。
	// 1行下なら2行目なので nextaddr は [0]、
	// 2行下なら3行目なので nextaddr は [1]。
	i = n - 1;

	if (0 <= i && i < nextaddr.size()) {
		return nextaddr[i] - (uint32)saddr;
	} else {
		// なければ仕方ないので適当に返す
		return n * md->inst_bytes;
	}
}

// 前ページ位置(-1)、次ページ位置(1) までのオフセットを返す。
int
DisasmMonitor::GetPageOffset(int n) const
{
	int row = nextaddr.size();

	if (n < 0) {
		// 戻る時は、命令長分ずつ戻るしかない。
		return (row - 1) * -md->inst_bytes;
	} else {
		// 次ページ開始位置。
		// 厳密には、次ページは最下行ではなくその次の(表示されていない)行から
		// だが、逆アセンブラでこれをやると連続性が失われて読みにくいので
		// 1行だぶらせることにする。
		return GetOffset(row - 1);
	}
}
