//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 逆アセンブラ (共通部分)
//

#pragma once

#include "monitor.h"
#include "debugger_memory.h"

// 逆アセンブラ(1行分)の基本クラス。
class DisasmLine
{
 public:
	virtual ~DisasmLine();

	// mem の位置を1命令逆アセンブルする。
	// 成功すれば true、失敗すれば false を返す?
	// 成功すれば text に出力文字列、ir に命令バイナリがセットされている。
	virtual bool Exec(DebuggerMemoryStream *memp) = 0;

	// 出力文字列(ダンプ)
	std::string dump {};

	// 出力文字列(命令)
	std::string text {};

	// 別名 (m88k で使う)
	std::string alttext {};

	// この命令のバイナリ列
	// (メモリイメージ順)
	std::vector<uint8> bin {};

	// メモリアクセス
	DebuggerMemoryStream *mem {};

 protected:
	// 継承先の Exec() の先頭で呼び出すこと。
	void ExecInit(DebuggerMemoryStream *mem_) {
		mem = mem_;
		dump.clear();
		text.clear();
		alttext.clear();
		bin.clear();
	}
};

// 逆アセンブルする人。基本クラスは CLI (デバッガ) 用。
class Disasm : public Object
{
	using inherited = Object;

 public:
	Disasm();
	Disasm(DebuggerMD *md_);
	~Disasm() override;

	void InitMD(DebuggerMD *md_);

	DebuggerMD *GetMD() const { return md; }

	uint32 GetAddr() const	{ return saddr.GetAddr(); }

	// saddr を代入
	void SetAddr(const saddr_t& saddr_);
	// 個別に設定
	void SetAddr(uint32 laddr_);
	void SetSuper(bool super_);
	void SetData(bool data_);
	// モードを設定
	void SetAccessMode(MemoryMode access_mode_);
	void SetLookupMode(MMULookupMode lookup_mode_);

	// アドレスをバイト単位で加算(減算)
	void Offset(int bytes);

	// コンソールに出力。
	void Print(FILE *cons, int row);

 protected:
	std::unique_ptr<DisasmLine> line {};
	DebuggerMD *md {};

	// アドレス等
	saddr_t saddr {};
	MemoryMode access_mode {};
	MMULookupMode lookup_mode {};
};

// 逆アセンブラモニタ
class DisasmMonitor : public Disasm
{
	using inherited = Disasm;
 public:
	DisasmMonitor(int idx_);
	~DisasmMonitor() override;

	// 現在位置(1行目)より n 行下の命令までのオフセットを返す。
	int GetOffset(int n) const;

	// 前ページ位置(-1)、次ページ位置(1)までのオフセットを返す。
	int GetPageOffset(int n) const;

 private:
	DECLARE_MONITOR_CALLBACK(MonitorUpdate);

	// 次の命令開始アドレス。少し変則的だが、一番上の行を1行目として、
	// [0] が2行目のアドレス、
	// [1] が3行目のアドレス、
	// :
	// [row - 2] が表示されてる最下行のアドレス、
	// [row - 1] が最下行の次のアドレス、
	// を指している。
	std::vector<uint32> nextaddr {};

	Monitor monitor { this };
};
