//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// メモリダンプ
//

#include "memdump.h"
#include "bus.h"
#include "debugger_memory.h"
#include "debugger_private.h"

//
// メモリダンプ (基本クラス)
//

// コンストラクタ
Memdump::Memdump()
	: inherited("Memdump")
{
	// 初期値
	saddr.SetSuper(true);
	saddr.SetData(true);
	SetAccessMode(MemoryMode::Physical);
	SetFormat(Word);
}

// コンストラクタ (MD 指定)
Memdump::Memdump(DebuggerMD *md_)
	: Memdump()	// 移譲
{
	InitMD(md_);
}

// デストラクタ
Memdump::~Memdump()
{
}

// md を初期化(代入)する。
void
Memdump::InitMD(DebuggerMD *md_)
{
	md = md_;
}

// saddr を設定する。
void
Memdump::SetAddr(const saddr_t& saddr_)
{
	saddr = saddr_;
}

// アドレスを設定する。
void
Memdump::SetAddr(uint32 laddr_)
{
	saddr.SetAddr(laddr_);
}

// User/Super を設定する。
void
Memdump::SetSuper(bool super_)
{
	saddr.SetSuper(super_);
}

// Data/Inst を設定する。
void
Memdump::SetData(bool data_)
{
	saddr.SetData(data_);
}

// アクセスモードを設定する。
void
Memdump::SetAccessMode(MemoryMode access_mode_)
{
	access_mode = access_mode_;
}

// MMU ルックアップモードを設定する。
void
Memdump::SetLookupMode(MMULookupMode lookup_mode_)
{
	lookup_mode = lookup_mode_;
}

// アドレスを行単位で加算(減算)する。
void
Memdump::OffsetLine(int offset)
{
	saddr += (uint32)(offset * stride);
}

// 表示フォーマットを設定する。
void
Memdump::SetFormat(Memdump::Format fmt_)
{
	fmt = fmt_;

	switch (fmt) {
	 case Byte:
	 case Word:
	 case Long:
		stride = 16;
		break;
	 case M68KMMU_SHORT:
	 case M88KMMU:
		stride = 4;
		break;
	 case M68KMMU_LONG:
		stride = 8;
		break;
	}

	// 表示形式を切り替えたことによってアドレスに端数が出るのを防ぐ
	saddr = saddr.GetAddr() & ~(stride - 1);
}

// 物理メモリ paddr 以降の内容を越権的に読み出す。bytes は 1, 2, 4 を想定。
// バスエラーなら負数を返す。
uint64
Memdump::Peek(uint32 paddr, int bytes) const
{
	uint64 data = 0;

	while (--bytes >= 0) {
		data = (data << 8) | vm_phys_peek_8(paddr++);
	}
	return data;
}

// 物理メモリ paddr 以降の内容を越権的に書き換える。
// bytes は 1, 2, 4 を想定しており、paddr はこれに整列しているはず。
// バスエラーが起きたらその場で処理を中止して (uint64)-1 を返す。
uint64
Memdump::Poke(uint32 paddr, uint32 data, int bytes)
{
	IODevice *dev = gMainbus->FindDevice(paddr);
	assert(dev);

	while (--bytes >= 0) {
		uint32 d = data >> (bytes * 8);
		uint64 berr = dev->Poke8(paddr++, d & 0xff);
		if ((int64)berr < 0) {
			return berr;
		}
	}

	return 0;
}

// この paddr が Poke() 可能なら true を返す。
bool
Memdump::CanPoke(uint32 paddr) const
{
	IODevice *dev = gMainbus->FindDevice(paddr);
	assert(dev);

	// -1 でチェック
	uint64 r = dev->Poke8(paddr, -1);
	return (bool)(r == 0);
}

// コンソールに出力。
// 出力後 saddr は更新されている。
void
Memdump::Print(FILE *cons, int row)
{
	DebuggerMemoryStream mem(md, saddr, access_mode, lookup_mode);

	for (int y = 0; y < row; y++) {
		std::string addrbuf;
		bool accessible;

		accessible = mem.FormatAddr(addrbuf);
		if (accessible == false) {
			fprintf(cons, "%s\n", addrbuf.c_str());
			mem.Offset(stride);
			continue;
		}
		// addrbuf は "VA:" もしくは "VA:PA:" の形式で返ってくる。
		// 本来は、どちらの形式であってもダンプは同じ桁から始めたいが、
		// バイト表示では 80桁に収まらない。
		// VA!=PA の場合はどうやっても 80桁に収まらないので諦めてそのまま
		// 出力するが、VA==PA の時は PA 用の空き地を削れば 80桁に収められる
		// ので、その努力はしてみる。
		if (fmt == Byte) {
			fprintf(cons, "%s ", addrbuf.c_str());
		} else {
			fprintf(cons, "%-18s ", addrbuf.c_str());
		}

		std::vector<int> vec = Read(mem, stride);

		switch (fmt) {
		 case Byte:
		 case Word:
		 case Long:
		 {
			std::string dumpbuf = DumpHex(vec);
			std::string charbuf = DumpChar(vec);
			// 左詰めでいいか
			fprintf(cons, "%s %s\n", dumpbuf.c_str(), charbuf.c_str());
			break;
		 }

		 default:
			fprintf(cons, "fmt=%d not supported\n", (int)fmt);
			break;
		}
	}

	// アドレスを更新
	saddr = mem.laddr;
}

// メモリストリームから bytes バイト分読み込んでベクタにして返す。
// 各要素はバイトデータで、バスエラーは -1 になる。
std::vector<int>
Memdump::Read(DebuggerMemoryStream& mem, int bytes)
{
	std::vector<int> vec(bytes);

	for (int i = 0; i < bytes; i++) {
		vec[i] = mem.Read8();
	}

	return vec;
}

// vec[] の文字ダンプ文字列を返す。
std::string
Memdump::DumpChar(const std::vector<int>& vec)
{
	std::string charbuf;

	for (auto c : vec) {
		if (c < 0) {
			charbuf += ' ';
		} else {
			charbuf += (0x20 <= c && c <= 0x7e) ? c : '.';
		}
	}

	return charbuf;
}

// vec[] の HEX ダンプ文字列を返す。
// CLI, GUI 共通。
std::string
Memdump::DumpHex(const std::vector<int>& vec)
{
	std::string dumpbuf;
	int span;

	switch (fmt) {
	 case Byte:	span = 1;	break;
	 case Word: span = 2;	break;
	 case Long: span = 4;	break;
	 default:
		__unreachable();
	}

	int s = span;
	for (auto c : vec) {
		if (c < 0) {
			dumpbuf += "--";
		} else {
			dumpbuf += string_format("%02x", c);
		}

		if (--s == 0) {
			dumpbuf += ' ';
			s = span;
		}
	}

	return dumpbuf;
}

// 68030 のショートフォーマットディスクリプタのダンプ文字列を返す。
std::string
Memdump::Dump68030PageShort(const std::vector<int>& vec)
{
	std::string dumpbuf;

	//  2         3         4         5         6         7
	// 9012345678901234567890123456789012345678901234567890123
	// 00000000  DT=0(Invalid)
	// ffffff00  DT=1(Page)       $00000000 CI M U WP
	// fffffff0  DT=2(Next Short) $00000000      U WP

	uint32 data = 0;
	for (auto b : vec) {
		if (b < 0) {
			return "BusErr";
		}
		data = (data << 8) | (uint8)b;
	}

	uint32 dt = (data & 0x03);
	dumpbuf = string_format("%08x  DT=%d%s", data, dt, dtname[dt]);

	if (dt != 0) {
		if (dt == 1) {
			dumpbuf += string_format(" $%08x %s %c",
				(data & 0xffff'ff00),
				(data & 0x40) ? "CI" : "- ",
				(data & 0x10) ? 'M' : '-');
		} else {
			dumpbuf += string_format(" $%08x    ", data & 0xffff'fff0);
		}
		dumpbuf += string_format(" %c %s",
			(data & 0x08) ? 'U' : '-',
			(data & 0x04) ? "WP" : "- ");
	}

	return dumpbuf;
}

// 88200 のセグメントディスクリプタとページディスクリプタ
std::string
Memdump::Dump88200Page(const std::vector<int>& vec)
{
	std::string dumpbuf;

	//  2         3         4         5         6         7
	// 9012345678901234567890123456789012345678901234567890123
	// 00000000  $00000'000 WT SP G CI M U WP V

	uint32 data = 0;
	for (auto b : vec) {
		if (b < 0) {
			return "BusErr";
		}
		data = (data << 8) | (uint8)b;
	}

	dumpbuf = string_format("%08x  $%05x'000 %s %s %c %s %c %c %s %c",
		data,
		(data >> 12),
		(data & 0x200) ? "WT" : "- ",
		(data & 0x100) ? "SP" : "- ",
		(data & 0x080) ? 'G'  : '-',
		(data & 0x040) ? "CI" : "- ",
		(data & 0x010) ? 'M'  : '-',
		(data & 0x008) ? 'U'  : '-',
		(data & 0x004) ? "WP" : "- ",
		(data & 0x001) ? 'V'  : '-');

	return dumpbuf;
}

/*static*/ const char * const
Memdump::dtname[4] = {
	"(Invalid)   ",
	"(Page)      ",
	"(Next Short)",
	"(Next Long) ",
};


//
// メモリダンプモニタ
//

// コンストラクタ
MemdumpMonitor::MemdumpMonitor(int idx_)
	: inherited()
{
	monitor.func = ToMonitorCallback(&MemdumpMonitor::MonitorUpdate);
	monitor.SetSize(84, 16);
	monitor.Regist(ID_MONITOR_MEMDUMP(idx_));
}

// デストラクタ
MemdumpMonitor::~MemdumpMonitor()
{
}

// モニタ更新
void
MemdumpMonitor::MonitorUpdate(Monitor *, TextScreen& screen)
{
	DebuggerMemoryStream mem(md, saddr, access_mode, lookup_mode);

	screen.Clear();

	// モニタでは 80桁を超えて必要サイズ確保してあるので
	// "VA:" も "VA:PA:" も同じ幅を占有し、
	// ダンプフィールドは常に 19桁目から開始する。(see Print())
	int dx = 19;

	// 文字列ダンプは桁数によらず右詰め
	int cx = screen.GetCol() - stride;

	int row = screen.GetRow();
	for (int y = 0; y < row; y++) {
		std::string addrbuf;
		bool accessible;

		accessible = mem.FormatAddr(addrbuf);
		screen.Puts(0, y, addrbuf.c_str());
		if (accessible == false) {
			mem.Offset(stride);
			continue;
		}

		std::vector<int> vec = Read(mem, stride);
		std::string charbuf = DumpChar(vec);
		switch (fmt) {
		 case Byte:
		 case Word:
		 case Long:
		 {
			std::string dumpbuf = DumpHex(vec);
			screen.Puts(dx, y, dumpbuf.c_str());
			break;
		 }

		 case M68KMMU_SHORT:
		 case M88KMMU:
			if (fmt == M68KMMU_SHORT) {
				MonitorUpdate68030PageShort(screen, y, vec);
			} else {
				MonitorUpdate88200Page(screen, y, vec);
			}
			break;

		 case M68KMMU_LONG:
			screen.Puts(dx, y, "M68KMMU_LONG not supported");
			break;
		}

		screen.Puts(cx, y, charbuf.c_str());
	}
}

void
MemdumpMonitor::MonitorUpdate68030PageShort(TextScreen& screen, int y,
	const std::vector<int>& vec)
{
	//  2         3         4         5         6         7
	// 9012345678901234567890123456789012345678901234567890123
	// 00000000  $00000000 CI M U WP DT=1(Page)
	// 00000000  $00000000      U WP DT=2(Next Short)

	uint32 data = 0;
	for (auto b : vec) {
		if (b < 0) {
			screen.Puts(19, y, "BusErr");
			return;
		}
		data = (data << 8) | (uint8)b;
	}

	screen.Print(19, y, "%08x", data);

	uint32 dt = (data & 0x03);
	TA attr = (dt != 0) ? TA::Normal : TA::Disable;

	if (dt == 1) {
		screen.Print(29, y, attr, "$%08x", data & 0xffff'ff00);
		screen.Puts(39, y, attr, (data & 0x40) ? "CI" : "-");
		screen.Putc(42, y, attr, (data & 0x10) ? 'M'  : '-');
	} else {
		screen.Print(29, y, attr, "$%08x", data & 0xffff'fff0);
	}
	screen.Putc(44, y, attr, (data & 0x08) ? 'U'  : '-');
	screen.Puts(46, y, attr, (data & 0x04) ? "WP" : "-");
	screen.Print(49, y, attr, "DT=%d%s", dt, dtname[dt]);
}

void
MemdumpMonitor::MonitorUpdate88200Page(TextScreen& screen, int y,
	const std::vector<int>& vec)
{
	//  2         3         4         5         6         7
	// 9012345678901234567890123456789012345678901234567890123
	// 00000000  $00000'000 WT SP G CI M U WP V

	uint32 data = 0;
	for (auto b : vec) {
		if (b < 0) {
			screen.Puts(19, y, "BusErr");
			return;
		}
		data = (data << 8) | (uint8)b;
	}

	screen.Print(19, y, "%08x", data);

	TA attr = (data & 0x001) ? TA::Normal : TA::Disable;
	screen.Print(29, y, attr, "$%05x'000 %s %s %c %s %c %c %s %c",
		(data >> 12),
		(data & 0x200) ? "WT" : "- ",
		(data & 0x100) ? "SP" : "- ",
		(data & 0x080) ? 'G'  : '-',
		(data & 0x040) ? "CI" : "- ",
		(data & 0x010) ? 'M'  : '-',
		(data & 0x008) ? 'U'  : '-',
		(data & 0x004) ? "WP" : "- ",
		(data & 0x001) ? 'V'  : '-');
}
