//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// ディスクイメージ (生イメージ形式)
//

#include "diskimage_raw.h"
#include <fcntl.h>
#include <sys/stat.h>

// コンストラクタ
DiskImageRaw::DiskImageRaw(const std::string& pathname_)
	: inherited(pathname_)
{
	filesize = (off_t)-1;
}

// デストラクタ
DiskImageRaw::~DiskImageRaw()
{
	Close();
}

// このディスクイメージのサイズと書き込み可能かどうかを返す。
// Open() 前でも動作する。
bool
DiskImageRaw::GetInfo(off_t *sizep, bool *w_ok)
{
	// ファイルサイズは未取得なら取得
	if (filesize == (off_t)-1) {
		struct stat st;

		if (stat(pathname.c_str(), &st) < 0) {
			warn("%s: stat failed", pathname.c_str());
			return false;
		}
		filesize = st.st_size;
	}
	*sizep = filesize;

	// ファイルのパーミッション
	int r = access(pathname.c_str(), W_OK);
	if (r == 0) {
		*w_ok = true;
	} else if (errno == EACCES) {
		*w_ok = false;
	} else {
		warn("%s: access failed", pathname.c_str());
		return false;
	}

	return true;
}

// オープン
bool
DiskImageRaw::Open(bool read_only)
{
	int openmode;

	assert(pathname.empty() == false);

	if (read_only) {
		openmode = O_RDONLY;
	} else {
		openmode = O_RDWR;
	}

	fd = open(pathname.c_str(), openmode);
	if (fd < 0) {
		warn("%s", pathname.c_str());
		return false;
	}

	// ファイルサイズは未取得なら取得
	if (filesize == (off_t)-1) {
		struct stat st;

		if (fstat(fd, &st) < 0) {
			warn("%s: stat failed", pathname.c_str());
			fd.Close();
			return false;
		}
		filesize = st.st_size;
	}

	return true;
}

void
DiskImageRaw::Close()
{
	fd.Close();
	filesize = (off_t)-1;
}

off_t
DiskImageRaw::GetSize() const
{
	return filesize;
}

bool
DiskImageRaw::Read(void *buf, off_t offset, size_t size)
{
	if (__predict_false(fd < 0)) {
		return false;
	}
	if (__predict_false(offset + size > GetSize())) {
		return false;
	}

	if (lseek(fd, offset, SEEK_SET) == (off_t)-1) {
		return false;
	}

	if (read(fd, buf, size) < 0) {
		return false;
	}
	return true;
}

bool
DiskImageRaw::Write(const void *buf, off_t offset, size_t size)
{
	if (__predict_false(fd < 0)) {
		return false;
	}
	if (__predict_false(offset + size > GetSize())) {
		return false;
	}

	if (lseek(fd, offset, SEEK_SET) == (off_t)-1) {
		return false;
	}

	if (write(fd, buf, size) < 0) {
		return false;
	}
	return true;
}
