//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ネットワークのホストデバイス
//

#pragma once

#include "hostdevice.h"
#include "macaddr.h"
#include "monitor.h"
#include "netdriver.h"
#include "spscqueue.h"

// パケット
class NetPacket : public std::array<uint8, 1600>
{
	// ジャンボフレームでないイーサネットフレームは最大 1518バイトだが
	// このパケットバッファの長さは 1519 バイト以上必要。
	// vm/lance が 1518 バイトを越えたことを検出するのに使っている。

 public:
	// バッファをクリアする
	void Clear() { length = 0; }

	// パケットに1バイト追加する
	void Append(uint8 data) { (*this)[length++] = data; }

	// バッファ内の有効長
	int length {};
};

class HostNetDevice : public HostDevice
{
	using inherited = HostDevice;
	using NetTxQueue = SPSCQueue<NetPacket, 32>;
	using NetRxQueue = SPSCQueue<NetPacket, 64>;

	// 統計情報
	struct stat_t {
		uint64 tx_pkts;
		uint64 tx_bytes;
		uint64 rx_pkts;
		uint64 rx_bytes;
		uint64 read_pkts;
		uint64 read_bytes;
		uint64 write_pkts;
		uint64 write_bytes;
		int txqfull_pkts;			// キューが一杯だった回数
		uint64 txqfull_bytes;		// キューに入れられなかったバイト数
		uint64 rxdisable_pkts;		// 上位層が受信無効
		uint64 rxdisable_bytes;		// 上位層が受信無効
		uint64 rxjumbo_pkts;		// 受信したジャンボパケット数
		uint64 rxjumbo_bytes;		// 受信したジャンボパケットのバイト数
		int rxqfull_pkts;			// キューが一杯だった回数
		uint64 rxqfull_bytes;		// キューに入れられなかったバイト数
		int rxfilter_pkts;			// 受信フィルタで破棄したパケット数
		uint64 rxfilter_bytes;		// フィルタされたバイト数

		int txq_peak;				// キューのおおよその最大使用量
		int rxq_peak;				// キューのおおよその最大使用量
	};

 public:
	HostNetDevice();
	~HostNetDevice() override;

	void SetLogLevel(int loglevel_) override;
	bool Init() override;

	// パケットを送信する。
	// data はプリアンブルを含まず、イーサネットフレームヘッダ(14バイト)、
	// データ(46-1500バイト) からなる。FCS(CRC) (4バイト) は含まない。
	// VM から引き取ったパケットについて tx_pkts, tx_bytes を更新すること
	// 送信キューに追加できれば true、出来なければ false を返す。
	bool Tx(const NetPacket& packet);

	// パケットを引き取るために VM から呼ばれる。
	// パケットが受信キューにあるときは packet にコピーして true を返す。
	// キューが空なら false を返す。
	bool Rx(NetPacket *packet);

	// 受信キューへの投入を許可/禁止する。
	void EnableRx(bool enable);

	// 自身の MAC アドレスを設定する。
	void SetMyAddr(const macaddr_t& myaddr_);

	// プロミスキャスモードを設定する。
	void SetPromisc(bool promisc_);

	// ドライバ名を返す (ステータスパネルから呼ばれる)
	const std::string GetDriverName() const;

	// 統計情報を返す (ステータスパネルから呼ばれる)
	uint64 GetTXPkts() const { return stat.tx_pkts; }
	uint64 GetRXPkts() const { return stat.rx_pkts; }

	// コンパイル済みのドライバ名一覧を返す
	static std::vector<std::string> GetDrivers();

 private:
	bool SelectDriver();
	bool CreateNone();
	bool CreateTap();
	bool CreateBPF();
	bool CreateAFPacket();

	int Read() override;

	// 外部への書き出し(パケットを送信)
	void Write(uint32 data) override;

	DECLARE_MONITOR_CALLBACK(MonitorUpdate);

	std::unique_ptr<NetDriver> driver {};

	bool rx_enable {};
	macaddr_t myaddr {};		// 自己 MAC アドレス
	bool promisc {};			// プロミスキャスモードなら true

	// キュー
	NetTxQueue txq {};
	NetRxQueue rxq {};

	// 統計情報
	struct stat_t stat {};

	std::string errmsg {};

	Monitor monitor { this };
};
