//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// ビットマップ
//

#include "bitmap.h"
#include <cmath>

//
// ビットマップ (基本クラス)
//

// コンストラクタ
BitmapBase::BitmapBase()
{
}

// コンストラクタ
BitmapBase::BitmapBase(int width_, int height_, int depth_)
{
	Create(width_, height_, depth_);
}

// デストラクタ
BitmapBase::~BitmapBase()
{
}

// 作成。width_ * height_ が面積を持っていれば true を返す。
bool
BitmapBase::Create(int width_, int height_, int depth_)
{
	assert(depth_ == 1 || depth_ == 8 || depth_ == 24);

	width = width_;
	height = height_;

	if (depth_ == 1) {
		stride = (width + 7) / 8;
	} else {
		stride = width * depth_ / 8;
	}

	return (width > 0 && height > 0);
}


//
// 1bpp ビットマップ
//

// コンストラクタ
BitmapI1::BitmapI1()
{
}

// コンストラクタ
BitmapI1::BitmapI1(int width_, int height_)
{
	Create(width_, height_);
}

// コンストラクタ
BitmapI1::BitmapI1(const void *buf_, int width_, int height_)
	: inherited(width_, height_, 1)
{
	buf = static_cast<uint8 *>(const_cast<void *>(buf_));
}

// デストラクタ
BitmapI1::~BitmapI1()
{
}

// 作成
void
BitmapI1::Create(int width_, int height_)
{
	if (inherited::Create(width_, height_, 1)) {
		owned_buf.reset(new uint8[stride * height]);
		buf = owned_buf.get();
	} else {
		owned_buf.reset();
		buf = NULL;
	}
}

// カラーコード cc で全体を塗りつぶす。
// 未使用ビットも変更される。
void
BitmapI1::Fill(int cc)
{
	assert(buf);

	memset(buf, (cc ? 0xff : 0), stride * height);
}


//
// 8bpp ビットマップ
//

// コンストラクタ
BitmapI8::BitmapI8()
{
}

// コンストラクタ
BitmapI8::BitmapI8(int width_, int height_)
{
	Create(width_, height_);
}

// コンストラクタ
BitmapI8::BitmapI8(const void *buf_, int width_, int height_)
{
	Create(buf_, width_, height_);
}

// デストラクタ
BitmapI8::~BitmapI8()
{
}

// 作成
void
BitmapI8::Create(int width_, int height_)
{
	if (inherited::Create(width_, height_, 8)) {
		owned_buf.reset(new uint8[stride * height]);
		buf = owned_buf.get();
	} else {
		owned_buf.reset();
		buf = NULL;
	}
}

// 元データを指定して作成
void
BitmapI8::Create(const void *buf_, int width_, int height_)
{
	if (inherited::Create(width_, height_, 8)) {
		buf = static_cast<uint8 *>(const_cast<void *>(buf_));
	} else {
		buf = NULL;
	}
}

// カラーコード cc で全体を塗りつぶす。
// 未使用ビットも変更される。
void
BitmapI8::Fill(int cc)
{
	assert(buf);

	memset(buf, cc, stride * height);
}


//
// 24bpp ビットマップ
//

// コンストラクタ
BitmapRGB::BitmapRGB()
{
}

// コンストラクタ
BitmapRGB::BitmapRGB(int width_, int height_)
{
	Create(width_, height_);
}

// コンストラクタ
BitmapRGB::BitmapRGB(const void *buf_, int width_, int height_)
	: inherited(width_, height_, 24)
{
	buf = static_cast<uint8 *>(const_cast<void *>(buf_));
}

// デストラクタ
BitmapRGB::~BitmapRGB()
{
}

// 作成
void
BitmapRGB::Create(int width_, int height_)
{
	if (inherited::Create(width_, height_, 24)) {
		owned_buf.reset(new uint8[stride * height]);
		buf = owned_buf.get();
	} else {
		owned_buf.reset();
		buf = NULL;
	}
}

// 色 c で全体を塗りつぶす。
void
BitmapRGB::Fill(Color c)
{
	FillRect(c, 0, 0, width, height);
}

// (dx, dy) にビットマップ src を描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawBitmap(int dx, int dy, const BitmapI1& src, const Color *palette)
{
	assert(buf);

	const uint8 *srcbuf = src.GetBuf();

	uint8 data = 0;
	for (int y = 0; y < src.GetHeight(); y++) {
		uint8 *d = GetPtr(dx, dy + y);
		for (int x = 0; x < src.GetWidth(); x++) {
			if (__predict_false(x % 8 == 0)) {
				data = *srcbuf++;
			}
			int cc;
			if ((int8)data < 0) {
				cc = 1;
			} else {
				cc = 0;
			}
			data <<= 1;

			Color c = palette[cc];
			*d++ = c.r;
			*d++ = c.g;
			*d++ = c.b;
		}
	}
}

// (dx, dy) にビットマップ src の全域を描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawBitmap(int dx, int dy, const BitmapI8& src, const Color *palette)
{
	Rect sr(0, 0, src.GetWidth(), src.GetHeight());
	DrawBitmap(dx, dy, src, palette, sr);
}

// (dx, dy) にビットマップ src の範囲 sr を描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawBitmap(int dx, int dy, const BitmapI8& src, const Color *palette,
	const Rect& sr)
{
	for (int y = 0; y < sr.h; y++) {
		uint8 *d = GetPtr(dx, dy + y);
		const uint8 *s = src.GetPtr(sr.x, sr.y + y);
		for (int x = 0; x < sr.w; x++) {
			int cc = *s++;
			Color c = palette[cc];
			*d++ = c.r;
			*d++ = c.g;
			*d++ = c.b;
		}
	}
}

// (dx, dy) にビットマップ src を描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawBitmap(int dx, int dy, const BitmapRGB& src)
{
	assert(buf);

	const uint8 *s = src.GetBuf();

	for (int y = 0; y < src.GetHeight(); y++) {
		uint8 *d = GetPtr(dx, dy + y);
		for (int x = 0; x < src.GetWidth(); x++) {
			*d++ = *s++;
			*d++ = *s++;
			*d++ = *s++;
		}
	}
}

// (dx, dy) にビットマップ src を scale 倍にして描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawScaleUp(int dx, int dy, int scale, const BitmapI1& src,
	const Color *palette)
{
	assert(buf);

	const uint8 *srcbuf = src.GetBuf();
	uint8 data = 0;
	int dx0 = dx;

	for (int y = 0; y < src.GetHeight(); y++) {
		dx = dx0;
		for (int x = 0; x < src.GetWidth(); x++) {
			if (__predict_false(x % 8 == 0)) {
				data = *srcbuf++;
			}
			int cc;
			if ((int8)data < 0) {
				cc = 1;
			} else {
				cc = 0;
			}
			data <<= 1;

			FillRect(palette[cc], dx, dy, scale, scale);
			dx += scale;
		}
		dy += scale;
	}
}

// 分数で表現可能な拡大縮小描画。
// dr にビットマップ src の (sxN / sxD, syN / syD) からを
// dst の 1 ピクセルに対して (sxS / sxD, syS / syD) ずつ動かした位置を
// 読み出し位置として描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawScale(const Rect& dr, const BitmapI8& src, const Color *palette,
	int sxN, int sxD, int sxS, int syN, int syD, int syS)
{
	assert(buf);
	assert(sxD >= 1);
	assert(syD >= 1);
	assert(sxS != 0);
	assert(syS != 0);

	int sxI = sxN / sxD;
	int syI = syN / syD;

	sxN %= sxD;
	syN %= syD;

	int sxI0 = sxI;
	int sxN0 = sxN;

	int sx;
	int sy = syI - 1;

	for (int y = 0; y < dr.h; y++) {
		uint8 *d = GetPtr(dr.x, dr.y + y);

		if (sy != syI) {
			sy = syI;

			sx = sxI0 - 1;
			sxI = sxI0;
			sxN = sxN0;

			uint8 cc = 0;
			for (int x = 0; x < dr.w; x++) {
				if (sx != sxI) {
					sx = sxI;
					const uint8 *s = src.GetPtr(sx, sy);
					cc = *s;
				}
				const Color& c = palette[cc];
				*d++ = c.r;
				*d++ = c.g;
				*d++ = c.b;

				sxN += sxS;
				if (sxN >= sxD) {
					sxI += sxN / sxD;
					sxN %= sxD;
				}
			}
		} else {
			// 前のラスタの中からコピーすればよい
			memcpy(d, d - stride, dr.w * 3);
		}

		syN += syS;
		if (syN >= syD) {
			syI += syN / syD;
			syN %= syD;
		}
	}
}

// sr 範囲の色の平均を返す
Color
BitmapRGB::Mean(const RectD& sr) const
{
	int x1 = std::floor(sr.x);
	int y1 = std::floor(sr.y);
	int x2 = std::ceil(sr.x + sr.w);
	int y2 = std::ceil(sr.y + sr.h);
	if (__predict_false(x2 >= GetWidth())) {
		x2 = GetWidth();
	}
	if (__predict_false(y2 >= GetHeight())) {
		y2 = GetHeight();
	}

	ColorD sum;
	// サブピクセルの辺の長さ
	double rx, ry;
	double s = 0;
	for (int y = y1; y < y2; y++) {
		if (y == y1) {
			ry = 1 - (sr.y - y1);
		} else if (y == y2 - 1) {
			ry = 1 - (y2 - (sr.y + sr.h));
		} else {
			ry = 1;
		}

		const uint8 *p = GetPtr(x1, y);
		for (int x = x1; x < x2; x++) {
			if (x == x1) {
				rx = 1 - (sr.x - x1);
			} else if (y == x2 - 1) {
				rx = 1 - (x2 - (sr.x + sr.w));
			} else {
				rx = 1;
			}

			Color c;
			c.r = *p++;
			c.g = *p++;
			c.b = *p++;

			// 面積比をかけて合成
			sum.r += c.r * rx * ry;
			sum.g += c.g * rx * ry;
			sum.b += c.b * rx * ry;
			s += rx * ry;
		}
	}

	// 平均にする
	sum.r /= s;
	sum.g /= s;
	sum.b /= s;

	return Color(sum.r, sum.g, sum.b);
}

// 伸縮して描画
void
BitmapRGB::StretchDraw(Rect dr, const BitmapRGB& src)
{
	if (__predict_false(dr.w == src.GetWidth() && dr.h == src.GetHeight())) {
		DrawBitmap(dr.x, dr.y, src);
		return;
	}

	// dst の 1 pixel が src の 何 pixel に相当するか
	double tx = (double)src.GetWidth() / dr.w;
	double ty = (double)src.GetHeight() / dr.h;

	// src 上を動くウィンドウ
	RectD sr(0, 0, tx, ty);

	for (int y = 0; y < dr.h; y++) {
		uint8 *d = GetRowPtr(y);
		sr.x = 0;
		for (int x = 0; x < dr.w; x++) {
			Color c = src.Mean(sr);

			*d++ = c.r;
			*d++ = c.g;
			*d++ = c.b;
			sr.x += tx;
		}
		sr.y += ty;
	}
}

// 直線 (x1, y1) - (x2, y2) を描画する。終点は開区間。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawLine(Color c, int x1, int y1, int x2, int y2)
{
	if (y1 == y2) {
		DrawLineH(c, x1, y1, x2);
	} else if (x1 == x2) {
		DrawLineV(c, x1, y1, y2);
	} else {
		assert(false);
	}
}

// 水平の直線 (x1, y1) - (x2, y1) を描画する。終点は開区間。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawLineH(Color c, int x1, int y1, int x2)
{
	assert(buf);

	if (__predict_false(x1 > x2)) {
		std::swap(x1, x2);
		x1++;
		x2++;
	}

	uint8 *d = GetPtr(x1, y1);

	// 12バイト境界に整列するまで
	for (; (uintptr_t)d % 4 != 0 && x1 < x2 ; x1++) {
		*d++ = c.r;
		*d++ = c.g;
		*d++ = c.b;
	}

	// 12バイト境界から 12バイト以上残ってる間
	int cnt = (x2 - x1) / 4;
	if (__predict_true(cnt > 0)) {
		uint32 pat[3];
		for (uint8 *p = (uint8 *)pat, *pend = p + sizeof(pat); p < pend; ) {
			*p++ = c.r;
			*p++ = c.g;
			*p++ = c.b;
		}
		for (int i = 0; i < cnt; i++) {
			memcpy(d, pat, sizeof(pat));
			d += sizeof(pat);
		}
		x1 += cnt * sizeof(pat) / 3;
	}

	// 残り
	for (; x1 < x2; x1++) {
		*d++ = c.r;
		*d++ = c.g;
		*d++ = c.b;
	}
}

// 垂直の直線 (x1, y1) - (x1, y2) を描画する。終点は開区間。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawLineV(Color c, int x1, int y1, int y2)
{
	assert(buf);

	if (__predict_false(y1 > y2)) {
		std::swap(y1, y2);
		y1++;
		y2++;
	}

	for (; y1 < y2; y1++) {
		uint8 *d = GetPtr(x1, y1);
		*d++ = c.r;
		*d++ = c.g;
		*d++ = c.b;
	}
}

// 矩形 (dx, dy) - (dw * dh) を色 c で塗りつぶす。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::FillRect(Color c, int dx, int dy, int dw, int dh)
{
	FillRect(c, Rect(dx, dy, dw, dh));
}

// 矩形 rect を色 c で塗りつぶす。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::FillRect(Color c, const Rect& rect)
{
	assert(buf);

	const int l = rect.x;
	const int t = rect.y;

	DrawLineH(c, l, t, l + rect.w);

	uint8 *s = GetPtr(l, t);
	uint8 *d = s + stride;
	const int len = rect.w * 3;

	for (int y = t + 1; y < t + rect.h; y++) {
		memcpy(d, s, len);
		d += stride;
	}
}

// 矩形 (dx, dy) - (dw * dh) を色 c で描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawRect(Color c, int dx, int dy, int dw, int dh)
{
	DrawRect(c, Rect(dx, dy, dw, dh));
}

// 矩形 rect の辺を色 c で描画する。
// 描画範囲をはみ出さないこと。
void
BitmapRGB::DrawRect(Color c, const Rect& rect)
{
	assert(buf);

	int l = rect.x;
	int t = rect.y;
	int r = rect.GetRight();
	int b = rect.GetBottom();

	DrawLineH(c, l, t, r);	// 上辺→
	DrawLineV(c, r, t, b);	// 右辺↓
	DrawLineH(c, r, b, l);	// 下辺←
	DrawLineV(c, l, b, t);	// 左辺↑
}
