//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// ビットマップ
//

#pragma once

#include "header.h"
#include "color.h"
#include "rect.h"

// ビットマップ (基本クラス)
class BitmapBase
{
 protected:
	BitmapBase();
	BitmapBase(int width_, int height_, int depth_);

	bool Create(int width_, int height_, int depth_);
 public:
	virtual ~BitmapBase();

	int GetWidth() const { return width; }
	int GetHeight() const { return height; }
	int GetStride() const { return stride; }

	uint8 *GetBuf() { return buf; }
	const uint8 *GetBuf() const { return buf; }

 protected:
	int width {};
	int height {};
	int stride {};

	uint8 *buf {};
	std::unique_ptr<uint8[]> owned_buf {};
};

// 1bpp ビットマップ
class BitmapI1 : public BitmapBase
{
	using inherited = BitmapBase;
 public:
	BitmapI1();
	BitmapI1(int width_, int height_);
	BitmapI1(const void *buf_, int width_, int height_);
	~BitmapI1() override;

	void Create(int width_, int height_);

	// カラーコード cc で全体を塗りつぶす
	void Fill(int cc);
};

// 8bpp ビットマップ
class BitmapI8 : public BitmapBase
{
	using inherited = BitmapBase;
 public:
	BitmapI8();
	BitmapI8(int width_, int height_);
	BitmapI8(const void *buf_, int width_, int height_);
	~BitmapI8() override;

	void Create(int width_, int height_);
	void Create(const void *buf_, int width_, int height_);

	// カラーコード cc で全体を塗りつぶす
	void Fill(int cc);

	uint8 *GetRowPtr(int y) const {
		assertmsg(y < height, "y=%d height=%d", y, height);
		return &buf[y * stride];
	}
	uint8 *GetPtr(int x, int y) const {
		assertmsg(x < width, "x=%d width=%d", x, width);
		assertmsg(y < height, "y=%d height=%d", y, height);
		return &buf[y * stride + x];
	}
};

// 24bpp ビットマップ
class BitmapRGB : public BitmapBase
{
	using inherited = BitmapBase;
 public:
	BitmapRGB();
	BitmapRGB(int width_, int height_);
	BitmapRGB(const void *buf_, int width_, int height_);
	~BitmapRGB() override;

	void Create(int width_, int height_);

	// 色 c で全体を塗りつぶす
	void Fill(Color c);

	void DrawLine(Color c, int x1, int y1, int x2, int y2);
	void DrawLineH(Color c, int x1, int y1, int x2);
	void DrawLineV(Color c, int x1, int y1, int y2);

	void FillRect(Color c, const Rect& rect);
	void FillRect(Color c, int dx, int dy, int dw, int dh);

	void DrawRect(Color c, const Rect& rect);
	void DrawRect(Color c, int dx, int dy, int dw, int dh);

	void DrawBitmap(int dx, int dy, const BitmapI1& src, const Color *palette);
	void DrawBitmap(int dx, int dy, const BitmapI8& src, const Color *palette);
	void DrawBitmap(int dx, int dy, const BitmapI8& src, const Color *palette,
		const Rect& sr);
	void DrawBitmap(int dx, int dy, const BitmapRGB& src);
	void DrawScaleUp(int dx, int dy, int scale, const BitmapI1& src,
		const Color *palette);
	void DrawScale(const Rect& dr, const BitmapI8& src, const Color *palette,
		int sxN, int sxD, int sxS, int syN, int syD, int syS);

	void StretchDraw(Rect dr, const BitmapRGB& src);

	uint8 *GetRowPtr(int y) const {
		assertmsg(y < height, "y=%d height=%d", y, height);
		return &buf[y * stride];
	}
	uint8 *GetPtr(int x, int y) const {
		assertmsg(x < width, "x=%d width=%d", x, width);
		assertmsg(y < height, "y=%d height=%d", y, height);
		return &buf[y * stride + x * 3];
	}

 private:
	Color Mean(const RectD& sr) const;
};
