//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// モニタ管理
//

#pragma once

#include "geometric.h"
#include "mainapp.h"
#include "object.h"
#include "textscreen.h"

// モニタ識別子
//
// テキストモニタの識別子だがメニューID (とウィンドウの識別ID) としても使う
// ので wxmainframe あたりも参照。
enum {
#include "monitor_id.h"
};

// 識別子 (n番目で指定)
static constexpr int ID_MONITOR_ATC(int n) {
	return ID_MONITOR_ATC0 + n;
}
static constexpr int ID_MONITOR_CMMU(int n) {
	return ID_MONITOR_CMMU0 + n;
}
static constexpr int ID_SUBWIN_CACHE(int n) {
	return ID_SUBWIN_CACHE0 + n;
}
static constexpr int ID_MONITOR_DISASM(int n) {
	return ID_MONITOR_DISASM0 + n;
}
static constexpr int ID_MONITOR_MEMDUMP(int n) {
	return ID_MONITOR_MEMDUMP0 + n;
}
// 判定
static constexpr bool IS_MONITOR_DISASM(int id) {
	return (ID_MONITOR_DISASM0 <= id &&
	        id < ID_MONITOR_DISASM(MAX_MEMDUMP_MONITOR));
}
static constexpr bool IS_MONITOR_MEMDUMP(int id) {
	return (ID_MONITOR_MEMDUMP0 <= id &&
	        id < ID_MONITOR_MEMDUMP(MAX_MEMDUMP_MONITOR));
}

class Monitor;
using MonitorCallback_t = void (Object::*)(Monitor *, TextScreen&);
#define ToMonitorCallback(f) static_cast<MonitorCallback_t>(f)

// モニタコールバックを宣言するマクロ。
// 強制力はないけどヘッダで宣言するところで使うこと。
// コールバック関数の型が変わった時にコンパイルエラーで検出させるため。
#define DECLARE_MONITOR_CALLBACK(name)	void name (Monitor *, TextScreen&)
// モニタ更新を呼び出すマクロ
#define MONITOR_UPDATE(m, scr) ((((m).obj)->*((m).func))(&(m), (scr)))

// モニタ
class Monitor
{
 public:
	Monitor();
	Monitor(Object *obj_);
	virtual ~Monitor();

	Object *obj {};				// オブジェクト
	MonitorCallback_t func {};	// コールバック関数

	int GetId() const { return id; }

	// モニターの表示サイズを設定・取得。
	void SetSize(nnSize size_) { size = size_; }
	void SetSize(int col_, int row_) { SetSize(nnSize(col_, row_)); }
	virtual nnSize GetSize() const { return size; }

	// モニターが縦に可変長の場合の最大表示行数。(固定長なら 0)
	void SetMaxHeight(int maxh_) { maxh = maxh_; }
	int GetMaxHeight() const { return maxh; }

	// 登録。
	// モニタの登録はコンストラクタおよび Create フェーズまでに行うこと。
	// Create() の後、Init() より前に -M オプションの処理があるため。
	// see wx/wxapp.cpp
	void Regist(int id);

	// 削除。
	void Unregist();

	// ID を表示用文字列にして返す。
	static const char *GetIdText(int id);

	// このモニターの ID を表示用文字列にして返す。
	const char *GetIdText() const { return GetIdText(id); }

 private:
	// モニター識別子
	int id;

	// モニターサイズ
	nnSize size {};

	// 縦に可変長の場合の最大表示行数 (固定長なら 0)
	int maxh {};

	// ID を表示用のテキストにしたものとの対応表。
	static const std::vector<std::pair<int, const char *>> idtext;
};

//
// モニターマネージャ
//
class MonitorManager
{
	friend class Monitor;

	// モニター情報 (事前定義)
	struct MonitorInfo {
		int id;
		VMCap vmcap;
		std::vector<std::string> aliases;
	};

 public:
	MonitorManager();
	~MonitorManager();

	// 現在登録されているモニターのリストを返す。
	const std::vector<Monitor *>& GetList() const { return list; }

	// 指定された id を持つモニターを返す。
	// こっちは、なければ assert する。
	Monitor& Get(int id) const {
		for (auto mon : list) {
			if (mon->GetId() == id) {
				return *mon;
			}
		}
		PANIC("id=%d not found", id);
	}

	// 指定された id を持つモニターを返す。
	// こっちは、なければ NULL を返す。
	Monitor *Find(int id) const {
		for (auto mon : list) {
			if (mon->GetId() == id) {
				return mon;
			}
		}
		return NULL;
	}

	// 指定された id に対応する機種ケーパビリティを返す。
	VMCap GetVMCap(int id) const {
		assert(0 <= id && id < info.size());
		return info[id].vmcap;
	}

	// 指定された id に対応する名前のリストを返す。
	// id が範囲外だと assert する。
	const std::vector<std::string>& GetAliases(int id) const {
		assert(0 <= id && id < info.size());
		return info[id].aliases;
	}

	// id リストから一覧表示用の文字列を生成して返す
	static std::string MakeListString(std::vector<int> list);

	// 以降の追加を禁止する (プログラムミスを避けるため)
	void Fix();

 private:
	// 登録 (friend の Monitor からのみ呼べる)
	void Regist(Monitor *mon);

	// 削除 (friend の Monitor からのみ呼べる)
	void Unregist(Monitor *mon);

	// 現在登録されているモニターのリスト
	std::vector<Monitor *> list {};

	// モニタの登録は -M オプションの処理が始まるまでに終えないといけない。
	bool fixed {};

	// 事前に設定されているモニター情報
	static const std::vector<MonitorInfo> info;
};

extern MonitorManager *gMonitorManager;
