//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 基本オブジェクト
//

#pragma once

#include "mystring.h"
#include <functional>
#include <vector>

class Logger;

// ログ表示に指定するラムダ式の便利マクロ。
// fmt 内では __func__ が使えないので代わりに lam_func を使うこと。
// lstr("%s value=%d", lam_func, value) のようになる。
// 他のキャプチャ指定子が必要になったらまた考える。
#define lstr(fmt...)	\
	[&, lam_func = __func__] { (void)lam_func; return string_format(fmt); }

// 基本オブジェクト
class Object
{
 public:
	// 引数はオブジェクト名(ログ表示名)で、これはログに前置される。
	// またコンストラクタ指定の場合はエイリアスにも登録する。
	// これは表示名と指示名が同じことが多いための便宜措置。
	// 指示名に表示名を使いたくない場合は ClearAlias() してから
	// AddAlias() すること。
	Object(const std::string& objname_);

	virtual ~Object();

	// オブジェクト名(ログ表示名)を取得する
	const std::string& GetName() const { return objname; }

	// エイリアスリストを返す
	const std::vector<std::string>& GetAliases() const { return log_aliases; }

 protected:
	// オブジェクト名を newname に設定(更新)する。
	// 通常、オブジェクト名はコンストラクタで指定するが、コンストラクタ時点
	// ではまだ名前が決まらないような場合にだけ使う。
	// こちらはエイリアスには関与しない。
	void SetName(const std::string& newname) {
		objname = newname;
	}

	// エイリアスをクリアする。
	// エイリアスは初期値として、コンストラクタで設定したオブジェクト名が
	// 1つ入っているが、これを削除したい時に使う。
	void ClearAlias() {
		log_aliases.clear();
	}

	// エイリアスを追加する。
	void AddAlias(const std::string& alias) {
		log_aliases.emplace_back(alias);
	}

 private:
	// オブジェクト名。
	// こちらはログの先頭に追加される。
	std::string objname {};

	// エイリアスのリスト。
	// こちらはログ名を指定する時に使う。
	std::vector<std::string> log_aliases {};

 public:
	//
	// ログ
	//

	// ログレベルを設定する
	virtual void SetLogLevel(int loglevel_);

	// メッセージ出力
	// メッセージは仮想時刻も現在の PC も表示しない。Init() など VM 実行前や
	// 仮想マシン(特に仮想時刻、PC) に関係ないあたりはこちらを使用。

	// レベル指定なし、フォーマット版。(レベルを評価した後で使う)
	void putmsgn(const char *fmt, ...) const __printflike(2, 3);

	// レベル指定あり、フォーマット版。
#define putmsg(lv, fmt...)	do {	\
	if (__predict_false(loglevel >= (lv))) {	\
		putmsgn(fmt);	\
	}	\
} while (0)

	// レベル指定あり、ラムダ式版。
	void putmsgf(int lv, std::function<std::string()> msgf) const {
		if (__predict_false(loglevel >= lv)) {
			putmsgn("%s", msgf().c_str());
		}
	}

	// ログ出力
	// ログは仮想時刻と現在の PC も表示する。VM 実行中は基本的にこれで表示。

	// レベル指定なし、フォーマット版。(レベルを評価した後で使う)
	virtual void putlogn(const char *fmt, ...) const __printflike(2, 3);

	// レベル指定あり、フォーマット版。
#define putlog(lv, fmt...)	do {	\
	if (__predict_false(loglevel >= (lv))) {	\
		putlogn(fmt);	\
	}	\
} while (0)

	// レベル指定あり、ラムダ式版。
	void putlogf(int lv, std::function<std::string()> msgf) const {
		if (__predict_false(loglevel >= lv)) {
			putlogn("%s", msgf().c_str());
		}
	}

	// オブジェクト登録時に代入される
	Logger *logger {};

	// ログレベルの目安
	// 0: 未実装動作などのデバッグログではない情報
	// 1: 大雑把なデバッグログ
	// 2: 詳細なデバッグログ
	// 3: うるさいデバッグログ
	int loglevel {};

 protected:
	// ログ直接出力
	void WriteLog(const char *) const;
};
