//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

//
// 外部公開用
//

#include "branchhistory.h"

__BEGIN_DECLS
#include "fpu_emulate.h"
__END_DECLS

class m68kcpu;
class MPU680x0Device;

#define M68K_SR_T		(0xc000)
#define M68K_SR_S		(0x2000)
#define M68K_SR_M		(0x1000)
#define M68K_SR_IMASK	(0x0700)
#define M68K_CCR_X		(0x0010)
#define M68K_CCR_N		(0x0008)
#define M68K_CCR_Z		(0x0004)
#define M68K_CCR_V		(0x0002)
#define M68K_CCR_C		(0x0001)

// SSW
#define SSW_SIZE(x)		(((x) & 0x03) << 4)
#define SSW_SIZE_BYTE	SSW_SIZE(1)
#define SSW_SIZE_WORD	SSW_SIZE(2)
#define SSW_SIZE_3BYTES	SSW_SIZE(3)
#define SSW_SIZE_LONG	SSW_SIZE(0)
#define SSW_SIZE_MASK	SSW_SIZE(0x03)

#define SSW_FC_MASK		(0x0007)	// FC_*
#define SSW_RW			(0x0040)	// SSW_BUS_{R,W}
#define SSW_RM			(0x0080)
#define SSW_DF			(0x0100)	// Data Fault
#define SSW_RB			(0x1000)
#define SSW_RC			(0x2000)
#define SSW_FB			(0x4000)
#define SSW_FC			(0x8000)

#define SSW_BUS_R		(0x0040)
#define SSW_BUS_W		(0x0000)

#define FC_DATA			(1)
#define FC_PROG			(2)
#define FC_USER			(0)
#define FC_SUPER		(4)
#define FC_UD			(FC_USER  | FC_DATA)
#define FC_UP			(FC_USER  | FC_PROG)
#define FC_SD			(FC_SUPER | FC_DATA)
#define FC_SP			(FC_SUPER | FC_PROG)

#define M68K_CACR_EI	(0x0001)
#define M68K_CACR_FI	(0x0002)
#define M68K_CACR_CEI	(0x0004)
#define M68K_CACR_CI	(0x0008)
#define M68K_CACR_IBE	(0x0010)
#define M68K_CACR_ED	(0x0100)
#define M68K_CACR_FD	(0x0200)
#define M68K_CACR_CED	(0x0400)
#define M68K_CACR_CD	(0x0800)
#define M68K_CACR_DBE	(0x1000)
#define M68K_CACR_WA	(0x2000)

#define M68K_FPCR_MASK	(0x0000fff0)
#define M68K_FPSR_MASK	(0x0ffffff8)

#include "m68030ccr.h"
#include "m68030mmu.h"

// レジスタイメージを保持する構造体。
// この構造体はコピーとか比較をするのでポインタとかは置かないこと。
struct m68kreg {
 public:
	uint32 pc;
	uint32 da[16];

	m68030ccr ccr;
	bool s;				// SR の S ビット
	bool m;				// SR の M ビット
	int  intr_mask;		// SR の割り込みマスク (0-7)

	// 現在の SP は常に da[15]。da[15] が示しているときは対応する変数の
	// ほうは更新されないため無効。
	// ユーザ状態なら da[15] が USP を示し usp は無効、
	// isp, msp は ISP, MSP を示して有効。
	// スーパバイザ状態で SR_M なら da[15] が MSP を示し msp は無効、
	// isp, usp は ISP, USP を示して有効。
	// スーパバイザ状態で !SR_M なら da[15] が ISP を示し isp は無効、
	// msp, usp は MSP, USP を示して有効。
	uint32 usp;
	uint32 isp;
	uint32 msp;

	uint32 vbr;
	uint32 sfc;
	uint32 dfc;

	uint32 cacr;
	uint32 caar;

	union64 srp;
	union64 crp;
	uint32 tt[2];
	uint32 tc;
	uint16 mmusr;

	struct fpframe fpframe;

 public:
	// SR の上位8ビット部分のみを返す
	uint16 sr_h() {
		return (s ? M68K_SR_S : 0) | (m ? M68K_SR_M : 0) | (intr_mask << 8);
	}
};

// バス状態
struct m68kbus {
	// 論理アドレスと SSW。
	// SSW はアクセス前に SSW_BUS_{R,W} と FC[012] をセットすること。
	DEF_UNION64(, ssw_laddr, ssw, laddr);

	uint32 paddr;			// 対応する物理アドレス
	// データバス。
	// 今の所書き込みだけで使う。読み込みデータは戻り値のみ。
	uint32 data;

	// SSW のうち現在の FC2 だけ別途覚えておく。
	// FC2 が変化するのは SR:S による特権モードが変わった時だけなので。
	uint16 fc2;

	// バスが Write なら true を返す
	bool IsWrite() const { return (ssw & SSW_BUS_R) == 0; }

	// このへんどうするか

	// FC2 をキャッシュしておく。
	void SetFC2(bool newfc2) {
		fc2 = newfc2 ? 4 : 0;
	}
	// FC を取得する。
	// これは MMU ルーチンから呼ばれる。そのためこの時点ではすでに
	// FC0/1 と FC2 は ssw に合成済みなので、これでよい。
	uint GetFC() const { return ssw & SSW_FC_MASK; }
};

class m68kcpu final {
 public:
	m68kcpu() { }
	~m68kcpu() { }

 public:
	m68kreg reg {};
	uint32 ppc {};			// 現在実行中の命令の先頭アドレス
	uint16 ir {};
	uint16 ir2 {};

	uint32 save_pd[8] {};	// -(An) を保存
	uint32 save_pi[8] {};	// (An)+ を保存
	uint8  flag_pd {};		// -(An) のどのレジスタを保存したか
	uint8  flag_pi {};		// (An)+ のどのレジスタを保存したか

	// CPU の実行状態
	static const uint32 CPU_STATE_NORMAL	= 0;
	static const uint32 CPU_STATE_STOP		= 1;
	static const uint32 CPU_STATE_HALT		= 2;
	uint32 state {};

	// MMU 状態
	bool mmu_enable {};
	// SRP
	bool SetSRP(uint32, uint32);
	uint64 GetSRP()  const { return reg.srp.q; }
	uint32 GetSRPh() const { return reg.srp.h; }
	uint32 GetSRPl() const { return reg.srp.l; }
	// CRP
	bool SetCRP(uint32, uint32);
	uint64 GetCRP()  const { return reg.crp.q; }
	uint32 GetCRPh() const { return reg.crp.h; }
	uint32 GetCRPl() const { return reg.crp.l; }
	// TT
	void SetTT(int idx, uint32);
	uint64 GetTT(int idx) const { return reg.tt[idx]; }
	// TC
	bool SetTC(uint32);
	uint32 GetTC() const { return reg.tc; }
	// MMUSR
	void SetMMUSR(uint16);
	uint16 GetMMUSR() const { return reg.mmusr; }
	// MMU work
	m68030TT mmu_tt[2] {};
	uint8 mmu_tt_quick[256] {};
	m68030TC mmu_tc {};
	m68030ATC atc {this};

 private:
	void mmu_enable_changed();

 public:
	int has_fpu {};			// 0:FPUなし、1:68881、2:68882(?)
	bool fpu_dirty {};
	struct fpemu fe {};

	// 現在の外部割り込みレベル
	int intr_level {};

	// 電源オンからの累積消費サイクル
	uint64 used_cycle {};

	int32 cycle_nocache {};		// ノーキャッシュケースの時の追加分

	// ブランチ履歴 (例外履歴を含む)
	BranchHistory_m680x0 brhist { BranchHistory::BrHist };
	// 例外履歴のみ
	BranchHistory_m680x0 exhist { BranchHistory::ExHistOnly };

	// バス
	m68kbus bus {};

	// A-Line, F-Line 命令エミュレーションのコールバックとユーザ引数。
	// ユーザ引数には関与しないので呼び出し側が自由に指定してよい。
	// コールバック側で処理完了したら true を返す。
	// デフォルトの処理(すなわちm68k例外)にする場合は false を返す。
	bool (*aline_callback)(m68kcpu *, void *) {};
	void *aline_arg {};
	bool (*fline_callback)(m68kcpu *, void *) {};
	void *fline_arg {};

	// RESET 命令のコールバックとユーザ引数
	void (*resetop_callback)(void *) {};
	void *resetop_arg {};

	// ダブルバスフォールトのコールバックとユーザ引数。
	void (*halt_callback)(void *) {};
	void *halt_arg {};

	// VM 側デバイス
	MPU680x0Device *dev {};
};

#define RegD(n)		(cpu->reg.da[(n)])
#define RegA(n)		(cpu->reg.da[(n) + 8])
#define RegR(n)		(cpu->reg.da[(n)])
#define RegPC			(cpu->reg.pc)
#define RegPPC			(cpu->ppc)
#define RegIR			(cpu->ir)
#define RegIR2			(cpu->ir2)
#define RegSR			(cpu->reg.sr_h() | cpu->reg.ccr.Get())
#define RegVBR			(cpu->reg.vbr)
#define RegUSP			(cpu->reg.usp)
#define RegISP			(cpu->reg.isp)
#define RegMSP			(cpu->reg.msp)
#define RegSFC			(cpu->reg.sfc)
#define RegDFC			(cpu->reg.dfc)
#define RegCACR		(cpu->reg.cacr)
#define RegCAAR		(cpu->reg.caar)

#define RegFP(n)	(cpu->reg.fpframe.fpf_regs[n])
#define RegFPCR		(cpu->reg.fpframe.fpf_fpcr)
#define RegFPSR		(cpu->reg.fpframe.fpf_fpsr)
#define RegFPIAR	(cpu->reg.fpframe.fpf_fpiar)

#define RegIRX			((RegIR >> 9) & 7)
#define RegIRY			(RegIR & 7)
#define RegDX			RegD(RegIRX)
#define RegAX			RegA(RegIRX)
#define RegDY			RegD(RegIRY)
#define RegAY			RegA(RegIRY)

// 外部公開ルーチン
extern m68kcpu *m68030_init(MPU680x0Device *dev_);
extern int m68030_exec(m68kcpu *cpu);
extern void m68030_power_on(m68kcpu *cpu);
extern bool m68030_interrupt(m68kcpu *cpu, int level);
extern bool m68030_check_interrupt(m68kcpu *cpu);
extern int  m68030_exception_reset(m68kcpu *cpu);
extern int  m68030_exception_interrupt(m68kcpu *cpu);

// 外部で定義されるコールバック
extern void m68030_reset_inst(m68kcpu *cpu);
extern int  m68030_interrupt_acknowledge(int level);
