//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m68030core.h"
#include "m68030bus.h"
#include "m68030ccr.h"
#include "m68030acc.h"

// インライン展開するにはでかいし使用頻度も少ない人たちはこっち。
// インライン展開できそうな人たちは m68030acc.h のほうへ。


// ビットフィールド命令

// コンストラクタ
// 呼び出し側でここまでに RegIR2 へのフェッチはしておくこと。
acc_bf::acc_bf(m68kcpu *cpu_)
{
	cpu = cpu_;

	//  F  E  D  C  B  A  9  8  7  6  5  4  3  2  1  0
	//  0  0  0  0 DO <-- offset --> DW <-- width --->

	offset = (RegIR2 >> 6) & 0x1f;
	if ((RegIR2 & 0x0800)) {
		offset = (int32)RegD(offset & 7);
		local_offset = offset & 0x1f;
	} else {
		local_offset = offset;
	}

	width = RegIR2;
	if ((RegIR2 & 0x0020)) {
		width = RegD(width & 7);
	}
	// モジュロ 32 を取り、0 なら 32 を表す
	width = ((width - 1) & 31) + 1;
	// mask は32ビットの上位から width ビット 1 が並んだもの
	mask = 0xffffffff << (32 - width);
}

// データレジスタから値を読み込む。n はデータレジスタの番号。
void
acc_bf::LoadReg(uint n)
{
	data = RegD(n);
	data = ROL32(data, local_offset);
	data &= mask;
}

// メモリから値を読み込む。
void
acc_bf::LoadMem(uint32 addr)
{
	// EA の計算
	ea = addr + offset / 8;
	local_offset = offset % 8;
	if (local_offset < 0) {
		local_offset += 8;
		ea--;
	}

	// アクセスするバイト数
	bytes = (local_offset + width + 7) / 8;

	// 40ビットバッファに読み込む
	// m68030_read_NN() はミスアラインドでもアクセスできる
	switch (bytes) {
	 case 1:
		buf = ((uint64)m68030_read_8(cpu, ea)) << 32;
		break;
	 case 2:
		buf = ((uint64)m68030_read_16(cpu, ea)) << 24;
		break;
	 case 3:
		if ((ea & 1) == 0) {
			// 偶数アドレスから3バイト
			buf = ((uint64)m68030_read_16(cpu, ea)) << 24;
			buf |= ((uint64)m68030_read_8(cpu, ea + 2)) << 16;
		} else {
			// 奇数アドレスから3バイト
			buf = ((uint64)m68030_read_8(cpu, ea)) << 32;
			buf |= (uint64)m68030_read_16(cpu, ea + 1) << 16;
		}
		break;
	 case 4:
		buf = ((uint64)m68030_read_32(cpu, ea)) << 8;
		break;
	 case 5:
		if ((ea & 1) == 0) {
			// 偶数アドレスから5バイト
			buf = ((uint64)m68030_read_32(cpu, ea)) << 8;
			buf |= ((uint64)m68030_read_8(cpu, ea + 4));
		} else {
			// 奇数アドレスから5バイト
			buf = ((uint64)m68030_read_8(cpu, ea)) << 32;
			buf |= ((uint64)m68030_read_32(cpu, ea + 1));
		}
		break;
	 default:
		__unreachable();
	}

	// 32ビット左詰めにする
	data = (buf >> (8 - local_offset)) & mask;
}

// データレジスタに値を書き出す。
void
acc_bf::StoreReg(uint n)
{
	uint32 m = ROR32(mask, local_offset);
	uint32 d = ROR32(data, local_offset);

	RegD(n) = (RegD(n) & ~m) | (d & m);
}

// メモリに値を書き出す
void
acc_bf::StoreMem()
{
	// 32ビット左詰めデータを 40ビットバッファにマージ
	uint64 m = (uint64)mask << (8 - local_offset);
	uint64 d = (uint64)data << (8 - local_offset);
	buf = (buf & ~m) | (d & m);

	// 40ビットバッファを書き戻す
	// m68030_write_NN() はミスアラインドでもアクセスできる
	switch (bytes) {
	 case 1:
		m68030_write_8(cpu, ea, buf >> 32);
		break;
	 case 2:
		m68030_write_16(cpu, ea, buf >> 24);
		break;
	 case 3:
		if ((ea & 1) == 0) {
			// 偶数アドレスから3バイト
			m68030_write_16(cpu, ea, buf >> 24);
			m68030_write_8(cpu, ea + 2, buf >> 16);
		} else {
			// 奇数アドレスから3バイト
			m68030_write_8(cpu, ea, buf >> 32);
			m68030_write_16(cpu, ea + 1, buf >> 16);
		}
		break;
	 case 4:
		m68030_write_32(cpu, ea, buf >> 8);
		break;
	 case 5:
		if ((ea & 1) == 0) {
			// 偶数アドレスから5バイト
			m68030_write_32(cpu, ea, buf >> 8);
			m68030_write_8(cpu, ea + 4, buf);
		} else {
			// 奇数アドレスから5バイト
			m68030_write_8(cpu, ea, buf >> 32);
			m68030_write_32(cpu, ea + 1, buf);
		}
		break;
	 default:
		__unreachable();
	}
}

void
acc_chk2cmp2_32(m68kcpu *cpu, uint32 compare, uint32 lower, uint32 upper)
{
	// 実機の実行結果をじっと睨んだ結果、このロジックで
	// cmp2 は実行していると考えられる。
	// オペランド順はフラグに影響する。
	// なぜ 68k のデザイナが順序を入れ替えたのかは不明。
	// なんかの都合があったんだろうな...

	// この cmp は通常の if で代替できるけど、
	// あえて MPU 内部の動作をエミュレート。
	acc_cmp_32(cpu, lower, upper);

	if (RegCondCS) {
		acc_cmp_32(cpu, compare, upper);
		if (RegCondEQ) {
			// break
		} else if (RegCondCS) {
			acc_cmp_32(cpu, lower, compare);
		} else {
			acc_cmp_32(cpu, compare, lower);
		}
	} else {
		acc_cmp_32(cpu, lower, compare);
		if (RegCondEQ) {
			// break
		} else if (RegCondCS) {
			// break
		} else {
			acc_cmp_32(cpu, compare, upper);
		}
	}
}

// dst(u32) / src(u16) を計算して、NZVC フラグをセットする。
// src の上位ビットはクリアしておくこと。
// true なら rem(16) と quo(16) が有効で、
// rem, quo の上位ビットはクリアされている。
// false なら結果を書き戻さずに終了する。
bool
acc_divu_32_16(m68kcpu *cpu, uint32 src, uint32 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}
	*quo = dst / src;
	*rem = dst % src;
	if (*quo > 0xffff) {
		// オーバーフローの場合、
		// X は変化しない
		// N,Z は未定義
		// V はセット
		// C は常にクリア
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}
	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	*rem &= 0xffff;
	acc_move_16(cpu, *quo);
	return true;
}

// dst(u32) / src(u32) を計算して、NZVC フラグをセットする。
// true なら rem(32), quo(32) が有効、
// false なら結果を書き戻さずに終了する。
bool
acc_divu_32_32(m68kcpu *cpu, uint32 src, uint32 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}
	*quo = dst / src;
	*rem = dst % src;
	// このケースでオーバーフローは起きないはず

	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	acc_move_32(cpu, *quo);
	return true;
}

// dst(u64) / src(u32) を計算して、NZVC フラグをセットする。
// true なら rem(32), quo(32) が有効、
// false なら結果を書き戻さずに終了する。
bool
acc_divu_64_32(m68kcpu *cpu, uint32 src, uint64 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}
	uint64 q64 = dst / src;
	*rem = dst % src;
	if (q64 > 0xffffffffUL) {
		// オーバーフローの場合、
		// X は変化しない
		// N,Z は未定義
		// V はセット
		// C は常にクリア
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}

	*quo = q64;

	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	acc_move_32(cpu, *quo);
	return true;
}

// dst(s32) / src(s16) を計算して、NZVC フラグをセットする。
// src の上位ビットはクリアしておくこと。
// true なら rem(16) と quo(16) が有効で、
// rem, quo の上位ビットはクリアされている。
// false なら結果を書き戻さずに終了する。
bool
acc_divs_32_16(m68kcpu *cpu, uint32 src, uint32 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}

	// ホストでオーバーフローが起きるのはこのケースのみ
	if (dst == 0x80000000 && src == 0xffff) {
		// オーバーフローの場合、
		// X は変化しない
		// N,Z は未定義
		// V はセット
		// C は常にクリア
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}

	int32 squo = (int32)dst / (int32)(int16)src;
	int32 srem = (int32)dst % (int32)(int16)src;
	if ((int16)squo != squo) {
		// オーバーフローの場合
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}
	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	*quo = (uint32)squo & 0xffff;
	*rem = (uint32)srem & 0xffff;
	acc_move_16(cpu, *quo);
	return true;
}

// dst(s32) / src(s32) を計算して、NZVC フラグをセットする。
// true なら rem(32), quo(32) が有効、
// false なら結果を書き戻さずに終了する。
bool
acc_divs_32_32(m68kcpu *cpu, uint32 src, uint32 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}

	// ホストでオーバーフローが起きるのはこのケースのみ
	if (dst == INT32_MIN && src == -1) {
		// オーバーフローの場合、
		// X は変化しない
		// N,Z は未定義
		// V はセット
		// C は常にクリア
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}

	*quo = (int32)dst / (int32)src;
	*rem = (int32)dst % (int32)src;
	// ここでオーバーフローは起きない

	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	acc_move_32(cpu, *quo);
	return true;
}

// dst(s64) / src(s32) を計算して、NZVC フラグをセットする。
// true なら rem(32), quo(32) が有効、
// false なら結果を書き戻さずに終了する。
bool
acc_divs_64_32(m68kcpu *cpu, uint32 src, uint64 dst, uint32 *quo, uint32 *rem)
{
	if (src == 0) {
		// ゼロ除算の場合、
		// X は変化しない
		// C は常にクリア
		// N,Z,V は未定義
		CCR.c = 0;
		m68030_exception(cpu, M68K_EXCEP_ZERODIV);
		return false;
	}

	// ホストでオーバーフローが起きるのはこのケースのみ
	if (dst == INT64_MIN && src == -1) {
		// オーバーフローの場合、
		// X は変化しない
		// N,Z は未定義
		// V はセット
		// C は常にクリア
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}

	int64 sq64 = (int64)dst / (int32)src;
	*rem = (int64)dst % (int32)src;
	if ((int32)sq64 != sq64) {
		// オーバーフロー
		CCR.v = m68030ccr::V_BIT;
		CCR.c = 0;
		return false;
	}

	// 正常の場合、
	// X は変化しない
	// V, C はクリア
	// N, Z は quo の結果。
	*quo = (uint32)sq64;
	acc_move_32(cpu, *quo);
	return true;
}

// src_10 + dst_10 + X を計算し、XNZVC フラグをセットして結果を返す。
// src, dst の上位ビットはクリアしておくこと。
// 戻り値の上位ビットはクリアされている。
uint32
acc_abcd_8(m68kcpu *cpu, uint32 src, uint32 dst)
{
	uint32 res;
	uint32 tmp;
	uint32 adj;
	uint32 low;
	uint32 c;

	adj = 0;
	c = 0;

	// X フラグは LSB なので直接足してよい
	tmp = src + dst + (CCR.x & m68030ccr::X_BIT);

	low = (src & 0x0f) + (dst & 0x0f) + (CCR.x & m68030ccr::X_BIT);
	if (low > 0x09) {
		adj = 0x06;
	}

	if (tmp > 0x99) {
		adj += 0x60;
		c = m68030ccr::C_BIT;
	}

	res = tmp + adj;
	c |= res >> 8;
	CCR.x = CCR.c = c;

	res &= 0xff;
	CCR.z |= res;

	// V フラグは不定だが 68030 では常にクリア
	CCR.v = 0;
	// N フラグは不定だが 68030 実機合わせ
	CCR.n = res << 24;

	return res;
}

// src_10 - dst_10 - X を計算し、XNZVC フラグをセットして結果を返す。
// src, dst の上位ビットはクリアしておくこと。
// 戻り値の上位ビットはクリアされている。
uint32
acc_sbcd_8(m68kcpu *cpu, uint32 src, uint32 dst)
{
	uint32 res;
	int32 tmp;
	int32 adj;
	int32 low;
	uint32 c;

	adj = 0;
	c = 0;

	// X フラグは LSB なので直接引いてよい
	tmp = dst - src - (CCR.x & m68030ccr::X_BIT);

	low = (dst & 0x0f) - (src & 0x0f) - (CCR.x & m68030ccr::X_BIT);
	if (low < 0) {
		adj = 0x06;
	}

	if (tmp < 0) {
		adj += 0x60;
		c = m68030ccr::C_BIT;
	}

	res = tmp - adj;
	c |= res >> 8;
	CCR.x = CCR.c = c;

	res &= 0xff;
	CCR.z |= res;

	// V フラグは不定だが 68030 では常にクリア
	CCR.v = 0;
	// N フラグは不定だが 68030 実機合わせ
	CCR.n = res << 24;

	return res;
}
