//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m68030core.h"
#include "vectortable.h"

#define OP_PROTO(name)	static inline void __CONCAT(op_,name)(m68kcpu *cpu)
#define OP_DEF(name)	static inline void __CONCAT(op_,name)(m68kcpu *cpu)
#define OP_FUNC(name)	__CONCAT(op_,name)(cpu)

#include "m68030bus.h"
#include "m68030bus.cpp"
#include "m68030excep.h"
#include "m68030ea.h"
#include "m68030acc.h"
#include "m68030ops.h"
#include "m68030mmu.h"
#include "m68030fpu.h"
#include "m68030ops.cpp"

// 1命令実行する。
// 戻り値は消費したサイクル数。
int
m68030_exec(m68kcpu *cpu)
{
	uint64 cycle_start = cpu->used_cycle;

	try {
		// 1命令版では switch-case を使うしかない。
		RegPPC = RegPC;
		cpu->flag_pi = 0;
		cpu->flag_pd = 0;
		RegIR = m68030_fetch_16(cpu);

		// uint16 を 6ビット右シフトしたら 1024 以上になるはずは
		// ないのだが、デフォルトでは gcc/clang どちらも律儀に
		// 1024 以上かどうか比較して分岐しようとする。
		// gcc は __assume() を入れることでこの比較・分岐を出力
		// しないようにすることが出来た(がパフォーマンスの差は
		// 見えない程度でしかなかった)。
		// clang ではどうやったら効くか不明。
		uint16 ir = RegIR >> 6;
		__assume(ir < 1024);
		switch (ir) {
#include "m68030switch.h"
		}

		cpu->used_cycle += cpu->cycle_nocache;
		cpu->cycle_nocache = 0;
	} catch (int vector) {
		// ここで(C++の)例外スローをキャッチするのは
		// - メモリ空間をアクセスした先でのバスエラー例外
		// - EA パース中の不当命令例外
		// のみ。それ以外の命令途中例外はその場で処理できる。
		if (__predict_false(gMPU->loglevel >= 2)) {
			const char *name = gVectorTable->GetExceptionName(vector);
			if (__predict_true(name)) {
				gMPU->putlogn("%s", name);
			} else {
				gMPU->putlogn("vector 0x%x", vector);
			}
		}
		m68030_exception(cpu, vector);
	}

	// 外部割り込みの起動は VM 側で処理している。

	return cpu->used_cycle - cycle_start;
}
