//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m68030core.h"
#include "m68030bus.h"
#include "m68030ea.h"
#include "m68030fpu.h"

//
// EA の拡張ワード
//

uint32
internal_ea_ix(m68kcpu *cpu, uint32 anpc)
{
	uint16 ext = m68030_fetch_16(cpu);
	int rn = (ext >> 12) & 15;
	bool idx_long = (ext & 0x0800);
	int scale = (ext >> 9) & 3;

	if ((ext & 0x0100) == 0) {
		// Brief extension
		//
		//   F   E   D   C   B   A   9   8   7   6   5   4   3   2   1   0
		// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
		// | DA|     n     | WL| SCALE | 0 |             disp              |
		// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
		//
		// WL: インデックスレジスタ(Rn)のサイズ 1=.L 0=.W
		// SCALE: スケールファクタ %00=*1 %01=*2 %10=*4 %11=*8

		CYCLE(4);
		int32 disp = (int32)(int8)(ext & 0xff);
		uint32 idx;

		if (idx_long) {
			idx = RegR(rn);
		} else {
			idx = (int32)(int16)(RegR(rn) & 0xffff);
		}
		idx <<= scale;
		return anpc + (int32)idx + disp;
	} else {
		// Full extension
		//
		//   F   E   D   C   B   A   9   8   7   6   5   4   3   2   1   0
		// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
		// | DA|     n     | WL| SCALE | 1 | BS| IS|  BDSZ | 0 |   I/IS    |
		// +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
		//
		// BS=1: ベースレジスタ・サプレス
		// IS=1: インデックスオペランド・サプレス
		// BDSZ: ベースディスプレースメントのサイズ %01=null %10=.W %11=.L
		// I/IS: モード

		// クロックは概略値なので完全に正確ではないがまず基本は以下の2通り。
		//  メモリ間接なし			(6,  6)
		//  メモリ間接あり			(10, 10)
		// これを元に必要に応じて以下をそれぞれ加算する。
		//  bd16					(+2, +3)
		//  bd32 (メモリ間接なし)	(+4, +6)
		//  bd32 (メモリ間接あり)	(+6, +6)
		//  outer (W/L 不問)		(+2, +3)
		// ただしここでは実行順の都合で bd32 はいずれも (+4, +6) とした。

		bool bs = (ext & 0x0080);
		bool is = (ext & 0x0040);
		int bdsize = (ext >> 4) & 3;
		int iis = (ext & 7);

		// ベースレジスタ
		if (bs) {
			anpc = 0;
		}

		// ベース・ディスプレースメント
		uint32 base = 0;
		switch (bdsize) {
		 case 0:	// 予約
			// XXX 未定義命令列になるんだろうか?
			PANIC("BDSIZE==0 not impl");
		 case 1:	// ベースはヌル
			break;
		 case 2:	// ベース.W
			base = (int32)(int16)m68030_fetch_16(cpu);
			CYCLE2(2, 3);
			break;
		 case 3:	// ベース.L
			base = m68030_fetch_32(cpu);
			CYCLE2(4, 6);
			break;
		 default:
			__unreachable();
		}

		uint32 outer;
		if (is == 0) {
			// インデックス
			uint32 idx = 0;
			if (idx_long) {
				idx = RegR(rn);
			} else {
				idx = (int32)(int16)(RegR(rn) & 0xffff);
			}
			idx <<= scale;

			switch (iis) {
			 case 0:	// no memory indirection
				CYCLE(6);
				return anpc + base + idx;

			 case 1:	// indirect pre-index with null outer
				CYCLE(10);
				return m68030_read_32(cpu, anpc + base + idx);
			 case 2:	// indirect pre-index with word outer
				CYCLE2(10+2, 10+3);
				outer = (int32)(int16)m68030_fetch_16(cpu);
				return m68030_read_32(cpu, anpc + base + idx) + outer;
			 case 3:	// indirect pre-index with long outer
				CYCLE2(10+2, 10+3);
				outer = m68030_fetch_32(cpu);
				return m68030_read_32(cpu, anpc + base + idx) + outer;

			 case 4:
				PANIC("iis==4 reserved");

			 case 5:	// indirect post-index with null outer
				CYCLE(10);
				return m68030_read_32(cpu, anpc + base) + idx;
			 case 6:	// indirect post-index with word outer
				CYCLE2(10+2, 10+3);
				outer = (int32)(int16)m68030_fetch_16(cpu);
				return m68030_read_32(cpu, anpc + base) + idx + outer;
			 case 7:	// indirect post-index with long outer
				CYCLE2(10+2, 10+3);
				outer = m68030_fetch_32(cpu);
				return m68030_read_32(cpu, anpc + base) + idx + outer;
			 default:
				__unreachable();
			}
		} else {
			switch (iis) {
			 case 0:	// no memory indirection
				CYCLE(6);
				return anpc + base;

			 case 1:	// memory indirect with null outer
				CYCLE(10);
				return m68030_read_32(cpu, anpc + base);
			 case 2:	// memory indirect with word outer
				CYCLE2(10+2, 10+3);
				outer = (int32)(int16)m68030_fetch_16(cpu);
				return m68030_read_32(cpu, anpc + base) + outer;
			 case 3:	// memory indirect with long outer
				CYCLE2(10+2, 10+3);
				outer = m68030_fetch_32(cpu);
				return m68030_read_32(cpu, anpc + base) + outer;
			 default:
				PANIC("is==1 && iis>3 reserved");
			}
		}
	}
}

//
// Calc EA
//

uint32
cea_ctrl(m68kcpu *cpu)
{
	switch (eamode(RegIR)) {
	 case 2:	// (An)
		return cea_anin(cpu, eanum(RegIR));
	 case 5:	// d16(An)
		return cea_andi(cpu, eanum(RegIR));
	 case 6:	// (An,IX)
		return cea_anix(cpu, eanum(RegIR));
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			return cea_absw(cpu);
		 case 1:	// Abs.L
			return cea_absl(cpu);
		 case 2:	// d16(PC)
			return cea_pcdi(cpu);
		 case 3:	// (PC,IX)
			return cea_pcix(cpu);
		 default:
			break;
		}
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
cea_data_8(m68kcpu *cpu)
{
	switch (eamode(RegIR)) {
	 case 2:	// (An)
		return cea_anin(cpu, eanum(RegIR));
	 case 3:	// (An)+
		return cea_anpi_8(cpu, eanum(RegIR));
	 case 4:	// -(An)
		return cea_anpd_8(cpu, eanum(RegIR));
	 case 5:	// d16(An)
		return cea_andi(cpu, eanum(RegIR));
	 case 6:	// (An,IX)
		return cea_anix(cpu, eanum(RegIR));
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			return cea_absw(cpu);
		 case 1:	// Abs.L
			return cea_absl(cpu);
		 default:
			break;
		}
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
cea_data_16(m68kcpu *cpu)
{
	switch (eamode(RegIR)) {
	 case 2:	// (An)
		return cea_anin(cpu, eanum(RegIR));
	 case 3:	// (An)+
		return cea_anpi_16(cpu, eanum(RegIR));
	 case 4:	// -(An)
		return cea_anpd_16(cpu, eanum(RegIR));
	 case 5:	// d16(An)
		return cea_andi(cpu, eanum(RegIR));
	 case 6:	// (An,IX)
		return cea_anix(cpu, eanum(RegIR));
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			return cea_absw(cpu);
		 case 1:	// Abs.L
			return cea_absl(cpu);
		 default:
			break;
		}
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
cea_data_32(m68kcpu *cpu)
{
	switch (eamode(RegIR)) {
	 case 2:	// (An)
		return cea_anin(cpu, eanum(RegIR));
	 case 3:	// (An)+
		return cea_anpi_32(cpu, eanum(RegIR));
	 case 4:	// -(An)
		return cea_anpd_32(cpu, eanum(RegIR));
	 case 5:	// d16(An)
		return cea_andi(cpu, eanum(RegIR));
	 case 6:	// (An,IX)
		return cea_anix(cpu, eanum(RegIR));
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			return cea_absw(cpu);
		 case 1:	// Abs.L
			return cea_absl(cpu);
		 default:
			break;
		}
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

// MMU からの呼び出し用。
uint32
cea_copro(m68kcpu *cpu)
{
	switch (eamode(RegIR)) {
	 case 2:	// (An)
		CYCLE(4);
		return internal_ea_anin(cpu, eanum(RegIR));
	 case 5:	// d16(An)
		CYCLE(4);
		return internal_ea_andi(cpu, eanum(RegIR));
	 case 6:	// (An,IX)
		CYCLE(4);
		return internal_ea_anix(cpu, eanum(RegIR));
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			CYCLE(4);
			return internal_ea_absw(cpu);
		 case 1:	// Abs.L
			CYCLE(6);
			return internal_ea_absl(cpu);
		 default:
			break;
		}
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

//
// Fetch EA
//

uint32
fea_data_8(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
		return RegD(eanum(RegIR)) & 0xff;
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_8(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_8(cpu);
		return m68030_read_8(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_8(cpu);
		return m68030_read_8(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_8(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_8(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_8(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_8(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_8(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_8(cpu, ea);
		 case 4:	// #imm
			fea_imm_16(cpu);
			return m68030_fetch_16(cpu) & 0xff;
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
fea_data_16(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
		return RegD(eanum(RegIR)) & 0xffff;
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_16(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_16(cpu);
		return m68030_read_16(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_16(cpu);
		return m68030_read_16(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_16(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_16(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_16(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_16(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_16(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_16(cpu, ea);
		 case 4:	// #imm
			fea_imm_16(cpu);
			return m68030_fetch_16(cpu);
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
fea_data_32(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
		return RegD(eanum(RegIR));
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_32(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_32(cpu);
		return m68030_read_32(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_32(cpu);
		return m68030_read_32(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_32(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_32(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_32(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_32(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_32(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_32(cpu, ea);
		 case 4:	// #imm
			fea_imm_32(cpu);
			return m68030_fetch_32(cpu);
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
fea_all_8(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
		return RegD(eanum(RegIR)) & 0xff;
	 case 1:	// An
		// An.B は不当命令
		break;
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_8(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_8(cpu);
		return m68030_read_8(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_8(cpu);
		return m68030_read_8(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_8(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_8(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_8(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_8(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_8(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_8(cpu, ea);
		 case 4:	// #imm
			fea_imm_16(cpu);
			return m68030_fetch_16(cpu) & 0xff;
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
fea_all_16(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
	 case 1:	// An
		return RegR(RegIR & 15) & 0xffff;
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_16(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_16(cpu);
		return m68030_read_16(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_16(cpu);
		return m68030_read_16(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_16(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_16(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_16(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_16(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_16(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_16(cpu, ea);
		 case 4:	// #imm
			fea_imm_16(cpu);
			return m68030_fetch_16(cpu);
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}

uint32
fea_all_32(m68kcpu *cpu)
{
	uint32 ea;

	switch (eamode(RegIR)) {
	 case 0:	// Dn
	 case 1:	// An
		return RegR(RegIR & 15);
	 case 2:	// (An)
		ea = fea_anin(cpu);
		return m68030_read_32(cpu, ea);
	 case 3:	// (An)+
		ea = fea_anpi_32(cpu);
		return m68030_read_32(cpu, ea);
	 case 4:	// -(An)
		ea = fea_anpd_32(cpu);
		return m68030_read_32(cpu, ea);
	 case 5:	// d16(An)
		ea = fea_andi(cpu);
		return m68030_read_32(cpu, ea);
	 case 6:	// (An,IX)
		ea = fea_anix(cpu);
		return m68030_read_32(cpu, ea);
	 case 7:
		switch (eanum(RegIR)) {
		 case 0:	// Abs.W
			ea = fea_absw(cpu);
			return m68030_read_32(cpu, ea);
		 case 1:	// Abs.L
			ea = fea_absl(cpu);
			return m68030_read_32(cpu, ea);
		 case 2:	// d16(PC)
			ea = fea_pcdi(cpu);
			return m68030_read_32(cpu, ea);
		 case 3:	// (PC,IX)
			ea = fea_pcix(cpu);
			return m68030_read_32(cpu, ea);
		 case 4:	// #imm
			fea_imm_32(cpu);
			return m68030_fetch_32(cpu);
		 default:
			break;
		}
		break;
	 default:
		break;
	}
	// 不当命令
	throw M68K_EXCEP_ILLEGAL;
}
