//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// EA

#pragma once

// EA は大別して5種類。
//
// |d m  rxwp |	制御	CalcCtrl		サイズ不問
// |dam+-rxw  |	可変	CalcData		B/W/L
// |  m  rxwp |	MMU		CalcMMU			サイズ不問
// |dam+-rxwpi|	FPU		CalcFPU			B/W/L/S/D/X/P
// |d m+-rxwpi|	データ	FetchData		B/W/L
// |dam+-rxwpi|	All		FetchAll		B/W/L
//
// CalcCtrl はユーザーズマニュアルで「制御(Control)」と呼ばれている
// グループで、デスティネーション側のため EA の計算のみ。
// また (An)+/-(An) を含まないため、サイズに関係なく1種類だけでよい。
//
// CalcData はユーザーズマニュアルで「可変」と呼ばれているグループで、
// デスティネーション側のため EA の計算のみ。B/W/L
//
// CalcMMU は対 MMU 用に勝手に用意したもの。
//
// CalcFPU は対 FPU 用。ソースであっても CalcFPU を使用する。
// サイズは B/W/L/S/D/X/P。
// フェッチサイクルは別途存在するはずだが不詳。
//
// FetchData はユーザーズマニュアルで「データ(Data)」と呼ばれている
// グループで、ソース側のため EA 計算および値の取得。B/W/L
//
// FetchAll はユーザーズマニュアルのここには記載がないグループで、
// ソース側のため EA 計算および値の取得。B/W/L
//
// CalcCtrl、CalcData でアドレッシングモードに D や A が含まれている場合は
// その命令のデスティネーションに Dn、An が指定できるという意味であり、
// EA 処理であるここには関係ない。
// EA 計算より前に呼出側で条件分けして処理すること。
//
// CalcCtrl、CalcData と FetchData には (*) 印の例外があるので留意。
//
// (CalcCtrl)
// |  m  rxwp | CMP2/CHK2
// |  m  rxwp | JMP/JSR
// |  m  rxwp | LEA/PEA
// |d m  rxwp | BFEXT*/BFFFO/BFTST
// |d m  rxw  | (*) BFCHG/BFCLR/BFINS/BFSET <ea>{#o:#w}
//
// (CalcData)
// |  m+-rxw  | ADD/SUB/CMP Dx,<ea>
// |  m+-rxw  | CAS/TAS <ea>
// |  m+-rxw  | MOVES <ea>,Rn
// |  m+-rxw  | shift/rotate.W <ea>
// |  m+-rxw  | BCHG/BCLR/BSET src,<ea>
// |d m+-rxw  | ADDI/SUBI Dx,<ea>
// |d m+-rwx  | NBCD <ea>
// |d m+-rxw  | AND/OR/EOR Dx,<ea>
// |d m+-rxw  | CLR/NEG/NEGX/NOT/TST/Scc <ea>
// |d m+-rxw  | MOVE CCR/SR,<ea>
// |dam+-rxw  | ADDQ/SUBQ #qqq,<ea>
// |dam+-rxw  | (*) MOVE の dst 側。EA の分類的には CalcData だが、
//              下位6ビットの IR ではないので、cea_data_NN() ではなく
//              各々 cea_XXXX() のほうを個別に呼ぶこと。
//
// (CalcMMU)
// |  m  rxw  | MMU
//
// (CalcFPU)
// |d m+-rxw  | FScc.B <ea>
// |dam+-rxw  | FMOVE.f FPn,<ea>
// |dam+-rxw  | FMOVE.L ctl,<ea>
// |dam+-rxw  | FMOVEM.L ctl,<ea>
// |d m+-rxwpi| Fgen <ea>,FPn
// |d m+-rxwpi| FMOVE.f <ea>,FPn
// |dam+-rwxpi| FMOVE.L <ea>,ctl
// |dam+-rxwpi| FMOVEM.L <ea>,ctl
// 実際にどのアドレッシングモードが有効かどうかは極力個別の命令内で判断する?
// ここは実行サイクル数的な分類。
//
// (FetchData)
// |d m+-rxwpi| AND/OR/CHK <ea>,Dn
// |d m+-rxwpi| MOVE.B <ea>,dst
// |d m+-rxwpi| MOVE <ea>,CCR/SR
// |d m+-rxwpi| MUL*/DIV* <ea>,Dn
// |d m+-rxwp | (*) TST.B <ea>
// |d m+-rxwp | (*) CMPI #<data>,<ea>
// |  m+-rxwp | (*) BTST.B #<data>,<ea>
// |  m+-rxwpi| (*) BTST.B Dx,<ea>
//
// (FetchAll)
// |dam+-rxwpi| ADD/SUB/CMP <ea>,Dn
// |dam+-rxwpi| MOVEA/ADDA/SUBA/CMPA <ea>,An
// |dam+-rxwpi| TST.WL <ea>
// |dam+-rxwpi| MOVE.WL <ea>,dst
//
// (たぶん別対応)
// |  m -rxw  | MOVEM list,<ea>
// |  m -rxw  | FMOVEM.X reg,<ea>
// |  m -rxw  | FSAVE <ea>
// |  m+ rxwp | MOVEM <ea>,list
// |  m+ rxwp | FMOVEM.X <ea>,reg
// |  m+ rxwp | FRESTORE <ea>

extern uint32 internal_ea_ix(m68kcpu *cpu, uint32 anpc);

extern uint32 cea_ctrl(m68kcpu *);
extern uint32 cea_data_8(m68kcpu *);
extern uint32 cea_data_16(m68kcpu *);
extern uint32 cea_data_32(m68kcpu *);
extern uint32 cea_copro(m68kcpu *);
extern uint32 fea_data_8(m68kcpu *);
extern uint32 fea_data_16(m68kcpu *);
extern uint32 fea_data_32(m68kcpu *);
extern uint32 fea_all_8(m68kcpu *);
extern uint32 fea_all_16(m68kcpu *);
extern uint32 fea_all_32(m68kcpu *);

// mmmrrr のうち mmm の部分を取り出す
#define eamode(x)	(((x) >> 3) & 7)

// mmmrrr のうち rrr の部分を取り出す
#define eanum(x)	((x) & 7)

//
// EA を計算だけするマクロ
// MOVE の DST 側からも呼ばれる
//

// (An)
static inline uint32 internal_ea_anin(m68kcpu *cpu, int n)
{
	return RegA(n);
}

// (An)+ .B
static inline uint32 internal_ea_anpi_8(m68kcpu *cpu, int n)
{
	m68030_save_reg_pi(cpu, n);
	if (n == 7) {
		RegA(n) += 2;
		return RegA(n) - 2;
	} else {
		RegA(n) += 1;
		return RegA(n) - 1;
	}
}

// (An)+ .W
static inline uint32 internal_ea_anpi_16(m68kcpu *cpu, int n)
{
	m68030_save_reg_pi(cpu, n);
	RegA(n) += 2;
	return RegA(n) - 2;
}

// (An)+ .L
static inline uint32 internal_ea_anpi_32(m68kcpu *cpu, int n)
{
	m68030_save_reg_pi(cpu, n);
	RegA(n) += 4;
	return RegA(n) - 4;
}

// -(An) .B
static inline uint32 internal_ea_anpd_8(m68kcpu *cpu, int n)
{
	m68030_save_reg_pd(cpu, n);
	if (n == 7) {
		RegA(n) -= 2;
	} else {
		RegA(n) -= 1;
	}
	return RegA(n);
}

// -(An) .W
static inline uint32 internal_ea_anpd_16(m68kcpu *cpu, int n)
{
	m68030_save_reg_pd(cpu, n);
	RegA(n) -= 2;
	return RegA(n);
}

// -(An) .L
static inline uint32 internal_ea_anpd_32(m68kcpu *cpu, int n)
{
	m68030_save_reg_pd(cpu, n);
	RegA(n) -= 4;
	return RegA(n);
}

// d16(An)
static inline uint32 internal_ea_andi(m68kcpu *cpu, int n)
{
	int32 disp = (int32)(int16)m68030_fetch_16(cpu);
	return RegA(n) + disp;
}

// (An,IX)
static inline uint32 internal_ea_anix(m68kcpu *cpu, int n)
{
	return internal_ea_ix(cpu, RegA(n));
}

// Abs.W
static inline uint32 internal_ea_absw(m68kcpu *cpu)
{
	return (int32)(int16)m68030_fetch_16(cpu);
}

// Abs.L
static inline uint32 internal_ea_absl(m68kcpu *cpu)
{
	return m68030_fetch_32(cpu);
}

// d16(PC)
static inline uint32 internal_ea_pcdi(m68kcpu *cpu)
{
	int32 pc = RegPC;
	int32 disp = (int32)(int16)m68030_fetch_16(cpu);
	return pc + disp;
}

// (PC,IX)
static inline uint32 internal_ea_pcix(m68kcpu *cpu)
{
	return internal_ea_ix(cpu, RegPC);
}

//
// Calc EA
//

static inline uint32
cea_anin(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anin(cpu, n);
}

static inline uint32
cea_anpi_8(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpi_8(cpu, n);
}

static inline uint32
cea_anpi_16(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpi_16(cpu, n);
}

static inline uint32
cea_anpi_32(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpi_32(cpu, n);
}

static inline uint32
cea_anpd_8(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpd_8(cpu, n);
}

static inline uint32
cea_anpd_16(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpd_16(cpu, n);
}

static inline uint32
cea_anpd_32(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_anpd_32(cpu, n);
}

static inline uint32
cea_andi(m68kcpu *cpu, int n)
{
	CYCLE(2);
	return internal_ea_andi(cpu, n);
}

static inline uint32
cea_anix(m68kcpu *cpu, int n)
{
	CYCLE(10);	// XXX 4,4-18,20
	return internal_ea_anix(cpu, n);
}

static inline uint32
cea_absw(m68kcpu *cpu)
{
	CYCLE(2);
	return internal_ea_absw(cpu);
}

static inline uint32
cea_absl(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_absl(cpu);
}

static inline uint32
cea_pcdi(m68kcpu *cpu)
{
	CYCLE(2);
	return internal_ea_pcdi(cpu);
}

static inline uint32
cea_pcix(m68kcpu *cpu)
{
	CYCLE(10);	// XXX 4,4-18,20
	return internal_ea_pcix(cpu);
}

//
// Fetch EA
//

static inline uint32
fea_anin(m68kcpu *cpu)
{
	CYCLE(3);
	return internal_ea_anin(cpu, eanum(RegIR));
}

static inline uint32
fea_anpi_8(m68kcpu *cpu)
{
	CYCLE(3);
	return internal_ea_anpi_8(cpu, eanum(RegIR));
}

static inline uint32
fea_anpi_16(m68kcpu *cpu)
{
	CYCLE(3);
	return internal_ea_anpi_16(cpu, eanum(RegIR));
}

static inline uint32
fea_anpi_32(m68kcpu *cpu)
{
	CYCLE(3);
	return internal_ea_anpi_32(cpu, eanum(RegIR));
}

static inline uint32
fea_anpd_8(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_anpd_8(cpu, eanum(RegIR));
}

static inline uint32
fea_anpd_16(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_anpd_16(cpu, eanum(RegIR));
}

static inline uint32
fea_anpd_32(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_anpd_32(cpu, eanum(RegIR));
}

static inline uint32
fea_andi(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_andi(cpu, eanum(RegIR));
}

static inline uint32
fea_anix(m68kcpu *cpu)
{
	CYCLE2(12, 14);	// XXX 6,6-18,21
	return internal_ea_anix(cpu, eanum(RegIR));
}

static inline uint32
fea_absw(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_absw(cpu);
}

static inline uint32
fea_absl(m68kcpu *cpu)
{
	CYCLE2(4, 5);
	return internal_ea_absl(cpu);
}

static inline uint32
fea_pcdi(m68kcpu *cpu)
{
	CYCLE(4);
	return internal_ea_pcdi(cpu);
}

static inline uint32
fea_pcix(m68kcpu *cpu)
{
	CYCLE2(12, 14);	// XXX 6,6-18,21
	return internal_ea_pcix(cpu);
}

static inline void
fea_imm_16(m68kcpu *cpu)
{
	CYCLE(2);
}

static inline void
fea_imm_32(m68kcpu *cpu)
{
	CYCLE(4);
}
