//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

// ATC_SINGLE は過去との比較のために残してある、双方向リストによる
// FC 混合 22 エントリ実装。
// 定義しない場合(デフォルト)は、FC 別の世代付き配列みたいなもの。
//#define ATC_SINGLE 1

// SRP, CRP
class m68030RP
{
	//        6                   5                   4               3
	//  3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2
	// +-+-----------------------------+---------------------------+---+
	// |U|           LIMIT             |          unused           |DT |
	// +-+-----------------------------+---------------------------+---+
	//
	//    3                   2                   1                   0
	//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
	// +-------------------------------------------------------+-------+
	// |           TABLE ADDRESS (PA31-PA4)                    |unused |
	// +-------------------------------------------------------+-------+
 public:
	static const uint32 DT_MASK	= 0x00000003;
	static const uint32 H_MASK	= 0xffff0003;
	static const uint32 L_MASK	= 0xfffffff0;
};

// TT レジスタの各フィールドを束ねた構造体のようなもの。
// m68kcpu::SetTT(), GetTT() で読み書きする。
class m68030TT
{
	//    3                   2                   1                   0
	//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
	// +---------------+---------------+-+-+-+-+-+-+-+-+-+-----+-+-----+
	// |LogicalAddrBase|LogicalAddrMask|E|       |C|R|M| | FCB | | FCM |
	// +---------------+---------------+-+-+-+-+-+-+-+-+-+-----+-+-----+

 public:
	static const uint32 MASK		= 0xffff8777;

	static const uint32 LBASE_MASK	= 0xff000000;
	static const uint32 LMASK_MASK	= 0x00ff0000;
	static const uint32 E			= 0x00008000;
	static const uint32 CI			= 0x00000400;
	static const uint32 RW			= 0x00000200;
	static const uint32 RWM			= 0x00000100;
	static const uint32 FCBASE_MASK	= 0x00000070;
	static const uint32 FCMASK_MASK	= 0x00000007;

	uint32 e {};			// 00008000 イネーブル
	uint32 ci {};			// 00000400 キャッシュ禁止
	uint64 base {};			// ベース
	uint64 mask {};			// マスク

	// アクセスがこの TT と一致するか調べる。
	bool Match(uint64 ssw_laddr);
};

// TC レジスタの各フィールドを束ねた構造体のようなもの。
// m68kcpu::SetTC(), GetTC() で読み書きする。
class m68030TC
{
 public:
	static const uint32 MASK	= 0x83ffffff;
	static const uint32 TC_E	= 0x80000000;
	static const uint32 TC_SRE	= 0x02000000;
	static const uint32 TC_FCL	= 0x01000000;

	// レジスタの各フィールド
	uint32 e {};		// TC_E   の位置
	uint32 sre {};		// TC_SRE の位置
	uint32 fcl {};		// TC_FCL の位置
	uint ps {};
	uint is {};
	union {
		uint tix[4] {};
		struct {
			uint tia;
			uint tib;
			uint tic;
			uint tid;
		} __packed;
	};

	int shift[4] {};		// TIA-TID の各部分のシフト量
	uint32 mask[4] {};		// TIA-TID の各部分のマスク
	uint32 lmask {};		// TIA-TID 全体のマスク
	uint32 pgmask {};		// PS 部分のマスク

 public:
	void Set(uint32);
	bool Check() const;
	void MakeMask();
};

class m68030MMUSR
{
 public:
	static const uint16 MASK = 0xee43;
	static const uint16 B    = 0x8000;	// バスエラー
	static const uint16 L    = 0x4000;	// リミット違反
	static const uint16 S    = 0x2000;	// スーパバイザ専用
	static const uint16 W    = 0x0800;	// 書き込み保護
	static const uint16 I    = 0x0400;	// 無効
	static const uint16 M    = 0x0200;	// 修正
	static const uint16 T    = 0x0040;	// 透過アクセス
	static const uint16 N    = 0x0007;	// レベル数
};

// ATC の1エントリ
struct m68030ATCLine
{
	// 1エントリはタグとデータの計64ビット。
	//
	// タグ
	//  LLLLLLLL_LLLLLLLL_LLLLLLLL_0000000I : L は論理アドレス
	//                                    +-- invalid (%1で無効)
	//
	//   FC はテーブル選択の時点で確定しているのでタグには不要。
	//
	// データ
	//  PPPPPPPP_PPPPPPPP_PPPPPPPP_0000BCPM : P は物理アドレス
	//                                 |||+-- Modified
	//                                 ||+--- Write Protect
	//                                 |+---- Cache Inhibit
	//                                 +----- Bus Error

	static const uint32 LADDR_MASK	= 0xffffff00;
	static const uint32 INVALID		= 0x00000001;
	static const uint32 PADDR_MASK	= 0xffffff00;
	static const uint32 BUSERROR	= 0x00000008;
	static const uint32 CINHIBIT	= 0x00000004;
	static const uint32 WPROTECT	= 0x00000002;
	static const uint32 MODIFIED    = 0x00000001;

	uint32 tag;
#if defined(ATC_SINGLE)
	uint32 fc;
#endif
	uint32 paddr;
	uint32 data;

	uint32 GetTag() const		{ return tag; }

	void Invalidate()			{ tag |= INVALID; }
	bool   IsInvalid() const	{ return tag & INVALID; }
	bool   IsValid() const		{ return !IsInvalid(); }

	uint32 GetLAddr() const		{ return tag & LADDR_MASK; }
	uint32 GetPAddr() const		{ return paddr; }
	bool   IsBusError() const	{ return data & BUSERROR; }
	bool   IsCInhibit() const	{ return data & CINHIBIT; }
	bool   IsWProtect() const	{ return data & WPROTECT; }
	bool   IsModified() const	{ return data & MODIFIED; }

#if !defined(ATC_SINGLE)
	// LRU 用カウンタ
	uint32 age;
#else
	m68030ATCLine *prev;
	m68030ATCLine *next;
#endif

	// デバッグ表示
	std::string fill_print(uint fc);
};

// ATC のテーブル
class m68030ATCTable final
{
 public:
	static const int LINES = 22;

	m68030ATCTable();
	~m68030ATCTable() { }

#if !defined(ATC_SINGLE)
	// 空きエントリを取得。
	m68030ATCLine *Lookup();
	// a をエントリの先頭に
	m68030ATCLine *MoveHead(m68030ATCLine *a);
#else
	// 空きエントリを取得。
	m68030ATCLine *lookup();
	// LRU の先頭、2番目に追加
	void insert_head(m68030ATCLine *);
	void insert_second(m68030ATCLine *);
	// LRU の先頭、2番目に移動
	void move_head(m68030ATCLine *);
	void move_second(m68030ATCLine *);
	// LRU の末尾に追加
	void insert_tail(m68030ATCLine *);
	// LRU から削除
	void remove(m68030ATCLine *);
#endif

	// フラッシュ
	// flush() はこのテーブルのすべてのエントリが対象。
	// flush(addr) はこのテーブルのアドレスが一致するエントリが対象。
	// flush_fc(fc) はこのテーブルの FC が fc であるエントリが対象。
	// flush(fc, addr) はこのテーブルの fc と addr が一致するエントリが対象。
	// いずれも ATC クラスから呼ばれるが、ATC_SINGLE によって、一致する FC が
	// 呼び出し前に決定されているかこちらで比較しないといけないかが変わる。
	void flush();
#if !defined(ATC_SINGLE)
	void flush(uint32 addr);
#else
	void flush_fc(uint32 fc);
	void flush(uint32 fc, uint32 addr);
#endif

	// ATC 実体
	m68030ATCLine line[LINES] {};

	// 先頭から何エントリ目でヒットしたかの統計。最後の +1 はキャッシュミス。
	uint64 hit[LINES + 1] {};

	// ATC_SINGLE ならリストの先頭。
	// !ATC_SINGLE なら注目エントリ。
	m68030ATCLine *head {};

#if !defined(ATC_SINGLE)
	// LRU 用の更新カウンタ
	uint32 counter = 0;
	// head のヒット回数。
	uint64 hit_head = 0;
#endif
};

// ATC (Address Translation Cache)
class m68030ATC final
{
	friend class MPU680x0Device;
 public:
	m68030ATC(m68kcpu *);
	~m68030ATC() { }

	// エントリをサーチ&更新
	const m68030ATCLine& fill_fetch();
	const m68030ATCLine& fill_read();
	const m68030ATCLine& fill_write();
	// デバッグ用にエントリを検索
	const m68030ATCLine *fill_peek(uint32 laddr, uint fc);

	// フラッシュ (pflush* 命令から呼ばれる)
	void flush();
	void flush(uint32 fc, uint32 mask);
	void flush(uint32 fc, uint32 mask, uint32 addr);

 private:
#if !defined(ATC_SINGLE)
	// FC ごとのテーブル。
	// 8個確保してあるが実際に使うのはこのうち4つだけ。
	m68030ATCTable tables[8] {};
#else
	// FC 混合のテーブル。
	m68030ATCTable tables[1] {};
#endif

	m68kcpu *cpu {};
};

extern bool m68030_mmu_translate_fetch(m68kcpu *cpu);
extern bool m68030_mmu_translate_read(m68kcpu *cpu);
extern bool m68030_mmu_translate_write(m68kcpu *cpu);
extern uint64 m68030_mmu_translate_peek(m68kcpu *cpu, uint32 laddr, uint fc,
	bool do_search);
extern void mmu_op_pflush(m68kcpu *cpu);
extern void mmu_op_pflush_ea(m68kcpu *cpu);
extern void mmu_op_pload(m68kcpu *cpu);
extern void mmu_op_ptest(m68kcpu *cpu);
