//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// Bt454/458
//

// 仕様書には $c100_0000 と書いてあるが、PROM も NetBSD カーネルも
// $c110_0000 にアクセスしている。実際にはアドレスバスの上位寄りは未接続の
// 箇所が多くほとんどの場所でミラーが見えているので問題はない。
//
// LUNA-I、LUNA-88K どちらでも (というかここはビデオボードなので本体関係なく)
// Bt454 ならバイトポートが連続配置 (BusIO_B 相当)、
// Bt458 ならバイトポートがロングワード配置 (たぶん BusIO_BFFF 相当、未確認)
// となっているようだ。このため、ここでは BusIO を使わず自力でデコードする。
//
// ソフトウェアから判別は以下の要領で行える。
// 先頭の1バイトは 1or4bpp ボード(連続配置)、8bpp ボード(ロングワード配置)の
// どちらでもアドレスレジスタなので、ここに $00 を書き込んでから読み返す。
// $0f が読み出せたら Bt454 なので 1or4bpp ボード、$00 が読み出せたら Bt458
// なので 8bpp ボードと分かる。Bt458 のデータバスは8ビットクリーンだが、
// Bt454 のデータバスは4ビットしかなく下位ニブルは常に $f になるため。
//
// LUNA-I の PROM v4.22 は 8bpp ボードに対応していない。

#include "bt454.h"
#include "config.h"
#include "mainapp.h"
#include "mpu.h"
#include "planevram.h"
#include "prom.h"
#include "renderer.h"
#include "uimessage.h"

// Bt454 の下位ニブル用マスク
#define LOW_NIBBLE	(0x0f)

// グローバル参照用
BT454Device *gBT454;

// コンストラクタ
BT454Device::BT454Device()
	: inherited("Bt45x")
{
	// func, size は Init() で決まる
	monitor.Regist(ID_MONITOR_BT454);
}

// デストラクタ
BT454Device::~BT454Device()
{
	gBT454 = NULL;
}

bool
BT454Device::Init()
{
	// 設定からプレーン数を取得
	const ConfigItem& item = gConfig->Find("luna-video-plane");
	if (item.TryInt(&nplane) == false) {
		item.Err();
		return false;
	}
	if (nplane != 1 && nplane != 4 && nplane != 8) {
		item.Err("Must be 1, 4, or 8");
		return false;
	}

	// それによってモデル分け。1bpp/4bpp なら Bt454、8bpp なら Bt458。
	if (nplane <= 4) {
		bt454 = true;
		palettes = 16;
		low_nibble = LOW_NIBBLE;
		readmask = 0x0f;

		monitor.func = ToMonitorCallback(&BT454Device::MonitorUpdate454);
		monitor.SetSize(63, 3);
	} else {
		bt454 = false;
		palettes = 256;
		low_nibble = 0;
		// 初期化されないそうなのでゴミを入れておく
		readmask = 0xcc;

		monitor.func = ToMonitorCallback(&BT454Device::MonitorUpdate458);
		monitor.SetSize(116, 18);

		// LUNA-I の PROM v4.22 は 8bpp ボードをサポートしておらず PROM の
		// 画面は真っ暗のままになる。その後 OS が再初期化すれば使えるようには
		// なるが、PROM が映らずびっくりするのでメッセージだけ表示して、
		// 実行は継続する。他の ROM バージョンは未調査。
		auto prom = dynamic_cast<PROMDevice *>(gPROM);
		if (gMainApp.IsLUNA1() && prom != NULL) {
			int ver = prom->GetROMVer();
			if (ver <= 422) {
				warnx("Warning: PROM v%d.%02d does not support 8bpp board.",
					ver / 100, ver % 100);
				// 実行は継続
			}
		}
	}

	color.resize(palettes * 3);
	hostcolor.resize(palettes);
	return true;
}

// リセット
void
BT454Device::ResetHard(bool poweron)
{
	idx = 0;

	// ReadMask、BlinkMask は初期化されない (初期値も不明)

	// XXX Bt45x のレジスタの初期値は不明なのでとりあえず全部黒にしておく
	memset(&color[0], 0, color.size());
	for (int i = 0; i < countof(color); i++) {
		MakeHostColor(i);
	}
	Notify();
}

/* static */ const char
BT454Device::rgbstr[3] = { 'R', 'G', 'B' };

// LUNA のビデオボードは Bt454 か Bt458 かでバイト配置が異なるようだ。

uint64
BT454Device::Read8(uint32 addr)
{
	if (bt454) {
		// バイトポートが連続、4バイトでミラー
		return Read(addr & 3);
	} else {
		// バイトポートがロングワード配置、隙間は何が見えるか不明、
		// たぶん16バイトでミラー。
		addr &= 15;
		if (addr % 4 == 0) {
			return Read(addr / 4);
		} else {
			return 0xff;
		}
	}
}

uint64
BT454Device::Write8(uint32 addr, uint32 data)
{
	if (bt454) {
		// バイトポートが連続、4バイトでミラー
		return Write(addr & 3, data);
	} else {
		// バイトポートがロングワード配置、隙間は何が見えるか不明、
		// たぶん16バイトでミラー。
		addr &= 15;
		if (addr % 4 == 0) {
			return Write(addr / 4, data);
		} else {
			return 0;
		}
	}
}

uint64
BT454Device::Peek8(uint32 addr)
{
	if (bt454) {
		// バイトポートが連続、4バイトでミラー
		return Peek(addr & 3);
	} else {
		// バイトポートがロングワード配置、隙間は何が見えるか不明、
		// たぶん16バイトでミラー。
		addr &= 15;
		if (addr % 4 == 0) {
			return Peek(addr / 4);
		} else {
			return 0xff;
		}
	}
}

// こっちがマッピング非依存のバイトポートアクセス
uint64
BT454Device::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:	// アドレスレジスタ
		data = ReadAddress();
		break;

	 case 1:	// カラーパレットレジスタ
		data = ReadPalette();
		break;

	 case 2:	// Bt454 ならリセット(?)、Bt458 なら各種制御レジスタ
		data = ReadControl();
		break;

	 case 3:
		putlog(0, "Read Overlay Register (NOT IMPLEMENTED)");
		data = 0xff;
		break;

	 default:
		__unreachable();
	}

	return data;
}

// こっちがマッピング非依存のバイトポートアクセス
uint64
BT454Device::Write(uint32 offset, uint32 data)
{
	// Bt454 なら書き込みデータのうち有効なのは上位ニブルだけ

	switch (offset) {
	 case 0:	// アドレスレジスタ
		WriteAddress(data);
		break;

	 case 1:	// カラーパレットレジスタ
		WritePalette(data);
		break;

	 case 2:	// Bt454 ならリセット(?)、Bt458 なら各種制御レジスタ
		WriteControl(data);
		break;

	 case 3:
		putlog(0, "Overlay Register <- $%02x (NOT IMPLEMENTED)", data);
		break;

	 default:
		__unreachable();
	}
	return 0;
}

// こっちがマッピング非依存のバイトポートアクセス
uint64
BT454Device::Peek(uint32 offset)
{
	uint8 data;

	switch (offset) {
	 case 0:	// アドレスレジスタ
		data = idx / 3;
		data &= readmask;
		break;

	 case 1:	// カラーパレットレジスタ
		// 今読み出そうとしたら読み出せるやつ
		data = color[idx];
		break;

	 case 2:	// Bt454 ならリセット(?)、Bt458 なら各種制御レジスタ
		data = PeekControl();
		break;

	 case 3:
		// 未実装
		data = 0xff;
		break;

	 default:
		__unreachable();
	}

	if (bt454) {
		data = (data << 4) | LOW_NIBBLE;
	}
	return data;
}

// Bt454
void
BT454Device::MonitorUpdate454(Monitor *, TextScreen& screen)
{
	screen.Clear();
	screen.Puts(0, 0, "Model: Bt454");

	// 01234567
	// $0  $1
	// RGB RGB

	for (int i = 0; i < 16; i++) {
		screen.Print(i * 4, 1, "$%x", i);
		screen.Print(i * 4, 2, "%03x", Palette454ToInt(i));
	}
}

// Bt458
void
BT454Device::MonitorUpdate458(Monitor *, TextScreen& screen)
{
	screen.Clear();
	screen.Puts(0, 0, "Model: Bt458");

	// 012345678901
	//       +$0    +$1
	// $00: RRGGBB RRGGBB

	for (int i = 0; i < 16; i++) {
		screen.Print(5 + i * 7, 1, "+$%x", i);
	}
	int c = 0;
	for (int y = 0; y < 16; y++) {
		screen.Print(0, y + 2, "$%02x:", y * 16);
		for (int x = 0; x < 16; x++) {
			screen.Print(5 + x * 7, y + 2, "%06x", Palette458ToInt(c));
			c++;
		}
	}
}

// idx 番目のレジスタ値からホストパレットを作成。
void
BT454Device::MakeHostColor(int idx_)
{
	assert(idx_ < color.size());
	int col = idx_ / 3;
	int ab  = idx_ % 3;

	if (nplane == 1) {
		// 1bpp (モノクロプレーン) なら R,G,B のうち G の値だけを使用
		if (ab == 1) {
			hostcolor[col].r = color[col * 3 + 1];
			hostcolor[col].g = color[col * 3 + 1];
			hostcolor[col].b = color[col * 3 + 1];
		}
	} else {
		// カラープレーンなら代入するだけ
		switch (ab) {
		 case 0:
			hostcolor[col].r = color[col * 3 + ab];
			break;
		 case 1:
			hostcolor[col].g = color[col * 3 + ab];
			break;
		 case 2:
			hostcolor[col].b = color[col * 3 + ab];
			break;
		 default:
			__unreachable();
		}
	}

	// 通知が必要なら別途行うこと。
}

// 通知
void
BT454Device::Notify()
{
	// パレット変更が起きたことをレンダラに通知
	gRenderer->Invalidate2();

	// UI にも通知
	UIMessage::Post(UIMessage::PALETTE);
}

// アドレスレジスタ読み込み。
uint32
BT454Device::ReadAddress()
{
	uint32 data;

	// 今の(今からアクセスする)パレット番号を返す。
	data = idx / 3;
	// Bt458 ならこれをマスクする?
	data &= readmask;
	if (bt454) {
		data = (data << 4) | LOW_NIBBLE;
	}
	putlog(2, "Address Register -> $%02x", data);

	// アドレスレジスタアクセスで ab カウンタは 0 にリセットされる
	idx = (idx / 3) * 3;

	return data;
}

// パレットレジスタ読み込み。
uint32
BT454Device::ReadPalette()
{
	uint32 data;

	// Bt454 の場合 color[] の上位下位4ビットは同じ値にしてある
	data = color[idx] | low_nibble;
	putlog(2, "Color Palette $%x(%c) -> $%02x",
		idx / 3, rgbstr[idx % 3], data);

	// アクセスするたびにインクリメントする。
	idx++;
	if (idx >= color.size()) {
		idx = 0;
	}

	return data;
}

// Bt454 ならリセット(?)、
// Bt458 なら各種制御レジスタの読み込み。
uint32
BT454Device::ReadControl()
{
	uint32 data;

	if (bt454) {
		// ADDRa,b を 0 にする? D0-D3 は 0 になる??
		// データシートが雑でよく分からん…
		idx = (idx / 3) * 3;
		data = 0x00 | LOW_NIBBLE;
		putlog(2, "reset ab");
	} else {
		// Bt458 なら、この時点のアドレスによってコマンド処理
		uint addr = idx / 3;
		switch (addr) {
		 case 0x04:
			data = readmask;
			putlog(2, "Read ReadMask -> $%02x", data);
			break;
		 case 0x05:
			data = blinkmask;
			putlog(2, "Read BlinkMask -> $%02x", data);
			break;
		 case 0x06:
			data = command;
			putlog(2, "Read Command -> $%02x", data);
			break;
		 case 0x07:
			data = testreg;
			putlog(2, "Read Test -> $%02x", data);
			break;
		 default:
			data = 0;
			putlog(0, "Read $%02x (NOT IMPLEMENTED)", addr);
			break;
		}
	}

	return data;
}

// アドレスレジスタ書き込み。
void
BT454Device::WriteAddress(uint32 data)
{
	// 今からアクセスするパレット番号。
	// アドレスレジスタアクセスで ab カウンタは 0 にリセットされる。
	// data はパレット番号だが idx は配列のインデックスなので 3倍。
	putlog(2, "Address Register <- $%02x", data);
	if (bt454) {
		data >>= 4;
	}
	idx = data * 3;
}

// パレットレジスタ書き込み。
void
BT454Device::WritePalette(uint32 data)
{
	// Bt454 なら上位下位4ビットに同じ値を置くことで 0 .. 255 にする。
	if (bt454) {
		data = (data & 0xf0) | ((data >> 4) & 0x0f);
	}
	color[idx] = data;

	if (loglevel == 1) {
		// ログレベル1なら RGB が書き込まれた時点で表示。
		// 大体はこれくらいの表示で構わないはず。
		if (idx % 3 == 2) {
			if (bt454) {
				putlogn("Color Palette $%x = (%x,%x,%x)",
					idx / 3,
					color[idx - 2] >> 4,
					color[idx - 1] >> 4,
					color[idx    ] >> 4);
			} else {
				putlogn("Color Palette $%02x = (%02x,%02x,%02x)",
					idx / 3,
					color[idx - 2],
					color[idx - 1],
					color[idx    ]);
			}
		}
	} else if (loglevel >= 2) {
		if (bt454) {
			putlogn("Color Palette $%x(%c) <- $%02x",
				idx / 3, rgbstr[idx % 3], data);
		} else {
			putlogn("Color Palette $%02x(%c) <- $%02x",
				idx / 3, rgbstr[idx % 3], data);
		}
	}

	// ホストカラーを計算。
	// (Bt458 では?) Blue を書き込んだところで色情報がストアされるっぽい?
	MakeHostColor(idx);
	Notify();

	// ホストカラー計算後にインクリメントする。
	idx++;
	if (idx >= color.size()) {
		idx = 0;
	}
}

// Bt454 ならリセット(?)、
// Bt458 なら各種制御レジスタへの書き込み。
void
BT454Device::WriteControl(uint32 data)
{
	if (bt454) {
		// ab カウンタを 0 にする? データは無視されるようだ
		idx = (idx / 3) * 3;
		putlog(2, "reset ab");
	} else {
		// Bt458 なら、この時点のアドレスによってコマンド処理。
		uint addr = idx / 3;
		switch (addr) {
		 case 0x04:
			WriteRMask(data);
			break;
		 case 0x05:
			WriteBMask(data);
			break;
		 case 0x06:
			WriteCommand(data);
			break;
		 case 0x07:
			WriteTest(data);
			break;
		 default:
			putlog(1, "Unknown command $%02x <- $%02x", addr, data);
			break;
		}
	}
}

// Bt458 Read Mask の書き込み。
void
BT454Device::WriteRMask(uint32 data)
{
	readmask = data;
	putlog(1, "Write ReadMask <- $%02x", readmask);
}

// Bt458 Blink Mask の書き込み。
void
BT454Device::WriteBMask(uint32 data)
{
	// 未実装
	blinkmask = data;
	if (blinkmask != 0x00) {
		putlog(0, "Write BlinkMask <- $%02x (NOT IMPLEMENTED)", blinkmask);
	} else {
		putlog(1, "Write BlinkMask <- $%02x", blinkmask);
	}
}

// Bt458 コマンドレジスタへの書き込み。
void
BT454Device::WriteCommand(uint32 data)
{
	command = data;
	putlog(1, "Write Command <- $%02x", command);

	// 未実装
	if ((command & 0x80)) {
		putlog(0, "5:1 Multiplex (NOT IMPLEMENTED)");
	}
	if ((command & 0x40) == 0) {
		putlog(0, "Use overlay color0 (NOT IMPLEMENTED)");
	}
	if ((command & 0x0c)) {
		putlog(0, "Blink enable (NOT IMPLEMENTED)");
	}
	if ((command & 0x03)) {
		putlog(0, "Display enable (NOT IMPLEMENTED)");
	}
}

// テストレジスタへの書き込み。
void
BT454Device::WriteTest(uint32 data)
{
	// 未実装
	testreg = data;
	if (testreg != 0x00) {
		putlog(0, "Write Test <- $%02x (NOT IMPLEMENTED)", testreg);
	} else {
		// カーネルが初期化時にアクセスに来る
		putlog(1, "Write Test <- $%02x (NOT IMPLEMENTED)", testreg);
	}
}

// 制御レジスタを読み出す。(Peek() 用)
uint32
BT454Device::PeekControl() const
{
	if (bt454) {
		return 0x00;
	} else {
		uint addr = idx / 3;
		switch (addr) {
		 case 0x04:
			return readmask;
		 case 0x05:
			return blinkmask;
		 case 0x06:
			return command;
		 case 0x07:
			return testreg;
		 default:
			// XXX ?
			return 0;
		}
	}
}

// ゲストパレットを uint32 形式にしたものの一覧を返す。(GUI 用)
const std::vector<uint32>
BT454Device::GetIntPalette() const
{
	std::vector<uint32> ipal;

	if (bt454) {
		for (int i = 0; i < palettes; i++) {
			ipal.push_back(Palette454ToInt(i));
		}
	} else {
		for (int i = 0; i < palettes; i++) {
			ipal.push_back(Palette458ToInt(i));
		}
	}
	return ipal;
}

// Bt454 のパレット番号 n の生データを uint32 で返す。
// Bt454 の場合は RGB 各4ビットを右詰め。$0000'0RGB。
uint32
BT454Device::Palette454ToInt(int n) const
{
	return ((color[n * 3 + 0] & 0xf0) << 4) |
	       ((color[n * 3 + 1] & 0xf0)     ) |
	       ((color[n * 3 + 2] & 0x0f)     );
}

// Bt458 のパレット番号 n の生データを uint32 で返す。
// Bt458 の場合は RGB 各8ビットを右詰め。$00RR'GGBB。
uint32
BT454Device::Palette458ToInt(int n) const
{
	return (color[n * 3 + 0] << 16) |
	       (color[n * 3 + 1] <<  8) |
	        color[n * 3 + 2];
}
