//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// Bt454/458
//

#pragma once

#include "device.h"
#include "color.h"
#include "monitor.h"

class BT454Device : public IODevice
{
	using inherited = IODevice;

 public:
	BT454Device();
	virtual ~BT454Device() override;

	bool Init() override;
	void ResetHard(bool poweron) override;

	uint64 Read8(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;

	// プレーン数を取得
	int GetPlaneCount() const { return nplane; }

	// ホストパレットのアドレスを取得
	const std::vector<ColorXBGR>& GetHostPalette() const { return hostcolor; }

	// 生パレットを uint32 形式にしたものを返す (GUI 用)
	const std::vector<uint32> GetIntPalette() const;

	// 内部アクセス
	// (内蔵 ROM からのアクセス用に特別に public にしてある)
	uint64 Write(uint32 offset, uint32 data);

 private:
	// BusIO に似てるが、BusIO ではない。
	uint64 Read(uint32 offset);
	uint64 Peek(uint32 offset);

	// レジスタ値からホストパレットを作成
	void MakeHostColor(int idx);

	// 通知
	void Notify();

	uint32 ReadAddress();
	uint32 ReadPalette();
	uint32 ReadControl();
	void WriteAddress(uint32 data);
	void WritePalette(uint32 data);
	void WriteControl(uint32 data);
	uint32 PeekControl() const;

	// Bt458 拡張
	void WriteRMask(uint32 data);
	void WriteBMask(uint32 data);
	void WriteCommand(uint32 data);
	void WriteTest(uint32 data);

	uint32 Palette454ToInt(int n) const;
	uint32 Palette458ToInt(int n) const;

	DECLARE_MONITOR_CALLBACK(MonitorUpdate454);
	DECLARE_MONITOR_CALLBACK(MonitorUpdate458);

	// モデル種別。Bt454 なら true。
	bool bt454 {};

	// ビットマップのプレーン数 (1, 4, 8)
	int nplane {};

	// パレット数 (16 or 256)
	int palettes {};

	// データの下位ニブルマスク。Bt454 なら LOW_NIBBLE。
	uint8 low_nibble {};

	// Bt458 拡張
	uint8 readmask {};		// パレットアドレスマスク
	uint8 blinkmask {};		// ブリンクマスク
	uint8 command {};		// コマンド
	uint8 testreg {};		// テストレジスタ

	// idx は次にアクセスされるべき color[] へのインデックス。
	// idx = 0 ならパレット0 の R。
	// idx = 4 ならパレット1 の G。
	// Bt454 では R/G/B をカウントするための 3値 2ビットのカウンタ(?)を
	// ADDRa,ADDRb (ADDRa,b) と呼んだりしているようなので、ここではその
	// modulo 3 を数えるカウンタを ab と呼ぶことにする。つまり
	// idx = 0 はパレット 0、ab 0。
	// idx = 4 はパレット 1、ab 1。
	int idx;

	// パレットは R, G, B の順に格納されている。
	// color[0] = パレット0 R
	// color[1] = パレット0 G
	// color[2] = パレット0 B
	// color[3] = パレット1 R
	// :
	// color[47] = パレット15 B
	//
	// Bt454 なら 16パレット * 3 = 48エントリ。
	// 色深度は 4ビットで上位4ビットと下位4ビットに同じものを格納する。
	// つまり $4 なら $44、$f なら $ff として格納しておく。
	//
	// Bt458 なら 256パレット * 3 = 768エントリ。
	// 色深度は 8ビットでそのまま格納。
	std::vector<uint8> color;

	// ホスト用に変換したパレットデータ。
	// ホスト用なので各8ビット、といいつつ元情報 (color[]) が8ビットすべて
	// 有効なので値はそのまま流用する。
	//
	// また 1bpp モードなら R,G,B のうち G だけが使われるので、ここで
	// 展開する。パレットのうち実際には #14 と #15 だけが使われるが、
	// それについてはレンダラ側で処理している。
	// see https://twitter.com/tsutsuii/status/356704682514714625
	std::vector<ColorXBGR> hostcolor;

	Monitor monitor { this };

	// ログ出力用
	static const char rgbstr[3];
};

extern BT454Device *gBT454;
