//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メインバス
//

#include "bus.h"
#include "mainram.h"

// グローバル参照用
MainbusDevice *gMainbus;

// コンストラクタ
MainbusDevice::MainbusDevice()
	: inherited("Mainbus")
{
}

// デストラクタ
MainbusDevice::~MainbusDevice()
{
	gMainbus = NULL;
}

uint64
MainbusDevice::Read8(uint32 addr)
{
	return vm_phys_read_8(addr);
}

uint64
MainbusDevice::Read16(uint32 addr)
{
	return vm_phys_read_16(addr);
}

uint64
MainbusDevice::Read32(uint32 addr)
{
	return vm_phys_read_32(addr);
}

uint64
MainbusDevice::Write8(uint32 addr, uint32 data)
{
	return vm_phys_write_8(addr, data);
}

uint64
MainbusDevice::Write16(uint32 addr, uint32 data)
{
	return vm_phys_write_16(addr, data);
}

uint64
MainbusDevice::Write32(uint32 addr, uint32 data)
{
	return vm_phys_write_32(addr, data);
}

uint64
MainbusDevice::Peek8(uint32 addr)
{
	return vm_phys_peek_8(addr);
}

// このアドレスを担当する次段の IODevice を返す (IOContainerDevice 用)
IODevice *
MainbusDevice::GetDevice(uint32 addr) const
{
	return devtable[addr >> 24];
}

// デバイスをセットする。すでにあれば更新する。
// idx は devtable[] のインデックス (16MB 単位)。
/*static*/ void
MainbusDevice::SetDevice(int idx, IODevice *dev)
{
	assert(idx < devtable.size());

	devtable[idx] = dev;
}

// デバイステーブルの src から len 個を dst にコピーする。(ミラーを作る)
/*static*/ void
MainbusDevice::CopyDevtable(int dst, int src, int len)
{
	memcpy(&devtable[dst], &devtable[src], sizeof(devtable[0]) * len);
}

// Mainbus は1つしかないので、スタティック変数にしておく。
/*static*/ std::array<IODevice *, 256> MainbusDevice::devtable;

// RAM へのアクセスは、アクセスアドレスが direct_ram_size 未満なら ram[] を
// 直接アクセスし、そうでなければデバイスクラス経由でアクセスする。
// 高速モードでは RAM デバイスを経由すること自体とその中でアクセスウェイトを
// 計算する分を回避したい。
/*static*/ int MainbusDevice::direct_ram_size;


//
// グローバル関数
//

uint64
vm_phys_read_8(uint32 addr)
{
	uint64 data;

	if (addr < MainbusDevice::direct_ram_size) {
		data = MainRAMDevice::mainram[HLB(addr)];
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		data = dev->Read8(addr);
	}
	return data;
}

uint64
vm_phys_read_16(uint32 addr)
{
	uint64 data;

	if (__predict_false((addr & 1) != 0))
		VMPANIC("unaligned access $%08x", addr);

	if (addr < MainbusDevice::direct_ram_size) {
		data = *(uint16 *)&MainRAMDevice::mainram[HLW(addr)];
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		data = dev->Read16(addr);
	}
	return data;
}

uint64
vm_phys_read_32(uint32 addr)
{
	uint64 data;

	if (__predict_false((addr & 3) != 0))
		VMPANIC("unaligned access $%08x", addr);

	if (addr < MainbusDevice::direct_ram_size) {
		data = *(uint32 *)&MainRAMDevice::mainram[addr];
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		data = dev->Read32(addr);
	}
	return data;
}

uint64
vm_phys_write_8(uint32 addr, uint32 data)
{
	if (addr < MainbusDevice::direct_ram_size) {
		MainRAMDevice::mainram[HLB(addr)] = data;
		return 0;
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		return dev->Write8(addr, data);
	}
}

uint64
vm_phys_write_16(uint32 addr, uint32 data)
{
	if (__predict_false((addr & 1) != 0))
		VMPANIC("unaligned access $%08x", addr);

	if (addr < MainbusDevice::direct_ram_size) {
		*(uint16 *)&MainRAMDevice::mainram[HLW(addr)] = data;
		return 0;
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		return dev->Write16(addr, data);
	}
}

uint64
vm_phys_write_32(uint32 addr, uint32 data)
{
	if (__predict_false((addr & 3) != 0))
		VMPANIC("unaligned access $%08x", addr);

	if (addr < MainbusDevice::direct_ram_size) {
		*(uint32 *)&MainRAMDevice::mainram[addr] = data;
		return 0;
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		return dev->Write32(addr, data);
	}
}

uint64
vm_phys_peek_8(uint32 addr)
{
	uint64 data;

	if (addr < MainbusDevice::direct_ram_size) {
		data = MainRAMDevice::mainram[HLB(addr)];
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		data = dev->Peek8(addr);
	}
	return data;
}

uint64
vm_phys_peek_32(uint32 addr)
{
	uint64 data;

	if (__predict_false((addr & 3) != 0))
		VMPANIC("unaligned access $%08x", addr);

	if (addr < MainbusDevice::direct_ram_size) {
		data  = *(uint32 *)&MainRAMDevice::mainram[addr];
	} else {
		IODevice *dev = MainbusDevice::devtable[addr >> 24];
		data  = dev->Peek8(addr++) << 24;
		data |= dev->Peek8(addr++) << 16;
		data |= dev->Peek8(addr++) << 8;
		data |= dev->Peek8(addr);
	}
	return data;
}
