//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// デバイスの基本クラスなど
//

#pragma once

#include "object.h"

#if BYTE_ORDER == LITTLE_ENDIAN
#define HB(addr)	((addr) ^ 1)
#define HLB(addr)	((addr) ^ 3)
#define HLW(addr)	((addr) ^ 2)
#else
#define HB(addr)	(addr)
#define HLB(addr)	(addr)
#define HLW(addr)	(addr)
#endif

class Device : public Object
{
	using inherited = Object;
 protected:
	Device(const std::string& objname_);
 public:
	virtual ~Device() override;

 public:
	// 初期化は以下の手順で行う。
	// 1. コンストラクタ
	//	メンバ変数の初期値代入などを行う。失敗するケースは考慮していない。
	//	gConfig の内部変数の追加はここ(か2.の動的コンストラクション)で
	//	行うこと。この時点ですでに gConfig の設定内容の参照は可能だが、
	//	エラーが出るものはここではなく Init() で処理すること。
	//	putmsg() は使えない。
	//
	// 2. Create() ... 動的コンストラクション
	//	必要なら設定ファイルによって可変なオブジェクトのコンストラクトを行う。
	//	例えば SCSI ドライブなどが該当する。ここでも gConfig の参照および
	//	内部変数の登録は可能、putmsg() は使えない。
	//
	// 3. parse_logopt() ... -L オプションの処理
	//	この時点でログの割り当てを受ける Object 由来のオブジェクトはすべて
	//	出揃っている必要がある。
	//	またここで --show-config も行うため、この時点で gConfig の内容は確定し
	//	以降 gConfig の変更は出来ない。
	//
	// (4.) ParseMonitors() ... -M オプションの処理
	//	(GUI 版のみ) この時点で Monitor がすべて登録されている必要がある。
	//
	// 5. Init() ... 初期化
	//	この時点で Object 由来のオブジェクトはすべて出揃っており、ログレベル
	//	も確定しているのでこれ以降 putmsg() は使える。
	//	gConfig の設定項目のうち内容によってエラー終了する可能性があるものは
	//	ここで処理する。
	//	ここでは Object 同士を相互にアクセス可能。ただしその際の依存関係に
	//	ついては、vm.cpp (や xiospace.cpp) 内のコンストラクタの順序で調整
	//	すること (循環参照不可)。
	//	ホストデバイスの個別ドライバはここで起動する。本来 Create() で行いたい
	//	ところだが、ネットワークドライバの選択などは複雑なので putmsg() を
	//	使いたい。そうするとここまで遅延せざるを得ない。そのため 3. を過ぎて
	//	からオブジェクトが増えることになるが、個別ドライバは新たなログ名を
	//	増やさず、ホストドライバのログレベルを継承するため、ぎりぎり大丈夫。
	//	スレッドはまだ起動していない。
	//
	// 6. StartThread() ... スレッド起動
	//
	// 7. Apply() ... 動的設定の適用
	//	実行中に変更可能なパラメータもここで一度適用する。
	//	ここで適用するのは実行中に変更可能なパラメータのみ。
	//	適用内容を表示したい時は putlog(1) が妥当か。
	//
	// 8. ResetHard(true) ... 電源オン

	// 動的コンストラクション。
	// 成功すれば true を返す。失敗なら warn() 系、あるいは gConfig->Err()
	// 系でエラーメッセージを表示して false を返すこと。
	virtual bool Create();

	// 初期化。
	// 成功すれば true を返す。失敗なら warn() 系、あるいは gConfig->Err()
	// 系でエラーメッセージを表示して false を返すこと。
	virtual bool Init();

	// 現在の設定を VM に反映させる。
	// 成功すれば true を返す。失敗なら warn() 系、あるいは gConfig->Err()
	// 系でエラーメッセージを表示して false を返すこと。
	// といっても false が返されたらどうするんだろうか。
	virtual bool Apply();

	// 電源オンもしくはリセット。
	// true なら電源オン、false ならリセット。
	// このデバイスがどのタイミングでリセットされるかは VM による。
	// X68030 なら vm_x68k.cpp 内のコメント参照。
	// LUNA は詳細不明。
	virtual void ResetHard(bool poweron);

	// 電源オフ
	virtual void PowerOff();

	// ログ出力
	void putlogn(const char *fmt, ...) const override __printflike(2, 3);
};

class IODevice : public Device
{
	using inherited = Device;
 protected:
	IODevice(const std::string& objname_);
 public:
	virtual ~IODevice() override;

 public:
	// アクセス関数。
	// Read*() は読み込んだ値を (uint32) で返す。
	// Write*() は書き込めたら 0 を返す。
	// いずれもバスエラーなら (uint64)-1 を返す。
	virtual uint64 Read8(uint32 addr);
	virtual uint64 Read16(uint32 addr);
	virtual uint64 Read32(uint32 addr);
	virtual uint64 Write8(uint32 addr, uint32 data);
	virtual uint64 Write16(uint32 addr, uint32 data);
	virtual uint64 Write32(uint32 addr, uint32 data);

	// Peek8() はデバッガアクセスで読み込んだ値を (uint32) で返す。
	// バイトアクセスした時に読める値ではなく、現在このアドレスに見えている
	// であろう値。バスエラーなら (uint64)-1 を返す。
	virtual uint64 Peek8(uint32 addr);

	// Poke8() はデバッガからこのアドレス(メモリ)へ書き込む。
	// メモリでない領域 (レジスタなど) は書き込み不可。
	// data が (int32 で) 非負なら data (の下位8ビット)を書き込む。
	// 書き込めれば 0、書き込めなければ (uint64)-1 を返す。
	// data が (int32 で) 負なら addr への書き込み可否の問い合わせで、
	// 書き込み可能なら 0、書き込み不可能なら (uint64)-1 を返す。
	virtual uint64 Poke8(uint32 addr, uint32 data);

 protected:
	// BusIO インタフェース。
	// BusIO を経由するデバイスはほとんどが Poke 不可なので、
	// ベースクラスに用意しておく。
	virtual uint64 Poke(uint32 addr, uint32 data);
};

// コンテナデバイス (複数の IODevice を格納するデバイス)。
class IOContainerDevice : public IODevice
{
	using inherited = IODevice;
 protected:
	IOContainerDevice(const std::string& objname_);
 public:
	virtual ~IOContainerDevice() override;

	// このアドレスを担当する IODevice を再帰的に探して返す。
	IODevice *FindDevice(uint32 addr) const;

 protected:
	// このアドレスを担当する次段の IODevice を返す。
	virtual IODevice *GetDevice(uint32 addr) const = 0;
};

class BusErrDevice : public IODevice
{
	using inherited = IODevice;
 public:
	BusErrDevice();
	virtual ~BusErrDevice() override;

 public:
	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;
};

// 何もしないデバイス。
// 読み込むと全域で 0xff が読めて、書き込んでも何も起きない。
class NopIODevice : public IODevice
{
	using inherited = IODevice;
 public:
	NopIODevice();
	virtual ~NopIODevice() override;

 public:
	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;
};

// 前方参照
class SCSIDevice;
class SCSIDisk;
class SCSITarget;

extern BusErrDevice *gBusErr;
extern NopIODevice *gNopIO;
