//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// Human68k .r .x .z console emulator
//

// メモリマップ
//
// 0000'0000 ベクタ
// 0000'0200 Humanエントリポイント
// 0000'8000 FCBエリア
// 0000'c000 コマンドライン引数
// 0000'e000 環境変数
// 0001'ff00 SSP 初期値
// 0001'ff00 PSP(プロセスエントリ)
// 0002'0000 ロードアドレス
// 00bf'ffff RAM_END

#include "human68k.h"
#include "autofd.h"
#include "iodevstream.h"
#include "mainapp.h"
#include "m68030bus.h"
#include "m68030core.h"
#include "mpu680x0.h"
#include "power.h"
#include "scheduler.h"
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

static bool human68k_fline_callback(m68kcpu *cpu, void *arg);

// グローバル参照用
Human68k *gHuman68k;

// コンストラクタ
Human68k::Human68k()
	: inherited("Human68k")
{
	assert(gMainApp.exec_file);
	human68k_file = gMainApp.exec_file;
	human68k_arg = gMainApp.exec_arg;

	Files[0].fd = 0;
	Files[0].filename = "|stdin";
	Files[1].fd = 1;
	Files[1].filename = "|stdout";
	Files[2].fd = 2;
	Files[2].filename = "|stderr";
}

// デストラクタ
Human68k::~Human68k()
{
	gHuman68k = NULL;
}

bool
Human68k::Init()
{
	bool r;
	uint32 psp_base;
	uint8 *file;

	// ショートカット用
	ram = gMainRAM;
	// Fライン命令をこちらで処理する
	gMPU680x0->SetFLineCallback(human68k_fline_callback, this);

	// ワークを用意 (どこでやるか)
	// $CBC.B MPU 種別 (3:68030)
	ram->Write8(0x0cbc, 3);
	// $CBD.B FPU 有無 (0xff:あり)
	if (gMPU680x0->HaveFPU()) {
		ram->Write8(0x0cbd, 0xff);
	}

	putmsg(1, "arg=%s", human68k_arg.c_str());

	// 実行ファイルオープン
	autofd file_fd = open(human68k_file, O_RDONLY);
	if (file_fd == -1) {
		warn("Human68k executable \"%s\" open failed", human68k_file);
		return false;
	}

	struct stat st;
	if (fstat(file_fd, &st) != 0) {
		warn("Human68k executable \"%s\" fstat failed", human68k_file);
		return false;
	}
	uint32 file_size;
	file_size = (uint32)st.st_size;
	putmsg(1, "file_size=%08x", file_size);

	// mmap
	file = (uint8 *)mmap(NULL, file_size, PROT_READ, MAP_PRIVATE, file_fd, 0);
	if (file == MAP_FAILED) {
		warn("Human68k executable \"%s\" mmap failed", human68k_file);
		return false;
	}

	// 拡張子判別
	if (isext(human68k_file, ".r")) {
		r = LoadR(file, file_size);
	} else if (isext(human68k_file, ".x")) {
		r = LoadX(file, file_size);
	} else if (isext(human68k_file, ".z")) {
		r = LoadZ(file, file_size);
	} else {
		// ELF .o のロードを試行する
		r = LoadELF(file, file_size);
		if (!r) {
			if (file[0] == 0x48 && file[1] == 0x55) {
				r = LoadX(file, file_size);
			} else if (file[0] == 0x60 && file[1] == 0x1a) {
				r = LoadZ(file, file_size);
			} else {
				warnx("Human68k executable \"%s\" cannot identify file type",
					human68k_file);
				munmap(file, file_size);
				return false;
			}
		}
	}
	if (!r) {
		warnx("Human68k executable \"%s\" invalid file format", human68k_file);
		munmap(file, file_size);
		return false;
	}

	munmap(file, file_size);

	uint32 last_addr = load_addr + load_size + bss_size;
	int ram_size = ram->GetSize();

	IODeviceStream ds(ram);
	boot_addr = 0x200;
	ds.Write32(0, ram_size - 4);	// 初期 SSP
	ds.Write32(4, boot_addr);

	// TRAP #15 ベクタを書き込む。
	ds.Write32(0xbc, 0x1000);
	// TRAP #15 ハンドラを書き込む。
	// エミュレータ内部命令 $f300 で処理して RTE するだけ。
	ds.Write16(0x1000, 0xf600);	// .dw $f300
	ds.Write16(0x1002, 0x4e73);	// rte

	// STOP による停止。終了時に CPU を止めておくため。
	ds.Write32(0x1010, 0x4e722700);	// stop #$2700

	// コマンドライン文字列を書き込む。LASCII 形式
	ds.SetAddr(0xc001);
	for (int i = 0; i < 255; i++) {
		uint8 c = human68k_arg[i];
		ds.Write8(c);
		if (c == '\0') {
			ds.Write8(0xc000, i);
			break;
		}
	}
	// XXX: 255文字以上のときどうなるか調べること

	// PSP (process entry) を書き込む。
	// instructiontest.x はコマンドラインを指定するだけで動くようだ
	// PSP はロードアドレス-256 の位置でなければならないようだ。
	psp_base = 0x1ff00;
	ds.SetAddr(psp_base);

	ds.Write32(-1);		// 1つ前のメモリ管理ポインタ
	ds.Write32(-1);		// このメモリを確保したプロセスのメモリ管理ポインタ
	ds.Write32(ram_size - 4 + 1);	// このメモリブロックの終わり+1 のアドレス
	ds.Write32(-1);		// 次のメモリ管理ポインタ

	ds.Write32(psp_base + 0x20, 0xc000);	// コマンドライン

	ds.SetAddr(boot_addr);
	ds.Write16(0x2c4f);		// move.l	a7,a6
	ds.Write16(0x2e7c);		// move.l	psp_base,a7
	ds.Write32(psp_base);
	ds.Write16(0x4e66);		// move.l	a6,usp
	ds.Write16(0x207c);		// move.l	psp_base,a0
	ds.Write32(psp_base);
	ds.Write16(0x227c);		// move.l	last_addr,a1
	ds.Write32(last_addr);
	ds.Write16(0x247c);		// move.l	#$c000,a2
	ds.Write32(0xc000);
	ds.Write16(0x267c);		// move.l	#$e000,a3
	ds.Write32(0xe000);
	ds.Write16(0x287c);		// move.l	exec_addr,a4
	ds.Write32(exec_addr);
	ds.Write16(0x2c49);		// move.l	a1,a6
	ds.Write16(0x46fc);		// move.w	#$0700,sr
	ds.Write16(0x0700);
	ds.Write16(0x4eb9);		// jsr.l	exec_addr
	ds.Write32(exec_addr);
	ds.Write16(0xff00);		// DOS		_EXIT

	putmsg(1, "LoadFile complete");
	return true;
}

bool
Human68k::LoadR(uint8 *file, uint size)
{
	// リロケータブルなのでどこでもいい
	load_addr = default_load_addr;
	load_size = size;
	LoadMem(load_addr, &file[0], load_size);

	// R 形式はファイル先頭が実行開始位置
	exec_addr = load_addr;
	text_size = load_size;
	putmsg(1, "r format");
	return true;
}

bool
Human68k::LoadX(uint8 *file, uint size)
{
	XFileHeader *hdr = (XFileHeader *)file;
	uint hdr_size = sizeof(*hdr);

	if (!(hdr->magic[0] == 'H' && hdr->magic[1] == 'U')) {
		errx(EXIT_FAILURE, "invalid magic");
	}
	base_addr = be32toh(hdr->base_addr);
	exec_addr = be32toh(hdr->exec_addr);
	text_size = be32toh(hdr->text_size);
	data_size = be32toh(hdr->data_size);
	bss_size = be32toh(hdr->bss_size);

	uint32 reloc_size = be32toh(hdr->reloc_size);

	// 最適配置位置も可能だが今回見送り
	load_addr = default_load_addr;
	load_size = text_size + data_size;
	LoadMem(load_addr, &file[hdr_size], load_size);

	// 再配置テーブルの file での位置
	uint32 reloc_pos = hdr_size + load_size;

	// 再配置
	IODeviceStream ds(ram);
	uint32 offset = load_addr - base_addr;
	uint32 reloc_end = reloc_pos + reloc_size;
	uint32 A = load_addr;
	uint32 B = base_addr;
	uint32 C = A - B;
	while (reloc_pos < reloc_end) {
		putlog(1, "reloc_pos=%08x reloc_end=%08x", reloc_pos, reloc_end);
		uint32 D;
		D = be16toh(*(uint16 *)&file[reloc_pos]);
		putlog(1, "D=%x", D);
		reloc_pos += 2;
		if (D == 1) {
			D = be32toh(*(uint32 *)&file[reloc_pos]);
			putlog(1, " odd, D=%x", D);
			reloc_pos += 4;
		}
		if ((D & 1) == 0) {
			A += D;
			uint32 old = ds.Read32(A);
			ds.Write32(A, old + C);
			putlog(1, " Write_L A=%x, old=%x new=%x", A, old, old + C);
		} else {
			A += D - 1;
			uint32 old = ds.Read16(A);
			ds.Write16(A, old + C);
			putlog(1, " Write_W A=%x, old=%x new=%x", A, old, old + C);
		}
	}

	exec_addr += offset;

	putmsg(1, "x format, base_addr=%08x, exec_addr=%08x", base_addr, exec_addr);

	return true;
}

bool
Human68k::LoadZ(uint8 *file, uint size)
{
	ZFileHeader *hdr = (ZFileHeader *)file;
	uint32 hdr_size = sizeof(*hdr);

	if (be16toh(hdr->magic1) != 0x601a) {
		errx(EXIT_FAILURE, "invalid magic");
	}
	text_size = be32toh(hdr->text_size);
	data_size = be32toh(hdr->data_size);
	bss_size = be32toh(hdr->bss_size);
	base_addr = be32toh(hdr->base_addr);
	exec_addr = base_addr;

	if (base_addr < 0x20000) {
		errx(EXIT_FAILURE, "base, unsupported");
	}

	load_addr = base_addr;
	load_size = text_size + data_size;
	LoadMem(load_addr, &file[hdr_size], load_size);

	putmsg(1, "z format, base_addr=%08x, exec_addr=%08x", base_addr, exec_addr);
	return true;
}

// なぜか m68k ELF .o が読める
bool
Human68k::LoadELF(uint8 *file, uint size)
{
	LoadInfo info(human68k_file, file, size);
	if (ram->LoadFromData(&info) == false) {
		return false;
	}

	// XXX load_size は esym のような気がする。
	// XXX text_size も今はやろうと思えば返せる構造になったが未対応。
	load_addr = info.entry;
	load_size = size;
	exec_addr = info.entry;
	text_size = size;
	return true;
}

bool
Human68k::LoadMem(uint32 addr, uint8 *src, uint size)
{
	if (addr + size >= ram->GetSize()) {
		errx(EXIT_FAILURE, "file too large");
	}

	IODeviceStream ds(ram, addr);
	for (uint32 i = 0; i < size; i++) {
		ds.Write8(*src++);
	}
	return true;
}

// file の拡張子が ext なら true を返す。ext は '.' を含む。
bool
Human68k::isext(const char *file, const char *ext)
{
	size_t len_file = strlen(file);
	size_t len_ext = strlen(ext);
	if (len_ext <= 0) {
		return false;
	}
	if (len_file < len_ext) {
		return false;
	}

	return strcasecmp(&file[len_file - len_ext], ext) == 0;
}


// 仮想マシンの中から終了させる
// XXX code は未対応。
void
Human68k::RequestExit(int code)
{
	// 電源を即切る
	gPower->PowerButton();
	gPower->SetSystemPowerOn(false);
}

// fileaddr: Human68k ファイル名のゲストVA
// atr: Human68k atr
// mode: unix open mode
// return: Human68k fileno
int32
Human68k::OpenFile(uint32 fileaddr, uint16 atr, int mode)
{
	// XXX: unix host only

	int32 fileno {};
	Human68k::File *f = NULL;

	// 開いているエントリを検索して Human fileno を取得
	for (int i = 0; i < FilesCount; i++) {
		if (Files[i].fd == -1) {
			fileno = i;
			f = &Files[fileno];
			break;
		}
	}
	if (f == NULL) {
		return -1;
	}

	char filename[256];
	char *p = filename;

	// ファイル名変換
	IODeviceStream ds(ram, fileaddr);
	for (int i = 0; i < countof(filename) - 3; i++) {
		uint8 c = ds.Read8();

		if (c == 0) {
			break;
		}
		if (c < 32 || c >= 127 || c == 0x5c) {
			p += sprintf(p, "%02X", c);
		} else {
			*p++ = c;
		}
	}
	*p = '\0';

	putmsg(1, "OpenFile: %d %s", fileno, filename);
	int fd = open(filename, mode, 0666);
	if (fd == -1) {
		warn("OpenFile.open");
		return -1;
	}

	f->fd = fd;
	f->filename = filename;

	return fileno;
}

// fileno: Human68k fileno
int32
Human68k::CloseFile(int32 fileno)
{
	Human68k::File *f;

	if (fileno <= 0 || fileno >= FilesCount) {
		warnx("CloseFile: unopened fileno=%d", fileno);
		return -1;
	}
	f = &Files[fileno];
	putmsg(1, "CloseFile: %d %s", fileno, f->filename.c_str());
	if (f->fd > 2) {
		// stdin/out/err は閉じない
		close(f->fd);
	}
	f->fd = -1;
	f->filename.clear();
	return 0;
}

// fileno: Human68k fileno
// dataaddr: data addr (guest VA)
// size: data length
int32
Human68k::WriteFile(int32 fileno, uint32 dataaddr, uint32 size)
{
	Human68k::File *f;

	if (fileno <= 0 || fileno >= FilesCount) {
		warnx("WriteFile: unopened fileno=%d", fileno);
		return -1;
	}

	f = &Files[fileno];
	if (f->fd < 0) {
		return -1;
	}

	IODeviceStream ds(ram, dataaddr);
	std::unique_ptr<uint8[]> buf(new uint8[size]);
	for (int i = 0; i < size; i++) {
		buf[i] = ds.Read8();
	}

	int32 rv = write(f->fd, buf.get(), size);
	return rv;
}

// fileno: Human68k fileno
// dataaddr: data addr (guest VA)
void
Human68k::FputsFile(int32 fileno, uint32 dataaddr)
{
	Human68k::File *f;

	if (fileno <= 0 || fileno >= FilesCount) {
		return;
	}

	f = &Files[fileno];
	if (f->fd < 0) {
		return;
	}

	IODeviceStream ds(ram, dataaddr);
	int size = 1024;
	std::vector<uint8> buf(size);
	bool eof = false;

	do {
		int len = 0;
		for (int i = 0; i < size; i++) {
			buf[i] = ds.Read8();
			if (buf[i] == 0) {
				eof = true;
				break;
			}
			len++;
		}
		if (len > 0) {
			write(f->fd, &buf[0], len);
		}
	} while (!eof);
}


// Human68k DOSCALL Host Emulation

void
Human68k::IOCS(m68kcpu *cpu)
{
	switch (RegD(0) & 0xff) {
	 case 0x21:	// _B_PRINT
	 {
		uint32 dataptr = RegA(1);
		DOS_PRINT(cpu, dataptr);
		// 戻り値 d0 は表示後の位置だが無視
		break;
	 }
	 case 0x7f:	// _ONTIME
	 {
		putmsg(1, "IOCS ONTIME");
		uint64 t = gScheduler->GetVirtTime();
		// nanosec to 10msec, in day
		RegD(0) = (t / 10_msec) % 8640000;
		// nanosec to day
		RegD(1) = t / 86400_sec;
		break;
	 }
	 case 0x82:	// _B_BPEEK
	 {
		uint32 data = m68030_read_8(cpu, RegA(1));
		RegD(0) = (RegD(0) & 0xffffff00) | data;
		RegA(1)++;
		break;
	 }
	 case 0xac: // _SYS_STAT (ROM1.3)
		switch (RegD(1)) {
		 case 0:
			// MPU 状態の取得
			RegD(0) =
			    ((250) << 16)	// 25.0MHz
			  | ((gMPU680x0->HaveFPU() ? 1 : 0) << 15)	// FPU
			  | ((0) << 14)		// MMU
			  | ((3) << 0);		// MPU Type
			break;
		 case 1:
			// キャッシュ状態の取得
			RegD(0) = 0;
			break;
		 case 2:
			// キャッシュを SRAM の設定値に設定
			RegD(0) = 0;
			break;
		 case 3:
			// キャッシュの消去
			RegD(0) = 0;
			break;
		 case 4:
			// キャッシュの設定
			RegD(0) = 0;
			break;
		 default:
			break;
		}
		break;

	 default:
		printf("IOCS $%02x (NOT IMPLEMENTED)\n", RegD(0) & 0xff);
		RequestExit(1);
		break;
	}
}

// F-Line 命令をこちらで処理する
bool
human68k_fline_callback(m68kcpu *cpu, void *arg)
{
	auto *human68k = (Human68k *)arg;
	return human68k->FLineOp(cpu);
}

// F-Line 命令
bool
Human68k::FLineOp(m68kcpu *cpu)
{
	switch (RegIR) {
	 case 0xf600:		// IOCS call emulation
		IOCS(cpu);
		break;

	 case 0xff00:		// EXIT
		putmsg(1, "DOS EXIT");
		// XXX CPU の実行を止める。方式自体を I/O 方式に見直したときには
		// ちゃんとした方法を取る予定。
		m68030_set_sr(cpu, RegSR | 0x2000);
		RegPC = 0x1010;
		RequestExit(0);
		break;

	 case 0xff09:		// PRINT
	 {
		uint32 dataptr = m68030_read_32(cpu, RegA(7));
		DOS_PRINT(cpu, dataptr);
		break;
	 }

	 case 0xff1e:		// FPUTS
	 {
		uint32 mesptr = m68030_read_32(cpu, RegA(7));
		uint16 fileno = m68030_read_16(cpu, RegA(7) + 4);
		FputsFile(fileno, mesptr);
		break;
	 }

	 case 0xff20:		// SUPER
	 {
		uint32 data = m68030_read_32(cpu, RegA(7));

		if (data == 0) {
			m68030_set_sr(cpu, RegSR | 0x2000);
			data = RegUSP;
			RegD(0) = RegA(7);
			RegA(7) = data;
			putmsg(1, "SUPERVISOR MODE");
		} else {
			m68030_set_sr(cpu, RegSR | 0x2000);
			RegA(7) = data;
			m68030_set_sr(cpu, RegSR & ~0x2000);
			putmsg(1, "USER MODE");
		}
		break;
	 }

	 case 0xff25:		// INTVCS
	 {
		uint16 intno = m68030_read_16(cpu, RegA(7));
		uint32 addr = m68030_read_32(cpu, RegA(7) + 2);

		uint32 vecaddr = (intno & 0xff) * 4;

		if (intno <= 0xff) {
			RegD(0) = m68030_read_32(cpu, vecaddr);
			m68030_write_32(cpu, vecaddr, addr);
		} else {
			vecaddr += 0xd000;
			RegD(0) = m68030_read_32(cpu, vecaddr);
			m68030_write_32(cpu, vecaddr, addr);
		}
		break;
	 }

	 case 0xff27:		// GETTIM2
	 {
		// DUMMY
		RegD(0) = 0;
		break;
	 }

	 case 0xff2a:		// GETDATE
	 {
		// DUMMY
		RegD(0) = 0;
		break;
	 }

	 case 0xff30:		// VERNUM
	 {
		RegD(0) = 0x36380302;		// ver3.02
		break;
	 }

	 case 0xff35:		// INTVCG
	 {
		uint16 intno = m68030_read_16(cpu, RegA(7));

		uint32 vecaddr = (intno & 0xff) * 4;

		if (intno <= 0xff) {
			RegD(0) = m68030_read_32(cpu, vecaddr);
		} else {
			vecaddr += 0xd000;
			RegD(0) = m68030_read_32(cpu, vecaddr);
		}
		break;
	 }

	 case 0xff3c:		// CREATE
	 {
		uint32 file = m68030_read_32(cpu, RegA(7));
		uint16 atr = m68030_read_32(cpu, RegA(7) + 4);

		int32 fileno = OpenFile(file, atr,
			O_CREAT | O_TRUNC | O_RDWR | O_SYNC);
		RegD(0) = fileno;
		break;
	 }

	 case 0xff3e:		// CLOSE
	 {
		uint16 fileno = m68030_read_16(cpu, RegA(7));

		CloseFile(fileno);
		RegD(0) = 0;
		break;
	 }

	 case 0xff40:		// WRITE
	 {
//		putmsg(1, "DOS _WRITE at %08x", RegPPC);
		uint16 fileno = m68030_read_16(cpu, RegA(7));
		uint32 dataptr = m68030_read_32(cpu, RegA(7) + 2);
		uint32 size = m68030_read_32(cpu, RegA(7) + 6);
//		putmsg(1, "WRITE(%d, 0x%08x, 0x%08x)\n", fileno, dataptr, size);
		RegD(0) = WriteFile(fileno, dataptr, size);
		break;
	 }

	 case 0xff44:		// IOCTRL
	 {
		uint32 mode = m68030_read_16(cpu, RegA(7));
		uint32 fileno;
		switch (mode) {
		 case 0:
			fileno = m68030_read_16(cpu, RegA(7) + 2);
			putmsg(1, "DOS IOCTRL(mode=%d, fileno=%d)", mode, fileno);
			if (fileno == 0) { // STDIN
				RegD(0) = 0x8081; // 100u'uuuu'100u'0001;
				break;
			} else if (fileno == 1) { // STDOUT
				RegD(0) = 0x8082; // 100u'uuuu'100u'0010;
				break;
			} else if (fileno == 2) { // STDERR
				RegD(0) = 0x8081; // 100u'uuuu'100u'0001;
				break;
			} else {
				RegD(0) = -1;
			}
			break;
		 default:
			errx(EXIT_FAILURE, "DOS IOCTRL(mode=%d) not impl.", mode);
		}
		break;
	 }

	 case 0xff4a:		// SETBLOCK
	 {
		uint32 memptr = m68030_read_32(cpu, RegA(7));
		uint32 len = m68030_read_32(cpu, RegA(7) + 4);
		putmsg(1, "DOS SETBLOCK(memptr=$%x len=$%x)", memptr, len);
		// なにもせずに、できたという
		RegD(0) = m68030_read_32(cpu, RegA(7) + 4);
		break;
	 }

	 case 0xff4c:		// EXIT2
	 {
		uint32 data = m68030_read_16(cpu, RegA(7));
		putmsg(1, "DOS EXIT2(%d)", data);
		RequestExit(data);
		break;
	 }

	 case 0xffac:		// GETFCB
	 {
		uint32 fileno = m68030_read_16(cpu, RegA(7));
		putmsg(1, "DOS GETFCB(fileno=%d)", fileno);
		if (fileno <= 4) {
			RegD(0) = 0x8000 + fileno * 0x60;
		} else {
			RegD(0) = -1;
		}
		break;
	 }

	 case 0xfff7:		// BUS_ERR
	 {
		uint32 p1 = m68030_read_32(cpu, RegA(7));
		uint32 p2 = m68030_read_32(cpu, RegA(7) + 4);
		uint16 size = m68030_read_16(cpu, RegA(7) + 8);
		putmsg(1, "DOS BUS_ERR(size=%d p1=$%x p2=$%x)", size, p1, p2);

		RegD(0) = -1;
		uint32 data;
		try {
			if (size == 1) {
				data = m68030_read_8(cpu, p1);
			} else if (size == 2 && ((p1 & 1) == 0)) {
				data = m68030_read_16(cpu, p1);
			} else if (size == 4 && ((p1 & 1) == 0)) {
				data = m68030_read_32(cpu, p1);
			} else {
				break;
			}
		} catch (int cause) {
			if (cause == M68K_EXCEP_BUSERR) {
				RegD(0) = 2;
				break;
			} else {
				throw;
			}
		}

		try {
			if (size == 1) {
				m68030_write_8(cpu, p2, data);
			} else if (size == 2 && ((p2 & 1) == 0)) {
				m68030_write_16(cpu, p2, data);
			} else if (size == 4 && ((p2 & 1) == 0)) {
				m68030_write_32(cpu, p2, data);
			} else {
				break;
			}
		} catch (int cause) {
			if (cause == M68K_EXCEP_BUSERR) {
				RegD(0) = 1;
				break;
			} else {
				throw;
			}
		}
		RegD(0) = 0;
		break;
	 }

	 default:
		printf("DOSCALL $%02x at $%08X (NOT IMPLEMENTED)\n", RegIR, RegPPC);
		RequestExit(1);
		break;
	}
	return true;
}

void
Human68k::DOS_PRINT(m68kcpu *cpu, uint32 dataptr)
{
	int c;
	while ((c = m68030_read_8(cpu, dataptr++)) != 0) {
		printf("%c", c);
	}
	RegD(0) = 0;
}
