//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 割り込みコントローラ
//

// IODevice
//    |
//    |  +-----------------+
//    +--| InterruptDevice | (割り込みコントローラの共通部)
//       +-----------------+
//          |  +-----------------+
//          +--| M680x0Interrupt | (m68k 割り込みコントローラの共通部)
//          |  +-----------------+
//          |     |
//          |     |  +-----------------+
//          |     +--| X68030Interrupt | (X68030 の割り込みコントローラ)
//          |     |  +-----------------+
//          |     |
//          |     |  +---------------+
//          |     +--| LunaInterrupt |   (LUNA-I の割り込みコントローラ)
//          |        +---------------+
//          |
//          +-- PEDECDevice   (X68030 の Lv1 担当コントローラ)
//          +-- SysCtlrDevice (LUNA-88K の割り込みコントローラ)

#include "interrupt.h"
#include "dmac.h"
#include "lance.h"
#include "pedec.h"
#include "m68030.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "newsctlr.h"
#include "nmi.h"
#include "scc.h"
#include "sio.h"
#include "spc.h"
#include "sysclk.h"

// グローバル参照用
InterruptDevice *gInterrupt;

//
// 仮想割り込みコントローラ (共通部)
//

// コンストラクタ(オブジェクト名指定あり)
InterruptDevice::InterruptDevice(const std::string& objname_)
	: inherited(objname_)
{
}

// コンストラクタ(オブジェクト名指定なし)
InterruptDevice::InterruptDevice()
	: InterruptDevice("Interrupt")
{
}

// デストラクタ
InterruptDevice::~InterruptDevice()
{
	gInterrupt = NULL;
}

// リセット
void
InterruptDevice::ResetHard(bool poweron)
{
	intmap = IntmapSentinel;
	memset(&counter, 0, sizeof(counter));
}

// 子デバイスが割り込み信号線をアサートした
void
InterruptDevice::AssertINT(Device *source)
{
	uint32 oldmap;
	uint32 srcmap;

	oldmap = intmap;
	srcmap = GetIntmap(source);
	intmap |= srcmap;

	if ((oldmap ^ intmap) != 0) {
		// 立ち上がりエッジでカウント
		counter[__builtin_clz(srcmap)]++;
		ChangeInterrupt();
	}
}

// 子デバイスが割り込み信号線をネゲートした
void
InterruptDevice::NegateINT(Device *source)
{
	uint32 oldmap;
	uint32 srcmap;

	oldmap = intmap;
	srcmap = GetIntmap(source);
	intmap &= ~srcmap;

	if ((oldmap ^ intmap) != 0) {
		ChangeInterrupt();
	}
}

//
// 仮想割り込みコントローラ (m680x0 共通部)
//

// CPU コアからの割り込みアクノリッジを受け付けるグローバル関数
int
m68030_interrupt_acknowledge(int lv)
{
	return gInterrupt->InterruptAcknowledge(lv);
}

// コンストラクタ
M680x0Interrupt::M680x0Interrupt()
{
}

// デストラクタ
M680x0Interrupt::~M680x0Interrupt()
{
}

// リセット
void
M680x0Interrupt::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);
	ipl = 0;
}

void
M680x0Interrupt::ChangeInterrupt()
{
	int newipl;

	// intmap は最下位ビットを常に立ててあるので 0 にならない
	newipl = (uint)(31 - __builtin_clz(intmap)) / 4;

	// 割り込みレベルが変わったら、新しいレベルを通知
	if (newipl != ipl) {
		ipl = newipl;
		gMPU->Interrupt(ipl);
	}
}


//
// X68030 仮想割り込みコントローラ
//

// コンストラクタ
X68030Interrupt::X68030Interrupt()
{
	monitor.func = ToMonitorCallback(&X68030Interrupt::MonitorUpdate);
	monitor.SetSize(25, 11);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
X68030Interrupt::~X68030Interrupt()
{
}

// デバイスから Intmap 値を引く
uint32
X68030Interrupt::GetIntmap(Device *source) const
{
	if (__predict_true(source == gMFP)) {
		return IntmapMFP;
	}
	if (__predict_true(source == gSCC)) {
		return IntmapSCC;
	}
	if (__predict_true(source == gDMAC)) {
		return IntmapDMAC;
	}
	if (__predict_true(source == gPEDEC)) {
		return IntmapPEDEC;
	}
	if (source == gNMI) {
		return IntmapNMI;
	}
	VMPANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジに応答する
int
X68030Interrupt::InterruptAcknowledge(int lv)
{
	switch (lv) {
	 case 7:
		return AutoVector;

	 case 6:
		return gMFP->InterruptAcknowledge();

	 case 5:
		return gSCC->InterruptAcknowledge();

	 case 3:
		return gDMAC->InterruptAcknowledge();

	 case 1:
		return gPEDEC->InterruptAcknowledge(lv);

	 default:
		break;
	}
	VMPANIC("Unknown interrupt acknowledge lv=%d", lv);
}

void
X68030Interrupt::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "NMI",	IntmapNMI },
		{ "MFP",	IntmapMFP },
		{ "SCC",	IntmapSCC },
		{ "DMAC",	IntmapDMAC },
		{ "PEDEC",	IntmapPEDEC },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	uint intr_mask = gMPU680x0->GetCPU()->reg.intr_mask;

	screen.Clear();

	y = 0;
	//                    0123456789012345678901234
	screen.Print(0, y++, "Lv Device     Count  IM=%d", intr_mask);
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%11" PRIu64, counter[clz]);
		if (lv <= intr_mask) {
			screen.Puts(21, y, "Mask");
		}
		y++;
	}

	// PEDEC 分も一緒に表示したい
	gPEDEC->MonitorUpdate(screen, intr_mask, y);
}


//
// LUNA-I 仮想割り込みコントローラ
//

// コンストラクタ
LunaInterrupt::LunaInterrupt()
{
	monitor.func = ToMonitorCallback(&LunaInterrupt::MonitorUpdate);
	monitor.SetSize(25, 6);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
LunaInterrupt::~LunaInterrupt()
{
}

// デバイスから Intmap 値を引く
uint32
LunaInterrupt::GetIntmap(Device *source) const
{
	if (__predict_true(source == gSIO)) {
		return IntmapSIO;
	}
	if (__predict_true(source == gSysClk)) {
		return IntmapSysClk;
	}
	if (__predict_true(source == gEthernet)) {
		return IntmapLance;
	}
	if (__predict_true(source == gSPC)) {
		return IntmapSPC;
	}
	if (source == gNMI) {
		return IntmapNMI;
	}
	VMPANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジに応答する
int
LunaInterrupt::InterruptAcknowledge(int lv)
{
	// LUNA は全てオートベクタ
	return AutoVector;
}

void
LunaInterrupt::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "NMI",	IntmapNMI },
		{ "SIO",	IntmapSIO },
		{ "Clock",	IntmapSysClk },
		{ "Lance",	IntmapLance },
		{ "SPC",	IntmapSPC },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	uint intr_mask = gMPU680x0->GetCPU()->reg.intr_mask;

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "Lv Device     Count  IM=%d", intr_mask);
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%11" PRIu64, counter[clz]);
		if (lv <= intr_mask) {
			screen.Puts(21, y, "Mask");
		}
		y++;
	}
}


//
// NEWS 仮想割り込みコントローラ
//

// コンストラクタ
NewsInterrupt::NewsInterrupt()
{
	monitor.func = ToMonitorCallback(&NewsInterrupt::MonitorUpdate);
	monitor.SetSize(25, 6);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
NewsInterrupt::~NewsInterrupt()
{
}

// デバイスから Intmap 値を引く
uint32
NewsInterrupt::GetIntmap(Device *source) const
{
	if (__predict_true(source == gNewsCtlr)) {
		return IntmapTimer;
	}
	if (__predict_true(source == gSCC)) {
		return IntmapSCC;
	}
	if (__predict_true(source == gEthernet)) {
		return IntmapLance;
	}
	if (source == gNMI) {
		return IntmapNMI;
	}
	VMPANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジに応答する
int
NewsInterrupt::InterruptAcknowledge(int lv)
{
	// NEWS では 5 の SCC だけベクタで、他はオートベクタ?
	switch (lv) {
	 case 5:
		return gSCC->InterruptAcknowledge();

	 default:
		return AutoVector;
	}
}

// ステータスバイトを読み出す。NewsCtlr から呼ばれる。
uint8
NewsInterrupt::PeekStatus() const
{
	uint8 data = 0;

	if ((intmap & IntmapSIC)) {
		data |= 0x80;
	}
	if ((intmap & IntmapLance)) {
		data |= 0x04;
	}

	return data;
}

void
NewsInterrupt::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "Timer",	IntmapTimer },
		{ "SCC",	IntmapSCC },
		{ "Lance",	IntmapLance },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	uint intr_mask = gMPU680x0->GetCPU()->reg.intr_mask;

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "Lv Device     Count  IM=%d", intr_mask);
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%11" PRIu64, counter[clz]);
		if (lv <= intr_mask) {
			screen.Puts(21, y, "Mask");
		}
		y++;
	}
}
