//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "device.h"

// IODevice クラスへの Read/Write をストリームっぽく扱う。
// Read/Write とも受け渡しはホストバイトオーダの値で行い、
// デバイスへのアクセスはビッグエンディアンで行う。
// アクセス中にバスエラーが起きることは考慮していない。
class IODeviceStream
{
 public:
	// コンストラクタ
	IODeviceStream(IODevice *d) {
		dev = d;
	}
	IODeviceStream(IODevice *d, uint32 a) {
		dev = d;
		addr = a;
	}
	// デストラクタ不要

	// アドレスを設定する
	void SetAddr(uint32 a) {
		addr = a;
	}
	// アドレスを取得する
	uint32 GetAddr() const {
		return addr;
	}

	// 1バイト読み込んでポインタを進める
	uint64 Read8() {
		uint64 data = dev->Read8(addr);
		addr += 1;
		return data;
	}

	// 2バイト読み込んでポインタを進める。addr はアラインしなくてよい
	uint64 Read16() {
		uint64 data;
		data  = Read8() << 8;
		data |= Read8();
		return data;
	}

	// 4バイト読み込んでポインタを進める。addr はアラインしなくてよい
	// データはホストエンディアンで返される。
	uint64 Read32() {
		uint64 data;
		data  = Read16(addr) << 16;
		data |= Read16(addr);
		return data;
	}

	// 1バイト書き込んでポインタを進める
	void Write8(uint32 data)
	{
		dev->Write8(addr, data);
		addr += 1;
	}

	// 2バイト書き込んでポインタを進める。addr はアラインしてなくてよい
	void Write16(uint32 data)
	{
		Write8(data >> 8);
		Write8(data & 0xff);
	}

	// 4バイト書き込んでポインタを進める。addr はアラインしてなくてよい
	void Write32(uint32 data)
	{
		Write16(data >> 16);
		Write16(data & 0xffff);
	}

	// アドレスも指定する版。ポインタは同様に進める
	uint64 Read8(uint32 a) {
		SetAddr(a);
		return Read8();
	}
	uint64 Read16(uint32 a) {
		SetAddr(a);
		return Read16();
	}
	uint64 Read32(uint32 a) {
		SetAddr(a);
		return Read32();
	}
	void Write8(uint32 a, uint32 data) {
		SetAddr(a);
		Write8(data);
	}
	void Write16(uint32 a, uint32 data) {
		SetAddr(a);
		Write16(data);
	}
	void Write32(uint32 a, uint32 data) {
		SetAddr(a);
		Write32(data);
	}

 private:
	IODevice *dev {};
	uint32 addr {};
};
