//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メインメモリ
//

#include "mainram.h"
#include "config.h"
#include "mainapp.h"
#include "mpu.h"
#include "prom.h"

// グローバル参照用
MainRAMDevice *gMainRAM;

// MainRAM はロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス、ワードアクセス時は HLB(), HLW() マクロを使用のこと。
/*static*/ std::unique_ptr<uint8[]> MainRAMDevice::mainram;

// コンストラクタ
MainRAMDevice::MainRAMDevice()
	: inherited("MainRAM")
{
	// アクセスウェイト
	switch (gMainApp.GetVMType()) {
	 case VMType::LUNA1:
		// LUNA-I は1ウェイト。
		read_wait  = 1;
		write_wait = 1;
		break;

	 case VMType::LUNA88K:
		// 取扱説明書p.21 に、CMMU からのアクセスタイムが
		// リード(4byte)で 200ns、バーストリード(16byte)で 320ns
		// ライト(4byte)で 160ns、バーストライト(16byte)で 280ns とある。
		// 25MHz は 40ns/clockで、この数値は CMMU からのアクセス全体の時間
		// らしいので、ここで加算するウェイトはそれから 2クロック引いたもの。
		read_wait  = 3;
		write_wait = 2;
		break;

	 case VMType::X68030:
	 default:
		// X68030 ではシステムポートが設定するのでここでは不要。
		break;
	}
}

// デストラクタ
MainRAMDevice::~MainRAMDevice()
{
	gMainRAM = NULL;
}

bool
MainRAMDevice::Init()
{
	int ram_size_MB;

	const ConfigItem& item = gConfig->Find("ram-size");
	ram_size_MB = item.AsInt();
	if (ram_size_MB < 1) {
		item.Err();
		return false;
	}
	// 下限以外の制約は機種による
	switch (gMainApp.GetVMType()) {
	 case VMType::X68030:
		// メインメモリは 12MB まで。
		// 内部は 8KB 単位で処理できるが設定が 1MB 単位。
		if (ram_size_MB > 12) {
			item.Err("Configurable maximum main memory size is 12 [MB]");
			return false;
		}
		break;

	 case VMType::LUNA1:
		// 16MB までは、実機に合わせて 4MB 単位とする。
		// see http://wiki.netbsd.org/ports/luna68k/luna68k_info/#hardware
		//
		// 16MB を超える増設 (or 魔改造) 部分は 1MB 単位で 255MB までとする。
		// 実 ROM のメモリチェックは 32KB だかそのくらいずつ行われ、
		// 最後に番兵として Null 領域 (BusError ではなく) が必要。
		// 現行の構成をあまり変えない範囲なら厳密には (256MB - 32KB) が
		// 上限だが、設定ファイルからの指定が 1MB 単位なので 255MB。
		if (ram_size_MB <= 16) {
			if (ram_size_MB % 4 != 0) {
				item.Err("For 16MB or less, must be specified in 4 [MB] unit");
				return false;
			}
		} else if (ram_size_MB > 255) {
			item.Err("Configurable maximum RAM size is 255 [MB]");
			return false;
		}
		break;

	 case VMType::LUNA88K:
	 {
		// 64MB までは、実機に合わせて 16MB 単位とする。
		//
		// 64MB を超える増設 (or 魔改造) 部分は未確認。
		if (ram_size_MB <= 64) {
			if (ram_size_MB % 16 != 0) {
				item.Err("For 64MB or less, must be specified in 16 [MB] unit");
				return false;
			}
		} else if (ram_size_MB > 255) {
			item.Err("Configurable maximum RAM size is 255 [MB]");
			return false;
		}

		// PROM 1.20 では 240MB を上限にする。
		// 0x0f00'0000 (ちょうど 240MB 目) へのアクセスでバスエラーが
		// 出ることを CMMU のアドレス変換チェックに使っているため。
		PROMDevice *prom = dynamic_cast<PROMDevice*>(gPROM);
		if (prom && prom->GetROMVer() == 120) {
			if (ram_size_MB > 240) {
				item.Err("Maximum RAM size with PROM 1.20 is 240 [MB]");
				return false;
			}
		}

		break;
	 }

	 case VMType::NEWS:
		// XXX とりあえず
		break;

	 default:
		__unreachable();
	}

	ram_size = ram_size_MB * 1024 * 1024;
	mainram.reset(new uint8[ram_size]);

	// devtable[] への RAM デバイスの配置はこの後 VM::Init 側で行っている

	return true;
}

// 電源オン/リセット
void
MainRAMDevice::ResetHard(bool poweron)
{
	if (poweron) {
		// XXX ノイズを用意する
	}

	// X68030 ではリセットで戻るはず。(LUNA は固定なので関係ない)
	if (gMainApp.IsX68030()) {
		read_wait  = 0;
		write_wait = 0;
	}
}

uint64
MainRAMDevice::Read8(uint32 addr)
{
	gMPU->AddCycle(read_wait);
	uint32 data;
	data = mainram[HLB(addr)];
	putlog(4, "$%08x.B -> $%02x", addr, data);
	return data;
}

uint64
MainRAMDevice::Read16(uint32 addr)
{
	gMPU->AddCycle(read_wait);
	uint32 data;
	data = *(uint16 *)&mainram[HLW(addr)];
	putlog(4, "$%08x.W -> $%04x", addr, data);
	return data;
}

uint64
MainRAMDevice::Read32(uint32 addr)
{
	gMPU->AddCycle(read_wait);
	uint32 data;
	data = *(uint32 *)&mainram[addr];
	putlog(4, "$%08x.L -> $%08x", addr, data);
	return data;
}

uint64
MainRAMDevice::Write8(uint32 addr, uint32 data)
{
	gMPU->AddCycle(write_wait);
	mainram[HLB(addr)] = data;
	putlog(3, "$%08x.B <- $%02x", addr, data);
	return 0;
}

uint64
MainRAMDevice::Write16(uint32 addr, uint32 data)
{
	gMPU->AddCycle(write_wait);
	*(uint16 *)&mainram[HLW(addr)] = data;
	putlog(3, "$%08x.W <- $%04x", addr, data);
	return 0;
}

uint64
MainRAMDevice::Write32(uint32 addr, uint32 data)
{
	gMPU->AddCycle(write_wait);
	*(uint32 *)&mainram[addr] = data;
	putlog(3, "$%08x.L <- $%08x", addr, data);
	return 0;
}

uint64
MainRAMDevice::Peek8(uint32 addr)
{
	return mainram[HLB(addr)];
}

uint64
MainRAMDevice::Poke8(uint32 addr, uint32 data)
{
	if ((int32)data >= 0) {
		mainram[HLB(addr)] = data;
	}
	return 0;
}

// アクセスウェイトを設定 (X68030 でシステムポートから呼ばれる)
void
MainRAMDevice::SetWait(uint32 wait)
{
	read_wait  = wait;
	write_wait = wait;
}

// info->path で指定されたホストの実行ファイルをゲストにロードする。
bool
MainRAMDevice::LoadFromFile(LoadInfo *info)
{
	BootLoader loader(this);

	return loader.LoadFromFile(info);
}

// info->data, size で指定されたバッファを実行ファイルとみなしてゲストに
// ロードする。
bool
MainRAMDevice::LoadFromData(LoadInfo *info)
{
	BootLoader loader(this);

	return loader.LoadFromData(info);
}
