//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// MPU (MC68030)
//

#pragma once

#include "event.h"
#include "mpu.h"
#include "m68030.h"

class MPU680x0Device : public MPUDevice
{
	using inherited = MPUDevice;
 public:
	MPU680x0Device();
	virtual ~MPU680x0Device() override;

	bool Init() override;
	void ResetHard(bool poweron) override;

	// 現在の PPC を取得
	uint32 GetPPC() const override { return RegPPC; }

	// 現在の VBR を取得
	uint32 GetVBR() const override { return RegVBR; }

	// 現在アクセス中の論理アドレス、物理アドレスを取得。
	uint32 GetLaddr() const override { return cpu->bus.laddr; }
	uint32 GetPaddr() const override { return cpu->bus.paddr; }

	// アクセスウェイトを加算
	void AddCycle(int32 cycle) override { cpu->used_cycle += cycle; }

	// 割り込みレベルが変わったことを MPU に通知
	void Interrupt(int level) override;

	// A-Line, F-Line 命令エミュレーションのコールバックを指定する。
	// (Human68k エミュレーションで使用)
	void SetALineCallback(bool (*callback)(m68kcpu*, void*), void *arg);
	void SetFLineCallback(bool (*callback)(m68kcpu*, void*), void *arg);

	// FPU を持っているか
	bool HaveFPU() const {
		return cpu->has_fpu;
	}

	// CPU コアを取得 (主にデバッグ用途)
	m68kcpu *GetCPU() const {
		return cpu;
	}

	// MPU からのアクセスをエミュレートする。
	// アドレスは論理アドレス。
	// 空間は現在のデータ空間。
	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;

	// コアからスケジューラへの動作モード変更通知
	static void NotifyCPUMode(uint32 new_mode);

	// コアからの次命令境界での割り込みチェック指示
	void CheckIntr();

	// 次の実行サイクルを Full にする。
	void MakeNextFull();

	// ホールト時のコアからのコールバック関数
	void Halt();

 private:
	// リセット例外コールバック
	void ResetCallback(Event& ev);

	// 命令実行系コールバック
	void ExecShort(Event& ev);
	void ExecFull(Event& ev);
	void ExecTrace(Event& ev);

	// トレース設定。デバッガから呼び出される。
	void SetTrace(bool trace_on) override;

	// 割り込みイベントコールバック
	void InterruptCallback(Event& ev);

	// 割り込みイベント
	Event intr_event { this };

	// レジスタモニター
	DECLARE_MONITOR_CALLBACK(MonitorUpdateReg);
	Monitor reg_monitor { this };

	// ATC モニター
	DECLARE_MONITOR_CALLBACK(MonitorUpdateATC);
	Monitor atc_monitor { this };

	// 割り込みペンディング
	bool intr_pending {};

	// トレースでコールバックを差し替えることができるように
	// イベントコールバックを変数に入れてある
	EventCallback_t exec_short {};
	EventCallback_t exec_full {};

	m68kcpu *cpu {};
};

#define gMPU680x0	(dynamic_cast<MPU680x0Device *>(gMPU))
