//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// NEWS の I/O 空間担当
//

// NWS-1750
//            +0 +1 +2 +3 +4 +5 +6 +7  +8 +9 +A +B +C +D +E +F'0000
// e000'0000 |
// e010'0000 |
// e020'0000 |
// e030'0000 |
// e040'0000 |
// e050'0000 |
// e060'0000 |
// e070'0000 |
// e080'0000 |
// e090'0000 |
// e0a0'0000 |
// e0b0'0000 |
// e0c0'0000 |            PAR?         FDC         SIC
// e0d0'0000 |KBC         SCC          RTC         LED
// e0e0'0000 |MEM                      DMAC
// e0f0'0000 |ETH
//
// MEM: Lance のメモリ

#include "newsio.h"
#include "busio.h"
#include "config.h"
#include "kbc.h"
#include "lance.h"
#include "mk48t02.h"
#include "newsctlr.h"
#include "scc.h"
#include "subram.h"
#include "vm.h"

// グローバル参照用
NewsIODevice *gNewsIO;

// コンストラクタ
NewsIODevice::NewsIODevice()
	: inherited("NewsIO")
{
	// SIC を無視するオプション(暫定)
	int ignore_sic = gConfig->Find("xxx-news-sic-ignore").AsInt();

	NEWDV(KBC, new KBCDevice());
	NEWDV(RTC, new BusIO_B<MK48T02Device>());
	NEWDV(SCC, new BusIO_B<SCCDevice>());
	NEWDV(Ethernet, new BusIO_W<LanceDevice>());
	NEWDV(SubRAM, new SubRAMDevice(16));
	NEWDV(NopIO, new NopIODevice());

	// まずバスエラーで埋める。
	// XXX バスエラーが起きるかどうかは知らないが
	for (int i = 0; i < table.size(); i++) {
		table[i] = gBusErr;
	}

	// 少ないので手動で置いていく。
	// table[] は $e0c0'0000 以降を 0x10000 (64KB) 単位にしたもの。
#define A(addr)	(addr - 0xe0c0)

	// FDC は現状サポートしていないが (それに機種が違うっぽい?)、
	// NetBSD の locore.s がリセットのためにアクセスしに来るので
	// NewsCtlr でついでに対処してある。
	table[A(0xe0c8)] = gNewsCtlr;	// FDC
	if (ignore_sic) {
		table[A(0xe0cc)] = gNopIO;	// SIC
	}
	table[A(0xe0d0)] = gKBC;
	table[A(0xe0d4)] = gSCC;
	table[A(0xe0d8)] = gRTC;
	table[A(0xe0dc)] = gNewsCtlr;	// LED
	table[A(0xe0e0)] = gSubRAM;
	if (ignore_sic) {
		table[A(0xe0e8)] = gNopIO;	// DMAC
	}
	table[A(0xe0f0)] = gEthernet;
}

// デストラクタ
NewsIODevice::~NewsIODevice()
{
	gNewsIO = NULL;
}

bool
NewsIODevice::Init()
{
	const ConfigItem& olditem = gConfig->Find("xxx-news-sci-ignore");
	const ConfigItem& newitem = gConfig->Find("xxx-news-sic-ignore");

	// 古いオプションだけが指定されていればエラー
	if (olditem.GetFrom() != ConfigItem::FromInitial &&
		newitem.GetFrom() == ConfigItem::FromInitial)
	{
		olditem.Err("The variable name was changed to \"xxx-news-sic-ignore\" "
			"('sci' to 'sic').  It was a typo :-)  "
			"Please catch up your configuration.");
		return false;
	}

	// 両方指定されていて値が異なればエラー
	if (olditem.GetFrom() != ConfigItem::FromInitial &&
		newitem.GetFrom() != ConfigItem::FromInitial &&
		olditem.AsString() != newitem.AsString())
	{
		newitem.Err("Old \"xxx-news-sci-ignore\" is specified and "
			"has a different value.  Please remove old one.");
		return false;
	}

	return true;
}

// アドレスから該当するデバイスを返す
inline IODevice *
NewsIODevice::SearchDevice(uint32 addr) const
{
	// アドレスを table のインデックスに変換。ここに来た時点で
	// $e0'000000 (かそのミラー) の 16MB 内にいることは分かっている。
	addr &= 0x00ff'ffff;
	uint idx = (addr - 0x00c0'0000) >> 16;
	return table[idx];
}

uint64
NewsIODevice::Read8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read8(addr);
}

uint64
NewsIODevice::Read16(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read16(addr);
}

uint64
NewsIODevice::Read32(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Read32(addr);
}

uint64
NewsIODevice::Write8(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write8(addr, data);
}

uint64
NewsIODevice::Write16(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write16(addr, data);
}

uint64
NewsIODevice::Write32(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	return d->Write32(addr, data);
}

uint64
NewsIODevice::Peek8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	return d->Peek8(addr);
}

// このアドレスを担当する次段の IODevice を返す (IOContainerDevice 用)
IODevice *
NewsIODevice::GetDevice(uint32 addr) const
{
	return SearchDevice(addr);
}
