//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// PIO/PPI (i8255)
//

// IODevice
//    |
//    |  +-------------+
//    +--| i8255Device | (8255 としての共通部分)
//       +-------------+
//          |
//          |  +-----------+
//          +--| PPIDevice |  (X68030)
//          |  +-----------+
//          |
//          |  +------------+
//          +--| PIO0Device | (LUNA)
//          |  +------------+
//          |
//          |  +------------+
//          +--| PIO1Device | (LUNA)
//             +------------+

#include "pio.h"
#include "bitops.h"
#include "dipsw.h"
#include "lcd.h"
#include "mainapp.h"
#include "mpu.h"
#include "power.h"
#include "scheduler.h"

// グローバル参照用
PPIDevice *gPPI;
PIO0Device *gPIO0;
PIO1Device *gPIO1;

//
// i8255
//

// コンストラクタ
i8255Device::i8255Device(const std::string& objname_)
	: inherited(objname_)
{
}

// デストラクタ
i8255Device::~i8255Device()
{
}

// リセット
void
i8255Device::ResetHard(bool poweron)
{
	// clears the control register
	// and place ports A, B, and C in input mode.
	// The input latches in ports A, B, and C are not cleared.

	ctrl = 0b10011011;
}

// コントロールポートへの書き込み。
void
i8255Device::WriteCtrl(uint32 data)
{
	if ((data & 0x80)) {
		// 最上位が立っていればモード設定
		ctrl = data;
		putlog(1,
			"Set mode: groupA=%d groupB=%d portA=%s portB=%s portC=%s/%s",
			GetGroupAMode(), GetGroupBMode(),
			PortADir() == DIR_IN ? "IN" : "OUT",
			PortBDir() == DIR_IN ? "IN" : "OUT",
			PortCHDir() == DIR_IN ? "IN" : "OUT",
			PortCLDir() == DIR_IN ? "IN" : "OUT");

		// モード設定すると PC はゼロクリアされる
		for (int pc = 0; pc < 8; pc++) {
			SetPC(pc, 0);
		}
	} else {
		// portC 出力
		//
		// 0  x  x  x  P  P  P  V
		//             |  |  |  +-- 書き込むビットデータ
		//             +--+--+----- 書き込むビット位置

		int pc = (data >> 1) & 7;
		int val = data & 1;
		SetPC(pc, val);
	}
}

//
// X68k PPI
//

// コンストラクタ
PPIDevice::PPIDevice()
	: inherited("PPI")
{
}

// デストラクタ
PPIDevice::~PPIDevice()
{
	gPPI = NULL;
}

uint64
PPIDevice::Read(uint32 offset)
{
	gMPU->AddCycle(19);	// InsideOut p.135

	switch (offset) {
	 case 0:
	 case 1:
		return 0xff;
	 case 2:
		putlog(0, "Read $e9a005 PortC (NOT IMPLEMENTED)");
		return 0xff;
	 case 3:
		putlog(0, "Read $e9a007 Ctrl (NOT IMPLEMENTED)");
		return 0xff;
	 default:
		__unreachable();
	}
}

uint64
PPIDevice::Write(uint32 offset, uint32 data)
{
	gMPU->AddCycle(19);	// InsideOut p.135

	switch (offset) {
	 case 0:
	 case 1:
		break;
	 case 2:	// PPI PortC
		for (int i = 0; i < 8; i++) {
			SetPC(i, data & 1);
			data >>= 1;
		}
		break;
	 case 3:
		WriteCtrl(data);
		break;
	 default:
		__unreachable();
	}
	return 0;
}

uint64
PPIDevice::Peek(uint32 offset)
{
	// XXX
	return 0xff;
}

// PortC 書き込み
void
PPIDevice::SetPC(int pc, int val)
{
	switch (pc) {
	 case 0:	// ADPCM PAN Left Off
		pan_left = !val;
		break;
	 case 1:	// ADPCM PAN Right Off
		pan_right = !val;
		break;
	 case 2:	// ADPCM サンプリングレート
		adpcm_rate = (adpcm_rate & 2) | val;
		break;
	 case 3:	// ADPCM サンプリングレート
		adpcm_rate = (val << 1) | (adpcm_rate & 1);
		break;

	 case 4:	// ジョイスティック #1 コントロール
	 case 5:	// ジョイスティック #2 コントロール
	 case 6:	// ジョイスティック #1 ボタン A
	 case 7:	// ジョイスティック #1 ボタン B
		putlog(1, "SetPC(%d) <- %d (NOT IMPLEMENTED)", pc, val);
		break;
	 default:
		__unreachable();
	}
}

//
// LUNA PIO0
//
// 仕様は 0x49000000 から 4バイトだが、
// LUNA-I   では 0x48000000 から 0x4bffffff まで 4バイト単位のミラーが見える。
// LUNA-88K では 0x48000000 から 0x49ffffff までが
// (こちらはロングワード配置なので) 16バイト単位のミラーに見える。
//
// PIO から読み出せる値のビットの並び順が機種によって異なる。
// LUNA-I   では DIPSW 1番が LSB、8番が MSB の並び順、
// LUNA-88K では DIPSW 1番が MSB、8番が LSB の並び順。
// 値は、並び順に関わらずどちらも
// %0 が false = down = "0"、%1 が true = up = "1" である。
//
//    #1 #2 #3 #4 #5 #6 #7 #8
//  +-------------------------+
//  |    [] [] [] [] [] [] [] | 本体外観
//  | []                      |
//  +-------------------------+
//
// luna-dipsw1 = "01111111"    ← 設定ファイル
//                ||||||||     ↓ このクラスのメンバ変数
//                |||||||+- dipsw1[7] = true;
//                ||||||+-- dipsw1[6] = true;
//                |||||+--- dipsw1[5] = true;
//                ||||+---- dipsw1[4] = true;
//                |||+----- dipsw1[3] = true;
//                ||+------ dipsw1[2] = true;
//                |+------- dipsw1[1] = true;
//                +-------- dipsw1[0] = false;
//
//              b7   b6   b5   b4   b3   b2   b1   b0
//            +----+----+----+----+----+----+----+----+
// LUNA-I     | #8 | #7 | #6 | #5 | #4 | #3 | #2 | #1 |
// PortA      |  1 |  1 |  1 |  1 |  1 |  1 |  1 |  0 | %1=UP, %0=DOWN
//            +----+----+----+----+----+----+----+----+
//
//              b7   b6   b5   b4   b3   b2   b1   b0
//            +----+----+----+----+----+----+----+----+
// LUNA-88K   | #1 | #2 | #3 | #4 | #5 | #6 | #7 | #8 |
// PortA      |  0 |  1 |  1 |  1 |  1 |  1 |  1 |  1 | %1=UP, %0=DOWN
//            +----+----+----+----+----+----+----+----+
//
// LUNA-I:
// #1-1 up なら UNIX をロードして移行。down なら ROM モニタで起動。
// #1-2 up ならディスプレイコンソール。down ならシリアルコンソール。
// #1-3 up - color display, down - force to have monochrome display
// #1-4 up - no write verification,
//      down - verification on every harddisk write operation
// #1-5 up - OS is UniOS-U (SystemV/COFF),
//      down - OS is UniOS (4.3BSD/a.out OMAGIC)
// #1-6 down - force monochrome display?
// #1-7 up - boot from local device, down - boot from network
// #1-8 up - normal boot, down - start diagnostics と書いてあるが
//      down にすると画面真っ暗で詳細不明。
//
// #2-x すべて up にすると /boot が自動的にカーネルを起動。一つでも down が
//      あれば /boot がプロンプトで停止。これは ROM ではなく NetBSD のブート
//      ローダの話。
//      どこか由来のソースコードらしいがたぶんビット演算を間違えている。
//
// LUNA-88K:
// #1-1 up なら ROM モニタで停止、down ならマルチユーザブート。
//      OpenBSD カーネルも up なら UserKernelConfig で止まる模様?。
// #1-2 up なら外付けシリアルコンソール、down なら内蔵ビットマップ。
// #1-3..#1-7 は不明?
// #1-8 down なら自己診断後電源オフするようだ。
// #2-x はユーザ(?)

// コンストラクタ
PIO0Device::PIO0Device()
	: inherited("PIO0")
{
	// LUNA-I と LUNA-88K とでポートのビット並びが逆。どうして…
	if (gMainApp.IsLUNA88K()) {
		msb_first = true;
	} else {
		msb_first = false;
	}

	monitor.func = ToMonitorCallback(&PIO0Device::MonitorUpdate);
	monitor.SetSize(22, 11);
	monitor.Regist(ID_MONITOR_PIO0);
}

// デストラクタ
PIO0Device::~PIO0Device()
{
	gPIO0 = NULL;
}

// リセット
void
PIO0Device::ResetHard(bool poweron)
{
	// XXX 本当は XP から割り込みが上がったタイミングとからしいけど、
	//     XP が出来るまでは当面リセットでラッチにしておく。
	for (int i = 0; i < 8; i++) {
		dipsw1[i] = gDipsw->Get(i);
		dipsw2[i] = gDipsw->Get(i + 8);
	}
}

// XXX ResetHard 未調査

uint64
PIO0Device::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:
		// PortA
		data = GetPA();
		putlog(1, "DIP-SW1 -> $%02x", data);
		return data;

	 case 1:
		// PortB
		data = GetPB();
		putlog(1, "DIP-SW2 -> $%02x", data);
		return data;

	 case 2:
		// PortC
		data = GetPC();
		putlog(1, "PortC -> $%02x", data);
		return data;

	 case 3:
		data = ReadCtrl();
		putlog(1, "Ctrl -> $%02x", data);
		return data;

	 default:
		__unreachable();
	}
}

uint64
PIO0Device::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// PIO0 PortA: DIPSW1 (入力のみ)
	 case 1:	// PIO0 PortB: DIPSW2 (入力のみ)
		return 0;

	 case 2:	// PIO0 PortC: ホスト割り込み制御
		for (int i = 0; i < 8; i++) {
			SetPC(i, data & 1);
			data >>= 1;
		}
		return 0;

	 case 3:	// コントロール
		WriteCtrl(data);
		return 0;

	 default:
		__unreachable();
	}
}

uint64
PIO0Device::Peek(uint32 offset)
{
	switch (offset) {
	 case 0:
		return GetPA();

	 case 1:
		return GetPB();

	 case 2:
		return GetPC();

	 case 3:
		return ReadCtrl();

	 default:
		__unreachable();
	}
}

void
PIO0Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "PortA(DIPSW1):%c%c%c%c%c%c%c%c",
		(dipsw1[0] ? '1' : '0'),
		(dipsw1[1] ? '1' : '0'),
		(dipsw1[2] ? '1' : '0'),
		(dipsw1[3] ? '1' : '0'),
		(dipsw1[4] ? '1' : '0'),
		(dipsw1[5] ? '1' : '0'),
		(dipsw1[6] ? '1' : '0'),
		(dipsw1[7] ? '1' : '0'));
	screen.Print(0, y++, "PortB(DIPSW2):%c%c%c%c%c%c%c%c",
		(dipsw2[0] ? '1' : '0'),
		(dipsw2[1] ? '1' : '0'),
		(dipsw2[2] ? '1' : '0'),
		(dipsw2[3] ? '1' : '0'),
		(dipsw2[4] ? '1' : '0'),
		(dipsw2[5] ? '1' : '0'),
		(dipsw2[6] ? '1' : '0'),
		(dipsw2[7] ? '1' : '0'));
	screen.Puts(0, y++, "PortC");
	screen.Puts(1, y++, TA::OnOff(xp_reset), "b7: XP Reset");
	screen.Puts(1, y++, TA::OnOff(parity),   "b6: Parity");
	screen.Puts(1, y++, TA::Disable,         "b5: ---");
	screen.Puts(1, y++, TA::OnOff(int5en),   "b4: Int5 Enable");
	screen.Puts(1, y++, TA::OnOff(int5rq),   "b3: Int5 Request");
	screen.Puts(1, y++, TA::OnOff(int1en),   "b2: Int1 Enable");
	screen.Puts(1, y++, TA::Disable,         "b1: ---");
	screen.Puts(1, y++, TA::OnOff(int1rq),   "b0: Intq Request");
}

// PortA (DIP-SW 1) 読み込み
uint32
PIO0Device::GetPA() const
{
	uint32 data = 0;
	for (int i = 0; i < 8; i++) {
		data |= ((int)dipsw1[i]) << i;
	}
	if (msb_first) {
		data = bitrev(data);
	}
	return data & 0xff;
}

// PortB (DIP-SW 2) 読み込み
uint32
PIO0Device::GetPB() const
{
	uint32 data = 0;
	for (int i = 0; i < 8; i++) {
		data |= ((int)dipsw2[i]) << i;
	}
	if (msb_first) {
		data = bitrev(data);
	}
	return data & 0xff;
}

// PortC 読み込み
uint32
PIO0Device::GetPC() const
{
	uint32 data;

	data = 0x00	// XXX 空きビットは何が読めるか?
		 | (int1rq ? 0x01 : 0)
		 | (int1en ? 0x04 : 0)
		 | (int5rq ? 0x08 : 0)
		 | (int5en ? 0x10 : 0)
		 | (parity ? 0x40 : 0)
		 | (xp_reset?0x80 : 0);
	return data;
}

// PortC 書き込み
void
PIO0Device::SetPC(int pc, int val)
{
	switch (pc) {
	 case 0:
		if ((val ^ int1rq)) {
			int1rq = val;
			putlog(0, "INT1 Interrupt Request %d (NOT IMPLEMENTED)", int1rq);
		}
		return;
	 case 1:
		// Not Connected.
		return;
	 case 2:
		if ((val ^ int1en)) {
			int1en = val;
			putlog(0, "INT1 Interrupt Enable %d (NOT IMPLEMENTED)", int1en);
		}
		return;
	 case 3:
		if ((val ^ int5rq)) {
			int5rq = val;
			putlog(0, "INT5 Interrupt Request %d (NOT IMPLEMENTED)", int5rq);
		}
		return;
	 case 4:
		if ((val ^ int5en)) {
			int5en = val;
			putlog(0, "INT5 Interrupt Enable %d (NOT IMPLEMENTED)", int5en);
		}
		return;
	 case 5:
		// Not Connected.
		return;
	 case 6:
		// パリティ有効/無効
		// 使わないので無視するが読み返しのために覚えておく
		parity = val;
		return;
	 case 7:
		// XP リセット
		if ((val ^ xp_reset)) {
			xp_reset = val;
			putlog(0, "XP Reset %d (NOT IMPLEMENTED)", xp_reset);
		}
		return;
	}
	VMPANIC("invalid pc=%d", pc);
}


//
// LUNA PIO1
//
// 仕様は 0x4d000000 から 4バイトだが、
// LUNA-I   では 0x4c000000 から 0x4fffffff まで 4バイト単位でミラーが見える。
// LUNA-88K では 0x4c000000 から 0x4dffffff までが
// (こちらはロングワード配置なので) 16バイト単位のミラーに見える。

// コンストラクタ
PIO1Device::PIO1Device()
	: inherited("PIO1")
{
}

// デストラクタ
PIO1Device::~PIO1Device()
{
	gPIO1 = NULL;
}

// リセット
void
PIO1Device::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	// デフォルトで PIO 出力は「1msec 後に電源オフ」という状態なので、
	// 電源オンから 1msec 以内にソフトウェアがこれをキャンセルしないと
	// いけないらしい。どうしてこうなった…。
	gPower->SetSystemPowerOn(false);
}

uint64
PIO1Device::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:
		// PIO1 PortA: LCD データ
		return gLCD->Read();

	 case 1:
		// PIO1 PortB: 無効
		return 0xff;

	 case 2:
		// PortC
		data = gLCD->Get();
		data |= xp_intreq ? 0 : 0x02;
		// 残りのビットは %1 だろうか?
		data |= 0x0d;
		putlog(1, "PortC -> $%02x", data);
		return data;

	 case 3:
		data = ReadCtrl();
		putlog(1, "Ctrl -> $%02x", data);
		return data;

	 default:
		__unreachable();
	}
}

uint64
PIO1Device::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// PIO1 PortA: LCD データ
		gLCD->Write(data);
		return 0;

	 case 1:	// PIO1 PortB: ローカル割り込み
		break;

	 case 2:	// PIO1 PortC: LCD/パワーオフ制御
		for (int i = 0; i < 8; i++) {
			SetPC(i, data & 1);
			data >>= 1;
		}
		return 0;

	 case 3:	// コントロール
		WriteCtrl(data);
		return 0;

	 default:
		__unreachable();
	}
	VMPANIC("$%08X <- $%02X 未サポート書き込み", gMPU->GetPaddr(), data);
	return 0;
}

uint64
PIO1Device::Peek(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:	// PIO1 PortA LCD データ
		return gLCD->Peek();

	 case 1:	// PIO1 PortB 無効
		return 0xff;

	 case 2:	// PIO1 PortC
		data = gLCD->Get();
		data |= xp_intreq ? 0 : 0x02;
		// 残りのビットは %1 だろうか?
		data |= 0x0d;
		return data;

	 case 3:
		return ReadCtrl();

	 default:
		__unreachable();
	}
}

void
PIO1Device::SetPC(int pc, int val)
{
	switch (pc) {
	 case 0:
	 case 1:
	 case 2:
	 case 3:
		// PortC の下半分は入力ポートとして使われることを想定して
		// いるので、ここへの書き込みは起きないだろうし、無視する。
		return;

	 case 4:
		power = val;
		if (power) {
			// 電源オフ取り消し
			putlog(1, "Power-off canceled");
			gPower->SetSystemPowerOn(true);
		} else {
			// 1msec 後に電源オフ
			putlog(1, "Power-off scheduled");
			gPower->SetSystemPowerOn(false);
		}
		return;
	 case 5:
		gLCD->SetRW(val);
		return;
	 case 6:
		gLCD->SetRS(val);
		return;
	 case 7:
		gLCD->SetE(val);
		return;

	 default:
		break;
	}
	VMPANIC("PC%d 未実装", pc);
}
