//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM (垂直 VRAM)
// Lunafb と X68k TVRAM の共通部分
//

#include "planevram.h"
#include "mybswap.h"
#if defined(__x86_64__)
#include <x86intrin.h>
#endif

// グローバル参照用
PlaneVRAMDevice *gPlaneVRAM;

// コンストラクタ
PlaneVRAMDevice::PlaneVRAMDevice(const std::string& objname_,
	int width_, int height_)
	: inherited(objname_)
{
	// ポカ避け。継承クラス側が設定する。
	nplane = -1;

	composite.Create(width_, height_);

	dirty.resize(height_);

	// レンダリング時に使うテーブルを事前に計算。
	InitDeptable();
}

// デストラクタ
PlaneVRAMDevice::~PlaneVRAMDevice()
{
	gPlaneVRAM = NULL;
}

// リセット
void
PlaneVRAMDevice::ResetHard(bool poweron)
{
	// XXX ここ?
	Invalidate();
}

// レンダリング時に使うテーブルを事前に計算。
void
PlaneVRAMDevice::InitDeptable()
{
	deptable.reset(new uint64[256]);

	// %abcdefgh の8ビットを
	// リトルエンディアンホストでは
	// %0000000h'0000000g'0000000f'0000000e'0000000d'0000000c'0000000b'0000000a
	// ビッグエンディアンホストでは
	// %0000000a'0000000b'0000000c'0000000d'0000000e'0000000f'0000000g'0000000h
	// の64ビットに伸張する。
	// (8ビットを各バイトの最下位ビットに対応させる)
	for (int i = 0; i < 256; i++) {
		uint64 res = 0;
		for (int bit = 0; bit < 8; bit++) {
			int n = bit * 8;
#if BYTE_ORDER == LITTLE_ENDIAN
			n = 56 - n;
#endif
			res |= (uint64)((i >> bit) & 1) << n;
		}
		deptable[i] = res;
	}
}

// 全画面の更新フラグを立てる
void
PlaneVRAMDevice::Invalidate()
{
	std::fill(dirty.begin(), dirty.end(), 1);
}

// 全画面の更新フラグを下ろす
// VM スレッドから呼ばれる。
void
PlaneVRAMDevice::ClearDirty()
{
	std::fill(dirty.begin(), dirty.end(), 0);
}

// 更新情報を取得。
// VM スレッドから呼ばれる。
ModifyInfo
PlaneVRAMDevice::GetModify() const
{
	ModifyInfo mod;
	memcpy(&mod.bits[0], &dirty[0], mod.bits.size());
	// TODO: 互換性のために n_dirty 計算。廃止するかも。
	for (int i = 0; i < dirty.size(); i++) {
		if (dirty[i]) {
			mod.n_dirty++;
		}
	}

	return mod;
}

// 画面合成。
// レンダラスレッドから呼ばれる。
bool
PlaneVRAMDevice::Render(BitmapRGB& bitmap, const ModifyInfo& modify)
{
	bool updated = false;

	// VRAM に更新があれば composite を更新
	if (modify.IsDirty()) {
		RenderVRAMToComposite(modify);
		updated = true;
	}

	// composite に更新があるか(update)、
	// 無条件に更新するか(invalidate2) なら bitmap を更新。
	if (updated || modify.invalidate2) {
		RenderCompositeToRGB(bitmap, modify);
		updated = true;
	}

	return updated;
}

// VRAM から composite 画面を合成する。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderVRAMToComposite(const ModifyInfo& modify)
{
	const int width  = composite.GetWidth();
	const int height = composite.GetHeight();
	const int planesz = (width / 8) * height;

	// 1bpp はパレット的にはほぼ 4bpp と同じ扱いになる。
	const int nplane4 = (nplane == 1) ? 4 : nplane;

	// テキスト合成画面を描く
	for (int y = 0; y < height; y++) {
		uint8 mod = modify.bits[y];
		if (mod == 0) {
			continue;
		}

		const uint32 *src = (const uint32 *)&mem[0];
		src += y * width / 32;
		uint64 *dst = (uint64 *)composite.GetPtr(0, y);

		for (int x = 0; x < width / 32; x++) {
			std::array<uint32, 8> data;

			// VRAM はホストエンディアンに関係なく 32bit で読み込んだ
			// 時の MSB が左端ピクセル。
			// ここでは左側から 8ピクセル(8ビット)ずつ 4回処理する
			// のを最下位バイトから順に右シフトで行いたいので、
			// (バイト内のビット順は維持したまま) バイトスワップする。
			//
			//          31     24  23     16  15      8   7      0
			//        +----------+----------+----------+----------+
			// VRAM   |+00 .. +07|+08 .. +15|+16 .. +23|+24 .. +31|
			//        +----------+----------+----------+----------+
			//  ↓
			//        +----------+----------+----------+----------+
			// data[] |+24 .. +31|+16 .. +23|+08 .. +15|+00 .. +07|
			//        +----------+----------+----------+----------+
			if (__predict_false(nplane == 1)) {
				// LUNA 1bpp の場合残りの3プレーンは ff で埋める。
				// 実際にも存在しないプレーンは ff が読み出せて
				// カラーインデックスは $e か $f の二択になるので。
				data[0] = bswap32(src[0]);
				data[1] = 0xffffffff;
				data[2] = 0xffffffff;
				data[3] = 0xffffffff;
			} else {
				int offset = planesz / sizeof(uint32);
				for (int plane = 0; plane < nplane; plane++) {
					data[plane] = bswap32(src[plane * offset]);
				}
			}
			src++;

			// data[] の下位 8bit ずつを取り出して deptable[] で変換。
			// p0..p3 は1バイト x 8個。
			//
			// data[] = %xxxxxxxx'xxxxxxxx'xxxxxxxx'ABCDEFGH
			// ↓
			// (リトルエンディアンホストの場合)
			// p.H = %0000000H'0000000G'0000000F'0000000E
			// p.L = %0000000D'0000000C'0000000B'0000000A
			for (int b = 0; b < 4; b++) {
				uint64 cc = 0;

				for (int plane = 0; plane < nplane4; plane++) {
					uint64 p;
					p = deptable[data[plane] & 0xff];
					p <<= plane;
					cc |= p;
					data[plane] >>= 8;
				}

				// cc (64bit = 8バイト) は各バイトが各ピクセルの
				// カラーインデックスになっているので
				// BitmapI8 に 8バイト一気に書き込める。
				// (エンディアンの影響は deptable[] で処理してある)
				*dst++ = cc;
			}
		}
	}
}

// composite 画面とパレット情報から RGB 画面を合成する。
// modify は composite 上での変更箇所を示している。
// レンダラスレッドから呼ばれる。
void
PlaneVRAMDevice::RenderCompositeToRGB(BitmapRGB& bitmap,
	const ModifyInfo& modify)
{
	const int width  = bitmap.GetWidth();
	const int height = bitmap.GetHeight();

	// view_mod は表示領域 view における更新ラスタ
	BitmapI8 view(width, height);
	std::vector<uint8> view_mod(height);

	// この view に対する modify を計算する
	if (__predict_false(modify.invalidate2)) {
		std::fill(view_mod.begin(), view_mod.end(), 1);
	} else {
		// Y scroll 方向の modify 計算
		int sy = yscroll;
		for (int y = 0; y < height; y++, sy++) {
			if (sy >= composite.GetHeight()) {
				sy -= composite.GetHeight();
			}
			view_mod[y] = modify.bits[sy];
		}

		// 右がはみ出す場合は、はみ出した先のラスタも加味する。
		if (xscroll + view.GetWidth() > composite.GetWidth()) {
			sy = yscroll;
			for (int y = 0; y < height; y++, sy++) {
				if (sy >= composite.GetHeight()) {
					sy -= composite.GetHeight();
				}
				int sy2 = GetWrapY(sy);
				view_mod[y] |= modify.bits[sy2];
			}
		}
	}

	// composite からスクロールを加味した表示領域 view を作る
	int sy = yscroll;
	for (int y = 0; y < height; y++, sy++) {
		if (sy >= composite.GetHeight()) {
			sy -= composite.GetHeight();
		}
		if (view_mod[y] == 0) {
			continue;
		}

		uint8 *v = view.GetPtr(0, y);
		uint8 *c = composite.GetPtr(xscroll, sy);
		if (xscroll + view.GetWidth() <= composite.GetWidth()) {
			memcpy(v, c, view.GetWidth());
		} else {
			// はみ出しの折り返し
			int len1 = composite.GetWidth() - xscroll;
			memcpy(v, c, len1);
			v += len1;
			int sy2 = GetWrapY(sy);
			c = composite.GetPtr(0, sy2);
			int len2 = view.GetWidth() - len1;
			memcpy(v, c, len2);
		}
	}

	// view (I8) を Bitmap (RGB) に展開する
	for (int y = 0; y < view_mod.size(); y++) {
		if (view_mod[y] == 0) {
			continue;
		}

		const uint8 *s8 = view.GetPtr(0, y);
		uint32 *d = (uint32 *)bitmap.GetPtr(0, y);

		const uint32 *s = (const uint32 *)s8;
		const uint32 *send = (const uint32 *)(s8 + view.GetWidth());
		for (; s < send;) {
			// s は BitmapI8 なのでビッグエンディアン並びに相当

#if defined(__AVX2__) && !defined(NO_AVX)	// オプション -mavx2 が必要
			// 35.1msec

			// 8ピクセルずつ処理する

			// idx0 = $00000000'00000000'I7I6I5I4'I3I2I1I0
			__m128i idx0 = _mm_loadl_epi64((const __m128i *)s);
			s += 2;

			// idx  = $000000I7'000000I6'000000I5'000000I4'
			//         000000I3'000000I2'000000I1'000000I0
			__m256i idx = _mm256_cvtepu8_epi32(idx0);

			// 8個の 32bit インデックスから XBGR を取得
			// c  = $00B7G7R7'00B6G6R6'00B5G5R5'00B4G4R4'
			//       00B3G3R3'00B2G2R2'00B1G1R1'00B0G0R0
			__m256i c = _mm256_i32gather_epi32(
				(const int *)palette, idx, 4);

			// 256bit リニアに使えれば下詰めで済んだんだが、
			// 上位 128bit と下位 128bit をまたいでシャッフルは
			// 出来なくて、かつこの後のストアは連続領域への
			// 書き込みになるので、中央でつなげておく。うーん。
			// c  = $00000000'B7G7R7B6'G6R6B5G5'R5B4G4R4'
			//       B3G3R3B2'G2R2B1G1'R1B0G0R0'00000000
			__m256i shuff = _mm256_set_epi8(
				-1,-1,-1,-1, 14,13,12,10,  9, 8, 6, 5,  4, 2, 1, 0,
				14,13,12,10,  9, 8, 6, 5,  4, 2, 1, 0, -1,-1,-1,-1
			);
			__m256i res = _mm256_shuffle_epi8(c, shuff);

			// 先頭(下位)と末尾(上位)の 32bit ずつを除いたところを
			// 書き出す。
			__m256i mask = _mm256_set_epi32(
				0, -1, -1, -1,  -1, -1, -1, 0
			);
			_mm256_maskstore_epi32((int *)(d - 1), mask, res);
			d += 6;
#else
			// 46.5msec

			// 4ピクセルずつ処理する

			uint32 cc4 = *s++;
			// 左側ピクセルが下位バイトになるよう並び替える
			// (le32toh() はビッグエンディアンなら bswap32 の意)
			cc4 = le32toh(cc4);

			// 4ピクセルをそれぞれ xBGR32 に変換
			uint32 c0 = palette[ cc4        & 0xff].xbgr;
			uint32 c1 = palette[(cc4 >>  8) & 0xff].xbgr;
			uint32 c2 = palette[(cc4 >> 16) & 0xff].xbgr;
			uint32 c3 = palette[ cc4 >> 24        ].xbgr;

			// xBGR32 4つ(16バイト)を RGB24 3つ(12バイト)に変換して
			// ロングワードで書き出す。
			// c0   = $x0B0G0R0
			// c1   = $x1B1G1R1
			// c2   = $x2B2G2R2
			// c3   = $x3B3G3R3
			//
			// d[0] = $00B0G0R0 | $R1000000 = $R1B0G0R0
			// d[1] = $0000B1G1 | $G2R20000 = $G2R2B1G1
			// d[2] = $000000B2 | $B3G3R300 = $B3G3R3B2

			// (le32toh() はビッグエンディアンなら bswap32 の意)
			*d++ = le32toh((c0      ) | (c1 << 24));
			*d++ = le32toh((c1 >>  8) | (c2 << 16));
			*d++ = le32toh((c2 >> 16) | (c3 <<  8));
#endif
		}
	}
}
