//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM (垂直 VRAM)
// Lunafb と X68k TVRAM の共通部分
//

#pragma once

#include "device.h"
#include "renderer.h"
#include <vector>

class PlaneVRAMDevice : public IODevice
{
	using inherited = IODevice;

 public:
	PlaneVRAMDevice(const std::string& objname_, int width_, int height_);
	virtual ~PlaneVRAMDevice() override;

	void ResetHard(bool poweron) override;

	// レンダリング時に使うテーブルを事前に計算する。
	void InitDeptable();

	// プレーン数を取得
	int GetPlaneCount() const { return nplane; }

	// 更新情報を取得
	ModifyInfo GetModify() const;

	// 更新フラグをクリア
	void ClearDirty();

	// 画面を合成する
	bool Render(BitmapRGB& dst, const ModifyInfo& modify);

	// 合成ビットマップを取得する
	const BitmapI8& GetComposite() const { return composite; }

	// 座標位置の情報を出力 (GUI 用)
	virtual void UpdateInfo(TextScreen&, int x, int y) const = 0;

 protected:
	// 全画面の更新フラグを立てる
	void Invalidate();

	// 垂直 VRAM から BitmapI8 バッファを作成する
	void RenderVRAMToComposite(const ModifyInfo& modify);

	// BitmapI8 バッファから BitmapRGB を作成する。
	void RenderCompositeToRGB(BitmapRGB& dst, const ModifyInfo& modify);

	// X 方向にはみ出したときに対応する仮想画面の Y 座標を返す。
	virtual int GetWrapY(int) const = 0;

	// ビットマッププレーン。
	// メモリを確保する時のサイズ計算のため型を uint8 にしているが、
	// 運用はロングワードでホストバイトオーダー。
	// Lunafb では 1枚 256KBで 1 or 4枚。TVRAM では 512KB。
	std::unique_ptr<uint8[]> mem {};

	// プレーン数 (TVRAM なら 4 固定、LUNA は設定による)
	int nplane {};

	// テキスト画面の合成ビットマップ
	BitmapI8 composite {};

	// 仮想画面の更新フラグ。仮想 Y 座標に対応するので
	// Lunafb も TVRAM も 1024個。
	std::vector<uint8> dirty {};

	// レンダリング時に使う変換テーブル
	std::unique_ptr<uint64[]> deptable {};

	// ホストパレット
	const ColorXBGR *palette {};

	// スクロール
	int xscroll {};
	int yscroll {};
};

extern PlaneVRAMDevice *gPlaneVRAM;
