//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// 電源周り
//

#include "power.h"
#include "mainapp.h"
#include "mfp.h"
#include "scheduler.h"
#include "sync.h"
#include "uimessage.h"

// グローバル参照用
PowerDevice *gPower;

// コンストラクタ
PowerDevice::PowerDevice()
	: inherited("Power")
{
	event.func = ToEventCallback(&PowerDevice::PowerCallback);
	event.time = 1_msec;
	event.Regist("Power Off");

	gScheduler->ConnectMessage(MessageID::RESET, this,
		ToMessageCallback(&PowerDevice::ResetCallback));
	gScheduler->ConnectMessage(MessageID::POWEROFF_RESTART, this,
		ToMessageCallback(&PowerDevice::PowerOffCallback));
	gScheduler->ConnectMessage(MessageID::POWER_BUTTON, this,
		ToMessageCallback(&PowerDevice::PowerButtonCallback));
}

// デストラクタ
PowerDevice::~PowerDevice()
{
	gPower = NULL;
}

// リセットをスケジューラに指示する。(外部から呼ばれる用)
void
PowerDevice::MakeResetHard()
{
	gScheduler->SendMessage(MessageID::RESET);
}

// リスタートをスケジューラに指示する。(外部から呼ばれる用)
void
PowerDevice::MakeRestart()
{
	gScheduler->SendMessage(MessageID::POWEROFF_RESTART);
}

// 電源ボタンを押す。 (外部から呼ばれる用)
void
PowerDevice::PowerButton()
{
	gScheduler->SendMessage(MessageID::POWER_BUTTON);
}

// リセットのメッセージコールバック
void
PowerDevice::ResetCallback(MessageID msgid, uint32 arg)
{
	DoResetHard();
}

// 電源オフのメッセージコールバック
void
PowerDevice::PowerOffCallback(MessageID msgid, uint32 arg)
{
	DoPowerOff(msgid);
}

// 電源ボタンのメッセージコールバック
void
PowerDevice::PowerButtonCallback(MessageID msgid, uint32 arg)
{
	DoPowerButton();
}

// 電源オン操作 (Device::ResetHard ではない)
void
PowerDevice::DoPowerOn()
{
	putlog(1, "PowerOn");

	if (ispower == false) {
		// 全デバイス電源オン
		ispower = true;
		for (auto obj : gMainApp.GetObjects()) {
			auto dev = dynamic_cast<Device *>(obj);
			if (dev) {
				dev->ResetHard(true);
			}
		}

		// 動作モード変更
		gSync->RequestPowerOffMode(false);
	}
}

// リセット操作 (Device::ResetHard ではない)
void
PowerDevice::DoResetHard()
{
	putlog(1, "Reset");

	if (ispower) {
		// 全デバイスをリセット
		for (auto obj : gMainApp.GetObjects()) {
			auto dev = dynamic_cast<Device *>(obj);
			if (dev) {
				dev->ResetHard(false);
			}
		}
	}
}

// 電源オフ操作 (Device::PowerOff ではない)
void
PowerDevice::DoPowerOff(MessageID msgid)
{
	putlog(1, "PowerOff");

	assertmsg(msgid == MessageID::POWEROFF_EXIT ||
	          msgid == MessageID::POWEROFF_RESTART,
		"msgid=%d", (int)msgid);

	if (ispower) {
		// 全デバイス電源オフ
		for (auto obj : gMainApp.GetObjects()) {
			auto dev = dynamic_cast<Device *>(obj);
			if (dev) {
				dev->PowerOff();
			}
		}
		ispower = false;

		// 動作モード変更
		gSync->RequestPowerOffMode(true);
	}

	// 電源オフ後の動作は引数によって異なる
	if (msgid == MessageID::POWEROFF_RESTART) {
		// 時間を始め直す
		gScheduler->StartTime();
		// 再び電源オンにする (リスタート)
		DoPowerOn();
	} else {
		// そうでなければ、GUI に終了を通知
		UIMessage::Post(UIMessage::APPEXIT);
	}
}

// 電源ボタン操作
void
PowerDevice::DoPowerButton()
{
	putlog(1, "PowerButton");

	if (gMainApp.IsX68030()) {
		// X68030 の電源ボタンはオルタネート動作
		power_button = !power_button;
		Change();
		// MFP の POWSW 信号は負論理
		gMFP->SetPowSW(!power_button);
	} else {
		// LUNA の電源ボタンはモーメンタリ動作
		power_button = true;
		Change();
		power_button = false;
	}
}

// RP5C15 の ALARM_OUT を受け取る
void
PowerDevice::SetAlarmOut(bool alarm_out_)
{
	alarm_out = alarm_out_;
	Change();
}

// システム内の電源オン信号を受け取る。
//                __
// X68030 だと PW ON/OFF 信号(の反転)、
// LUNA だと PIO1 の POWF* 信号で、どちらの場合も true がオン。
void
PowerDevice::SetSystemPowerOn(bool poweron)
{
	system_poweron = poweron;
	Change();
}

// 電源状態の変更。
// power_button、alarm_out、system_poweron のいずれかの変更で呼ぶこと。
void
PowerDevice::Change()
{
	bool old_power = ispower;
	bool new_power = power_button | alarm_out | system_poweron;

	if (old_power == false && new_power) {
		DoPowerOn();
	} else if (old_power && new_power == false) {
		if (gMainApp.IsX68030()) {
			DoPowerOff(MessageID::POWEROFF_EXIT);
		} else {
			gScheduler->RestartEvent(event);
		}
	}

	bool old_led = led;
	if (gMainApp.GetVMType() == VMType::X68030) {
		// X68030 の電源制御部の回路による。
		led = power_button | alarm_out;
	} else {
		// LUNA の場合は電源の状態そのまま?
		led = ispower;
	}
	if (old_led != led) {
		// LED の状態が変わったことを UI に通知
		UIMessage::Post(UIMessage::LED);
	}
}

// LUNA の電源オフが信号状態の変化に対し 1msec 遅れて応答することを
// エミュレートするためのイベントコールバック
void
PowerDevice::PowerCallback(Event& ev)
{
	if (system_poweron == false) {
		DoPowerOff(MessageID::POWEROFF_EXIT);
	}
}
