//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// レンダラ
//

#include "renderer.h"
#include "lunafb.h"
#include "sync.h"
#include "uimessage.h"

// グローバル参照用
Renderer *gRenderer;

// コンストラクタ
Renderer::Renderer()
	: inherited("Renderer")
{
	monitor.func = ToMonitorCallback(&X68030Renderer::MonitorUpdate);
	monitor.SetSize(68, 33);
	monitor.Regist(ID_MONITOR_RENDERER);
}

// デストラクタ
Renderer::~Renderer()
{
	TerminateThread();
	gRenderer = NULL;
}

// リセット
void
Renderer::ResetHard(bool poweron)
{
	next_refresh_rtime = gSync->GetRealTime();

	// 統計情報をクリア
	render_time.Reset();
	render_exec = 0;
	last_mod_blocks = 0;
	total_mod_blocks = 0;
	total_invalidate2 = 0;
}

// 電源オフ
void
Renderer::PowerOff()
{
	if (enable) {
		// レンダリングスレッドに指示
		std::unique_lock<std::mutex> lock(mtx);
		request |= REQ_REND_POWEROFF;
		cv.notify_one();
	}
}

// 合成処理イネーブル
// 一度だけ呼び出すことができる
void
Renderer::Enable()
{
	assert(enable == false);

	enable = true;
	bitmap.Create(width, height);
	// 初期値灰色にしておく
	bitmap.Fill(Color(0x70, 0x70, 0x70));
}

// 画面合成処理 (VM スレッドで CRTC/CRTC2 から呼ばれる)。
// 今は一画面単位で合成を行なっている。ラスター云々は未対応。
void
Renderer::Render()
{
	// CLI 版なら何もしない。
	if (enable == false) {
		return;
	}

	// 更新されていればレンダラに引き渡す
	ModifyInfo mod = gPlaneVRAM->GetModify();
	if (__predict_false(invalidate2 == true)) {
		mod.invalidate2 = true;
	}

	if (mod.IsDirty() || mod.invalidate2) {
		if (modq.Enqueue(mod)) {
			// 更新情報を引き取ったのでクリアする
			gPlaneVRAM->ClearDirty();
			invalidate2 = false;
		} else {
			// VM スレッドが速すぎるため次回に持ち越し
		}
	}

	// スケジューラ都合による間引きとかがあればここらへんで評価?
	// XXX どっちのスレッドでやるか

	// レンダリングスレッドに指示
	std::unique_lock<std::mutex> lock(mtx);
	request |= REQ_REND_NORMAL;
	cv.notify_one();
}

// 全画面更新の指示 (パレット変更やスクロールなど VRAM の変更を伴わないもの)。
// ここでは無効化だけ受け付けておいて、実際の処理はレンダリングの際に行う。
void
Renderer::Invalidate2()
{
	invalidate2 = true;
}

// レンダリングスレッド
void
Renderer::ThreadRun()
{
	for (;;) {
		uint32 req;

		// 何か起きるまで待つ
		{
			std::unique_lock<std::mutex> lock(mtx);
			cv.wait(lock, [&] { return (request != 0); });
			req = request;
			request = 0;
		}

		if ((req & REQ_TERMINATE)) {
			// 終了するので、他のフラグは無視
			break;
		}
		if ((req & REQ_REND_POWEROFF)) {
			// 電源オフ時の描画 (通常描画フラグは無視)

			// 電源オフされると画面を灰色にする。
			// 黒でもいいけど、電源オンで黒いままなのと区別したい。
			bitmap.Fill(Color(0x70, 0x70, 0x70));
			// UI に通知
			UIMessage::Post(UIMessage::RENDER);
			continue;
		}
		if ((req & REQ_REND_NORMAL)) {
			// 通常描画
			if (modq.Dequeue(&modify)) {
				// 更新ブロック数を数える (統計用)
				last_mod_blocks = modify.n_dirty;
				total_mod_blocks += last_mod_blocks;
				if (__predict_false(modify.invalidate2)) {
					total_invalidate2++;
				}

				// レンダラを駆動する。
				// out_dirty は前回 UI に通知した後一度でも更新があれば true。
				render_time.Start();
				out_dirty |= RenderMain();
				render_time.Stop();
				render_exec++;
			}

			if (out_dirty) {
				uint64 rtime = gSync->GetRealTime();
				if (rtime >= next_refresh_rtime) {
					UIMessage::Post(UIMessage::RENDER);
					next_refresh_rtime = rtime + refresh_interval;
					out_dirty = false;
				}
			}
		}
	}
}

// スレッドの終了を指示
void
Renderer::Terminate()
{
	std::unique_lock<std::mutex> lock(mtx);
	request |= REQ_TERMINATE;
	cv.notify_one();
}

// 描画更新(通知)間隔を設定
void
Renderer::SetRefreshInterval(uint64 t)
{
	refresh_interval = t;
}

// モニタ更新
void
Renderer::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int x;
	int y;

	screen.Clear();

	// modify
	screen.Print(0, 0, "Last Update Region");
	const uint8 *m = &modify.bits[0];
	for (int i = 0; i < 32; i++) {
		char buf[33];
		uint8 c;
		for (int j = 0; j < 32; j++) {
			if (__predict_false(*m++ != 0)) {
				c = '*';
			} else {
				c = '.';
			}
			buf[j] = c;
		}

		buf[32] = '\0';
		screen.Print(0, 1 + i, "%4d:%s", i * 32, buf);
	}
	screen.Puts(0, 1, "Y=");

	x = 43;
	y = 0;
	uint64 nsec = render_time.Elapsed();
	screen.Print(x, y++, "RenderTime %3d.%06d sec",
		(int)(nsec / 1000 / 1000 / 1000),
		(int)(nsec / 1000) % (1000 * 1000));

	screen.Print(x, y++, "RenderExec      %9" PRIu64, render_exec);

	y++;
	screen.Print(x, y++, "Last Mod Blks   %9d", last_mod_blocks);
	screen.Print(x, y++, "Total Mod Blks  %9" PRIu64, total_mod_blocks);
	screen.Print(x, y++, "Total Invalid2  %9" PRIu64, total_invalidate2);
}


//
// X68030 レンダラ
//

// コンストラクタ
X68030Renderer::X68030Renderer()
{
	// XXX とりあえず
	width = 768;
	height = 512;
}

// デストラクタ
X68030Renderer::~X68030Renderer()
{
}

// 画面合成
bool
X68030Renderer::RenderMain()
{
	// 今はテキスト画面のみを合成バッファに直接描いている
	return gPlaneVRAM->Render(bitmap, modify);
}


//
// LUNA レンダラ
//

// コンストラクタ
LunaRenderer::LunaRenderer()
{
	width = 1280;
	height = 1024;
}

// デストラクタ
LunaRenderer::~LunaRenderer()
{
}

// 画面合成
bool
LunaRenderer::RenderMain()
{
	return gPlaneVRAM->Render(bitmap, modify);
}


//
// NEWS レンダラ
//

// コンストラクタ
NewsRenderer::NewsRenderer()
{
	// XXX サポートしてない
	width = 640;
	height = 480;
}

// デストラクタ
NewsRenderer::~NewsRenderer()
{
}

// 画面合成
bool
NewsRenderer::RenderMain()
{
	return false;
}
