//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA-88K の ROM エミュレーション
//

// IODevice
//    |
//    +-- ROMDevice (LoadROM()、ウェイト、マスク等を持つ)
//    |     |
//    |     +-- PROMDevice    (LUNA* の PROM)
//    |     +-- IPLROM1Device (X680x0 の IPLROM 後半)
//    |     +-- IPLROM2Device (X680x0 の IPLROM 前半)
//    |     +-- CGROMDevice   (X680x0 の CGROM)
//    |     +-- LunaPROMEmuDevice (LUNA PROM エミュレーションの共通部分)
//    |           |
//    |           +-- Luna1PROMEmuDevice     (LUNA-I の PROM エミュレーション)
//    |           |
//    |           |  +----------------------+
//    |           +--| Luna88kPROMEmuDevice |(LUNA-88K の PROM エミュレーション)
//    |              +----------------------+
//    |
//    +-- PROM0Device   (LUNA* のブートページ切り替え用プロキシ)
//    +-- IPLROM0Device (X680x0 のブートページ切り替え用プロキシ)

#include "romemu_luna88k.h"
#include "interrupt.h"
#include "iodevstream.h"
#include "mainapp.h"
#include "mainram.h"
#include "memorystream.h"
#include "mk48t02.h"
#include "mpu88xx0.h"
#include "pio.h"

static bool luna88k_fline_callback(MPU88xx0Device *cpu, void *arg);

// m88k の命令コードを少し読みやすく書きたい。
// 必要なものは適宜足すこと。
enum class OPr {
	r0,  r1,  r2,  r3,  r4,  r5,  r6,  r7,
	r8,  r9,  r10, r11, r12, r13, r14, r15,
	r16, r17, r18, r19, r20, r21, r22, r23,
	r24, r25, r26, r27, r28, r29, r30, r31,
};
#define r0	OPr::r0
#define r1	OPr::r1
#define r2	OPr::r2
#define r3	OPr::r3
#define r4	OPr::r4
#define r5	OPr::r5
#define r6	OPr::r6
#define r7	OPr::r7
#define r8	OPr::r8
#define r9	OPr::r9
#define r10	OPr::r10
enum OPm5 {
	eq0 = 0x02,
	ge0 = 0x03,
	ne0 = 0x0d,
};
enum OPcr {
	psr = 0x01,
};

#define USC  __unused static constexpr

USC uint32 OP3(uint32 base, uint32 rd, uint32 rs1, uint32 rs2) {
	return (base) | (rd << 21) | (rs1 << 16) | rs2;
}
USC uint32 OP4(uint32 base, uint32 rd, uint32 rs1,
	uint32 w, uint32 o) {
	return (base) | (rd << 21) | (rs1 << 16) | (w << 5) | o;
}

USC uint32 OP_addu(OPr rd, OPr rs1, uint32 imm) {
	return OP3(0x60000000, (uint32)rd, (uint32)rs1, imm);
}
USC uint32 OP_addu(OPr rd, OPr rs1, OPr rs2) {
	return OP3(0xf4006000, (uint32)rd, (uint32)rs1, (uint32)rs2);
}
USC uint32 OP_bcnd(OPm5 m5, OPr rs1, uint32 d16) {
	return OP3(0xe8000000, (uint32)m5, (uint32)rs1, d16 & 0xffff);
}
USC uint32 OP_bsr(uint32 d26) {
	return (0xc8000000 | (d26 & 0x03ffffff));
}
USC uint32 OP_clr(OPr rd, OPr rs1, uint32 w, uint32 o) {
	return OP4(0xf0008000, (uint32)rd, (uint32)rs1, w, o);
}
USC uint32 OP_jmp(OPr rs2) {
	return OP3(0xf400c000, 0, 0, (uint32)rs2);
}
USC uint32 OP_ld(OPr rd, OPr rs1, uint32 imm) {
	return OP3(0x14000000, (uint32)rd, (uint32)rs1, imm);
}
USC uint32 OP_ldcr(OPr rd, uint32 crs) {
	return OP4(0x80004000, (uint32)rd, 0, crs, 0);
}
USC uint32 __unused OP_or(OPr rd, OPr rs1, OPr rs2) {
	return OP3(0xf4005800, (uint32)rd, (uint32)rs1, (uint32)rs2);
}
USC uint32 OP_or(OPr rd, OPr rs1, uint32 imm) {
	return OP3(0x58000000, (uint32)rd, (uint32)rs1, imm);
}
USC uint32 OP_or_u(OPr rd, OPr rs1, uint32 imm) {
	return OP3(0x5c000000, (uint32)rd, (uint32)rs1, imm);
}
USC uint32 OP_rte() {
	return 0xf400fc00;
}
USC uint32 OP_st(OPr rd, OPr rs1, OPr rs2) {
	return OP3(0xf4002400, (uint32)rd, (uint32)rs1, (uint32)rs2);
}
USC uint32 OP_stcr(OPr rs1, uint32 crd) {
	return OP4(0x80008000, 0, (uint32)rs1, crd, (uint32)rs1);
}
USC uint32 OP_doscall(OPr rd, OPr rs1, uint32 imm) {
	return OP3(0xfc000400, (uint32)rd, (uint32)rs1, imm);
}
#define DOS_GETCHAR	(1)
#define DOS_PUTCHAR	(2)


// コンストラクタ
Luna88kPROMEmuDevice::Luna88kPROMEmuDevice()
{
	// ROM の特定アドレスを読み出す人がいるかも知れないので、
	// サイズは実 ROM と同じにしておく。LUNA-I からの類推。
	uint devlen = 256 * 1024;
	imagebuf.reset(new uint8 [devlen]);
	memset(&imagebuf[0], 0, devlen);
	mem = &imagebuf[0];

	mask = devlen - 1;

	// プロンプトを初期化。
	prompt = default_prompt;
}

// デストラクタ
Luna88kPROMEmuDevice::~Luna88kPROMEmuDevice()
{
}

bool
Luna88kPROMEmuDevice::Init()
{
	// 親クラス
	if (inherited::Init() == false) {
		return false;
	}

	// DOS call エミュレーションのコールバックを登録
	gMPU88xx0->SetFLineCallback(luna88k_fline_callback, this);

	MemoryStreamBE ms(imagebuf.get());

	// リセット時のブートページには ROM (これ) が見えている。
	// ROMIO_INIT へのアクセスでブートページを RAM に切り替えて、同時に RAM の
	// 0番地付近のベクタも作り直す。そのため、ここのベクタ1 以降 ($41000008〜)
	// をベクタとして参照することはないので、ベクタ0 の命令列がそのまま使い潰
	// してよい。というか2ワードでは 0番地から ROM まで飛べないので使わないと
	// 面倒くさい。一体どういう設計してんだ…。

	// ここからややアセンブラっぽいインデント

	// _START:
		// ここは 0番地。ここから _ROM_START へジャンプ
		ms.Write32(OP_or_u	(r6, r0, 0x4100));
		ms.Write32(OP_or	(r6, r6, 0x000c));
		ms.Write32(OP_jmp	(r6));

	// _ROM_START: (ここは ROM)
		// RAM に切り替え
		ms.Write32(OP_st	(r0, r6, r0));

		// 元の PSR を r10 に保存し、r9 に割り込み許可の値を作る
		ms.Write32(OP_ldcr	(r10, psr));
		ms.Write32(OP_clr	(r9, r10, 2, 0));	// clear IND,SFRZ

		ms.Write32(OP_or_u	(r3, r0, ROMIO_BASE >> 16));
		ms.Write32(OP_or	(r3, r3, ROMIO_BASE & 0xffff));
		// ROMIO_INIT を読み出す
		ms.Write32(OP_ld	(r2, r3, (ROMIO_INIT - ROMIO_BASE)));
		// 起動先があれば GO
		ms.Write32(OP_bcnd	(ge0, r2, 7));	// 7 = _go

	// _prompt:
		// 疑似 STOP でキー入力を待つ
		ms.Write32(OP_stcr	(r9, psr));		// 割り込み許可
		ms.Write32(OP_bsr	(1));			// 1 = _stop
		// この時点で r1 = _stop

	// _stop:
		ms.Write32(OP_jmp	(r1));

	// _back:
		uint32 back = baseaddr + ms.GetOffset();
		// 割り込み禁止に戻す
		ms.Write32(OP_stcr	(r10, psr));

		ms.Write32(OP_ld	(r2, r3, (ROMIO_ENTRY - ROMIO_BASE)));
		// 起動先がなければループ
		ms.Write32(OP_bcnd	(eq0, r2, -5));	// -5 = _prompt

	// _go:
		// 起動先があれば後始末して
		ms.Write32(OP_ld	(r0, r3, (ROMIO_CLOSE - ROMIO_BASE)));
		// ジャンプ
		ms.Write32(OP_jmp	(r2));

	// ここから割り込みハンドラ。割り込みは SIO のみ許可にしてある。
	// キーボード割り込みが上がったら KEYIN をコールして _back に戻る。
	ms.SetOffset(0x200);
		ms.Write32(OP_ld	(r0, r3, (ROMIO_KEYIN - ROMIO_BASE)));
		// 戻ったところの jmp を抜ける
		ms.Write32(OP_or_u	(r1, r0, back >> 16));
		ms.Write32(OP_or	(r1, r1, back & 0xffff));
		ms.Write32(OP_rte	());

	// ROM GETC エミュレータ
	ms.SetOffset(0x300);
		ms.Write32(OP_doscall(r2, r0, DOS_GETCHAR));
		ms.Write32(OP_jmp	(r1));

	// ROM PUTC エミュレータ
	// 表示する文字は r2 に置かれている。
	ms.SetOffset(0x400);
		ms.Write32(OP_doscall(r0, r2, DOS_PUTCHAR));
		ms.Write32(OP_jmp	(r1));


	// 変なインデントここまで

	return true;
}

// 起動時の処理を行う。
// 次段実行ファイルの読み込みに成功すればそのエントリポイントを返す。
// そうでなければ 0 を返すので、プロンプトに処理を移すこと。
uint32
Luna88kPROMEmuDevice::ROM_Init()
{
	IODeviceStream rs(gMainRAM);
	uint32 entry;

	b_entrypoint = -1;

	// デバイスの初期化 (親クラス)
	InitDevice();

	// 割り込みマスクで SIO だけ許可
	gInterrupt->Write32(0x65000000, 0x40000000);

	rs.SetAddr(0);
	// ベクタテーブルを rte で埋めておく
	for (int i = 0; i < 512 * 2; i++) {
		rs.Write32(OP_rte());
	}
	// +$08: 割り込みベクタ
	rs.Write32(8, 0xc00003fe);		// br	$00001000

	// 割り込みトランポリン (0x1000)
	// ベクタテーブルは各2ワードしかないのに、0 番地付近にあるベクタテーブル
	// から2ワードでは ROM まで飛べない。一体どういう設計してんだ…。
	rs.SetAddr(0x1000);
		rs.Write32(OP_or_u	(r6, r0, 0x4100));
		rs.Write32(OP_or	(r6, r6, 0x0200));
		rs.Write32(OP_jmp	(r6));

	// ROM の機能テーブル
	// 0x00001100 ?
	// 0x00001104 ?
	// 0x00001108 ?
	// 0x0000110c GETC() のアドレス
	// 0x00001110 PUTC(c) のアドレス
	// 0x00001114 ビットマッププレーンの有効状態が入っている。
	//  bit0: plane0 有効
	//  bit1..7: plane1..7 有効
	rs.SetAddr(0x110c);
		rs.Write32(0x41000300);
		rs.Write32(0x41000400);
		rs.Write32(0x00000001);

	// スクリーンを初期化
	InitScreen();

	entry = -1;

	// 表示環境が整ったところで NVRAM を照合
	if (VerifyNVRAM() == false) {
		// PROM 1.20 は初期化するかどうか確認してくるけど、とりあえず。
		InitNVRAM();
		errmsg = "NVRAM Initialized.";
		// NVRAM をクリアしたら DIPSW によらずプロンプトに降りる
		goto done;
	}

	// -X file
	//  DIPSW の状態に関わらず、ファイルをロードし実行する。
	// --load-only file
	//  DIPSW の状態に関わらず、ファイルをロードし、プロンプトで止まる。
	if (gMainApp.exec_file) {
		// ホストのファイルを読み込んでエントリポイントアドレスを
		// この Read アクセスの読み出し値として返す。
		// エラーがあればエラーメッセージは呼び出し先側で表示してある。
		b_entrypoint = LoadHostFile();
		if (gMainApp.load_and_exec) {
			entry = b_entrypoint;
		}
	} else {
		// DIPSW#1-1 が UP なら自動起動、DOWN なら ROM モニタ起動なので、
		// これを真似する。
		if (gPIO0->IsDIPSW11Up()) {
			// NVRAM の設定に従ってブート先を探す。
			errmsg.clear();

			bootinfo_t bootinfo;
			GetNVRAM(bootinfo);
			entry = LoadGuestFile(bootinfo);
		}
	}

 done:
	if (entry == -1) {
		PrintTitle("LUNA88K");
		if (errmsg.empty() == false) {
			Print("** %s\n\n", errmsg.c_str());
		}
		ClearPrompt();
	}

	return entry;
}

// ROM 処理。エントリポイントを返す。
uint32
Luna88kPROMEmuDevice::ROM_Entry()
{
	if (b_entrypoint != -1) {
		return b_entrypoint;
	} else {
		return 0;
	}
}

void
Luna88kPROMEmuDevice::ProcessChar(char asciicode)
{
	if (prompt_idx == 0) {
		// 行入力モード (画面更新もしてある)
		inherited::ProcessChar(asciicode);
	} else {
		// 0 以外ならノンブロッキングモード

		// 入力されたキーを表示
		if (' ' <= asciicode && asciicode < 0x7f) {
			Putc(asciicode);
		}

		switch (prompt_idx) {
		 case PROMPT_NVRAM_UPDATE:
			CommandNvramUpdate2(asciicode);
			break;

		 case PROMPT_NVRAM_NEW:
			CommandNvramNew2(asciicode);
			break;

		 case PROMPT_NVRAM_INIT:
			CommandNvramInit2(asciicode);
			break;
		}

		// プロンプトを戻す
		prompt = default_prompt;
		prompt_idx = 0;
		prompt_y = cursor_y;
		ClearPrompt();
	}
}

// 入力コマンド処理 (入力行が揃ったところで呼ばれる)。
//
// b					現在のパラメータで起動
// b ...				パラメータを指定して起動
// nvram				全パラメータを表示
// nvram [<name>]		name のみ表示?
// nvram <name> <val>	値を設定
// nvram - <name>		削除
// nvram--				NVRAM 初期化

void
Luna88kPROMEmuDevice::Command()
{
	std::vector<char> buf(inputbuf.length() + 1);
	std::vector<std::string> arg;

	// 1ワードごとに分解
	strcpy(&buf[0], inputbuf.c_str());
	char *p = &buf[0];
	for (char *ap; (ap = strsep(&p, " ")) != NULL; ) {
		if (*ap != '\0') {
			arg.push_back(std::string(ap));
		}
	}
	if (loglevel >= 1) {
		for (int i = 0; i < arg.size(); i++) {
			putmsgn("arg[%d]=|%s|", i, arg[i].c_str());
		}
	}

	// 空行
	if (arg.size() == 0) {
		return;
	}

	// 1ワード目がコマンド
	if (arg[0] == "h") {
		CommandH(arg);
		return;
	}
	if (arg[0] == "b") {
		CommandB(arg);
		return;
	}
	if (arg[0] == "nvram") {
		CommandNvram(arg);
		return;
	}

	Print("** Unknown command: %s\n", arg[0].c_str());
}

// "h" コマンド
// 引数なしなら、全体のヘルプメッセージ。
// 引数ありなら、個別のヘルプメッセージ。
void
Luna88kPROMEmuDevice::CommandH(const std::vector<std::string>& arg)
{
	if (arg.size() < 2) {
		// ヘルプメッセージ
 mainhelp:
		Print(
"         *** Help Message ***\n"
"  b      : load and execute program\n"
"  h      : show this message\n"
"  nvram  : show/set nvram parameters\n"
		);
		return;
	}

	const std::string& key = arg[1];
	if (key == "nvram") {
		Print(
"         *** Help Message ***\n"
" nvram [<symbol>]       : show\n"
" nvram <symbol> <value> : set\n"
" nvram - <symbol>       : delete\n"
" nvram --               : initialize\n"
		);
		return;
	} else {
		goto mainhelp;
	}
}

// "b" コマンド
// 引数なしなら NVRAM の情報に従ってプログラムをロードして実行。
// 引数ありのほうは未実装。
void
Luna88kPROMEmuDevice::CommandB(const std::vector<std::string>& arg)
{
	errmsg.clear();

	// XXX とりあえず boot を起動するだけ
	bootinfo_t bootinfo;
	GetNVRAM(bootinfo);
	b_entrypoint = LoadGuestFile(bootinfo);
	if (b_entrypoint == -1) {
		Print("** %s\n", errmsg.c_str());
	}
}

// "bh" コマンド (勝手に拡張)
// -X オプションで指定されたホストプログラムをロードして実行。
void
Luna88kPROMEmuDevice::CommandBH()
{
	// 指定されてたら読み込む
	if (gMainApp.exec_file == NULL) {
		Print("** Cannot loaded; host program is not specified.\n");
		return;
	}
	b_entrypoint = LoadHostFile();
}

// "nvram" コマンド
// nvram             .. 全表示
// nvram <key>       .. 表示
// nvram <key> <val> .. 追加/更新
// nvram - <key>     .. 削除
// nvram --          .. 初期化
void
Luna88kPROMEmuDevice::CommandNvram(const std::vector<std::string>& arg)
{
	switch (arg.size()) {
	 case 1:
		// 引数なし
		CommandNvramShowAll();
		break;
	 case 2:
		// 引数1つ
		if (arg[1] == "--") {
			CommandNvramInit();
		} else {
			CommandNvramShow(arg);
		}
		break;
	 case 3:
		// 引数2つ
		if (arg[1] == "-") {
			CommandNvramDelete(arg);
		} else {
			CommandNvramUpdate(arg);
		}
		break;
	 default:
		Print("** Syntax error\n");
	}
}

// NVRAM は 32バイトを1エントリとする。
// [0] はヘッダに相当し、フォーマットは別記。
// [1] 以降が key (16バイト)、val(16バイト) の並び。こんな感じ。
//
// 00000020  62 6f 6f 74 5f 64 65 76 69 63 65 00 00 00 00 00 |boot_device.....|
// 00000030  73 64 00 00 00 00 00 00 00 00 00 00 00 00 00 00 |sd..............|
//
// key, val とも '\0' で終端もしくは16文字で終端のようだ。
// val に17文字を設定しようとすると怒られる。
// key が "" ならこのエントリは空きで val は無視、
// 空きエントリは無視して次へ (FAT のディレクトリエントリみたいな感じ)。
// val が "" は有効。
//
// オフセット 0x540 から始まるエントリが最後で、これ以上追加しようとすると
// "nvram set error: no area" (PROM 1.20) と言われる。
// LUNA-88K にはないが LUNA-I では 0x560 にぽつんと ENADDR があったので、
// その辺りとの互換性だろうか。

// NVRAM 全項目表示
void
Luna88kPROMEmuDevice::CommandNvramShowAll()
{
	putmsg(1, "nvram ShowAll");
	for (uint32 offset = 0x20; offset < 0x560; offset += 0x20) {
		std::string key = nvram->PeekString(offset);
		if (key.empty()) {
			continue;
		}
		std::string val = nvram->PeekString(offset + 16);
		Print("%-15s : %s\n", key.c_str(), val.c_str());
	}
}

// NVRAM 指定項目を表示
void
Luna88kPROMEmuDevice::CommandNvramShow(const std::vector<std::string>& arg)
{
	assert(arg.size() >= 2);
	const std::string& key = arg[1];

	putmsg(1, "nvram Show |%s|", key.c_str());
	uint32 offset = FindNVRAM(key);
	if (offset > 0) {
		std::string val = nvram->PeekString(offset + 16);
		Print("%-15s : %s\n", key.c_str(), val.c_str());
	} else {
		Print("** Not found\n");
	}
}

// NVRAM 項目を追加/更新
void
Luna88kPROMEmuDevice::CommandNvramUpdate(const std::vector<std::string>& arg)
{
	// キーがあれば更新、なければ追加。
	// いずれも事前にプロンプトで確認する。

	assert(arg.size() >= 3);
	nvram_key = arg[1];
	nvram_val = arg[2];
	putmsg(1, "nvram Update |%s|%s|", nvram_key.c_str(), nvram_val.c_str());

	if (nvram_key.size() > 16) {
		Print("** Key too long\n");
		return;
	}
	if (nvram_val.size() > 16) {
		Print("** Val too long\n");
		return;
	}

	nvram_offset = FindNVRAM(nvram_key);
	if (nvram_offset > 0) {
		// キーがすでにあれば更新。
		// Update2() に申し送りするのは nvram_offset と nvram_val のみでいい。

		std::string oldval = nvram->PeekString(nvram_offset + 16);
		prompt = string_format("Update %s : \"%s\" -> \"%s\" (Y/[N]):",
			nvram_key.c_str(), oldval.c_str(), nvram_val.c_str());
		prompt_idx = PROMPT_NVRAM_UPDATE;
	} else {
		// キーがなければ追加
		// New2() に申し送りするのは nvram_offset, nvram_key, nvram_val の3つ。

		// 先に空き領域をチェックしている
		nvram_offset = FindNVRAM("");
		if (nvram_offset == 0) {
			Print("** No area available\n");
			return;
		}

		prompt = string_format("Set new %s : \"%s\" (Y/[N]):",
			nvram_key.c_str(), nvram_val.c_str());
		prompt_idx = PROMPT_NVRAM_NEW;
	}

	// 入力待ちの状態で戻る
	// (prompt_idx != 0 なのでキー入力後に指定の関数へ飛んで続きを実行する)
}

// NVRAM 項目を更新 (キー入力後)
void
Luna88kPROMEmuDevice::CommandNvramUpdate2(char asciicode)
{
	assert(prompt_idx == PROMPT_NVRAM_UPDATE);
	assert(nvram_offset > 0);
	assert(nvram_val.size() <= 16);

	Putc('\n');

	if (asciicode != 'y' && asciicode != 'Y') {
		Print("** Not updated\n");
		return;
	}

	if (nvram->WriteString(nvram_offset + 16, nvram_val) == false) {
		Print("** Write error\n");
		return;
	}
	WriteNVRAMCsum();

	Print("Updated\n");
}

// NVRAM 項目を追加 (キー入力後)
void
Luna88kPROMEmuDevice::CommandNvramNew2(char asciicode)
{
	assert(prompt_idx == PROMPT_NVRAM_NEW);
	assert(nvram_offset > 0);
	assert(!nvram_key.empty());
	assert(nvram_key.size() <= 16);
	assert(nvram_val.size() <= 16);

	Putc('\n');

	if (asciicode != 'y' && asciicode != 'Y') {
		Print("** Not added\n");
		return;
	}

	if (nvram->WriteString(nvram_offset, nvram_key) == false
	 || nvram->WriteString(nvram_offset + 16, nvram_val) == false) {
		Print("** Write error\n");
		return;
	}
	WriteNVRAMCsum();

	Print("Added\n");
}

// NVRAM 項目削除
//
// 削除する時は確認がないらしい… (PROM 1.20)
// 成功しても何も言わないらしい… (PROM 1.20)
// 追加削除に比べてなんか雑じゃない?
void
Luna88kPROMEmuDevice::CommandNvramDelete(const std::vector<std::string>& arg)
{
	assert(arg.size() >= 3);
	assert(arg[1] == "-");
	const std::string& key = arg[2];
	putmsg(1, "nvram Delete |%s|", key.c_str());

	if (key.size() > 16) {
		Print("** key too long\n");
		return;
	}

	uint32 offset = FindNVRAM(key);
	if (offset == 0) {
		Print("** Not found\n");
		return;
	}

	// キーを消せば削除になる (値のほうは消してないようだ)
	nvram->WriteString(offset, "");
	WriteNVRAMCsum();
}

// NVRAM から key を探してそのエントリのオフセットを返す。
// key == "" なら空きエントリを探す。
// 見付からなければ 0 を返す (オフセット 0 はパラメータには使わないため)
uint32
Luna88kPROMEmuDevice::FindNVRAM(const std::string& key) const
{
	for (uint32 offset = 0x20; offset < 0x560; offset += 0x20) {
		std::string k = nvram->PeekString(offset);
		if (k == key) {
			return offset;
		}
		// XXX 同じキーが複数あったらどうなるか
	}
	return 0;
}

// NVRAM 初期化
void
Luna88kPROMEmuDevice::CommandNvramInit()
{
	putmsg(1, "nvram Init");

	prompt = "Initialize (Y/[N]):";
	prompt_idx = PROMPT_NVRAM_INIT;

	// 入力待ちの状態で戻る
}

// NVRAM 初期化 (キー入力後)
void
Luna88kPROMEmuDevice::CommandNvramInit2(char asciicode)
{
	assert(prompt_idx == PROMPT_NVRAM_INIT);

	Putc('\n');

	if (asciicode != 'y' && asciicode != 'Y') {
		Print("** Not initialized\n");
		return;
	}

	InitNVRAM();
	Print("** Initialized\n");
}

// NVRAM を初期化する
bool
Luna88kPROMEmuDevice::InitNVRAM()
{
	nvram->ClearAll();

	nvram->WriteString(0x0004, "<nv>");
	nvram->WriteString(0x0020, "boot_device");
	nvram->WriteString(0x0030, "sd");
	nvram->WriteString(0x0040, "boot_spec");
	nvram->WriteString(0x0050, "0");
	nvram->WriteString(0x0060, "boot_unit");
	nvram->WriteString(0x0070, "0");
	nvram->WriteString(0x0080, "boot_partition");
	nvram->WriteString(0x0090, "0");
	nvram->WriteString(0x00a0, "boot_filename");
	nvram->WriteString(0x00b0, "vmunix");
	nvram->WriteString(0x00c0, "main_mem_check");
	nvram->WriteString(0x00d0, "1");
	nvram->WriteString(0x00e0, "3port_mem_check");
	nvram->WriteString(0x00f0, "1");

	// チェックサム再計算
	WriteNVRAMCsum();

	return true;
}

// NVRAM のブート設定を bootinfo に取得する。
void
Luna88kPROMEmuDevice::GetNVRAM(bootinfo_t& bootinfo)
{
	// dkunit (装置番号) はたぶん boot_unit (0..)
	// dkpart (パーティション) は boot_partition (0..)
	// dkfile (ファイル名) は boot_filename
	// sd 以外のデバイスは未対応
	bootinfo.dkunit = nvram->Peek(0x0070) - '0';
	bootinfo.dkpart = nvram->Peek(0x0090) - '0';
	bootinfo.dkfile = nvram->PeekString(0x00b0);
	putmsg(1, "%s unit=%d", __func__, bootinfo.dkunit);
	putmsg(1, "%s part=%c", __func__, bootinfo.dkpart + 'a');
	putmsg(1, "%s file=|%s|", __func__, bootinfo.dkfile.c_str());
}

// コールバック関数 (C のグローバル関数)
bool
luna88k_fline_callback(MPU88xx0Device *cpu, void *arg)
{
	Luna88kPROMEmuDevice *dev = (Luna88kPROMEmuDevice *)arg;
	return dev->FLineCallback(cpu);
}

// コールバック関数 (こっちが実体)
bool
Luna88kPROMEmuDevice::FLineCallback(MPU88xx0Device *cpu)
{
	// 111111_DDDDDSSSSS_000001_00nnnnnnnn
	//
	// 引数は rS1 から順番に格納。
	// m68k 的なバイト値やワード値の場合は下位詰めしてあるが、不要な上位
	// ビットの処置はゲストコードよりホストコードで行うほうが楽なので、
	// すべてこちらで責任持って処理すること。
	// 戻り値がある DOS call では戻り値は rD に格納。

	uint32 D   = (cpu->reg.opX >> 21) & 0x1f;
	uint32 S1  = (cpu->reg.opX >> 16) & 0x1f;
	uint32 num =  cpu->reg.opX        & 0xff;

	switch (num) {
	 case 0x00:	// _EXIT
		break;

	 case 0x01:	// _GETCHAR
		cpu->reg.r[D] = Getc();
		break;

	 case 0x02:	// _PUTCHAR
		Putc(cpu->reg.r[S1++] & 0xff);
		break;

	 default:
		putlog(0, "DOS call #$%02x (NOT IMPLEMENTED)", num);
		cpu->reg.r[D] = 0xffffffff;
		break;
	}

	// 存在しないコール番号、未実装コールも含めて処理済みとする。
	return true;
}
