//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// NEWS の ROM エミュレーション (-X のみ対応)
//

#include "romemu_news.h"
#include "mainram.h"
#include "memorystream.h"
#include "mpu680x0.h"
#include "newsfb.h"

// コンストラクタ
NewsROMEmuDevice::NewsROMEmuDevice()
	: inherited("PROM")
{
	// 適当
	uint devlen = 4096;
	imagebuf.reset(new uint8 [devlen]);
	memset(&imagebuf[0], 0, devlen);
	mem = &imagebuf[0];

	mask = devlen - 1;
}

// デストラクタ
NewsROMEmuDevice::~NewsROMEmuDevice()
{
}

// 初期化
bool
NewsROMEmuDevice::Init()
{
	MemoryStreamBE ms(imagebuf.get());
	ms.Write32(0x00000400);		// +00: リセットベクタ SP
	ms.Write32(0xf8000400);		// +04: リセットベクタ PC
	// ベクタ
	for (int i = 2; i < 256; i++) {
		ms.Write32(0);
	}
	ms.SetOffset(32 * 4);
	ms.Write32(0xf8000500);		// Trap#0 で ROM のサービスコール

	// リセット時に実行する命令
	ms.SetOffset(0x400);
	ms.Write16(0x2039);			//	move.l	(ROMIO_INIT),d0
	ms.Write32(ROMIO_INIT);
	ms.Write16(0x2040);			//	move.l	d0,a0	; ジャンプ先を a0 にセット
	ms.Write16(0x4ed0);			//	jmp		(a0)	; ジャンプ

	// ROM サービスコール
	ms.SetOffset(0x500);
	ms.Write16(0x0c80);			//	cmpi.l	#3,d0
	ms.Write32(0x00000003);		//
	ms.Write16(0x6708);			//	beq		@f
	ms.Write16(0x2039);			//	move.l	(ROMIO_SYS),d0
	ms.Write32(ROMIO_SYS);		//
	ms.Write16(0x4e73);			//	rte
								//@@:				; Read は対応してないので
	ms.Write32(0x4e722500);		//	stop	#0x2500	; STOP しつつ
	ms.Write16(0x60fa);			//	bra		@b		; 無限ループ。
	ms.Write16(0x4e73);			//	rte

	return true;
}

void
NewsROMEmuDevice::ResetHard(bool poweron)
{
	// 起動時に RAM にリセットベクタを書き込んでおく。(手抜き)
	for (int i = 0; i < 8; i++) {
		gMainRAM->Write8(i, imagebuf[i]);
	}
}

uint64
NewsROMEmuDevice::Read32(uint32 addr)
{
	switch (addr) {
	 case ROMIO_INIT:
		return ROM_Init();

	 case ROMIO_SYS:
		return ROM_Sys();

	 default:
		break;
	}

	return inherited::Read32(addr);
}

uint32
NewsROMEmuDevice::ROM_Init()
{
	LoadInfo info;

	// -X file
	if (gMainApp.exec_file) {
		// ホストのファイルを読み込んでエントリポイントアドレスを
		// この Read アクセスの読み出し値として返す。
		// エラーがあればエラーメッセージは呼び出し先側で表示してある。
		info.path = gMainApp.exec_file;
		if (gMainRAM->LoadFromFile(&info) == false) {
			// XXX どうすべ
			PANIC("** Couldn't load specified host program.\n");
			return -1;
		}
		putmsg(1, "Host program loaded.  Entry point = $%08x\n", info.entry);
	}

	// ROMCONS のために VBR を設定
	m68kcpu *cpu = gMPU680x0->GetCPU();
	RegVBR = 0xf8000000;

	// %d2 には MARK_END (たぶん最後の番地)
	// %d4 にメモリサイズ。
	// %d5 には読み込んだファイル名の先頭番地?
	// %d6 に bootdev、NetBSD/news68k の場合は
	// bootdev = $MCxUHPTT
	//            || |||++ Type? (SD=0)
	//            || ||+-- Partition?
	//            || |+--- Host: 機種? (NEWS1700 は 0)
	//            || +---- Unit?
	//            |+------ Controller?
	//            +------- magic ($5)
	// 下位28ビット(7桁) は ROM から渡されるようだ。
	// %d7 は boothowto。
	RegD(2) = info.sym;
	RegD(4) = gMainRAM->GetSize();
	RegD(6) = 0x50000000;
	RegD(7) = 0;

	return info.entry;
}

uint32
NewsROMEmuDevice::ROM_Sys()
{
	m68kcpu *cpu = gMPU680x0->GetCPU();

	if (RegD(0) == SYS_write) {
		uint32 a6 = RegA(6);
		uint32 fd = gMPU680x0->Read32(a6 + 8);
		uint32 buf = gMPU680x0->Read32(a6 + 12);
		uint32 len = gMPU680x0->Read32(a6 + 16);

		for (int i = 0; i < len; i++) {
			uint64 ch = gMPU680x0->Read8(buf++);
			if ((int64)ch < 0) {
				// XXX ?
				return 0;
			}
			if (ch == '\r') {
				continue;
			}
			gNewsfb->Putc(ch);
		}
		(void)fd;
		return len;
	}

	VMPANIC("Unknown service call $%02x", RegD(0));
	return 0;
}
