//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// X680x0 の各種 ROM
//

// IODevice
//    |
//    +-- ROMDevice (LoadROM()、ウェイト、マスク等を持つ)
//    |     |
//    |     +-- PROMDevice       (LUNA* の PROM)
//    |     |
//    |     |  +---------------+
//    |     +--| IPLROM1Device | (X680x0 の IPLROM 後半)
//    |     |  +---------------+
//    |     |
//    |     |  +---------------+
//    |     +--| IPLROM2Device | (X680x0 の IPLROM 前半)
//    |     |  +---------------+
//    |     |
//    |     |  +-------------+
//    |     +--| CGROMDevice |   (X680x0 の CGROM)
//    |     |  +-------------+
//    |     |
//    |     +-- LunaPROMEmuDevice (LUNA PROM エミュレーションの共通部分)
//    |           |
//    |           +-- Luna1PROMEmuDevice   (LUNA-I の PROM エミュレーション)
//    |           +-- Luna88kPROMEmuDevice (LUNA-88K の PROM エミュレーション)
//    |
//    +-- PROM0Device      (LUNA* のブートページ切り替え用プロキシ)
//    |
//    |  +---------------+
//    +--| IPLROM0Device | (X680x0 のブートページ切り替え用プロキシ)
//       +---------------+

#include "romimg_x68k.h"
#include "builtinrom.h"
#include "config.h"
#include "mainapp.h"
#include "vm.h"

// グローバル参照用
CGROMDevice *gCGROM;
IPLROM0Device *gIPLROM0;
IPLROM1Device *gIPLROM1;
ROMDevice *gIPLROM2;		// IPLROM2/ROM30Emu で共有するため

//
// IPLROM0
//
// IPLROM0 は IPLROM(1) のブートページ担当。
// これ自身はメモリ領域を持たず IPLROM1 へのプロキシとして動作する。
//
// X680x0 のブートページは $000000..$00ffff の 64KB で、
// ここへのアクセスは $ff0000..$ffffff (IPLROM1 の後半 64KB) が見える。
// ROM へアクセスがあればブートページを RAM に切り替える。
// そのため、RAM 側も ROM 側もこいつが応対する必要がある。

// コンストラクタ
IPLROM0Device::IPLROM0Device()
	: inherited("IPLROM(BOOT)")
{
	ClearAlias();
	AddAlias("IPLROMBOOT");
}

// デストラクタ
IPLROM0Device::~IPLROM0Device()
{
	gIPLROM0 = NULL;
}

bool
IPLROM0Device::Init()
{
	iplrom1 = gIPLROM1;
	return true;
}

// リセット
void
IPLROM0Device::ResetHard(bool poweron)
{
	gVM->SwitchBootPageToROM(this);
}

// Read の共通処理。
// アドレスのデコードと RAM への切り替え判定を行う。
uint32
IPLROM0Device::ReadDecoder(uint32 addr)
{
	// ROM アドレスがアクセスされたら切り替える
	if (addr >= 0xff0000) {
		gVM->SwitchBootPageToRAM(this);
	}

	// どちらにしてもアクセスは ROM のアドレス
	return addr | 0xff0000;
}

uint64
IPLROM0Device::Read8(uint32 addr)
{
	addr = ReadDecoder(addr);
	return iplrom1->Read8(addr);
}

uint64
IPLROM0Device::Read16(uint32 addr)
{
	addr = ReadDecoder(addr);
	return iplrom1->Read16(addr);
}

uint64
IPLROM0Device::Read32(uint32 addr)
{
	addr = ReadDecoder(addr);
	return iplrom1->Read32(addr);
}

// Write は RAM 領域へも ROM 領域へも発生しないはずなので基底のを使う。

uint64
IPLROM0Device::Peek8(uint32 addr)
{
	addr |= 0xff0000;
	return iplrom1->Peek8(addr);
}


//
// IPLROM(1) 0xfe0000..0xffffff 担当
//

// コンストラクタ
IPLROM1Device::IPLROM1Device()
	: inherited("IPLROM1")
{
	// X680x0 では ROM への書き込みはバスエラー
	write_op = -1;
}

// デストラクタ
IPLROM1Device::~IPLROM1Device()
{
	gIPLROM1 = NULL;
}

bool
IPLROM1Device::Init()
{
	std::string name;

	uint devlen = 0x020000;
	mask = devlen - 1;

	// ベンチマークモードは先に処理。この辺どうするか。
	if (gMainApp.benchmark_mode != 0) {
		name = string_format("BENCH%02d.ROM", gMainApp.benchmark_mode);
		return LoadROM(name.c_str(), devlen);
	}

	const ConfigItem& item = gConfig->Find("iplrom1-image");
	name = item.AsString();
	if (name.empty()) {
		// 指定がなければ、内蔵 IPLROM を使う。
		mem = Builtin::IPLROM30;
		return true;
	}

	// 外部ファイルが指定されれば、それを使う。
	return LoadROM(name.c_str(), devlen);
}


//
// IPLROM(2) 0xfc0000..0xfdffff 担当
//

// コンストラクタ
IPLROM2Device::IPLROM2Device()
	: inherited("IPLROM2")
{
	// X680x0 では ROM への書き込みはバスエラー
	write_op = -1;
}

// デストラクタ
IPLROM2Device::~IPLROM2Device()
{
	gIPLROM2 = NULL;
}

bool
IPLROM2Device::Init()
{
	uint devlen = 0x020000;
	mask = devlen - 1;

	if (gMainApp.benchmark_mode != 0) {
		return true;
	}

	const ConfigItem& item = gConfig->Find("iplrom2-image");
	std::string name = item.AsString();
	if (name.empty()) {
		printf("iplrom2: not yet\n");
		return false;
	}

	// 外部ファイルが指定されれば、それを使う。
	return LoadROM(name.c_str(), devlen);
}


//
// CGROM
//

// コンストラクタ
CGROMDevice::CGROMDevice()
	: inherited("CGROM")
{
	// X680x0 では ROM への書き込みはバスエラー
	write_op = -1;
}

// デストラクタ
CGROMDevice::~CGROMDevice()
{
	gCGROM = NULL;
}

bool
CGROMDevice::Init()
{
	uint devlen = 768 * 1024;
	mask = 1024 * 1024 - 1;

	const ConfigItem& item = gConfig->Find("cgrom-image");
	std::string name = item.AsString();
	if (name.empty()) {
		// 指定がなければ、内蔵 CGROM を使う。
		mem = Builtin::CGROM;
		return true;
	}

	// 外部ファイルが指定されれば、それを使う。
	if (LoadROM(name.c_str(), devlen) == false) {
		// 指定したファイルが読めなければエラー。
		return false;
	}

	// X68030 の CGROM には、X68000 のものには収録されていない 6x12 フォント
	// が収録されている。6x12 フォント以外は(たぶん)同じなのでほとんどの場面
	// で困ることはないが、それ故に忘れた頃に一部の文字だけ表示されなくて
	// 「???」ということになりがちなので、6x12 フォントが収録されてなさそう
	// なら警告を出しておく (処理は続行する)。
	uint8 val = 0;
	for (int i = 0; i < 12; i++) {
		val |= imagebuf[0xbf400 + ('0' * 12) + i];
	}
	if (val == 0) {
		warnx("warning: 6x12 font not found in CGROM30.DAT");
	}

	return true;
}
