//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// RTC (RP5C15)
//

#pragma once

#include "device.h"
#include "monitor.h"
#include "rtc.h"

struct RP5C15
{
	// レジスタ 13, 14, 15 は Bank0/Bank1 どちらも同じものが見える。
	// これらのレジスタは Bank1 だったら Bank0 を見せる。
	//
	// RTC の未使用ビットは、RTC への書き込み時に未使用ビットを
	// mask[] で '0' にマスクしてから変数に書き込むこと。
	union {
		uint8 r[16 + 13];
		struct {
			uint8 sec1;
			uint8 sec10;
			uint8 min1;
			uint8 min10;
			uint8 hour1;
			uint8 hour10;
			uint8 wday;
			uint8 day1;
			uint8 day10;
			uint8 mon1;
			uint8 mon10;
			uint8 year1;
			uint8 year10;
			uint8 mode;
			uint8 test;
			uint8 reset;

			uint8 clksel;
			uint8 adjust;
			uint8 alarm_min1;
			uint8 alarm_min10;
			uint8 alarm_hour1;
			uint8 alarm_hour10;
			uint8 alarm_wday;
			uint8 alarm_day1;
			uint8 alarm_day10;
			uint8 padding1;
			uint8 sel24;
			uint8 leap;
			uint8 padding2;
		} __packed;
	};

	// レジスタマスク (有効ビットを '1' としたもの)
	static const uint8 mask[16 + 13];

	// バンク0
	static const int SEC1			= 0x00;	// $E8A001
	static const int SEC10			= 0x01;	// $E8A003
	static const int MIN1			= 0x02;	// $E8A005
	static const int MIN10			= 0x03;	// $E8A007
	static const int HOUR1			= 0x04;	// $E8A009
	static const int HOUR10			= 0x05;	// $E8A00B
	static const int WDAY			= 0x06;	// $E8A00D
	static const int DAY1			= 0x07;	// $E8A00F
	static const int DAY10			= 0x08;	// $E8A011
	static const int MON1			= 0x09;	// $E8A013
	static const int MON10			= 0x0a;	// $E8A015
	static const int YEAR1			= 0x0b;	// $E8A017
	static const int YEAR10			= 0x0c;	// $E8A019
	static const int MODE			= 0x0d;	// $E8A01B
	static const int TEST			= 0x0e;	// $E8A01D
	static const int RESET			= 0x0f;	// $E8A01F
	// バンク1
	static const int CLKSEL			= 0x10;	// $E8A001
	static const int ADJUST			= 0x11;	// $E8A003
	static const int ALARM_MIN1		= 0x12;	// $E8A005
	static const int ALARM_MIN10	= 0x13;	// $E8A007
	static const int ALARM_HOUR1	= 0x14;	// $E8A009
	static const int ALARM_HOUR10	= 0x15;	// $E8A00B
	static const int ALARM_WDAY		= 0x16;	// $E8A00D
	static const int ALARM_DAY1		= 0x17;	// $E8A00F
	static const int ALARM_DAY10	= 0x18;	// $E8A011
	static const int notused1		= 0x19;	// $E8A013
	static const int SEL24			= 0x1a;	// $E8A015
	static const int LEAP			= 0x1b;	// $E8A017
	static const int notused2		= 0x1c;	// $E8A019
	// バンク1 の 13, 14, 15 はバンク0 と共通
	static const int MODE_BANK1		= 0x1d;	// $E8A01B
	static const int TEST_BANK1		= 0x1e;	// $E8A01D
	static const int RESET_BANK1	= 0x1f;	// $E8A01F

	static const int MODE_TIMER_EN	= 0x08;
	static const int MODE_ALARM_EN	= 0x04;
	static const int MODE_BANK		= 0x01;

	static const int RESET_1Hz		= 0x08;
	static const int RESET_16Hz		= 0x04;
	static const int RESET_TIMER	= 0x02;
	static const int RESET_ALARM	= 0x01;
};

class RP5C15Device : public RTCDevice
{
	using inherited = RTCDevice;
 public:
	RP5C15Device();
	virtual ~RP5C15Device() override;

	bool Init() override;

	uint GetSec() const override;	// 秒(0-59)
	uint GetMin() const override;	// 分(0-59)
	uint GetHour() const override;	// 時(0-23)
	uint GetWday() const override;	// 曜日(0が日曜)
	uint GetMday() const override;	// 日(1-31)
	uint GetMon() const override;	// 月(1-12)
	uint GetYear() const override;	// 年(西暦)
	uint GetLeap() const override;	// うるう年カウンタ(0..3)

 protected:
	// BusIO インタフェース
	static const uint32 NPORT = 16;
	uint64 Read(uint32 offset);
	uint64 Write(uint32 offset, uint32 data);
	uint64 Peek(uint32 offset);

	// 仮想 RTC からのクロック
	void Tick32Hz() override;
	void Tick2Hz() override;
	void Tick1Hz() override;

	void SetSec(uint v) override;
	void SetMin(uint v) override;
	void SetHour(uint v) override;
	void SetWday(uint v) override;
	void SetMday(uint v) override;
	void SetMon(uint v) override;
	void SetYear(uint v) override;
	void SetLeap(uint v) override;

 private:
	void ChangeAlarmOut();

	DECLARE_MONITOR_CALLBACK(MonitorUpdate);

	// レジスタ
	RP5C15 reg {};

	// CLKOUT 端子の状態 (true なら 'H')
	bool clkout {};

	// 分周段から秒カウンタへのカウントアップホールド内部信号。
	bool clock_hold {};

	// CLKOUT 1/60Hz 用の 60秒カウンタ。カレンダとは独立。
	uint32 cnt_60sec {};

	// アラーム端子出力。
	// ALARM 端子は負論理(Low active) だが、ここでは
	// アサート(L) のとき true とする。
	bool alarm_out {};

	// 内部アラーム 16Hz 信号。CLKOUT、alarm_1Hz とは独立。
	bool alarm_16Hz {};
	// 内部アラーム 1Hz 信号。CLKOUT、alarm_16Hz とは独立。
	bool alarm_1Hz {};
	// アラーム時刻一致信号。
	// XXX アラーム機能自体は未実装。
	bool alarm_on {};

	Monitor monitor { this };
};
