//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// SCSI デバイス
//

// ログレベルは
// 1: SCSI コマンドの概要。
//    デバイスの動作のうち大きな変更のもの
// 2: SCSI コマンドの詳細、ダンプ。
//    デバイスの動作のうち些細なもの、あるいは頻繁なもの
// 3: フェーズ遷移
// 4: データも含める

#include "scsicmd.h"
#include "scsidev.h"
#include "config.h"
#include "mainapp.h"
#include "scheduler.h"
#include "uimessage.h"

//
// SCSI デバイス共通部分
//

// コンストラクタ
SCSIDevice::SCSIDevice(const std::string& objname_)
	: inherited(objname_)
{
}

// デストラクタ
SCSIDevice::~SCSIDevice()
{
}


//
// SCSI ホストデバイス
//

// コンストラクタ
SCSIHostDevice::SCSIHostDevice(const std::string& objname_)
	: inherited(objname_)
{
	devtype = SCSI::DevType::Initiator;

	monitor_devs.func = ToMonitorCallback(&SCSIHostDevice::MonitorUpdateDevs);
	// サイズは Create で決まる
	monitor_devs.Regist(ID_MONITOR_SCSIDEVS);

	// バスデバイスを作成する。
	// ただし自身がバスに参加するのは継承側で行う
	// (例えば SPC なら BDID 書き込みで ID が決まった後、など)。
	scsibus.reset(new SCSIBus());
}

// デストラクタ
SCSIHostDevice::~SCSIHostDevice()
{
}

// 設定ファイルに使われるこのホストデバイスを示す接頭語を設定する
// (継承クラス側のコンストラクタで config_name を設定後に呼ぶこと)
void
SCSIHostDevice::SetConfigName(const char *name_)
{
	config_name = name_;
}

bool
SCSIHostDevice::Create()
{
	// 設定を読み込む。
	for (int id = 0; id < 8; id++) {
		const std::string key = string_format("%s-id%d-image",
			GetConfigName(), id);
		const ConfigItem& item = gConfig->Find(key);
		const std::string& image = item.AsString();

		// 空ならデバイスなし
		if (image.length() == 0) {
			// 未接続部分のパラメータは減らしておくか。
			// --show-config するとさすがに無駄に多くて邪魔なので。
			const std::string wikey = string_format("%s-id%d-writeignore",
				GetConfigName(), id);
			const std::string wpkey = string_format("%s-id%d-writeprotect",
				GetConfigName(), id);
			const std::string stkey = string_format("%s-id%d-seektime",
				GetConfigName(), id);
			gConfig->Delete(wikey);
			gConfig->Delete(wpkey);
			gConfig->Delete(stkey);
			continue;
		}

		// ',' より前がデバイス種別 (ここではパスは不要)
		auto v = string_split(image, ',');
		// 比較のため前後の空白を取り除いて小文字にする
		std::string type = string_trim(v[0]);
		type = string_tolower(type);

		SCSI::DevType target_devtype;
		if (type == "hd") {
			target_devtype = SCSI::DevType::HD;
		} else if (type == "cd") {
			target_devtype = SCSI::DevType::CD;
		} else if (type == "mo") {
			target_devtype = SCSI::DevType::MO;
		} else {
			item.Err("Invaild device type '%s'", type.c_str());
			return false;
		}
		target[id].reset(new SCSIDisk(this, id, target_devtype));
		scsibus->Attach(target[id].get(), id);
	}

	return true;
}

// 初期化。
// 継承クラスが connected_devices をセットした後で呼ぶこと。
bool
SCSIHostDevice::Init()
{
	// 初期化ではないけどログレベルがセットできたここでデバッグ表示。
	// -L オプションの処理は Create() の後なので。
	if (loglevel >= 1) {
		for (int id = 0; id < 8; id++) {
			std::string typestr;
			if (target[id]) {
				typestr = SCSI::GetDevTypeName(target[id]->GetDevType());
				putmsgn("%s target #%d %s", GetConfigName(), id,
					typestr.c_str());
			}
		}
	}

	// connected_devices によってモニタの大きさを決定。
	monitor_devs.SetSize(75, (connected_devices.size() - 1) * 4 + 1);

	return true;
}

// バスフリーコールバック (SCSIBus から呼ばれる)
void
SCSIHostDevice::BusFreeCB(int id)
{
	// 実際にはバスフリーを検出したデバイスは各自自身が送出している信号を
	// 下げるのだが、色々面倒なので、ここではホストのバスフリーコールバックが
	// 一括して全部の線を下げる。
	// ただしさすがにログがうるさいので、上がってる線だけ下げる。

	uint32 data;
	while ((data = GetBSY()) != 0) {
		bus->NegateBSY(DecodeID(data));
	}
	while ((data = GetSEL()) != 0) {
		bus->NegateSEL(DecodeID(data));
	}
	if (GetREQ()) {
		NegateREQ();
	}
	if (GetACK()) {
		NegateACK();
	}
	if (GetATN()) {
		NegateATN();
	}
	SetXfer(0);
	SetData(0);
}

// SCSI デバイスモニタ
// (どこでやるのがいいか分からんけど、SCSI デバイスを一括して表示したいので
// とりあえずここで)
void
SCSIHostDevice::MonitorUpdateDevs(Monitor *, TextScreen& screen)
{
	int y;

// 012345678901234567890123456789012345678901234567890123456789012345678901234
// ID6: HD(2048MB)
//      MediumLoaded   Off      LogicalBlock 2048    ImageSize   2,147,483,648
//      RemovalPrevent Off      WriteProtected


	screen.Clear();

	y = 0;
	for (const auto dev : connected_devices) {
		std::string desc = SCSI::GetDevTypeName(dev->GetDevType());
		std::string pathname;

		SCSIDisk *disk = dynamic_cast<SCSIDisk*>(dev);
		if (disk && disk->IsMediumLoaded()) {
			desc += string_format("(%dMB)",
				(int)(disk->GetSize() / 1024 / 1024));
			pathname = disk->GetPathName();
		}
		screen.Print(0, y, "ID%c: %s", '0' + dev->GetMyID(), desc.c_str());

#if 1
		(void)pathname;
#else
		// パス名の非ASCIIどうするか
		if (!pathname.empty()) {
			if (pathname.size() > screen.GetCol() - 16 - 3) {
				pathname = pathname.substr(0, screen.GetCol() - 16 - 3);
				pathname += "...";
			}
			screen.Puts(16, y, pathname.c_str());
		}
#endif
		y++;

		// イニシエータならここまで。(今の所イニシエータとディスクしかない)
		if (disk == NULL) {
			continue;
		}

		const int x1 = 5;
		const int x2 = 29;
		const int x3 = 50;

		// 1行目
		screen.Print(x1, y,
			(disk->IsRemovableDevice() ? TA::Normal : TA::Disable),
			"MediumLoaded   %s",
			disk->IsMediumLoaded() ? "On" : "Off");
		screen.Print(x2, y, "LogicalBlock %d", disk->GetBlocksize());
		std::string imagesize = format_number(disk->GetSize());
		screen.Print(x3, y, "ImageSize   %13s", imagesize.c_str());
		y++;

		// 2行目
		screen.Print(x1, y,
			(disk->IsRemovableDevice() ? TA::Normal : TA::Disable),
			"RemovalPrevent %s",
			disk->IsMediumRemovalPrevented() ? "On" : "Off");

		screen.Puts(x2, y,
			(disk->IsMediumLoaded() ? TA::Normal : TA::Disable),
			disk->GetWriteModeStr());
		y++;

		y++;
	}
}


//
// SCSI ターゲットデバイスの共通部分
//

// コンストラクタ
SCSITarget::SCSITarget(const std::string& objname_, SCSIHostDevice *h, int id)
	: inherited(objname_)
{
	// ホストへのリンク (所有関係のないポインタ)
	host = h;
	myid = id;
}

// デストラクタ
SCSITarget::~SCSITarget()
{
}

// 電源オン
void
SCSITarget::ResetHard(bool poweron)
{
	if (poweron) {
		ClearSense();
		cmdseq.clear();
		cmdlen = 0;
		cmd.reset();
		bytecount = 0;
		lastbyte = false;
	}
}

// バスリセットされた (SCSIBus から呼ばれる)
void
SCSITarget::BusResetCB()
{
	// センスキーをクリア
	ClearSense();

	// リセットされたので持ってる信号を全部解放。
	// 一応自分が変えていい信号線だけを落とす。
	switch (GetPhase()) {
	 case SCSI::Phase::BusFree:
		break;
	 case SCSI::Phase::Arbitration:
	 case SCSI::Phase::Selection:
	 case SCSI::Phase::Reselection:
		NegateSEL();
		NegateBSY();
		break;
	 case SCSI::Phase::Transfer:
		NegateBSY();
		break;
	}
}

// セレクションで選択された (SCSIBus から呼ばれる)
void
SCSITarget::SelectionCB()
{
	// ターゲットが、自身が選択されていることを認識してから BSY を
	// 立てるまでの時間はもうバス側でカウントしてあるので、ここでは
	// ただちに BSY を上げてよい。
	putlog(4, "Selection Callback");
	AssertBSY();
}

// 情報転送フェーズを開始する (SCSIBus から呼ばれる)
void
SCSITarget::StartTransferCB()
{
	// 最初はコマンドフェーズ。(ATN はとりあえず放置)
	putlog(4, "Start Command Phase");
	cmdseq.clear();
	cmdlen = 0;
	SetXfer(SCSI::XferPhase::Command);
	AssertREQ();
}

// 情報転送フェーズ間の遷移について
//
// 情報転送フェーズは TransferCB() で1バイトずつ転送を行い、所定のバイト数
// の転送が完了すればフェーズを遷移する。遷移方法は入出力方向別で6通り。
//  o IN  -> IN
//  o IN  -> OUT
//  o IN  -> END
//  o OUT -> OUT
//  o OUT -> IN
//  o OUT -> END
//
// IN -> IN の遷移。TransferCB() で前フェーズの最終バイト受信後、フェーズを
// 変更して REQ を立てるところまで。
//
// IN -> OUT の遷移。TransferCB() で前フェーズの最終バイト受信後、フェーズを
// 変更して (おそらく {Phase}Begin() を発行してから)、データをセットし REQ
// を立てる?
//
// IN -> END の遷移。TransferCB() で前フェーズの最終バイト受信後、BSY を
// 下げて終了。
//
// OUT -> OUT の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、フェーズを変更して (おそらく {Phase}Begin()
// を発行してから)、データをセットし REQ を立てる?
//
// OUT -> IN の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、フェーズを変更して (おそらく {Phase}Begin()
// を発行してから)、REQ を立てる?
//
// OUT -> END の遷移。前フェーズの最終バイトをイニシエータが受信後に
// TransferCB() が呼ばれるので、BSY を下げる?

// 情報転送フェーズでイニシエータが ACK を上げた。(SCSIBus から呼ばれる)
// OUT (イニシエータ → ターゲット方向) なら、ターゲットが上げた REQ に応答
// してイニシエータがデータをデータバスに乗せてからこれをコールしてくる。
// IN (ターゲット → イニシエータ方向) なら、ターゲットがデータをデータバスに
// 乗せて REQ を上げたのをイニシエータが受信した後でこれをコールしてくる。
void
SCSITarget::TransferCB()
{
	uint8 data;
	SCSI::XferPhase xfer;

	// このフェーズの最終バイトか
	lastbyte = false;

	// このフェーズの1バイト処理
	xfer = GetXfer();
	if (xfer == SCSI::XferPhase::Command) {
		// コマンドフェーズだけいろいろ特殊なので別処理。
		// 受信バッファが異なる、1バイト目で受信長が決まるなど。
		data = GetData();
		putlog(4, "TransferCB: cmdseq[%02d] = $%02x",
			(int)cmdseq.size(), data);
		cmdseq.push_back(data);
		if (cmdseq.size() == 1) {
			// 1バイト目ならこのコマンドが何バイトか調べる
			switch (cmdseq[0] >> 5) {
			 case 0:	cmdlen =  6;	break;
			 case 1:	cmdlen = 10;	break;
			 case 2:	cmdlen = 10;	break;
			 case 5:	cmdlen = 12;	break;
			 default:
				// 3,4 はリサーブ、6,7 はベンダー定義。XXX 来たらどうする?
				cmdlen = 6;
				break;
			}
		} else if (cmdseq.size() >= cmdlen) {
			if (loglevel >= 2
#if defined(NO_READWRITE_LOG)
				&& (cmdseq[0] != SCSI::Command::Read6 &&
				    cmdseq[0] != SCSI::Command::Read10 &&
				    cmdseq[0] != SCSI::Command::Write6 &&
				    cmdseq[0] != SCSI::Command::Write10)
#endif
			) {
				std::string cmdbuf;
				const char *name = SCSI::GetCommandName(cmdseq[0]);
				cmdbuf += string_format("Command \"%s\"", name ?: "?");
				for (const auto& v : cmdseq) {
					cmdbuf += string_format(" %02x", v);
				}
				putlogn("%s", cmdbuf.c_str());
			}
			// 最終バイトを受信したらコマンドを選択
			DispatchCmd();
			lastbyte = true;
		}
	} else if (GetIO() == false) {
		// OUT 方向 (ターゲットが受信側)
		// bytecount は減算方向。
		cmd->buf.push_back(GetData());
		bytecount--;
		if (bytecount == 0) {
			lastbyte = true;
		}
	} else {
		// IN 方向 (ターゲットが送信側)
		// bytecount は加算方向。
		// データは前のループの終わり (TransferAckCB()の下のほう) で
		// セットしているのでここでは引き取られた分をカウントする。
		bytecount++;
		if (bytecount >= cmd->buf.size()) {
			lastbyte = true;
		}
	}

	NegateREQ();
}

// 情報転送フェーズでイニシエータが ACK を下げた。(SCSIBus から呼ばれる)
void
SCSITarget::TransferAckCB()
{
	SCSI::XferPhase xfer = GetXfer();

	// これが最終バイトだったら、フェーズ完了 → フェーズ切り替え。
	if (lastbyte) {
		SCSI::XferPhase next;
		// フェーズ完了
		switch (xfer) {
		 case SCSI::XferPhase::Command:
			next = cmd->Command(cmdseq);
			break;
		 case SCSI::XferPhase::DataOut:
			next = cmd->DoneDataOut();
			break;
		 case SCSI::XferPhase::DataIn:
			next = cmd->DoneDataIn();
			break;
		 case SCSI::XferPhase::MsgOut:
			next = cmd->DoneMsgOut();
			break;
		 case SCSI::XferPhase::MsgIn:
			next = cmd->DoneMsgIn();
			break;
		 case SCSI::XferPhase::Status:
			next = cmd->DoneStatus();
			break;
		 default:
			PANIC("TransferCB unknown xfer done %d", (int)xfer);
		}
		putlog(3, "%s done", SCSI::GetXferPhaseName(xfer));

		// 簡略化のため、フェーズ完了時だけ REQ を引き伸ばすことができる
		if (cmd->optime != 0) {
			bus->SetOneshotReqWait(cmd->optime);
			cmd->optime = 0;
		}

		if (next != xfer) {
			// End ならバスフリーにして終了。
			if (next == SCSI::XferPhase::End) {
				NegateBSY();
				return;
			}

			// (必要なら)フェーズを更新
			SetXfer(next);
			xfer = next;

			// 新しいフェーズの開始処理
			// XXX ここで bytecount とかをリセットできればいいのだが
			// Command フェーズで Data フェーズの準備をすることが多いので
			// 悩ましい。
			switch (xfer) {
			 case SCSI::XferPhase::DataOut:
				cmd->buf.clear();
				bytecount = cmd->recvbytes;
				putlog(3, "DataOut begin (%d bytes)", bytecount);
				break;
			 case SCSI::XferPhase::DataIn:
				bytecount = 0;
				putlog(3, "DataIn begin (%d bytes)", (int)cmd->buf.size());
				break;
			 case SCSI::XferPhase::MsgOut:
				cmd->BeginMsgOut();
				cmd->buf.clear();
				bytecount = cmd->recvbytes;
				putlog(3, "MsgOut begin (%d bytes)", bytecount);
				break;
			 case SCSI::XferPhase::MsgIn:
				cmd->BeginMsgIn();
				bytecount = 0;
				putlog(3, "MsgIn begin (%d bytes)", (int)cmd->buf.size());
				break;
			 case SCSI::XferPhase::Status:
				cmd->BeginStatus();
				bytecount = 0;
				putlog(3, "Status begin (%d bytes)", (int)cmd->buf.size());
				break;
			 default:
				PANIC("TransferCB unknown xfer begin %d", (int)xfer);
			}
		}
	}

	// IN 方向 (ターゲットが送信側) ならここでデータをバスにセット。
	if (GetIO()) {
		SetData(cmd->buf[bytecount]);
		putlog(4, "SetData $%02x", GetData());
	}

	// 次のバイト転送を行うために REQ を上げる。
	AssertREQ();
}

// SCSI コマンドのディスパッチャ
void
SCSITarget::DispatchCmd()
{
	switch (cmdseq[0]) {
	 case SCSI::Command::TestUnitReady:	// 0x00
		cmd.reset(new SCSICmd(this));
		break;

	 case SCSI::Command::RezeroUnit:	// 0x01
		cmd.reset(new SCSICmd(this));
		break;

	 case SCSI::Command::RequestSense:	// 0x03
		cmd.reset(new SCSICmdRequestSense(this));
		break;

	 case SCSI::Command::Inquiry:		// 0x12
		cmd.reset(new SCSICmdInquiry(this));
		break;

	 case SCSI::Command::ModeSelect6:	// 0x15
	 case SCSI::Command::ModeSelect10:	// 0x55
		cmd.reset(new SCSICmdModeSelect(this));
		break;

	 case SCSI::Command::ModeSense6:	// 0x1a
	 case SCSI::Command::ModeSense10:	// 0x5a
		cmd.reset(new SCSICmdModeSense(this));
		break;

	 default:
		// サポートしていないコマンド
		const char *cname = SCSI::GetCommandName(cmdseq[0]);
		std::string name;
		if (cname) {
			name = string_format(" %s", cname);
		}
		putlog(0, "SCSI Command $%02x%s len=%d not supported",
			cmdseq[0], name.c_str(), (int)cmdseq.size());

		cmd.reset(new SCSICmdNotSupportedCommand(this));
		break;
	}
}

// コマンド列を指定して対応するコマンドを選択する。なければ NULL を返す。
// ROM30 エミュレーション用。
SCSICmd *
SCSITarget::SelectCommand(const std::vector<uint8>& cmdseq_)
{
	cmdseq = cmdseq_;
	DispatchCmd();
	return cmd.get();
}


//
// SCSI Disk
//

// コンストラクタ
SCSIDisk::SCSIDisk(SCSIHostDevice *h, int id, SCSI::DevType devtype_)
	: inherited("", h, id)
{
	devtype = devtype_;

	const char *devtypename = SCSI::GetDevTypeName(devtype);

	// コンストラクタの初期化子では devtypename を組み込みにくいので、
	// コンストラクタ本文でログ名などを設定する。その時の作法に
	// 従っているのですこし面倒な手法が必要。
	// lib/object.h 参照。
	SetName(string_format("SCSI%s%d", devtypename, id));
	ClearAlias();
	AddAlias(string_format("SCSI%d", id));
	AddAlias(string_format("%s%d", devtypename, id));

	switch (devtype) {
	 case SCSI::DevType::HD:
		removable_device = false;
		writeable_device = true;
		blocksize = 512;
		break;
	 case SCSI::DevType::CD:
		removable_device = true;
		writeable_device = false;
		blocksize = 2048;
		break;
	 case SCSI::DevType::MO:
		removable_device = true;
		writeable_device = true;
		blocksize = 512;
		break;
	 default:
		__unreachable();
	}

}

// デストラクタ
SCSIDisk::~SCSIDisk()
{
}

// 初期化
bool
SCSIDisk::Init()
{
	const std::string keybody = string_format("%s-id%d-",
		host->GetConfigName(), myid);
	const std::string imgkey = keybody + "image";
	const std::string wikey  = keybody + "writeignore";
	const std::string wpkey  = keybody + "writeprotect";
	const std::string stkey  = keybody + "seektime";

	// デバイス種別(v[0])とパス(v[1])に分解。
	// デバイス種別はすでにチェックしてあるのでここでは不要。
	const ConfigItem& itemimg = gConfig->Find(imgkey);
	const std::string& imageval = itemimg.AsString();
	auto v = string_split(imageval, ',', 2);
	std::string filename;
	if (v.size() > 1) {
		filename = string_trim(v[1]);
	}

	// 固定デバイス(HD)ならファイル名は省略不可
	if (!IsRemovableDevice() && filename.empty()) {
		itemimg.Err("imagefile not specified");
		return false;
	}

	// リムーバブルデバイスのみ UI からの通知を受け取るイベントを用意
	if (IsRemovableDevice()) {
		gScheduler->ConnectMessage(MessageID::SCSIDEV_LOAD(myid), this,
			ToMessageCallback(&SCSIDisk::LoadCallback));
		gScheduler->ConnectMessage(MessageID::SCSIDEV_UNLOAD(myid), this,
			ToMessageCallback(&SCSIDisk::UnloadCallback));
	}

	// オープン前に writeignore をチェック。
	// この後パス名を処理するとすぐに UI に通知を出すため、それより前で
	// 行わないといけない。うーん。
	//
	// 変数名は元々 -writeprotect だったが -writeignore に変更した。
	// 単にリネームするだけだと、開発中に古いバージョンを二分探索とかする
	// 必要が出来た場合に面倒なので、当面は古いほうも面倒を見る。
	// 十分いらなくなった頃に削除する。
	//
	// wi指定	wp指定
	// なし		なし	wi を採用。
	// なし		あり	wp を採用。ただし警告
	// あり		なし	wi を採用。
	// あり		あり	wi を採用。ここは警告不要でいいだろう
	//
	const ConfigItem& wi = gConfig->Find(wikey);
	const ConfigItem& wp = gConfig->Find(wpkey);
	if (wi.GetFrom() == ConfigItem::FromInitial &&
	    wp.GetFrom() != ConfigItem::FromInitial   )
	{
		// -writeignore (新)がなく -writeprotect (旧) だけが指定されると警告。
		// エラー停止まではしなくていいか。
		warnx("Warning: '%s' is obsolete. Use new '%s'",
			wpkey.c_str(), wikey.c_str());

		write_ignore = wp.AsInt();
	} else {
		write_ignore = wi.AsInt();
	}

	// 設定時点でファイルが指定されていればオープン
	if (filename.empty() == false) {
		std::string path;
		if (filename[0] == '/') {
			path = filename;
		} else {
			// 相対パスなら VM ディレクトリから
			path = gMainApp.GetVMDir() + filename;
		}

		if (LoadDisk(path) == false) {
			return false;
		}
	} else {
		// メディアがロードされないままでも起動時は必ず通知を投げる。
		// ステータスパネルはこの UIMessage でのみ状態を更新するため。
		MediaChanged();
	}

	// 設定は [msec]、変数は [nsec]。
	seektime = gConfig->Find(stkey).AsInt();
	seektime *= 1_msec;

	return true;
}

// 電源オン
void
SCSIDisk::ResetHard(bool poweron)
{
	inherited::ResetHard(poweron);

	if (poweron) {
		// メディアの取り出し禁止を解除
		PreventMediumRemoval(false);
	}
}

// バスリセット
void
SCSIDisk::BusResetCB()
{
	// メディアの取り出し禁止を解除
	PreventMediumRemoval(false);

	inherited::BusResetCB();
}

// バックエンドのディスクイメージを開く。
// 成功すれば true、失敗すれば false を返す。(ロードされれば true を返す
// のではないが、medium_loaded が同じ値になるので使いまわしている)
bool
SCSIDisk::LoadDisk(const std::string& pathname_)
{
	bool unloaded;
	off_t size;
	bool w_ok;
	bool read_only;

	// すでにあればクローズ (ここでは通知は行わない)
	unloaded = UnloadDisk_internal();

	// 新しいイメージをオープン(する準備)
	pathname = pathname_;
	if (image.CreateHandler(pathname) == false) {
		goto done;
	}

	if (image.GetInfo(&size, &w_ok) == false) {
		goto done;
	}

	// セクタ単位になっていること
	if (size % blocksize != 0) {
		warnx("%s: Bad image size(%ju): Not a multiple of blocksize(%d)",
			pathname.c_str(), (uintmax_t)size, blocksize);
		goto done;
	}
	// メディアごとの最大サイズを越えていないこと
	switch (GetDevType()) {
	 case SCSI::DevType::HD:
		// LBA (32bit) * 512 byte/sector
		if (size >= 0x200'0000'0000ULL) {
			warnx("%s: Bad image size(%ju): Too large",
				pathname.c_str(), (uintmax_t)size);
			goto done;
		}
		break;
	 case SCSI::DevType::CD:
		// https://cdrfaq.org/faq07.html#S7-6
		// 63min = 283,500 [sect] = 553.7MB CD-ROM = 635.9MB CD-DA
		// 74min = 333,000 [sect] = 650.3MB CD-ROM = 746.9MB CD-DA
		// 80min = 360,000 [sect] = 703.1MB CD-ROM = 807.4MB CD-DA
		// 90min = 405,000 [sect] = 791.0MB CD-ROM = 908.4MB CD-DA
		// 99min = 445,500 [sect] = 870.1MB CD-ROM = 999.3MB CD-DA
		if (size >= 445500 * 2048) {
			warnx("%s: Bad image size(%ju): Too Large",
				pathname.c_str(), (uintmax_t)size);
			goto done;
		}
		break;
	 case SCSI::DevType::MO:
		// 128M は 25 [sect] * 10,000 [sect/track] * 512 [byte/sect]
		// 230M は 25 [sect] * 17,853 [sect/track] * 512 [byte/sect]
		// 540M は 25 [sect] * 41,660 [sect/track] * 512 [byte/sect]
		// 640M は 17 [sect] * 18,256 [sect/track] * 2048 [byte/sect]
		// 1.3G は 17 [sect] * 35,638 [sect/track] * 2048 [byte/sect]
		// 2.3G は 17 [sect] * 62,538 [sect/track] * 2048 [byte/sect]
		//
		// 今の所 512 byte/sector フォーマットのみ対応しておく。
		if (size >= 25 * 41660 * 512) {
			warnx("%s: Bad image size(%ju): 640M or larger MO not supported",
				pathname.c_str(), (uintmax_t)size);
			goto done;
		}
		break;
	 default:
		__unreachable();
	}

	// 書き込みモードをここで確定
	if (IsWriteableDevice()) {
		// 書き込み可能デバイスの場合
		if (w_ok) {
			if (write_ignore) {
				write_mode = RW::WriteIgnore;
			} else {
				write_mode = RW::Writeable;
			}
		} else {
			write_mode = RW::WriteProtect;
		}
	} else {
		// 読み込み専用デバイスの場合
		write_mode = RW::ReadOnly;
	}

	// ここでオープン
	read_only = (GetWriteMode() != RW::Writeable);
	if (image.Open(read_only, write_ignore) == false) {
		goto done;
	}

	// 書き込み無視なら一応ログ出力
	if (GetWriteMode() == RW::WriteIgnore) {
		putmsg(0, "write is ignored");
	}

	medium_loaded = true;
	// リムーバブルデバイスでだけログを出す
	if (IsRemovableDevice()) {
		putlog(1, "Medium loaded");
	}

 done:
	if (medium_loaded == false) {
		// ここで状態をクリアする
		Clear();
	}

	// 変化があれば通知
	if (unloaded || medium_loaded) {
		MediaChanged();
	}
	return medium_loaded;
}

// バックエンドのディスクイメージを閉じる。
// オープンされてなければ何もしない。
void
SCSIDisk::UnloadDisk(bool force)
{
	if (IsMediumRemovalAllowed() || force) {
		if (UnloadDisk_internal()) {
			// 実際に閉じたら通知
			MediaChanged();
		}
	}
}

// バックエンドのディスクイメージを閉じる (内部用)。
// 実際に閉じれば true を返す。オープンされてなければ何もせず false を返す。
// ここでは通知は行わない。
bool
SCSIDisk::UnloadDisk_internal()
{
	if ((bool)image == false) {
		return false;
	}

	Clear();

	// 取り出し禁止はされてても解除する
	if (IsMediumRemovalPrevented()) {
		// XXX PreventMediumRemoval() を呼ぶかどうするか
		medium_removal_prevented = false;
	}

	putlog(1, "Medium unloaded");
	return true;
}

// ディスク状態をクリアする
// (メディアを取り外した時など)
void
SCSIDisk::Clear()
{
	medium_loaded = false;
	image.Close();
	pathname.clear();
	write_mode = RW::Writeable;
}

// メディア状態変更を UI に通知する
void
SCSIDisk::MediaChanged() const
{
	UIMessage::Post(UIMessage::SCSI_MEDIA_CHANGE, GetMyID());
}

// メディアを挿入する。
// UI スレッドで実行されるので、VM スレッドに通知するだけ。
void
SCSIDisk::LoadDiskUI(const std::string& pathname_)
{
	new_pathname = pathname_;
	gScheduler->SendMessage(MessageID::SCSIDEV_LOAD(GetMyID()));
}

// メディアを排出する。
// UI スレッドで実行されるので、VM スレッドに通知するだけ。
void
SCSIDisk::UnloadDiskUI(bool force)
{
	gScheduler->SendMessage(MessageID::SCSIDEV_UNLOAD(GetMyID()), force);
}

// メディア挿入メッセージコールバック
void
SCSIDisk::LoadCallback(MessageID msgid, uint32 arg)
{
	if (LoadDisk(new_pathname) == false) {
		// 失敗したら UI に通知
		UIMessage::Post(UIMessage::SCSI_MEDIA_FAILED);
	}
}

// メディア排出メッセージコールバック
void
SCSIDisk::UnloadCallback(MessageID msgid, uint32 arg)
{
	bool force = arg;
	UnloadDisk(force);
}

// SCSI コマンドのディスパッチャ
void
SCSIDisk::DispatchCmd()
{
	switch (cmdseq[0]) {
	 case SCSI::Command::Read6:			// 0x08
	 case SCSI::Command::Read10:		// 0x28
		cmd.reset(new SCSICmdRead(this));
		break;

	 case SCSI::Command::Write6:		// 0x0a
	 case SCSI::Command::Write10:		// 0x2a
		if (IsWriteableDevice()) {
			cmd.reset(new SCSICmdWrite(this));
		} else {
			cmd.reset(new SCSICmdNotSupportedCommand(this));
		}
		break;

	 case SCSI::Command::StartStopUnit:	// 0x1b
		cmd.reset(new SCSICmdStartStopUnit(this));
		break;

	 case SCSI::Command::PreventAllowMediumRemoval:	// 0x1e
		if (IsRemovableDevice() == false) {
			// HD は取り出せないので、とりあえず。
			// XXX 要調査
			cmd.reset(new SCSICmdNotSupportedCommand(this));
			return;
		}
		cmd.reset(new SCSICmdPreventAllowMediumRemoval(this));
		break;

	 case SCSI::Command::ReadCapacity:	// 0x25
		// DA デバイスでは Read Capacity (0x25)、
		// CD-ROM デバイスでは Read CDROM Capacity (0x25) だが
		// 今の所同じ動作なので区別せず処理する。
		cmd.reset(new SCSICmdReadCapacity(this));
		break;

	 case SCSI::Command::SynchronizeCache:	// 0x35
		// このクラスは、物理ドライブにあるようなディスクとの間の
		// キャッシュ機構は持っていないので、そのキャッシュをフラッシュ
		// しろという指示に対しては、黙って成功するだけでいい気がする。
		// このコマンドはオプションだが NetBSD カーネルが使う。
		//
		// CD-ROM でキャッシュの同期とは読み出し側のことだろうか?
		// とりあえず黙って成功しておく。
		cmd.reset(new SCSICmd(this));
		break;

	 case SCSI::Command::ReadTOC:		// 0x43
		if (devtype == SCSI::DevType::CD) {
			cmd.reset(new SCSICmdReadTOC(this));
		} else {
			cmd.reset(new SCSICmdNotSupportedCommand(this));
		}
		break;

	 case SCSI::Command::ReadFormatCapacities:	// 0x23
		// SCSI-2 規格でないので詳細不明。
		// ディスク用か共通かも分からないけど、とりあえず
	 case SCSI::Command::ReadDiscInformation:	// 0x51
		// CD-ROM デバイスのオプションコマンドで
		// CD-R ライター用(?)らしいので無視する。
		cmd.reset(new SCSICmdNotSupportedCommand(this));
		break;

	 default:
		// 見付からなければ基本クラス側に任せる
		inherited::DispatchCmd();
		break;
	}
}

// Inquiry データを返す
bool
SCSIDisk::Inquiry(std::vector<uint8>& buf, int lun)
{
	uint8 inqtype;
	const char *prodname;

	switch (devtype) {
	 case SCSI::DevType::HD:
		inqtype = SCSI::InquiryDeviceType::DirectAccess;
		prodname = "SCSIHD";
		break;
	 case SCSI::DevType::CD:
		inqtype = SCSI::InquiryDeviceType::CDROM;
		prodname = "SCSICD";
		break;
	 case SCSI::DevType::MO:
		inqtype = SCSI::InquiryDeviceType::MO;
		prodname = "SCSIMO";
		break;
	 default:
		PANIC("Unexpected devtype=%d", (int)devtype);
	}

	buf[0] = inqtype;
	if (lun != 0) {
		buf[0] |= SCSI::InquiryQualifier::LU_NotPreseted;
	}
	buf[1] = 0;
	if (IsRemovableDevice()) {
		buf[1] |= 0x80;
	}
	buf[2] = 2;		// SCSI-2
	buf[3] = 2;		// Response Data Format = SCSI-2
	buf[4] = buf.size() - 4;	// 追加データ長

	// 文字列はゼロ終端ではなく、余りは空白で埋める
	memset(&buf[8], ' ', 36 - 8);
	memcpy(&buf[8],  "NONO", 4);
	memcpy(&buf[16], prodname, strlen(prodname));
	memcpy(&buf[32], "0", 1);

	return true;
}

// ModeSense コマンドのデバイス固有パラメータを返す
uint8
SCSIDisk::GetDeviceSpecificParam() const
{
	uint8 param = 0;

	// DA, MO, CD のデバイス固有パラメータは以下のようになっている。
	// DPOFUA はキャッシュ制御フラグのサポート(%1)かどうかなので %0 でいい。
	//
	//      7   6   5     4    3   2   1   0
	//    +---+---+---+------+---+---+---+---+
	// DA |WP |   0   |DPOFUA|       0       | WP: Write Protect
	//    +---+---+---+------+---+---+---+---+
	//
	//    +---+---+---+------+---+---+---+---+
	// MO |WP |   0   |DPOFUA|     0     |EBC| EBC: Enable Blank Check
	//    +---+---+---+------+---+---+---+---+
	//
	//    +---+---+---+------+---+---+---+---+
	// CD |     0     |DPOFUA|       0       |
	//    +---+---+---+------+---+---+---+---+

	if (GetDevType() == SCSI::DevType::CD) {
		// CD には WP ビットの定義がない
	} else {
		// DA, MO は WP ビットの定義がある。
		if (IsMediumLoaded() && GetWriteMode() == RW::WriteProtect) {
			param |= 0x80;
		}
	}

	return param;
}

bool
SCSIDisk::Read(std::vector<uint8>& buf, uint64 start)
{
	if (__predict_false((bool)image == false)) {
		return false;
	}
	return image.Read(buf.data(), start, buf.size());
}

bool
SCSIDisk::Write(const std::vector<uint8>& buf, uint64 start)
{
	if (__predict_false((bool)image == false)) {
		return false;
	}

	if (GetWriteMode() >= RW::WriteProtect) {
		// 書き込み不可
		return false;
	}
	return image.Write(buf.data(), start, buf.size());
}

off_t
SCSIDisk::GetSize() const
{
	if (__predict_false((bool)image == false)) {
		return 0;
	}
	return image.GetSize();
}

// 指定の場所を自由に読み出す
// (エミュレータ特権アクセス)
bool
SCSIDisk::Peek(void *buf, uint64 start, uint32 len) const
{
	if (__predict_false((bool)image == false)) {
		return false;
	}
	return image.Read(buf, start, len);
}

// メディアの取り出し可否(取り出し禁止)を設定する
void
SCSIDisk::PreventMediumRemoval(bool val)
{
	// 状態が変わったところでログを表示
	if (medium_removal_prevented == false && val == true) {
		putlog(1, "Medium removal prevented");
	} else if (medium_removal_prevented == true && val == false) {
		putlog(1, "Medium removal allowed");
	}
	medium_removal_prevented = val;
}

// メディアの書き込みモードを文字列で返す
const char *
SCSIDisk::GetWriteModeStr() const
{
	switch (write_mode) {
	 case RW::ReadOnly:		return "ReadOnly";
	 case RW::WriteProtect:	return "WriteProtected";
	 case RW::Writeable:	return "Writeable";
	 case RW::WriteIgnore:	return "WriteIgnore";
	 default:
		__unreachable();
	}
}
