//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// スプライトコントローラ
//

#include "sprite.h"
#include "mpu.h"

// グローバル参照用
SpriteDevice *gSprite;

// コンストラクタ
SpriteDevice::SpriteDevice()
	: inherited("Sprite")
{
	mem.reset(new uint8[4 * 8192]);
}

// デストラクタ
SpriteDevice::~SpriteDevice()
{
	gSprite = NULL;
}

// XXX ResetHard() 未調査

uint64
SpriteDevice::Read8(uint32 addr)
{
	if (addr >= 0xeb8000) {
		return mem[HB(addr - 0xeb8000)];
	}
	VMPANIC("not impl");
}

uint64
SpriteDevice::Read16(uint32 addr)
{
	// スプライトの Read はレジスタ、VRAM とも 17 ウェイト。
	// InsideOut p.135
	gMPU->AddCycle(17);

	if (addr >= 0xeb8000) {
		return *(uint16 *)&mem[addr - 0xeb8000];
	}
	if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			return reg.sprite[n].xpos;
		 case 2:
			return reg.sprite[n].ypos;
		 case 4:
			return reg.sprite[n].col;
		 case 6:
			return reg.sprite[n].prw;
		}
	}
	if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			return reg.bg0x;
		 case 0xeb0802:
			return reg.bg0y;
		 case 0xeb0804:
			return reg.bg1x;
		 case 0xeb0806:
			return reg.bg1y;
		 case 0xeb0808:
			return reg.bgctrl;
		 case 0xeb080a:
			return reg.htotal;
		 case 0xeb080c:
			return reg.hdisp;
		 case 0xeb080e:
			return reg.vdisp;
		 case 0xeb0810:
			return reg.res;
		}
	}

	return (uint64)-1;
}

uint64
SpriteDevice::Write8(uint32 addr, uint32 data)
{
	if (addr >= 0xeb8000) {
		// スプライト VRAM への書き込み (InsideOut p.135)
		gMPU->AddCycle(18);
		mem[HB(addr - 0xeb8000)] = data;
		return 0;
	}

	// スプライトレジスタへの書き込み (InsideOut p.135)
	gMPU->AddCycle(20);

	VMPANIC("not impl");
}

uint64
SpriteDevice::Write16(uint32 addr, uint32 data)
{
	putlog(2, "$%06x <- $%04x", addr, data);
	if (addr >= 0xeb8000) {
		*(uint16 *)&mem[addr - 0xeb8000] = data;
		return 0;
	}
	if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			reg.sprite[n].xpos = data & 0x3ff;
			break;
		 case 2:
			reg.sprite[n].ypos = data & 0x3ff;
			break;
		 case 4:
			reg.sprite[n].col = data & 0xcfff;
			break;
		 case 6:
			reg.sprite[n].prw = data & 0x3;
			break;
		}
		return 0;
	}
	if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			reg.bg0x = data & 0x3ff;
			break;
		 case 0xeb0802:
			reg.bg0y = data & 0x3ff;
			break;
		 case 0xeb0804:
			reg.bg1x = data & 0x3ff;
			break;
		 case 0xeb0806:
			reg.bg1y = data & 0x3ff;
			break;
		 case 0xeb0808:
			reg.bgctrl = data & 0x23f;
			break;
		 case 0xeb080a:
			reg.htotal = data & 0xff;
			break;
		 case 0xeb080c:
			reg.hdisp = data & 0x3f;
			break;
		 case 0xeb080e:
			reg.vdisp = data & 0xff;
			break;
		 case 0xeb0810:
			reg.res = data & 0x1f;
			break;
		}
		return 0;
	}

	return (uint64)-1;
}

uint64
SpriteDevice::Peek8(uint32 addr)
{
	if (addr >= 0xeb8000) {
		return mem[HB(addr - 0xeb8000)];
	}
	if (addr < 0xeb0400) {
		int n = (addr - 0xeb0000) / 8;
		switch (addr % 8) {
		 case 0:
			return reg.sprite[n].xpos;
		 case 2:
			return reg.sprite[n].ypos;
		 case 4:
			return reg.sprite[n].col;
		 case 6:
			return reg.sprite[n].prw;
		}
	}
	if (0xeb0800 <= addr && addr < 0xeb0812) {
		switch (addr) {
		 case 0xeb0800:
			return reg.bg0x;
		 case 0xeb0802:
			return reg.bg0y;
		 case 0xeb0804:
			return reg.bg1x;
		 case 0xeb0806:
			return reg.bg1y;
		 case 0xeb0808:
			return reg.bgctrl;
		 case 0xeb080a:
			return reg.htotal;
		 case 0xeb080c:
			return reg.hdisp;
		 case 0xeb080e:
			return reg.vdisp;
		 case 0xeb0810:
			return reg.res;
		}
	}

	return (uint64)-1;
}
