//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// TVRAM
//

// TVRAM (PlaneVRAM) はロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス時は HLB()、ワードアクセス時は HLW() マクロを使用のこと。

#include "tvram.h"
#include "mpu.h"
#include "videoctlr.h"

// コンストラクタ
TVRAMDevice::TVRAMDevice()
	: inherited("TVRAM", 1024, 1024)
{
	uint devlen = 512 * 1024;

	// プレーン数は4枚固定
	nplane = NPLANE;

	mem.reset(new uint8[devlen]);
}

// デストラクタ
TVRAMDevice::~TVRAMDevice()
{
}

bool
TVRAMDevice::Init()
{
	// 加工済みパレットを取得
	palette = &(gVideoCtlr->GetHostPalette())[0];

	return true;
}

// リセット
void
TVRAMDevice::ResetHard(bool poweron)
{
	// XXX ここ?
	Invalidate();
}

inline uint64
TVRAMDevice::Decoder(uint32 addr) const
{
	return addr - baseaddr;
}

uint64
TVRAMDevice::Read8(uint32 addr)
{
	gMPU->AddCycle(9);	// Inside/Out p.135
	uint32 offset = Decoder(addr);
	return mem[HLB(offset)];
}

uint64
TVRAMDevice::Read16(uint32 addr)
{
	gMPU->AddCycle(9);	// Inside/Out p.135
	uint32 offset = Decoder(addr);
	return *(uint16 *)&mem[HLW(offset)];
}

uint64
TVRAMDevice::Write8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	uint8 *ptr;
	uint8 mask8;

	gMPU->AddCycle(9);	// Inside/Out p.135

	// XXX とりあえずね
	putlog(3, "Write $%06x.B <- $%02x", addr, data);

	if (tvram.men) {
		if ((addr & 1) == 0) {
			mask8 = tvram.mask >> 8;
		} else {
			mask8 = tvram.mask;
		}
	} else {
		mask8 = 0;
	}

	uint32 poffset = offset & 0x1ffff;
	if (tvram.sa) {
		// 同時アクセス
		for (int plane = 0; plane < 4; plane++) {
			if (tvram.ap[plane]) {
				ptr = &mem[HLB(0x20000 * plane + poffset)];
				*ptr = (data & ~mask8) | (*ptr & mask8);
			}
		}
	} else {
		// 通常アクセス
		ptr = &mem[HLB(offset)];
		*ptr = (data & ~mask8) | (*ptr & mask8);
	}
	SetDirty(poffset);
	return 0;
}

uint64
TVRAMDevice::Write16(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	uint16 *ptr;
	uint16 mask16;

	gMPU->AddCycle(9);	// Inside/Out p.135

	// XXX とりあえずね
	putlog(3, "Write $%06x.W <- $%04x", addr, data);

	if (tvram.men) {
		mask16 = tvram.mask;
	} else {
		mask16 = 0;
	}

	uint32 poffset = offset & 0x1ffff;
	if (tvram.sa) {
		// 同時アクセス
		for (int plane = 0; plane < 4; plane++) {
			if (tvram.ap[plane]) {
				ptr = (uint16 *)&mem[HLW(0x20000 * plane + poffset)];
				*ptr = (data & ~mask16) | (*ptr & mask16);
			}
		}
	} else {
		// 通常アクセス
		ptr = (uint16 *)&mem[HLW(offset)];
		*ptr = (data & ~mask16) | (*ptr & mask16);
	}
	SetDirty(poffset);
	return 0;
}

uint64
TVRAMDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	return mem[HLB(offset)];
}

uint64
TVRAMDevice::Poke8(uint32 addr, uint32 data)
{
	if ((int32)data >= 0) {
		uint32 offset = Decoder(addr);
		uint32 poffset = offset & 0x1ffff;
		mem[HLB(offset)] = data;
		// この場合も画面を更新する
		SetDirty(poffset);
	}
	return 0;
}

// offset の位置に対応する更新フラグを立てる。
// offset はプレーン先頭からのバイトオフセット。
inline void
TVRAMDevice::SetDirty(uint32 offset)
{
	// VRAM は横 1024 ドット(128バイト) x 縦 1024 ドット。
	//
	//              1                               0
	//        3 2 1 0 f e d c b a 9 8 7 6 5 4 3 2 1 0
	//       +-----------+-------+-----+-------+-----+
	// VRAM  |  Y座標 (1024dot)  | X (128byte) : 8bit|
	//       +-----------+-------+-----+-------+-----+
	//       :                   :
	//       +-------------------+
	// Dirty |                   |
	//       +-------------------+
	//        9 8 7 6 5 4 3 2 1 0

	// 更新フラグはラスタ単位
	int y = offset >> 7;
	dirty[y] = 1;
}

// X 方向にはみ出したときに対応する仮想画面の Y 座標を返す。
int
TVRAMDevice::GetWrapY(int y) const
{
	// はみ出すと次のラスタの先頭から表示するが、
	// TVRAM は 4 ラスタがひとかたまりで構成されていて、
	// 末尾 2bit が 3 のラスタの次は、末尾だけ 0 のラスタに戻る。ようだ。
	return (y & ~3) + ((y + 1) & 3);
}

void
TVRAMDevice::SetScrollX(int x)
{
	xscroll = x;
}

void
TVRAMDevice::SetScrollY(int y)
{
	yscroll = y;
	gRenderer->Invalidate2();
}

// CRTC R21 での同時アクセス設定を反映する。
// CRTC から呼ばれる。
void
TVRAMDevice::SetAccessPlane(uint16 data)
{
	tvram.cp[0] = (data & 0x01);
	tvram.cp[1] = (data & 0x02);
	tvram.cp[2] = (data & 0x04);
	tvram.cp[3] = (data & 0x08);
	tvram.ap[0] = (data & 0x10);
	tvram.ap[1] = (data & 0x20);
	tvram.ap[2] = (data & 0x40);
	tvram.ap[3] = (data & 0x80);
	tvram.sa = (data & 0x100);
	tvram.men = (data & 0x200);

	if (0) {
		if (tvram.sa) {
			putlog(2, "テキスト画面同時アクセス有効 %c%c%c%c",
				tvram.ap[0] ? '0' : '-',
				tvram.ap[1] ? '1' : '-',
				tvram.ap[2] ? '2' : '-',
				tvram.ap[3] ? '3' : '-');
		} else {
			putlog(2, "テキスト画面同時アクセス無効");
		}
	}
}

// CRTC R23 でのアクセスマスク設定を反映する。
// CRTC から呼ばれる。
void
TVRAMDevice::SetAccessMask(uint16 data)
{
	tvram.mask = data;
}

// ラスターコピーを開始する。
// src_ras, dst_ras はライン番号 (レジスタ設定値を4倍したもの)。
void
TVRAMDevice::RasterCopy(int src_ras, int dst_ras)
{
	for (int plane = 0; plane < 4; plane++) {
		if (tvram.cp[plane]) {
			uint8 *src = &mem[0x20000 * plane + src_ras * 128];
			uint8 *dst = &mem[0x20000 * plane + dst_ras * 128];
			memcpy(dst, src, 128 * 4);
		}
	}

	// dst からの4ラインを全部 dirty にする。
	for (int y = 0; y < 4; y++) {
		SetDirty((dst_ras + y) * 128);
	}
}

// (x, y) 位置のピクセル情報を screen に描画。(ビットマップモニタの情報欄)
void
TVRAMDevice::UpdateInfo(TextScreen& screen, int x, int y) const
{
	screen.Clear();

	//0         1         2         3         4         5         6
	//012345678901234567890123456789012345678901234567890123456789012345
	//X=1024 Y=2024: (P0) $e01234: $00 (%00000000) ColorCode=$d

	screen.Puts(0, 0, "X=     Y=");
	for (int i = 0; i < nplane; i++) {
		screen.Print(15, i, "(P%d)", i);
	}

	// カーソルが範囲外ならここまで。
	if (x < 0) {
		return;
	}

	screen.Print(2, 0, "%4d Y=%4d", x, y);

	// 1プレーン内のバイトオフセット
	uint32 planeoffset = ((uint)y * 128) + ((uint)x / 8);

	uint cc = 0;
	for (int i = 0; i < nplane; i++) {
		// プレーン0先頭からのバイトオフセット
		uint32 totaloffset = 0x20000 * i + planeoffset;

		uint8 data = mem[HLB(totaloffset)];
		screen.Print(15, i, "(P%d) $%06x: $%02x (%%",
			i, baseaddr + totaloffset, data);

		// ビットパターン
		for (int m = 0x80; m != 0; m >>= 1) {
			screen.Putc((data & m) ? '1' : '0');
		}
		screen.Putc(')');

		// カーソル位置をハイライト
		screen.Locate(35 + (x % 8), i);
		screen.SetAttr(TA::Em);

		// カラーコードを計算
		if ((data & (0x80U >> (x % 8))) != 0) {
			cc |= (1U << i);
		}
	}

	screen.Print(45, 0, "ColorCode=$%x", cc);
}
