//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ビデオコントローラ
//

// e82000..e823ff パレット
// e82400..e824ff R0(e82400.w) のミラー
// e82500..e825ff R1(e82500.w) のミラー
// e82600..e826ff R2(e82600.w) のミラー
// e82700..e82fff $00 が読めるようだ
// 以上の 4KB をミラー。

#include "videoctlr.h"
#include "mpu.h"
#include "renderer.h"
#include "uimessage.h"

// グローバル参照用
VideoCtlrDevice *gVideoCtlr;

// コンストラクタ
VideoCtlrDevice::VideoCtlrDevice()
	: inherited("VideoCtlr")
{
	hostcolor.resize(16);
}

// デストラクタ
VideoCtlrDevice::~VideoCtlrDevice()
{
	gVideoCtlr = NULL;
}

// リセット
void
VideoCtlrDevice::ResetHard(bool poweron)
{
	// XXX not yet
}

inline uint32
VideoCtlrDevice::Decoder(uint32 addr) const
{
	return addr & 0xfff;
}

uint64
VideoCtlrDevice::Read8(uint32 addr)
{
	uint32 data;

	gMPU->AddCycle(9);	// InsideOut p.135

	uint32 offset = Decoder(addr);
	if (offset < 0x400) {				// パレット
		data = palette[HB(offset)];
		putlog(3, "Palette $%06x.B -> $%02x", addr, data);

	} else if (offset < 0x700) {		// R0,R1,R2
		uint nr = Offset2NR(offset);
		data = GetNR(nr);
		putlog(3, "R%d.%c -> $%02x",
			(nr / 2),
			(nr % 2) == 0 ? 'H' : 'L',
			data);

	} else {
		data = 0;
	}

	return data;
}

uint64
VideoCtlrDevice::Read16(uint32 addr)
{
	uint32 data;

	gMPU->AddCycle(9);	// InsideOut p.135

	uint32 offset = Decoder(addr);
	if (offset < 0x400) {				// パレット
		data = *(uint16 *)&palette[offset];
		putlog(3, "Palette $%06x.W -> $%04x", addr, data);

	} else if (offset < 0x700) {		// R0,R1,R2
		uint nr = Offset2NR(offset);
		data  = GetNR(nr) << 8;
		data |= GetNR(nr + 1);
		putlog(3, "R%d -> $%04x", (nr / 2), data);

	} else {
		data = 0;
	}

	return data;
}

uint64
VideoCtlrDevice::Write8(uint32 addr, uint32 data)
{
	gMPU->AddCycle(11);	// InsideOut p.135

	uint32 offset = Decoder(addr);
	if (offset < 0x400) {				// パレット
		putlog(2, "Palette $%06x.B <- $%02x", addr, data);
		palette[HB(offset)] = data;
		MakePalette();

	} else if (offset < 0x700) {		// R0,R1,R2
		uint nr = Offset2NR(offset);
		putlog(2, "R%d.%c <- $%02x (NOT IMPLEMENTED)",
			(nr / 2),
			(nr % 2) == 0 ? 'H' : 'L',
			data);
		SetNR(nr, data);

	} else {
		// たぶん何も起きない?
	}

	return 0;
}

uint64
VideoCtlrDevice::Write16(uint32 addr, uint32 data)
{
	gMPU->AddCycle(11);	// InsideOut p.135

	uint32 offset = Decoder(addr);
	if (offset < 0x400) {				// パレット
		putlog(2, "Palette $%06x.W <- $%04x", addr, data);
		*(uint16 *)&palette[offset] = data;
		MakePalette();

	} else if (offset < 0x700) {		// R0,R1,R2
		uint nr = Offset2NR(offset);
		putlog(3, "R%d <- $%04x (NOT IMPLEMENTED)", (nr / 2), data);
		SetNR(nr,     data >> 8);
		SetNR(nr + 1, data & 0xff);

	} else {
		// たぶん何も起きない?
	}

	return 0;
}

uint64
VideoCtlrDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	if (offset < 0x400) {				// パレット
		return palette[HB(offset)];

	} else if (offset < 0x700) {		// R0,R1,R2
		uint nr = Offset2NR(offset);
		return GetNR(nr);

	} else {
		return 0x00;
	}
}

uint64
VideoCtlrDevice::Poke8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	if (offset < 0x400) {
		if ((int32)data >= 0) {
			palette[HB(offset)] = data;
			// Poke による編集でもパレットに反映させる
			MakePalette();
		}
		return 0;
	}

	return -1;
}

// R0, R1, R2 のオフセットから内部仮想レジスタ番号 NR0-NR5 に変換する。
// offset は R0, R1, R2 の領域を指していること。
inline uint
VideoCtlrDevice::Offset2NR(uint32 offset) const
{
	assert(0x400 <= offset && offset < 0x700);
	return ((offset / 0x100) - 4) * 2 + (offset & 1);
}

// 内部仮想レジスタ NR0-NR5 を読み出す。
uint32
VideoCtlrDevice::GetNR(uint reg) const
{
	return nreg[reg];
}

// 内部仮想レジスタ NR0-NR5 にバイトデータを書き込む。
void
VideoCtlrDevice::SetNR(uint reg, uint32 data)
{
	// XXX まだ値を保存する以外何もしていない
	assert(reg < 6);
	switch (reg) {
	 case 0:	// R0.H
		nreg[reg] = 0;
		break;
	 case 1:	// R0.L
		nreg[reg] = data & 0x03;
		break;

	 case 2:	// R1.H
		nreg[reg] = data & 0x3f;
		break;
	 case 3:	// R1.L
		nreg[reg] = data & 0xff;
		break;

	 case 4:	// R2.H
		nreg[reg] = data & 0xff;
		break;
	 case 5:	// R2.L
		nreg[reg] = data & 0x7f;
		break;

	 default:
		__unreachable();
	}
}

//
void
VideoCtlrDevice::MakePalette()
{
	const uint16 *p;

	// パレット前半 256 ワードはグラフィックパレット、
	// テキストパレットは後半。
	p = (uint16 *)&palette[512];

	// パレット情報から Color 形式の色データを作成
	// X680x0 のパレットは %GGGGG'RRRRR'BBBBB'I で並んでいる。
	for (int i = 0; i < 16; i++) {
		int I = (p[i] & 1) << 2;	// 輝度ビット
		int R = (((p[i] >> 6) & 0x1f) << 3) + I;
		int G = (((p[i] >> 11)) << 3) + I;
		int B = (((p[i] >> 1) & 0x1f) << 3) + I;
		hostcolor[i].r = R;
		hostcolor[i].g = G;
		hostcolor[i].b = B;
	}

	// パレット変更が起きたことをレンダラに通知
	gRenderer->Invalidate2();

	// UI にも通知
	UIMessage::Post(UIMessage::PALETTE);
}

// ゲストパレットを uint32 形式にしたものの一覧を返す。(GUI 用)
// XXX 今の所テキスト16色しか考えてない
const std::vector<uint32>
VideoCtlrDevice::GetIntPalette() const
{
	std::vector<uint32> ipal;
	const uint16 *p = (const uint16 *)&palette[0];

	for (int i = 0; i < 16; i++) {
		ipal.push_back(p[256 + i]);
	}
	return ipal;
}
