//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// VM (X68030)
//

#include "vm_x68k.h"
#include "bus.h"
#include "config.h"
#include "dmac.h"
#include "fdc.h"
#include "human68k.h"
#include "interrupt.h"
#include "mainapp.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "nmi.h"
#include "pedec.h"
#include "pluto.h"
#include "pio.h"
#include "power.h"
#include "renderer.h"
#include "rp5c15.h"
#include "romimg_x68k.h"
#include "scheduler.h"
#include "spc.h"
#include "x68kkbd.h"
#include "xiospace.h"

// コンストラクタ
VM_X68030::VM_X68030()
{
	// 機種固有パラメータ
	// MPU クロックの初期値は機種ごとに異なる。単位は MHz
	gConfig->SetDefault("mpu-clock", 25);
	// RAM 初期値は MAX にしておく
	gConfig->SetDefault("ram-size", 12);
	// SPC の入力クロック (5MHz = 200ns)
	gConfig->Add("!spc-tCLF", 200);

	// デバイス (順序に注意)
	// BusErrDevice は親コンストラクタで作成済み。
	NEWDV(MPU, new MPU680x0Device());
	NEWDV(Interrupt, new X68030Interrupt());
	NEWDV(MainRAM, new MainRAMDevice());			// required by XIOSpace
	NEWDV(XIOSpace, new XIOSpaceDevice());
	NEWDV(IPLROM0, new IPLROM0Device());

	NEWDV(NMI, new NMIDevice());
	NEWDV(Keyboard, new X68030Keyboard());
	NEWDV(Renderer, new X68030Renderer());

	if (gMainApp.human_mode) {
		NEWDV(Human68k, new Human68k());
	}

	// 割り当て
	// 今は拡張メモリがないので、今は上位8ビット全領域がミラーになっている。
	for (int i = 0; i < gMainbus->devtable.size(); i++) {
		gMainbus->SetDevice(i, gXIOSpace);
	}
}

// デストラクタ
VM_X68030::~VM_X68030()
{
}

// ブートページを切り替える。
void
VM_X68030::SwitchBootPage(Device *parent, bool isrom)
{
	if (gMainApp.human_mode) {
		// Human モードではブートページを RAM に見せる。
		// 今のところ。
		return;
	}

	// 共通処理
	inherited::SwitchBootPage(parent, isrom);

	// X68030 では XIOSpace が担当している
	gXIOSpace->SwitchBootPage(isrom);
}

// リセットについて
//
//           PWRRESET
// 電源オン ----------+--> VIPS
// (7505)             +--> VideoCtlr
//                    +--> OSCIAN2
//                    |
//                    |   +------+
//                    +-->|      |
// 本体      SWRESET      | YUKI |
// リセット ------------->|      |
// スイッチ               +------+
//                         | IPLRESET
//                         |
//                         +--> FPU
//                         |
//                         ▽
//                   RESET |
//                         |<-> MPU (68030)
//                         +--> PPI (8255)
//                         +--> SPC (MB89352)
//                         +--> ADPCM (MSM6258)
//                         +--> FDC (uPD72065B)
//                         +--> MFP (MC68901)
//                         |
//                         |   +-------+
//                         +-->| PEDEC |
//                         |   +-------+
//                         |     |
//                         |     +-- OPMIC -----> OPM(YM2151)
//                         |     +-- SCCRD/WR --> SCC(Z8530)
//                         |
//                         |      EXRESET
//                         +---|>-----------> 外部バス
//                         ▽
//                SYSRESET |
//                         |<-> SAKI  ---+
//                         |             | BEC0,BEC1
//                         +--> DMAC  <--+
//                         +--> CYNTHIA
// リセットされない?
// o CACHY

// MPU の RESET 命令によるリセット
void
VM_X68030::ResetByMPU()
{
	gDMAC->ResetHard(false);
	gFDC->ResetHard(false);
	gMFP->ResetHard(false);
	gPEDEC->ResetHard(false);
	gPluto->ResetHard(false);
	gPPI->ResetHard(false);
	gSPC->ResetHard(false);
}
