//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// X68030 のメイン空間 ($000000-$FFFFFF) 担当
//

#include "xiospace.h"
#include "adpcm.h"
#include "areaset.h"
#include "busio.h"
#include "config.h"
#include "crtc.h"
#include "dmac.h"
#include "fdc.h"
#include "gvram.h"
#include "mainram.h"
#include "mfp.h"
#include "opm.h"
#include "pedec.h"
#include "pio.h"
#include "pluto.h"
#include "printer.h"
#include "romemu_x68k.h"
#include "romimg_x68k.h"
#include "rp5c15.h"
#include "scc.h"
#include "spc.h"
#include "sprite.h"
#include "sram.h"
#include "sysport.h"
#include "tvram.h"
#include "videoctlr.h"
#include "vm.h"

// グローバル参照用
XIOSpaceDevice *gXIOSpace;

// コンストラクタ
XIOSpaceDevice::XIOSpaceDevice()
	: inherited("IOSpace")
{
#define ADD_DEVICE(NAME, ADDR, LEN)	\
	AddDevice(dynamic_cast<IODevice *>(__CONCAT(p,NAME).get()),	ADDR, LEN)

#define N(NAME, NEW_CTOR, ADDR, LEN)	do {	\
		NEWDV(NAME, NEW_CTOR);	\
		ADD_DEVICE(NAME, ADDR, LEN);	\
	} while (0)

	// デバイスクラス作成。
	// 初期化順の制約がない限りアドレス順。
	//                                      	開始アドレス 長さ
	N(GVRAM, new GVRAMDevice(),					0xc00000,0x200000);
	N(PlaneVRAM, new TVRAMDevice(),				0xe00000, 0x80000);
	N(CRTC, new BusIO_W<CRTCDevice>(),			0xe80000,  0x2000);
	N(VideoCtlr, new VideoCtlrDevice(),			0xe82000,  0x2000);
	N(DMAC, new DMACDevice(),					0xe84000,  0x2000);
	N(AreaSet, new AreaSetDevice(),				0xe86000,  0x2000);
	N(MFP, new BusIO_EB<MFPDevice>(),			0xe88000,  0x2000);
	N(RTC, new BusIO_EB<RP5C15Device>(),		0xe8a000,  0x2000);
	N(Printer, new PrinterDevice(),				0xe8c000,  0x2000);
	N(Sysport,new BusIO_FB<SysportDevice>(),	0xe8e000,  0x2000);
	N(OPM, new BusIO_EB<OPMDevice>(),			0xe90000,  0x2000);
	N(ADPCM, new BusIO_EB<ADPCMDevice>(),		0xe92000,  0x2000);
	N(FDC, new BusIO_EB<FDCDevice>(),			0xe94000,  0x2000);
	N(SPC, new BusIO_X68kSPC<SPCDevice>(),		0xe96000,  0x2000);
	N(SCC, new BusIO_EB<SCCDevice>(),			0xe98000,  0x2000);
	N(PPI, new BusIO_EB<PPIDevice>(),			0xe9a000,  0x2000);
	N(PEDEC, new BusIO_EB<PEDECDevice>(),		0xe9c000,  0x2000);
	N(Pluto, new PlutoDevice(),					0xeac000,  0x2000);
	N(Sprite, new SpriteDevice(),				0xeb0000, 0x10000);
	N(SRAM, new SRAMDevice(),					0xed0000,  0x4000);
	N(CGROM, new CGROMDevice(),					0xf00000, 0xc0000);
	if (gConfig->Find("iplrom2-image").AsString().empty()) {
		NEWDV(IPLROM2, new ROM30EmuDevice());
	} else {
		NEWDV(IPLROM2, new IPLROM2Device());
	}
	ADD_DEVICE(IPLROM2,							0xfc0000, 0x20000);
	N(IPLROM1, new IPLROM1Device(),				0xfe0000, 0x20000);
}

// デストラクタ
XIOSpaceDevice::~XIOSpaceDevice()
{
	gXIOSpace = NULL;
}

bool
XIOSpaceDevice::Init()
{
	// RAM のみ RamDevice::Init() で ram_size が決定してから行う。
	int ram_size = gMainRAM->GetSize();
	for (int i = 0; i < ram_size / 8192; i++) {
		table[i] = gMainRAM;
	}

	// 必要なデバイスは全部埋めたので、この時点で nullptr なエントリを
	// BusErr に差し替える。
	for (auto&& d : table) {
		if (d == nullptr) {
			d = gBusErr;
		}
	}

	// デバッグ表示
	if (0) {
		for (int i = 1536; i < table.size(); i++) {
			printf("[%3d] $%06x ", i, i * 8192);
			assert(table[i]);
			printf("%s\n", table[i]->GetName().c_str());
		}
	}

	return true;
}

// デバイステーブルに dev を追加する。
void
XIOSpaceDevice::AddDevice(IODevice *dev, uint devaddr, uint devlen)
{
	int idx = devaddr / 8192;
	for (int i = idx; i < idx + (devlen + 8191) /  8192; i++) {
		// 重複するケースは未対応
		assertmsg(table[i] == nullptr, "XIOSpaceDevice.AddDevice '%s'",
			table[i]->GetName().c_str());

		table[i] = dev;
	}
}

// アドレス addr に対応する IODevice* を返す。
// addr は事前に 24bit マスクしてあるもの。
inline IODevice *
XIOSpaceDevice::SearchDevice(uint32 addr) const
{
	uint idx = addr / 8192;
	__assume(idx < 2048);
	return table[idx];
}

uint64
XIOSpaceDevice::Read8(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Read8(addr);
}

uint64
XIOSpaceDevice::Read16(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Read16(addr);
}

uint64
XIOSpaceDevice::Read32(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Read32(addr);
}

uint64
XIOSpaceDevice::Write8(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Write8(addr, data);
}

uint64
XIOSpaceDevice::Write16(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Write16(addr, data);
}

uint64
XIOSpaceDevice::Write32(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Write32(addr, data);
}

uint64
XIOSpaceDevice::Peek8(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	return d->Peek8(addr);
}

// このアドレスを担当する次段の IODevice を返す (IOContainerDevice 用)
IODevice *
XIOSpaceDevice::GetDevice(uint32 addr) const
{
	return SearchDevice(addr & 0x00ffffff);
}

// ブートページを切り替える。
// (VM_X68030::SwitchBootPage() の下請け)
void
XIOSpaceDevice::SwitchBootPage(bool isrom)
{
	IODevice *ram;
	IODevice *rom;

	if (isrom) {
		// 0x000000 からの  64KB ( 8エントリ)、
		// 0xfe0000 からの 128KB (16エントリ) のどちらも IPLROM0 が担当する。
		ram = gIPLROM0;
		rom = gIPLROM0;
	} else {
		// 0x000000 からの  64KB ( 8エントリ) に RAM を見せる。
		// 0xfe0000 からの 128KB (16エントリ) に IPLROM1 を見せる。
		ram = gMainRAM;
		rom = gIPLROM1;
	}

	for (int i = 0; i < 8; i++) {
		table[i] = ram;
	}
	for (int i = 0; i < 16; i++) {
		table[table.size() - 16 + i] = rom;
	}
}
