//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// 逆アセンブルウィンドウ
//

#include "wxdisasmmonitor.h"
#include "bus.h"
#include "debugger_private.h"
//#include "mpu680x0.h"

enum {
	ID_TEXT = IDGROUP_DISASM,
	ID_PREVLINE,
	ID_PREVPAGE,
	ID_NEXTLINE,
	ID_NEXTPAGE,

	ID_local_end,	// 最後に置く (チェック用)
};
static_assert(ID_local_end - 1 <= (int)IDGROUP_DISASM_END, "ID exceeded");

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXDisasmWindow, inherited)
	EVT_TEXT_ENTER(ID_TEXT, WXDisasmWindow::OnTextEnter)
	EVT_COMMAND(ID_PREVLINE, NONO_EVT_BUTTON, WXDisasmWindow::OnPrevLine)
	EVT_COMMAND(ID_PREVPAGE, NONO_EVT_BUTTON, WXDisasmWindow::OnPrevPage)
	EVT_COMMAND(ID_NEXTLINE, NONO_EVT_BUTTON, WXDisasmWindow::OnNextLine)
	EVT_COMMAND(ID_NEXTPAGE, NONO_EVT_BUTTON, WXDisasmWindow::OnNextPage)
wxEND_EVENT_TABLE()

// コンストラクタ
// (タイトルは UpdateAddr() でセットする)
WXDisasmWindow::WXDisasmWindow(wxWindow *parent, int idx_)
	: inherited(parent, wxID_ANY, wxEmptyString,
		DEFAULT_STYLE | wxRESIZE_BORDER)
{
	idx = idx_;

	// ↓ topsizer
	//   +--------------------------------------------+
	//   |上枠 →                                     |
	//   |+-----+------+-----+----------+-----+------+|
	//   ||余白 | "<<" | "<" | アドレス | ">" | ">>" ||
	//   |+-----+------+-----+----------+-----+------+|
	//   +--------------------------------------------+
	//   | TextScreen                                 |
	//   +--------------------------------------------+

	auto *topsizer = new wxBoxSizer(wxVERTICAL);

	// 上枠用の下敷きパネル
	auto *ctrlpanel = new wxPanel(this);
	topsizer->Add(ctrlpanel, 0, wxEXPAND);

	// 上枠用の横 sizer
	ctrlbox = new wxBoxSizer(wxHORIZONTAL);

	// アドレスコントロール
	addrctrl = new wxTextCtrl(ctrlpanel, ID_TEXT,
		wxEmptyString, wxDefaultPosition, wxDefaultSize,
		wxTE_PROCESS_ENTER);
	buttons[0] = new WXButton(ctrlpanel, ID_PREVPAGE, wxDefaultSize, "<<");
	buttons[1] = new WXButton(ctrlpanel, ID_PREVLINE, wxDefaultSize, "<");
	buttons[2] = new WXButton(ctrlpanel, ID_NEXTLINE, wxDefaultSize, ">");
	buttons[3] = new WXButton(ctrlpanel, ID_NEXTPAGE, wxDefaultSize, ">>");

	ctrlbox->AddSpacer(3);
	ctrlbox->Add(buttons[0], 0, wxALIGN_CENTER);
	ctrlbox->Add(buttons[1], 0, wxALIGN_CENTER);
	ctrlbox->AddSpacer(3);
	ctrlbox->Add(addrctrl, 0, wxALIGN_CENTER);
	ctrlbox->AddSpacer(3);
	ctrlbox->Add(buttons[2], 0, wxALIGN_CENTER);
	ctrlbox->Add(buttons[3], 0, wxALIGN_CENTER);

	// 上枠の sizer と下敷きを紐付ける
	ctrlpanel->SetSizer(ctrlbox);
	ctrlbox->SetSizeHints(ctrlpanel);

	screen = new WXMonitorPanel(this,
		gMonitorManager->Get(ID_MONITOR_DISASM(idx)));
	topsizer->Add(screen, 1, wxEXPAND);

	SetSizer(topsizer);

	// ウィンドウサイズを確定させる
	fixed_height = ctrlbox->GetSize().y;
	FontChanged();
	DoSize();

	// パネルに来るマウスイベントをこっちに回す
	screen->Connect(wxEVT_MOUSEWHEEL,
		wxMouseEventHandler(WXDisasmWindow::OnMouseWheel), NULL, this);

	// 逆アセンブラオブジェクトを取得
	auto mon = screen->GetMonitor();
	disasm = dynamic_cast<DisasmMonitor *>(mon.obj);
	assert(disasm);

	// 最短の1命令長からアドレスマスクを決定。
	inst_bytes = disasm->GetMD()->inst_bytes;
	inst_mask = ~(inst_bytes - 1);

	// 初期値をコントロールに設定
	UpdateAddr();
}

// デストラクタ
WXDisasmWindow::~WXDisasmWindow()
{
}

bool
WXDisasmWindow::Layout()
{
	// 縦リサイズ可能レイアウト
	return LayoutTextVResize(screen, fixed_height);
}

void
WXDisasmWindow::FontChanged()
{
	inherited::FontChanged();

	// 最小 24x24 にしてみる
	wxSize sz(buttons[0]->GetFontWidth() * 2 + 4,
		      buttons[0]->GetFontHeight() + 4);
	if (sz.x < 24)
		sz.x = 24;
	if (sz.y < 24)
		sz.y = 24;
	for (auto *btn : buttons) {
		btn->SetClientSize(sz);
		btn->SetMinClientSize(sz);
		btn->FontChanged();
	}
}

// テキスト入力イベント
void
WXDisasmWindow::OnTextEnter(wxCommandEvent& event)
{
	// 入力文字列(16進文字列)を数値に変換
	const wxString& wstr = event.GetString();
	uint32 val = strtoul((const char *)wstr.mb_str(), NULL, 16);

	// 最小命令長に整列。
	val &= inst_mask;

	// アドレスを更新
	disasm->SetAddr(val);
	UpdateAddr();
}

// "<"(前行) ボタン押下イベント
void
WXDisasmWindow::OnPrevLine(wxCommandEvent& event)
{
	// 1ワード戻る。m68k だと辻褄が合わないかも知れないが仕方ない。
	disasm->Offset(-inst_bytes);
	UpdateAddr();

#if 0
	// 最短命令長
	int inst_bytes = md->inst_bytes;

	// 1ワード手前をひとまず候補とする
	uint32 begin = laddr - inst_bytes;
	uint32 candidate_begin = (uint32)-1;

	for (;; begin -= inst_bytes) {
		// ここから1回逆アセンブルする
		saddr_t saddr(begin, !user, false);
		DebuggerMemoryStream mem(md, saddr, access_mode, lookup_mode);
		std::string mnemonic;
		std::vector<uint8> local_ir;
		md->Disassemble(mem, mnemonic, local_ir);

		uint32 end = mem.laddr.GetAddr();
		if (end == laddr) {
			// ここから始めると laddr が命令境界になるのでここで確定。
			// 本当はここではなくてもっと手前から始まる長い命令の途中かも
			// 知れないが、きりがないというかどのみち難しいので気にしない。
			candidate_begin = begin;
			break;
		}
		if (end < laddr) {
			if (candidate_begin == (uint32)-1) {
				// 候補がまだ見付からないのに次命令境界が laddr より手前に
				// きてしまったら、そもそも laddr が正しい命令境界でなかった
				// 可能性が高い。ここは laddr (を含む命令) の2命令前だと
				// 思われるので、laddr を含む命令の laddr より手前の命令境界
				// (のはず) を返すほうがよいか。
				candidate_begin = begin + inst_bytes;
				break;
			} else {
				// 開始位置候補がある状態で、次の命令境界が laddr より手前に
				// なったら、さっきの候補が命令境界で確定してよさそう。
			}
			break;
		}
	}

	if (candidate_begin != (uint32)-1) {
		UpdateAddr(candidate_begin);
	}
#endif
}

// ">"(次行) ボタン押下イベント
void
WXDisasmWindow::OnNextLine(wxCommandEvent& event)
{
	disasm->Offset(disasm->GetOffset(1));
	UpdateAddr();
}

// "<<"(前ページ) ボタン押下イベント
void
WXDisasmWindow::OnPrevPage(wxCommandEvent& event)
{
	disasm->Offset(disasm->GetPageOffset(-1));
	UpdateAddr();
}

// ">>"(次ページ) ボタン押下イベント
void
WXDisasmWindow::OnNextPage(wxCommandEvent& event)
{
	disasm->Offset(disasm->GetPageOffset(1));
	UpdateAddr();
}

// アドレス欄の表示を更新する
void
WXDisasmWindow::UpdateAddr()
{
	uint32 addr = disasm->GetAddr();

	// EVT_TEXT を起こさず変更する
	addrctrl->ChangeValue(wxString(string_format("%08x", addr)));

	// カーソルを末尾に
	addrctrl->SetInsertionPointEnd();

	// タイトルを変更
	SetTitle(_("Disassemble") + string_format(" %d: $%08x", idx, addr));

	// 即更新
	Refresh();
}

// マウスホイールイベント
void
WXDisasmWindow::OnMouseWheel(wxMouseEvent& event)
{
	int delta = event.GetWheelRotation();

	// 1回の変化で +/-120 が来る。
	// たいていのアプリケーションで3行ずつの移動に換算することが多いようだが
	// ここでは 4行ずつにする。

	if (delta < 0) {
		disasm->Offset(disasm->GetOffset(4));
		UpdateAddr();
	} else if (delta > 0) {
		// 可変長だと戻り量は分からないので最小単位 x 4 にするしか。
		disasm->Offset(inst_bytes * -4);
		UpdateAddr();
	}
}
