//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メインウィンドウ
//

#include "wxmainframe.h"
#include "fontmanager.h"
#include "wxdipswwindow.h"
#include "wxdisasmmonitor.h"
#include "wxdumpmonitor.h"
#include "wxlcdwindow.h"
#include "wxlogmonitor.h"
#include "wxlogsetting.h"
#include "wxmainview.h"
#include "wxmonitor.h"
#include "wxmpumonitor.h"
#include "wxpalettemonitor.h"
#include "wxplanemonitor.h"
#include "wxromwindow.h"
#include "wxsoftkey.h"
#include "wxstatuspanel.h"
#include "wxuimessage.h"
#include "wxversion.h"
#include "config.h"
#include "debugger.h"
#include "fdc.h"
#include "fdd.h"
#include "mainapp.h"
#include "mpu680x0.h"
#include "mpu88xx0.h"
#include "netdriver.h"
#include "nmi.h"
#include "power.h"
#include "spc.h"
#include "sync.h"
#include "vm.h"
#include <map>

// ID は 0 から (wxID_LOWEST = 5000 未満まで) 使えるっぽい。
// メニュー ID のうちウィンドウのオープン状態を維持する必要があるものは
// (ID が 0 から使えるのなら) 0 から割り当てて配列で管理してしまうのが
// 手っ取り早い。
// ウィンドウに割り当てるウィンドウ ID とメニューイベントで飛んでくる
// メニュー ID は互いに干渉しないはずなので、同じ値を使うことでメニューと
// ウィンドウの紐付けをする。
//
// ウィンドウ(サブウィンドウ)は細かくは以下のように4つに分類できるが、
// 処理上区別が必要なのは A. (デバッガの show コマンドとも共用できる
// サブウィンドウ) とそれ以外のサブウィンドウだけ。
//
// -----------+------------+----
// モニター   | テキスト   | A. 単純なテキストモニタウィンドウ
// ウィンドウ | モニター   |    (SPC, MFP などのデバイスのモニター)
//            | ウィンドウ +----
//            |            | B. それ以外のテキストモニターウィンドウ
//            |            |    (ログ、メモリのように表示位置があるもの、
//            |            |     CMMU のように構造があるもの)
//            +------------+----
//            | それ以外の | C. ビデオモニターウィンドウなど
//            | モニター   |    (ビットマッププレーン、TVRAM など)
//            | ウィンドウ |
// -----------+------------+----
// それ以外のウィンドウ    | D. ソフトウェアキーボードウィンドウなど
// ------------------------+----
enum {
	ID_SUBWIN_END_DUMMY = ID_SUBWIN_END,

	// ここから (ウィンドウでない)メニュー ID
	ID_EXIT,
	ID_RESTART_VM,
	ID_FULL_SPEED,

	ID_INPUT_START,
	ID_INPUT_CHAR,
	ID_INPUT_JP,
	ID_INPUT_END,

	ID_SCALE_START,
	ID_SCALE_0_50,
	ID_SCALE_0_75,
	ID_SCALE_1_00,
	ID_SCALE_1_50,
	ID_SCALE_2_00,
	ID_SCALE_END,

	ID_FONTSIZE_START, // START 以降は fontsize_t と同じ順に並べること
	ID_FONTSIZE_12,
	ID_FONTSIZE_16,
	ID_FONTSIZE_24,
	ID_FONTSIZE_END,

	ID_MONRATE_START,
	ID_MONRATE_20,
	ID_MONRATE_5,
	ID_MONRATE_1,
	ID_MONRATE_END,

	ID_SHOW_STATPANEL,

	// FD
	ID_FD_INSERT_0,
	ID_FD_INSERT_1,
	ID_FD_INSERT_2,
	ID_FD_INSERT_3,
	ID_FD_MISINSERT_0,
	ID_FD_MISINSERT_1,
	ID_FD_MISINSERT_2,
	ID_FD_MISINSERT_3,
	ID_FD_EJECT_0,
	ID_FD_EJECT_1,
	ID_FD_EJECT_2,
	ID_FD_EJECT_3,
	ID_FD_FORCE_0,
	ID_FD_FORCE_1,
	ID_FD_FORCE_2,
	ID_FD_FORCE_3,

	// CD/MO (リムーバブルデバイス)
	ID_CD_INSERT_0,
	ID_CD_INSERT_1,
	ID_CD_INSERT_2,
	ID_CD_INSERT_3,
	ID_CD_INSERT_4,
	ID_CD_INSERT_5,
	ID_CD_INSERT_6,
	ID_CD_INSERT_7,
	ID_CD_EJECT_0,
	ID_CD_EJECT_1,
	ID_CD_EJECT_2,
	ID_CD_EJECT_3,
	ID_CD_EJECT_4,
	ID_CD_EJECT_5,
	ID_CD_EJECT_6,
	ID_CD_EJECT_7,
	ID_CD_FORCE_0,
	ID_CD_FORCE_1,
	ID_CD_FORCE_2,
	ID_CD_FORCE_3,
	ID_CD_FORCE_4,
	ID_CD_FORCE_5,
	ID_CD_FORCE_6,
	ID_CD_FORCE_7,

	ID_MOUSEMODE,
	ID_POWER_BUTTON,
	ID_RESET_BUTTON,
	ID_NMI_BUTTON,
	ID_KEYBOARD_CONNECT,

	ID_local_end,	// 最後に置く (チェック用)
};
static_assert(ID_local_end - 1 <= (int)IDGROUP_MAINFRAME_END, "ID exceeded");

#define ID_FD_INSERT(id)	(ID_FD_INSERT_0 + (id))
#define ID_FD_MISINSERT(id)	(ID_FD_MISINSERT_0 + (id))
#define ID_FD_EJECT(id)		(ID_FD_EJECT_0 + (id))
#define ID_FD_FORCE(id)		(ID_FD_FORCE_0 + (id))
#define ID_CD_INSERT(id)	(ID_CD_INSERT_0 + (id))
#define ID_CD_EJECT(id) 	(ID_CD_EJECT_0 + (id))
#define ID_CD_FORCE(id) 	(ID_CD_FORCE_0 + (id))

WXMainFrame *gMainFrame;

wxBEGIN_EVENT_TABLE(WXMainFrame, inherited)
	EVT_CLOSE(WXMainFrame::OnClose)
	EVT_MENU(ID_EXIT, WXMainFrame::OnExit)

	EVT_MENU(ID_RESTART_VM, WXMainFrame::OnRestartVM)
	EVT_MENU(ID_FULL_SPEED, WXMainFrame::OnFullSpeed)
	EVT_UPDATE_UI(ID_FULL_SPEED, WXMainFrame::OnFullSpeedUI)
	EVT_MENU_RANGE(ID_INPUT_START + 1, ID_INPUT_END - 1,
		WXMainFrame::OnMenuInput)
	EVT_UPDATE_UI_RANGE(ID_INPUT_START + 1, ID_INPUT_END -1,
		WXMainFrame::OnMenuInputUI)
	EVT_MENU_RANGE(ID_SCALE_START + 1, ID_SCALE_END - 1,
		WXMainFrame::OnMenuScale)
	EVT_UPDATE_UI_RANGE(ID_SCALE_START + 1, ID_SCALE_END - 1,
		WXMainFrame::OnMenuScaleUI)
	EVT_MENU_RANGE(ID_FONTSIZE_START + 1, ID_FONTSIZE_END - 1,
		WXMainFrame::OnMenuFontSize)
	EVT_UPDATE_UI_RANGE(ID_FONTSIZE_START + 1, ID_FONTSIZE_END - 1,
		WXMainFrame::OnMenuFontSizeUI)
	EVT_MENU_RANGE(ID_MONRATE_START + 1, ID_MONRATE_END - 1,
		WXMainFrame::OnMenuMonRate)
	EVT_UPDATE_UI_RANGE(ID_MONRATE_START + 1, ID_MONRATE_END - 1,
		WXMainFrame::OnMenuMonRateUI)
	EVT_MENU(ID_SHOW_STATPANEL, WXMainFrame::OnShowStatusPanel)
	EVT_UPDATE_UI(ID_SHOW_STATPANEL, WXMainFrame::OnShowStatusPanelUI)
	EVT_MENU_RANGE(ID_FD_INSERT_0, ID_FD_INSERT_3,
		WXMainFrame::OnFDInsert)
	EVT_UPDATE_UI_RANGE(ID_FD_INSERT_0, ID_FD_INSERT_3,
		WXMainFrame::OnFDInsertUI)
	EVT_MENU_RANGE(ID_FD_EJECT_0, ID_FD_EJECT_3,
		WXMainFrame::OnFDEject)
	EVT_UPDATE_UI_RANGE(ID_FD_EJECT_0, ID_FD_EJECT_3,
		WXMainFrame::OnFDEjectUI)
	EVT_MENU_RANGE(ID_FD_FORCE_0, ID_FD_FORCE_3,
		WXMainFrame::OnFDForce)
	EVT_UPDATE_UI_RANGE(ID_FD_FORCE_0, ID_FD_FORCE_3,
		WXMainFrame::OnFDForceUI)
	EVT_MENU_RANGE(ID_CD_INSERT_0, ID_CD_INSERT_7,
		WXMainFrame::OnCDInsert)
	EVT_UPDATE_UI_RANGE(ID_CD_INSERT_0, ID_CD_INSERT_7,
		WXMainFrame::OnCDInsertUI)
	EVT_MENU_RANGE(ID_CD_EJECT_0, ID_CD_EJECT_7,
		WXMainFrame::OnCDEject)
	EVT_UPDATE_UI_RANGE(ID_CD_EJECT_0, ID_CD_EJECT_7,
		WXMainFrame::OnCDEjectUI)
	EVT_MENU_RANGE(ID_CD_FORCE_0, ID_CD_FORCE_7,
		WXMainFrame::OnCDForce)
	EVT_UPDATE_UI_RANGE(ID_CD_FORCE_0, ID_CD_FORCE_7,
		WXMainFrame::OnCDForceUI)
	EVT_MENU(ID_MOUSEMODE, WXMainFrame::OnMouseMode)
	EVT_UPDATE_UI(ID_MOUSEMODE, WXMainFrame::OnMouseModeUI)
	EVT_MENU(ID_POWER_BUTTON, WXMainFrame::OnPowerButton)
	EVT_UPDATE_UI(ID_POWER_BUTTON, WXMainFrame::OnPowerButtonUI)
	EVT_MENU(ID_RESET_BUTTON, WXMainFrame::OnResetButton)
	EVT_MENU(ID_NMI_BUTTON, WXMainFrame::OnNMIButton)
	EVT_MENU(ID_KEYBOARD_CONNECT, WXMainFrame::OnKeyboardConnect)
	EVT_UPDATE_UI(ID_KEYBOARD_CONNECT, WXMainFrame::OnKeyboardConnectUI)
	EVT_MENU_RANGE(0, ID_SUBWIN_END, WXMainFrame::OnWindow)
	EVT_UPDATE_UI_RANGE(0, ID_SUBWIN_END, WXMainFrame::OnWindowUI)
	EVT_MENU(wxID_ABOUT, WXMainFrame::OnAbout)
wxEND_EVENT_TABLE()

WXMainFrame::WXMainFrame(wxWindow *parent)
	: inherited(parent, wxID_ANY, "nono", wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_FRAME_STYLE & ~wxMAXIMIZE_BOX & ~wxRESIZE_BORDER)
{
	VMType vmtype;
	vmtype = gMainApp.GetVMType();

	// メニュー
	wxMenuBar *menubar = new wxMenuBar();
	// メニュー > ファイル
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_RESTART_VM, _("&Restart VM"), "");
	menuFile->Append(ID_EXIT, _("E&xit"));
	menubar->Append(menuFile, _("&File"));

	// メニュー > 設定
	wxMenu *menuSet = new wxMenu();
	menuSet->Append(ID_FULL_SPEED, _("Full &Speed Mode"), "", wxITEM_CHECK);
	if (gKeyboard) {
		wxMenu *menuSetInput = new wxMenu();
		menuSetInput->AppendRadioItem(ID_INPUT_CHAR, _("By &Character"));
		menuSetInput->AppendRadioItem(ID_INPUT_JP, _("&JP Keyboard"));
		menuSet->AppendSubMenu(menuSetInput, _("&Key Input"));
	}
	menubar->Append(menuSet, _("&Setting"));

	// メニュー > 表示
	wxMenu *menuDisplay = new wxMenu();
	wxMenu *menuViewScale = new wxMenu();
	menuViewScale->Append(ID_SCALE_0_50, "x0.5", "", wxITEM_CHECK);
	menuViewScale->Append(ID_SCALE_0_75, "x0.75", "", wxITEM_CHECK);
	menuViewScale->Append(ID_SCALE_1_00, "x1.0", "", wxITEM_CHECK);
	menuViewScale->Append(ID_SCALE_1_50, "x1.5", "", wxITEM_CHECK);
	menuViewScale->Append(ID_SCALE_2_00, "x2.0", "", wxITEM_CHECK);
	menuDisplay->AppendSubMenu(menuViewScale, _("&Scale"));
	wxMenu *menuFontSize = new wxMenu();
	menuFontSize->AppendRadioItem(ID_FONTSIZE_12, "1&2");
	menuFontSize->AppendRadioItem(ID_FONTSIZE_16, "1&6");
	menuFontSize->AppendRadioItem(ID_FONTSIZE_24, "2&4");
	menuDisplay->AppendSubMenu(menuFontSize, _("&Font Size"));
	wxMenu *menuMonRate = new wxMenu();
	menuMonRate->AppendRadioItem(ID_MONRATE_20, _("High (&20Hz)"));
	menuMonRate->AppendRadioItem(ID_MONRATE_5, _("Middle (&5Hz)"));
	menuMonRate->AppendRadioItem(ID_MONRATE_1, _("Low (&1Hz)"));
	menuDisplay->AppendSubMenu(menuMonRate, _("&Monitor Speed"));
	menuDisplay->Append(ID_SHOW_STATPANEL, _("Show Status &Panel"), "",
		wxITEM_CHECK);
	menubar->Append(menuDisplay, _("&View"));

	// メニュー > デバイス
	wxMenu *menuDevice = new wxMenu();
	CreateRemovableMenu(menuDevice);
	if (gKeyboard) {
		menuDevice->Append(ID_SUBWIN_SOFTKEY, _("Software &Keyboard"),
			"", wxITEM_CHECK);
		menuDevice->Append(ID_MOUSEMODE, _("&Mouse Mode\tCTRL+F12"),
			"", wxITEM_CHECK);
	}
	if (gMainApp.Has(VMCap::LUNA)) {
		menuDevice->Append(ID_SUBWIN_LCDPANEL, _("Front &LCD"),
			"", wxITEM_CHECK);
	}
	menuDevice->AppendSeparator();
	wxMenu *menuOper = new wxMenu();
	wxItemKind power_button_kind;
	if (gMainApp.Has(VMCap::LUNA)) {
		power_button_kind = wxITEM_NORMAL;
	} else {
		power_button_kind = wxITEM_CHECK;
	}
	menuOper->Append(ID_POWER_BUTTON, _("Push &Power button"),
		"", power_button_kind);
	menuOper->Append(ID_RESET_BUTTON, _("Push &Reset button"));
	if (gMainApp.Has(VMCap::LUNA)) {
		menuOper->Append(ID_NMI_BUTTON, _("Push &ABORT button"));
	} else if (gMainApp.Has(VMCap::X68030)) {
		menuOper->Append(ID_NMI_BUTTON, _("Push &Interrupt button"));
	}
	// これ以降があればここにセパレータがほしい
	bool sep = false;
	if (gMainApp.Has(VMCap::LUNA | VMCap::NEWS)) {
		if (sep == false) {
			menuOper->AppendSeparator();
			sep = true;
		}
		menuOper->Append(ID_SUBWIN_DIPSW, "&DIPSW", "", wxITEM_CHECK);
	}
	if (gKeyboard) {
		if (sep == false) {
			menuOper->AppendSeparator();
			sep = true;
		}
		menuOper->Append(ID_KEYBOARD_CONNECT, _("Connect &Keyboard"),
			"", wxITEM_CHECK);
	}
	menuDevice->AppendSubMenu(menuOper, _("&Operation"));
	menubar->Append(menuDevice, _("&Device"));

	// 機種によってモニタウィンドウを追加する
#define AppendIf(menu, id, label)	do { \
	VMCap vmcap_ = gMonitorManager->GetVMCap(id); \
	if (gMainApp.Has(vmcap_)) { \
		(menu)->Append((id), (label), "", wxITEM_CHECK); \
	} \
} while (0)
#define AppendIfSeparator(menu, vmcap_)	do { \
	if (gMainApp.Has(vmcap_)) { \
		(menu)->AppendSeparator(); \
	} \
} while (0)

	// メニュー > モニター
	wxMenu *menuMonitor = new wxMenu();
	AppendIf(menuMonitor, ID_MONITOR_SCHEDULER,	_("&Scheduler"));
	AppendIf(menuMonitor, ID_SUBWIN_LOG,		_("&Log"));

	wxMenu *menuDebug = new wxMenu();
	AppendIf(menuDebug, ID_MONITOR_BREAKPOINT,	_("&Breakpoint"));
	AppendIf(menuDebug, ID_SUBWIN_BRHIST,		_("Branch &History"));
	AppendIf(menuDebug, ID_SUBWIN_EXHIST,		_("&Exception History"));
	AppendIf(menuDebug, ID_SUBWIN_VECTOR,		_("&Vector Table"));
	menuDebug->AppendSeparator();
	AppendIf(menuDebug, ID_SUBWIN_LOGSETTING,	_("&Log Level Setting"));
	menuMonitor->AppendSubMenu(menuDebug, _("&Debug"));

	wxMenu *menuMPU = new wxMenu();
	AppendIf(menuMPU, ID_MONITOR_MPUREG,		_("&Register"));
	AppendIf(menuMPU, ID_MONITOR_MPUATC,		  "&ATC");
	AppendIfSeparator(menuMPU, VMCap::M88K);
	AppendIf(menuMPU, ID_MONITOR_CMMU7,			 "CMMU#7" + _("(Inst)"));
	AppendIf(menuMPU, ID_MONITOR_ATC7,			  "ATC#7");
	AppendIf(menuMPU, ID_SUBWIN_CACHE7,			_("Cache") + "#7");
	AppendIfSeparator(menuMPU, VMCap::M88K);
	AppendIf(menuMPU, ID_MONITOR_CMMU6,			  "CMMU#6" + _("(Data)"));
	AppendIf(menuMPU, ID_MONITOR_ATC6,			  "ATC#6");
	AppendIf(menuMPU, ID_SUBWIN_CACHE6,			_("Cache") + "#6");
	menuMonitor->AppendSubMenu(menuMPU, "MP&U");

	// DISASMn、MEMDUMPn のラベルは UpdateUI で書き直すのでここでは適当に
	// 埋めておく。(empty だと怒られる)
	wxMenu *menuMemdump = new wxMenu();
	for (int i = 0; i < MAX_MEMDUMP_MONITOR; i++) {
		AppendIf(menuMemdump, ID_MONITOR_MEMDUMP(i), "memdump");
	}
	menuMonitor->AppendSubMenu(menuMemdump, _("&Memory Dump"));
	wxMenu *menuDisasm = new wxMenu();
	for (int i = 0; i < MAX_DISASM_MONITOR; i++) {
		AppendIf(menuDisasm, ID_MONITOR_DISASM(i), "disasm");
	}
	menuMonitor->AppendSubMenu(menuDisasm, _("Dis&assemble"));
	// MONITOR_RTC は X68k の RTC、LUNA では MONITOR_MK48T02 だが今の所
	// NVRAM 側のことは表示してないのでラベルも RTC にしてある。
	AppendIf(menuMonitor, ID_MONITOR_INTERRUPT,	_("Interrupt"));
	AppendIf(menuMonitor, ID_MONITOR_SPC,		  "SPC");
	AppendIf(menuMonitor, ID_MONITOR_SCSIDEVS,	_("SCSI Devices"));
	AppendIf(menuMonitor, ID_MONITOR_RTC,		  "RTC");
	AppendIf(menuMonitor, ID_MONITOR_MK48T02,	  "RTC");
	AppendIf(menuMonitor, ID_MONITOR_DMAC,		  "DMAC");
	AppendIf(menuMonitor, ID_MONITOR_MFP,		  "MFP");
	AppendIf(menuMonitor, ID_MONITOR_FDC,		  "FDC");
	AppendIf(menuMonitor, ID_MONITOR_SCC,		  "SCC");
	AppendIf(menuMonitor, ID_MONITOR_SIO,		  "SIO");
	AppendIf(menuMonitor, ID_MONITOR_PIO0,		  "PIO0");
	AppendIf(menuMonitor, ID_MONITOR_KEYBOARD,	_("Keyboard/Mouse"));
	AppendIf(menuMonitor, ID_MONITOR_LANCE,		  "Lance");
	AppendIf(menuMonitor, ID_MONITOR_LCD,		  "LCD");
	AppendIf(menuMonitor, ID_MONITOR_SYSCLK,	_("System Clock"));
	AppendIf(menuMonitor, ID_MONITOR_CRTC,		  "CRTC");
	AppendIf(menuMonitor, ID_MONITOR_LUNAVC,	_("Video Control"));
	AppendIf(menuMonitor, ID_MONITOR_BT454,		  "Bt454/458");
	AppendIf(menuMonitor, ID_SUBWIN_PALETTE,	_("Palette"));
	AppendIf(menuMonitor, ID_SUBWIN_TEXTPAL,	_("Text Palette"));
	AppendIf(menuMonitor, ID_MONITOR_RENDERER,	_("Renderer"));
	AppendIf(menuMonitor, ID_SUBWIN_LUNAFB,		_("Bitmap Plane"));
	AppendIf(menuMonitor, ID_SUBWIN_TVRAM,		_("Text VRAM"));
	AppendIf(menuMonitor, ID_MONITOR_ROMCONS,	_("ROM Console (provisional)"));
	AppendIf(menuMonitor, ID_SUBWIN_ROM,		  "ROM");

	wxMenu *menuHost = new wxMenu();
	AppendIf(menuHost, ID_MONITOR_WXHOSTINFO,	_("Host &Information"));
	AppendIf(menuHost, ID_MONITOR_HOSTNET,		_("Host &Network"));
	AppendIf(menuHost, ID_MONITOR_HOSTCOM,		_("Host &Serial Port"));
	menuMonitor->AppendSubMenu(menuHost, _("&Host"));

	menubar->Append(menuMonitor, _("&Monitor"));

	// メニュー > ヘルプ
#if defined(__WXOSX__)
	// OSX では wxID_ABOUT はリンゴメニューに入れるしきたりになっている。
	// Append() すると Quit より下に来てしまうので、Prepend() で上に追加。
	// 今のところヘルプメニューから About ダイアログを抜くとメニューが空に
	// なってしまうので、ヘルプメニューの追加ごと抑制する。
	menubar->OSXGetAppleMenu()->Prepend(wxID_ABOUT, _("About nono"));
#else
	wxMenu *menuHelp = new wxMenu();
	menuHelp->Append(wxID_ABOUT, _("About nono"));
	menubar->Append(menuHelp, _("&Help"));
#endif

	SetMenuBar(menubar);

	wxString title("nono");
	switch (vmtype) {
	 case VMType::X68030:
		title += "(X68030)";
		break;
	 case VMType::LUNA1:
		title += "(LUNA-I)";
		break;
	 case VMType::LUNA88K:
		title += "(LUNA-88K)";
		break;
	 case VMType::NEWS:
		title += "(NWS-1750)";
		break;
	 default:
		break;
	}
	SetTitle(title);

	// UpdateUI イベントは Idle 時に時折チェックされるらしいが、
	// (メインビューとサブウィンドウを合わせた) タイマーイベントの粒度が
	// 小さいためか、秒間120回とかいうレベルで UpdateUIEvent が飛んでくる
	// っぽい (全タイマーをとめたら治る模様)。
	// さすがにそんなに来られても気持ち悪いのと、どうせメニュー開いた瞬間
	// 表示しようとする前には一度飛んでくるのだから、アイドル時のチェックは
	// いらない気がする。間隔広げておいても、ぱっと見副作用はなさそうだけど
	// 念のため事象が分かりやすいように 10秒くらいにしておく。
	wxUpdateUIEvent::SetUpdateInterval(10000);

	// WXStatusPanel コンストラクタが gMainView を参照するので
	// gMainView のほうを先に作成すること。
	gMainView = new WXMainView(this);
	gStatusPanel = new WXStatusPanel(this);

	// SHOW イベントを最初にも一度起こしたいため
	// ステータスパネルを一旦非表示にする (これで SHOW(false) イベントが飛ぶ)
	gStatusPanel->Hide();
	const ConfigItem& item_show = gConfig->Find("show-statuspanel");
	if (item_show.AsInt() != 0) {
		// 必要なら改めて表示 (これで SHOW(true) イベントが飛ぶ)
		gStatusPanel->Show();
	}

	// Sizer 使わず自前でレイアウトする。
	SetAutoLayout(true);

	// モニタの更新頻度は初期値だけ設定から読んで、あとはここで管理する。
	// 不正値ならデフォルトの 15Hz にしておくか
	const ConfigItem& item_rate = gConfig->Find("monitor-rate");
	monitor_rate = item_rate.AsInt();
	if (monitor_rate < 1 || monitor_rate > 60) {
		monitor_rate = 15;
	}

	// Activate イベントは、このトップレベルウィンドウにウィンドウフォーカスが
	// 来た or 外れたことを示すので、wxFrame (というかたぶん wxTopLevelWindow)
	// を継承している WXMainFrameに飛んでくるのだが、処理の都合上 WXMainView で
	// 扱いたいので、そちらに回す。
	Connect(wxEVT_ACTIVATE, wxActivateEventHandler(WXMainView::OnActivate),
		NULL, gMainView);

	// ウィンドウ表示後に続きの処理をするための単発アイドルイベント。
	Connect(wxEVT_IDLE, wxIdleEventHandler(WXMainFrame::OnIdle));

	// VM からの通知を接続
	WXUIMessage::Connect(UIMessage::APPEXIT, this,
		wxCommandEventHandler(WXMainFrame::OnExit));
	WXUIMessage::Connect(UIMessage::SCSI_MEDIA_FAILED, this,
		wxCommandEventHandler(WXMainFrame::OnInsertFailed));
	WXUIMessage::Connect(UIMessage::FDD_MEDIA_FAILED, this,
		wxCommandEventHandler(WXMainFrame::OnInsertFailed));
	if (gMPU680x0) {
		WXUIMessage::Connect(UIMessage::HALT, this,
			wxCommandEventHandler(WXMainFrame::OnHaltNotify));
	}
}

WXMainFrame::~WXMainFrame()
{
	WXUIMessage::Disconnect(UIMessage::APPEXIT, this,
		wxCommandEventHandler(WXMainFrame::OnExit));
	WXUIMessage::Disconnect(UIMessage::SCSI_MEDIA_FAILED, this,
		wxCommandEventHandler(WXMainFrame::OnInsertFailed));
	WXUIMessage::Disconnect(UIMessage::FDD_MEDIA_FAILED, this,
		wxCommandEventHandler(WXMainFrame::OnInsertFailed));
	if (gMPU680x0) {
		WXUIMessage::Disconnect(UIMessage::HALT, this,
			wxCommandEventHandler(WXMainFrame::OnHaltNotify));
	}
}

// リムーバブルデバイス用のメニューを作成して menuDevice に追加する
void
WXMainFrame::CreateRemovableMenu(wxMenu *menuDevice)
{
	wxMenu *menu;
	bool added = false;

	// FD
	if (gFDC) {
		const std::vector<FDDDevice *>& fdd_list = gFDC->GetFDD();
		for (const auto fdd : fdd_list) {
			int unit = fdd->GetUnitNo();
			fdd_array[unit] = fdd;

			menu = CreateFDMenu(unit);
			std::string label = string_format("FD%d", unit);
			menuDevice->AppendSubMenu(menu, label);
			added = true;
		}
	}

	// CD, MO
	if (gSPC) {
		for (int id = 0; id < 8; id++) {
			scsidev[id] = gSPC->GetTarget(id);
			if (scsidev[id]) {
				menu = CreateSCSIRemovableMenu(id);
				if (menu) {
					std::string label = string_format("%s (ID%d)",
						SCSI::GetDevTypeName(scsidev[id]->GetDevType()), id);
					menuDevice->AppendSubMenu(menu, label);
					added = true;
				}
			}
		}
	}

	// 1つでも追加したらセパレータ
	if (added) {
		menuDevice->AppendSeparator();
	}
}

// FD 用のメニューを作成して返す。
wxMenu *
WXMainFrame::CreateFDMenu(int unit) const
{
	wxMenu *menu = new wxMenu();
	// Insert の実際のラベルは UpdateUI でセットする。
	// ただし何か入れておかないといけないので適当にセット。
	menu->Append(ID_FD_INSERT(unit), "Insert", "");
	// XXX 未実装
	//menu->Append(ID_FD_MISINSERT(unit), _("Insert incorrectly"), "");
	menu->Append(ID_FD_EJECT(unit), _("&Eject"), "");
	menu->Append(ID_FD_FORCE(unit), _("&Force Eject"), "");
	return menu;
}

// SCSI リムーバブルデバイス用のメニューを作成して返す。
// id はデバイスが接続されていること (リムーバブルデバイスであるかは問わない)。
// リムーバブルデバイスでなければ NULL を返す。
// メニューバーとポップアップメニュー共通。
wxMenu *
WXMainFrame::CreateSCSIRemovableMenu(uint id) const
{
	const auto *dev = scsidev[id];
	assert(dev);

	switch (dev->GetDevType()) {
	 case SCSI::DevType::CD:
	 case SCSI::DevType::MO:
	 {
		wxMenu *menu = new wxMenu();
		// Insert の実際のラベルは UpdateUI でセットする。
		// ただし何か入れておかないといけないので適当にセット。
		menu->Append(ID_CD_INSERT(id), "Insert", "");
		menu->Append(ID_CD_EJECT(id), _("&Eject"), "");
		menu->Append(ID_CD_FORCE(id), _("&Force Eject"), "");
		return menu;
	 }
	 default:
		return NULL;
	}
}

// (設定反映などの) 初期化。
// 成功すれば true、失敗すれば false を返す。
bool
WXMainFrame::Init()
{
	if (!gMainView->Init()) {
		return false;
	}

	return true;
}

bool
WXMainFrame::Layout()
{
	if (inherited::Layout() == false) {
		return false;
	}

	wxSize msize = gStatusPanel->GetMinClientSize();
	wxSize ssize = gStatusPanel->GetClientSize();
	wxSize vsize = gMainView->GetClientSize();

	if (gStatusPanel->IsShown()) {
		ssize.y = msize.y;
	} else {
		ssize.y = 0;
	}

	wxSize size;
	size.x = std::max(msize.x, vsize.x);
	size.y = ssize.y + vsize.y;

	if (ssize.y != 0) {
		gStatusPanel->SetSize(0, 0, size.x, ssize.y);
	}
	gMainView->SetSize(0, ssize.y, size.x, vsize.y);
	SetClientSize(size);

	return true;
}

// ウィンドウ作成イベント (を実現するためここではアイドルイベントを使う)
//
// 初期化(っぽい)もののうちコンストラクタでは早すぎるものは、コンストラクタ
// 直後に起こすこのイベント内で処理する。使えそうなイベントはいくつかあり、
// 自前の Create (今はもうない) -> WindowCreate -> Idle -> Paint の順で
// 起きるようだ。
// 自前の Create 時点では早すぎて、ここでサブウィンドウを作成するとサイズが 0
// みたいなウィンドウが一瞬出来てしまい GTK のワーニングが出る (その後サイズを
// 矯正すれば表示自体は出来る)。
// WindowCreate 時点ではその問題は起きないが、ここはまだメインウィンドウ表示
// 前なので、ここでサブウィンドウを表示した後メインウィンドウを表示すると
// (ウィンドウマネージャにもよるかもだが)、メインウィンドウがサブウィンドウを
// 隠してしまう。
// Idle イベント時点まで来ると上の2つの問題は回避できるようだ。
// Paint イベントの時点でもタイミング的には問題ないが、環境によっては(?)
// このイベント自体が飛んで来ないことがあるようだ。WXMainFrame の再描画は
// 子コントロールが行っており、そのため WXMainFrame 自身が再描画しなければ
// ならない領域は存在しないからかも知れないが詳細未調査。
void
WXMainFrame::OnIdle(wxIdleEvent& ev)
{
	// UIMessage の処理を開始する。
	// 少なくとも WXMainFrame のコンストラクタが完了した後のほうがいいはず。
	UIMessage::Attach(&WXUIMessage::Process);

	// ウィンドウを表示
	for (int i = 0; i < subwindow_start.size(); i++) {
		if (subwindow_start[i]) {
			DoWindow(i, true);
		}
	}

	// 電源投入
	gPower->PowerButton();

	// 一度処理したら、このイベントを外す
	Disconnect(wxEVT_IDLE, wxIdleEventHandler(WXMainFrame::OnIdle));
}

// ウィンドウクローズイベント
void
WXMainFrame::OnClose(wxCloseEvent& event)
{
	// ステータスパネルの更新を停止
	gStatusPanel->Close();
	// メインビューの描画を停止
	gMainView->Close();

	// サブウィンドウを全部閉じる
	for (int i = 0; i < ID_SUBWIN_MAX; i++) {
		if ((bool)windows[i]) {
			windows[i]->Close();
		}
	}

	// VM を解放
	gVM->Dispose();

	// 後はデフォルト処理
	event.Skip();
}

// メニューの「Restart VM」イベント
void
WXMainFrame::OnRestartVM(wxCommandEvent& event)
{
	gPower->MakeRestart();
}

// メニューの「終了」イベント
void
WXMainFrame::OnExit(wxCommandEvent& event)
{
	// ウィンドウを閉じる
	Close();
}

// メニューの「高速モード」イベント
void
WXMainFrame::OnFullSpeed(wxCommandEvent& event)
{
	if (event.IsChecked()) {
		// チェックが入ったので、高速モードにする
		gSync->RequestFullSpeed(true);
	} else {
		// チェックが消えたので、通常モードにする
		gSync->RequestFullSpeed(false);
	}
}

// メニューの「高速モード」UI
void
WXMainFrame::OnFullSpeedUI(wxUpdateUIEvent& event)
{
	bool is_fullspeed = gSync->GetFullSpeed();
	event.Check(is_fullspeed);
}

// メニューの「設定 > 入力 > *」の共通イベント
void
WXMainFrame::OnMenuInput(wxCommandEvent& event)
{
	bool char_mode = (event.GetId() == ID_INPUT_CHAR);

	gMainView->SetCharInputMode(char_mode);
}

// メニューの「設定 > 入力 > *」の共通 UI
void
WXMainFrame::OnMenuInputUI(wxUpdateUIEvent& event)
{
	bool char_mode = gMainView->GetCharInputMode();

	if (char_mode) {
		event.Check(event.GetId() == ID_INPUT_CHAR);
	} else {
		event.Check(event.GetId() == ID_INPUT_JP);
	}
}

// スケール一覧
static std::map<int, double> scaletable {
	{ ID_SCALE_0_50,	0.50 },
	{ ID_SCALE_0_75,	0.75 },
	{ ID_SCALE_1_00,	1.00 },
	{ ID_SCALE_1_50,	1.50 },
	{ ID_SCALE_2_00,	2.00 },
};

// メニューの「表示 > スケール > *」の共通イベント
void
WXMainFrame::OnMenuScale(wxCommandEvent& event)
{
	for (const auto& p : scaletable) {
		int id = p.first;
		double scale = p.second;

		// 一致するはず
		if (event.GetId() == id) {
			gMainView->DoResize(scale);

			// 再レイアウト
			Layout();
			return;
		}
	}
}

// メニューの「表示 > スケール > *」の共通 UI
void
WXMainFrame::OnMenuScaleUI(wxUpdateUIEvent& event)
{
	double current_scale = gMainView->GetScale();

	for (const auto& p : scaletable) {
		int id = p.first;
		double scale = p.second;

		if (event.GetId() == id) {
			event.Check((bool)(current_scale == scale));
			break;
		}
	}
}

// メニューの「表示 > フォントサイズ > *」の共通イベント
void
WXMainFrame::OnMenuFontSize(wxCommandEvent& event)
{
	FontId fontid;

	assertmsg(
		ID_FONTSIZE_12 <= event.GetId() && event.GetId() <= ID_FONTSIZE_24,
		"event.GetId()=%d", event.GetId());
	switch (event.GetId()) {
	 case ID_FONTSIZE_12:
		fontid = FontId::_6x12;
		break;
	 case ID_FONTSIZE_16:
		fontid = FontId::_8x16;
		break;
	 case ID_FONTSIZE_24:
		fontid = FontId::_12x24;
		break;
	 default:
		__unreachable();
		break;
	}

	if (fontid != gFontManager->GetFontId()) {
		gFontManager->SetFont(fontid);

		// アクティブなサブウィンドウ全員にフォント変更通知
		for (int i = 0; i < ID_SUBWIN_MAX; i++) {
			if ((bool)windows[i]) {
				WXSubWindow *subwin =
					dynamic_cast<WXSubWindow *>(windows[i].get());
				subwin->FontChanged();
			}
		}
		// メインウィンドウはここで直接ステータスパネルに通知
		gStatusPanel->FontChanged();

		// ステータスパネルの大きさが変わるので再レイアウト
		Layout();
	}
}

// メニューの「表示 > フォントサイズ > *」共通 UI
void
WXMainFrame::OnMenuFontSizeUI(wxUpdateUIEvent& event)
{
	FontId current_size = gFontManager->GetFontId();
	FontId event_size = (FontId)(event.GetId() - (ID_FONTSIZE_START+1));

	event.Check(current_size == event_size);
}

// 更新頻度 ID -> 数値 (メニューのテキスト、ID と一致させること)
/*static*/ const int
WXMainFrame::IDtoMonRate[] = {
	20,		// ID_MONRATE_20
	5,		// ID_MONRATE_5
	1,		// ID_MONRATE_1
};

// メニューの「表示 > モニタ更新頻度 > *」の共通イベント
void
WXMainFrame::OnMenuMonRate(wxCommandEvent& event)
{
	int id = event.GetId();
	assert(ID_MONRATE_START < id && id < ID_MONRATE_END);
	monitor_rate = IDtoMonRate[id - (ID_MONRATE_START + 1)];

	// アクティブなサブウィンドウのテキストパネル全員に通知
	for (int i = 0; i < ID_SUBWIN_MAX; i++) {
		if ((bool)windows[i]) {
			auto *subwin = dynamic_cast<WXSubWindow *>(windows[i].get());
			for (wxWindow *child : subwin->GetChildren()) {
				auto *panel = dynamic_cast<WXMonitorPanel*>(child);
				if (panel) {
					panel->SetRate(monitor_rate);
				}
			}
		}
	}
}

// メニューの「表示 > モニタ更新頻度 > *」共通 UI
void
WXMainFrame::OnMenuMonRateUI(wxUpdateUIEvent& event)
{
	int id = event.GetId();
	assert(ID_MONRATE_START < id && id < ID_MONRATE_END);
	int event_rate = IDtoMonRate[id - (ID_MONRATE_START + 1)];

	event.Check(monitor_rate == event_rate);
}

// メニューの「表示 > ステータスパネルを表示」
void
WXMainFrame::OnShowStatusPanel(wxCommandEvent& event)
{
	// enable ならチェックを付けようとして呼ばれた
	bool enable = event.IsChecked();

	gStatusPanel->Show(enable);

	// 再レイアウト
	Layout();
}

// メニューの「表示 > ステータスパネルを表示」UI
void
WXMainFrame::OnShowStatusPanelUI(wxUpdateUIEvent& event)
{
	event.Check(gStatusPanel->IsShown());
}

// メニューの「デバイス > FD > 挿入」
void
WXMainFrame::OnFDInsert(wxCommandEvent& event)
{
	int unit = event.GetId() - ID_FD_INSERT_0;
	const auto *fdd = fdd_array[unit];
	assert(fdd);

	std::string title(_("Open a FD image"));
	std::string wildcard("FD images (*.xdf;*.img;*.fs)"
		"|*.xdf;*.XDF;*.img;*.IMG;*.fs;*.FS"
		"|All files (*.*)|*.*");

	// ディレクトリは、まだ一度も使ってなければ VM ディレクトリにする
	if (current_dir.empty()) {
		current_dir = gMainApp.GetVMDir();
	}

	wxFileDialog dialog(this, title, current_dir, "", wildcard,
		wxFD_OPEN | wxFD_FILE_MUST_EXIST);
	if (dialog.ShowModal() == wxID_CANCEL) {
		return;
	}
	auto path = std::string(dialog.GetPath());
	InsertFD(unit, path);
}

// メニューの「デバイス > FD > 挿入」UI
void
WXMainFrame::OnFDInsertUI(wxUpdateUIEvent& event)
{
	int unit = event.GetId() - ID_FD_INSERT_0;
	const auto *fdd = fdd_array[unit];
	assert(fdd);

	if (fdd->IsMediumLoaded()) {
		event.SetText(_("Ex&change"));
	} else {
		event.SetText(_("&Insert"));
	}
}

// メニューの「デバイス > FD > 取り出し」
void
WXMainFrame::OnFDEject(wxCommandEvent& event)
{
	int unit = event.GetId() - ID_FD_EJECT_0;

	EjectFD(unit, false);
}

// メニューの「デバイス > FD > 取り出し」UI
void
WXMainFrame::OnFDEjectUI(wxUpdateUIEvent& event)
{
	int unit = event.GetId() - ID_FD_EJECT_0;
	FDDDevice *fdd = fdd_array[unit];
	assert(fdd);

	// 挿入されてて取り出し可能な時のみ有効
	event.Enable(fdd->IsMediumLoaded() && fdd->IsEjectEnable());
}

// メニューの「デバイス > FD > 強制取り出し」
void
WXMainFrame::OnFDForce(wxCommandEvent& event)
{
	int unit = event.GetId() - ID_FD_FORCE_0;

	EjectFD(unit, true);
}

// メニューの「デバイス > FD > 強制取り出し」UI
void
WXMainFrame::OnFDForceUI(wxUpdateUIEvent& event)
{
	int unit = event.GetId() - ID_FD_FORCE_0;
	FDDDevice *fdd = fdd_array[unit];
	assert(fdd);

	// 挿入されてて取り出し禁止の時のみ有効
	event.Enable(fdd->IsMediumLoaded() && fdd->IsEjectEnable() == false);
}

// メニューの「デバイス > CD/MO > 挿入」
void
WXMainFrame::OnCDInsert(wxCommandEvent& event)
{
	std::string title;
	std::string wildcard;

	int id = event.GetId() - ID_CD_INSERT_0;
	const auto *dev = scsidev[id];
	assert(dev);

	switch (dev->GetDevType()) {
	 case SCSI::DevType::CD:
		title = _("Open a CD image");
		wildcard = "ISO images (*.iso;*.img)|*.iso;*.img;*.ISO;*.IMG";
		break;
	 case SCSI::DevType::MO:
		title = _("Open an MO image");
		wildcard = "MO images (*.img;*.mos)|*.img;*.mos;*.IMG;*.MOS";
		break;
	 default:
		__unreachable();
	}
	wildcard += "|All files (*.*)|*.*";

	// ディレクトリは、まだ一度も使ってなければ VM ディレクトリにする
	if (current_dir.empty()) {
		current_dir = gMainApp.GetVMDir();
	}

	wxFileDialog dialog(this, title, current_dir, "", wildcard,
		wxFD_OPEN | wxFD_FILE_MUST_EXIST);
	if (dialog.ShowModal() == wxID_CANCEL) {
		return;
	}
	auto path = std::string(dialog.GetPath());
	InsertCD(id, path);
}

// メニューの「デバイス > CD/MO > 挿入」UI
void
WXMainFrame::OnCDInsertUI(wxUpdateUIEvent& event)
{
	int id = event.GetId() - ID_CD_INSERT_0;
	SCSIDisk *disk = dynamic_cast<SCSIDisk *>(scsidev[id]);
	assert(disk);

	if (disk->IsMediumLoaded()) {
		event.SetText(_("Ex&change"));
	} else {
		event.SetText(_("&Insert"));
	}
}

// メニューの「デバイス > CD/MO > 取り出し」
void
WXMainFrame::OnCDEject(wxCommandEvent& event)
{
	int id = event.GetId() - ID_CD_EJECT_0;

	EjectCD(id, false);
}

// メニューの「デバイス > CD/MO > 取り出し」UI
void
WXMainFrame::OnCDEjectUI(wxUpdateUIEvent& event)
{
	int id = event.GetId() - ID_CD_EJECT_0;
	SCSIDisk *disk = dynamic_cast<SCSIDisk *>(scsidev[id]);
	assert(disk);

	// 挿入されてて取り出し可能な時のみ有効
	event.Enable(disk->IsMediumLoaded() &&
	             disk->IsMediumRemovalAllowed());
}

// メニューの「デバイス > CD/MO > 強制取り出し」
void
WXMainFrame::OnCDForce(wxCommandEvent& event)
{
	int id = event.GetId() - ID_CD_FORCE_0;

	EjectCD(id, true);
}

// メニューの「デバイス > CD/MO > 強制取り出し」UI
void
WXMainFrame::OnCDForceUI(wxUpdateUIEvent& event)
{
	int id = event.GetId() - ID_CD_FORCE_0;
	SCSIDisk *disk = dynamic_cast<SCSIDisk *>(scsidev[id]);
	assert(disk);

	// 挿入されてて取り出し禁止の時のみ有効
	event.Enable(disk->IsMediumLoaded() &&
	             disk->IsMediumRemovalPrevented());
}

// メニューの「デバイス > マウスモード」
void
WXMainFrame::OnMouseMode(wxCommandEvent& event)
{
	// enable ならチェックを付けようとして呼ばれた
	bool enable = event.IsChecked();

	if (enable) {
		// メッセージを表示したい
	} else {
		// メッセージを表示したい
	}

	// マウスモードを設定
	wxCommandEvent newev(NONO_EVT_MOUSEMODE, GetId());
	newev.SetEventObject(this);
	newev.SetInt(enable);
	::wxPostEvent(gMainView, newev);
}

// メニューの「デバイス > マウスモード」UI
void
WXMainFrame::OnMouseModeUI(wxUpdateUIEvent& event)
{
	event.Check(gMainView->GetMouseMode());
}

// メニューの「デバイス > Push Power Button」イベント
void
WXMainFrame::OnPowerButton(wxCommandEvent& event)
{
	// LUNA はモーメンタリ動作、X68030 はオルタネート動作のスイッチ。
	// この違いは VM 側で吸収するのでここはボタンを押したという動作だけ
	// 通知する。
	gPower->PowerButton();
}

// メニューの「デバイス > Push Power Button」UI
void
WXMainFrame::OnPowerButtonUI(wxUpdateUIEvent& event)
{
	if (gMainApp.Has(VMCap::X68030)) {
		// X68030 の電源ボタンはオルタネート動作なので、
		// チェックで状態を表す。
		event.Check(gPower->IsPowerPressed());
	} else {
		// LUNA の電源ボタンはモーメンタリ動作で、電源オフ->オンの機能しかなく
		// さらに今の所電源オフ状態はないので、実質無効。
		event.Enable(false);
	}
}

// メニューの「デバイス > Push Reset Button」イベント
// こっちは本体リセットボタンによるリセット
void
WXMainFrame::OnResetButton(wxCommandEvent& event)
{
	gPower->MakeResetHard();
}

// メニューの「デバイス > Push ABORT(Interrupt) Button」イベント
void
WXMainFrame::OnNMIButton(wxCommandEvent& event)
{
	gNMI->PressNMI();
}

// メニューの「デバイス > Operation > Keyboard Connect」イベント
void
WXMainFrame::OnKeyboardConnect(wxCommandEvent& event)
{
	// enable ならチェックを付けようとして呼ばれた
	bool enable = event.IsChecked();

	// キーボード接続を設定
	gKeyboard->Connect(enable);
}

// メニューの「デバイス > Operation > Keyboard Connect」UI
void
WXMainFrame::OnKeyboardConnectUI(wxUpdateUIEvent& event)
{
	event.Check(gKeyboard->IsConnected());
}

// メニューのウィンドウ ID 共通イベント
// チェックが入る方なら、ウィンドウを作成して表示。
// チェックが消える方なら、ウィンドウを削除。
void
WXMainFrame::OnWindow(wxCommandEvent& event)
{
	int id = event.GetId();

	DoWindow(id, event.IsChecked());
}

void
WXMainFrame::DoWindow(int id, bool enable)
{
	// モニタウィンドウを作成する共通部分。
	// id はモニター ID であると同時に、windows[] 配列の添字、さらにこの
	// モニタウィンドウの wxWindowID としても使っている
	// (WXSubWindow::OnClose() 参照)。
#define SetMonitor(id, name)	\
	windows[id].reset(new WXMonitorWindow(this, (name), \
		gMonitorManager->Get(id)))

	if (enable) {
		// チェックが入ったのでウィンドウを作成
		switch (id) {
		 case ID_MONITOR_ATC0:
		 case ID_MONITOR_ATC1:
		 case ID_MONITOR_ATC2:
		 case ID_MONITOR_ATC3:
		 case ID_MONITOR_ATC4:
		 case ID_MONITOR_ATC5:
		 case ID_MONITOR_ATC6:
		 case ID_MONITOR_ATC7:
		 {
			int n = id - ID_MONITOR_ATC0;
			wxString name(string_format("ATC#%d", n));
			if ((n & 1) == 0) {
				name += _("(Data)");
			} else {
				name += _("(Inst)");
			}
			SetMonitor(id, name);
			break;
		 }

		 case ID_MONITOR_BREAKPOINT:
			SetMonitor(id, _("Breakpoint"));
			break;

		 case ID_MONITOR_BT454:
			SetMonitor(id, "Bt454/458");
			break;

		 case ID_MONITOR_CMMU0:
		 case ID_MONITOR_CMMU1:
		 case ID_MONITOR_CMMU2:
		 case ID_MONITOR_CMMU3:
		 case ID_MONITOR_CMMU4:
		 case ID_MONITOR_CMMU5:
		 case ID_MONITOR_CMMU6:
		 case ID_MONITOR_CMMU7:
		 {
			int n = id - ID_MONITOR_CMMU0;
			wxString name(string_format("CMMU#%d", n));
			if ((n & 1) == 0) {
				name += _("(Data)");
			} else {
				name += _("(Inst)");
			}
			SetMonitor(id, name);
			break;
		 }

		 case ID_MONITOR_CRTC:
			SetMonitor(id, "CRTC");
			break;

		 case ID_MONITOR_DISASM0 ...
		      ID_MONITOR_DISASM(MAX_DISASM_MONITOR - 1):
		 {
			int n = id - ID_MONITOR_DISASM0;
			windows[id].reset(new WXDisasmWindow(this, n));
			break;
		 }

		 case ID_MONITOR_DMAC:
			SetMonitor(id, "DMAC");
			break;

		 case ID_MONITOR_FDC:
			SetMonitor(id, "FDC");
			break;

		 case ID_MONITOR_HOSTCOM:
			SetMonitor(id, _("Host Serial Port"));
			break;

		 case ID_MONITOR_HOSTNET:
			SetMonitor(id, _("Host Network"));
			break;

		 case ID_MONITOR_INTERRUPT:
			SetMonitor(id, _("Interrupt"));
			break;

		 case ID_MONITOR_KEYBOARD:
			SetMonitor(id, _("Keyboard/Mouse"));
			break;

		 case ID_MONITOR_LANCE:
			SetMonitor(id, "LANCE");
			break;

		 case ID_MONITOR_LCD:
			windows[id].reset(new WXLCDMonitor(this));
			break;

		 case ID_MONITOR_LUNAVC:
			SetMonitor(id, _("Video Control"));
			break;

		 case ID_MONITOR_MEMDUMP0 ...
		      ID_MONITOR_MEMDUMP(MAX_MEMDUMP_MONITOR - 1):
		 {
			int n = id - ID_MONITOR_MEMDUMP0;
			windows[id].reset(new WXMemdumpWindow(this, n));
			break;
		 }

		 case ID_MONITOR_MFP:
			SetMonitor(id, "MFP");
			break;

		 case ID_MONITOR_MK48T02:
			SetMonitor(id, "MK48T02");
			break;

		 case ID_MONITOR_MPUATC:
			SetMonitor(id, _("MPU (ATC)"));
			break;

		 case ID_MONITOR_MPUREG:
			SetMonitor(id, _("MPU (Register)"));
			break;

		 case ID_MONITOR_ROMCONS:
			SetMonitor(id, _("ROM Console (provisional)"));
			break;

		 case ID_MONITOR_PIO0:
			SetMonitor(id, "PIO");
			break;

		 case ID_MONITOR_RENDERER:
			SetMonitor(id, _("Renderer"));
			break;

		 case ID_MONITOR_RTC:
			SetMonitor(id, "RTC");
			break;

		 case ID_MONITOR_SCC:
			SetMonitor(id, "SCC");
			break;

		 case ID_MONITOR_SCHEDULER:
			SetMonitor(id, _("Scheduler"));
			break;

		 case ID_MONITOR_SCSIDEVS:
			SetMonitor(id, _("SCSI Devices"));
			break;

		 case ID_MONITOR_SIO:
			SetMonitor(id, "SIO");
			break;

		 case ID_MONITOR_SPC:
			SetMonitor(id, "SPC");
			break;

		 case ID_MONITOR_SYSCLK:
			SetMonitor(id, _("System Clock"));
			break;

		 case ID_MONITOR_WXHOSTINFO:
			SetMonitor(id, _("Host Information"));
			break;

		 case ID_SUBWIN_BRHIST:
			windows[id].reset(new WXScrollMonitorWindow(this,
				_("Branch History"), gMonitorManager->Get(id)));
			break;

		 case ID_SUBWIN_EXHIST:
			windows[id].reset(new WXScrollMonitorWindow(this,
				_("Exception History"), gMonitorManager->Get(id)));
			break;

		 case ID_SUBWIN_CACHE0:
		 case ID_SUBWIN_CACHE1:
		 case ID_SUBWIN_CACHE2:
		 case ID_SUBWIN_CACHE3:
		 case ID_SUBWIN_CACHE4:
		 case ID_SUBWIN_CACHE5:
		 case ID_SUBWIN_CACHE6:
		 case ID_SUBWIN_CACHE7:
		 {
			int n = id - ID_SUBWIN_CACHE0;
			wxString name = _("Cache");
			name += string_format("#%d", n);
			if ((n & 1) == 0) {
				name += _("(Data)");
			} else {
				name += _("(Inst)");
			}
			windows[id].reset(new WXCacheWindow(this, name,
				gMonitorManager->Get(id)));
			break;
		 }

		 case ID_SUBWIN_DIPSW:
			windows[id].reset(new WXDipswWindow(this));
			break;

		 case ID_SUBWIN_LCDPANEL:
			windows[id].reset(new WXLCDWindow(this));
			break;
		 case ID_SUBWIN_LOG:
			windows[id].reset(new WXLogMonitor(this));
			break;
		 case ID_SUBWIN_LOGSETTING:
			windows[id].reset(new WXLogSettingWindow(this));
			break;

		 case ID_SUBWIN_LUNAFB:
		 case ID_SUBWIN_TVRAM:
		 {
			// 呼び方が違うので名前と ID が違う
			wxString name;
			if (id == ID_SUBWIN_LUNAFB) {
				name = _("Bitmap Plane");
			} else {
				name = _("Text VRAM");
			}
			windows[id].reset(new WXPlaneVRAMWindow(this, name));
			break;
		 }

		 case ID_SUBWIN_PALETTE:
		 case ID_SUBWIN_TEXTPAL:
		 {
			// 呼び方が違うので名前と ID が違う
			wxString name;
			if (id == ID_SUBWIN_TEXTPAL) {
				name = _("Text Palette");
			} else {
				name = _("Palette");
			}
			windows[id].reset(new WXPaletteWindow(this, name));
			break;
		 }

		 case ID_SUBWIN_ROM:
			windows[id].reset(new WXROMWindow(this));
			break;

		 case ID_SUBWIN_SOFTKEY:
			windows[id].reset(new WXSoftKeyWindow(this, gMainApp.GetVMType()));
			break;

		 case ID_SUBWIN_VECTOR:
			windows[id].reset(new WXScrollMonitorWindow(this,
				_("Vector Table"), gMonitorManager->Get(id)));
			break;

		 default:
			PANIC("'%s' not listed", Monitor::GetIdText(id));
		}
		windows[id]->Show();
	} else {
		// チェックが消えたのでウィンドウを終了
		windows[id]->Close();
		// (削除自体はこの少し下の DeleteWindow() で行なっている)
	}
}

// Update UI イベント
void
WXMainFrame::OnWindowUI(wxUpdateUIEvent& event)
{
	int id = event.GetId();

	// ウィンドウが存在していればチェックを表示
	event.Check((bool)windows[id]);

	if (IS_MONITOR_DISASM(id)) {
		// メニューに現在のアドレスを載せる
		// (ウィンドウの生存期間外にも呼ばれることに注意)
		auto mon = gMonitorManager->Get(id);
		auto *disasm = dynamic_cast<Disasm *>(mon.obj);
		assert(disasm);

		int n = id - ID_MONITOR_DISASM0;
		event.SetText(string_format("&%d: $%08x", n, disasm->GetAddr()));
	}

	if (IS_MONITOR_MEMDUMP(id)) {
		// メニューに現在のアドレスを載せる
		// (ウィンドウの生存期間外にも呼ばれることに注意)
		auto mon = gMonitorManager->Get(id);
		auto *memdump = dynamic_cast<Memdump *>(mon.obj);
		assert(memdump);

		int n = id - ID_MONITOR_MEMDUMP0;
		event.SetText(string_format("&%d: $%08x", n, memdump->GetAddr()));
	}
}

// バージョンダイアログ
void
WXMainFrame::OnAbout(wxCommandEvent& event)
{
	WXVersionDlg dlg(this);
	dlg.ShowModal();
}

// FD を挿入する。
void
WXMainFrame::InsertFD(int unit, const std::string& path)
{
	FDDDevice *fdd = fdd_array[unit];
	assert(fdd);

	// VM に依頼
	// (エラーが起きれば別途 UI メッセージが飛んでくる)
	fdd->LoadDiskUI(path);

	// 次回以降のダイアログのカレントディレクトリをここに更新
	auto pos = path.rfind('/');
	current_dir = path.substr(0, pos);
}

// FD をイジェクトする。
void
WXMainFrame::EjectFD(int unit, bool force)
{
	FDDDevice *fdd = fdd_array[unit];
	assert(fdd);

	fdd->UnloadDiskUI(force);
}

// CD/MO を挿入する。
void
WXMainFrame::InsertCD(int id, const std::string& path)
{
	SCSIDisk *scsidisk = dynamic_cast<SCSIDisk*>(scsidev[id]);
	assert(scsidisk);

	// VM に依頼
	// (エラーが起きれば別途 UI メッセージが飛んでくる)
	scsidisk->LoadDiskUI(path);

	// 次回以降のダイアログのカレントディレクトリをここに更新
	auto pos = path.rfind('/');
	current_dir = path.substr(0, pos);
}

// CD/MO をイジェクトする。
void
WXMainFrame::EjectCD(int id, bool force)
{
	SCSIDisk *scsidisk = dynamic_cast<SCSIDisk*>(scsidev[id]);
	assert(scsidisk);

	scsidisk->UnloadDiskUI(force);
}

// CD/MO/FD の挿入失敗メッセージ。
// CD/MO と FD でメッセージが同じなので共用する。
void
WXMainFrame::OnInsertFailed(wxCommandEvent& event)
{
	::wxMessageBox(_("Cannot load the image file"));
}

// ウィンドウをリストから削除する
// (ウィンドウの Close() から呼ばれる)
void
WXMainFrame::DeleteWindow(wxWindow *win)
{
	for (int id = 0; id < ID_SUBWIN_MAX; id++) {
		if (windows[id].get() == win) {
			windows[id].reset();
			break;
		}
	}
}

// m680x0 ダブルバスフォールト処理
void
WXMainFrame::OnHaltNotify(wxCommandEvent& event)
{
	::wxMessageBox(_("Double bus fault has occurred in VM"),
		wxEmptyString,
		wxICON_EXCLAMATION);
}

// 起動時にオープンするサブウィンドウ (-M オプション)
/*static*/ std::array<bool, ID_SUBWIN_MAX> WXMainFrame::subwindow_start;
