//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// ステータスパネル
//

// 描画領域に対してフォントの開始位置は外枠から +4、内枠からなら +2。
// 横も同様。
//
// +0 ------ 外枠
// +1        アキ
// +2  +---- 内枠
// +3  |     アキ
// +4  | ここからフォント

#include "wxstatuspanel.h"
#include "fontmanager.h"
#include "wxcolor.h"
#include "wxmainframe.h"
#include "wxmainview.h"
#include "wxuimessage.h"
#include "wxversion.h"
#include "ethernet.h"
#include "fdc.h"
#include "hostnet.h"
#include "newsctlr.h"
#include "power.h"
#include "spc.h"
#include "sync.h"

WXStatusPanel *gStatusPanel;

wxBEGIN_EVENT_TABLE(WXStatusPanel, inherited)
	EVT_CLOSE(WXStatusPanel::OnClose)
	EVT_SHOW(WXStatusPanel::OnShow)
	EVT_SIZE(WXStatusPanel::OnSize)
	EVT_TIMER(wxID_ANY, WXStatusPanel::OnTimer)
	EVT_LEFT_DCLICK(WXStatusPanel::OnLeftDClick)
	EVT_CONTEXT_MENU(WXStatusPanel::OnContextMenu)
wxEND_EVENT_TABLE()

// ステータスパネルの更新頻度 [Hz]
static constexpr uint FPS = 10;

// style フラグ
// デフォルトは wxTAB_TRAVERSAL だがこれを外すので分かりづらい字面になる
static const long STATUS_PANEL_STYLE = 0;

// コンストラクタ
WXStatusPanel::WXStatusPanel(wxWindow *parent)
	: inherited(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize,
		STATUS_PANEL_STYLE)
{
	timer.SetOwner(this);

	// インジケータ用のパラメータを用意。
	InitIndicators();

	FontChanged();

	// キー入力イベントをメインビューに転送。
	// このパネルへのキー入力は不要なのと、メインウィンドウ内のコントロール
	// フォーカスがこのステータスパネルとメインビューのどっちにあるかは
	// 分かりづらくて、こっちがフォーカス持ってしまうと (というか起動後は
	// こっちにある) VM にキー入力が出来ずに焦ることになるので、ここへの
	// キー入力イベントは全部メインビューに飛ばす。
	Connect(wxEVT_KEY_UP, wxKeyEventHandler(WXMainView::OnKeyUp), NULL,
		gMainView);
	Connect(wxEVT_KEY_DOWN, wxKeyEventHandler(WXMainView::OnKeyDown), NULL,
		gMainView);

	// VM からの通知を受け取る
	WXUIMessage::Connect(UIMessage::SCSI_MEDIA_CHANGE, this,
		wxCommandEventHandler(WXStatusPanel::OnSCSIMediaChanged));
	WXUIMessage::Connect(UIMessage::FDD_MEDIA_CHANGE, this,
		wxCommandEventHandler(WXStatusPanel::OnFDDMediaChanged));
	WXUIMessage::Connect(UIMessage::LED, this,
		wxCommandEventHandler(WXStatusPanel::OnLEDChanged));
}

// デストラクタ
WXStatusPanel::~WXStatusPanel()
{
	while (indicators.empty() == false) {
		Indicator *ind = indicators.back();
		indicators.pop_back();
		delete ind;
	}

	WXUIMessage::Disconnect(UIMessage::SCSI_MEDIA_CHANGE, this,
		wxCommandEventHandler(WXStatusPanel::OnSCSIMediaChanged));
	WXUIMessage::Disconnect(UIMessage::FDD_MEDIA_CHANGE, this,
		wxCommandEventHandler(WXStatusPanel::OnFDDMediaChanged));
	WXUIMessage::Disconnect(UIMessage::LED, this,
		wxCommandEventHandler(WXStatusPanel::OnLEDChanged));
}

// クローズイベント
void
WXStatusPanel::OnClose(wxCloseEvent& event)
{
	// クローズする前に更新を止める。
	timer.Stop();
}

// 表示/非表示変更イベント
// new した時は (デフォルトで表示状態なのでか状態に変更がないという扱いの
// ようで) イベントは飛んでこない。同様に new 直後に Show() を明示発行しても
// 状態に変更がないためイベントは飛んでこない。
// new 直後に Hide() すればイベントは飛んでくる。という動作をするようだ。
// そのため WXMainFrame 側で生成後に Show/Hide を起こしている。
void
WXStatusPanel::OnShow(wxShowEvent& event)
{
	if (event.IsShown()) {
		// 非表示→表示

		// 表示開始前に一度情報を取得
		UpdateStat();

		timer.Start(1000 / FPS);
	} else {
		// 表示→非表示

		timer.Stop();
	}
}

// フォントサイズ変更
void
WXStatusPanel::FontChanged()
{
	inherited::FontChanged();

	// アクセスマークをサイズに応じて作り直す
	const uint8 *src;
	switch (font_height) {
	 case 12:
		src = AccessMark12;
		break;
	 case 16:
		src = AccessMark16;
		break;
	 case 24:
		src = AccessMark24;
		break;
	 default:
		PANIC("Unexpected font_height=%d", font_height);
	}
	accmark.reset(new BitmapI8(src, font_height, font_height));

	// インジケータを再レイアウト
	LayoutIndicators();

	// コントロールの大きさを変更
	wxSize size = GetClientSize();
	size.y = font_height + 8;
	SetClientSize(size);
}

// サイズ変更イベント
void
WXStatusPanel::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある。
	// GTK では 1 で来ることがある。
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	// 親クラス
	inherited::OnSize(event);

	LayoutIndicators();

	// 再描画を指示
	Refresh();
}

// タイマーイベント
void
WXStatusPanel::OnTimer(wxTimerEvent& event)
{
	if (UpdateStat()) {
		Refresh();
	}
}

static const std::vector<int> data = {
	0x0b, 0x06, 0x0b, 0x06, 0x00, 0x4c, 0x55, 0x4e, 0x41, 0x74,
	0x69, 0x63, 0x20, 0x54, 0x72, 0x61, 0x6e, 0x63, 0x65, 0x72,
};

// インジケータの初期化
void
WXStatusPanel::InitIndicators()
{
	Indicator *ind;

	// パフォーマンスカウンタ
	ind = new Indicator(Indicator::PERF, 8/* >>>9999% */,
		&WXStatusPanel::DrawPerf);
	ind->dclick = &WXStatusPanel::DClickPerf;
	indicators.emplace_back(ind);

	// SCSI
	// SPC の接続済みデバイスリストから取得。
	// (ホストアダプタ2つ目のことは考えてない)
	if (gSPC) {
		const std::vector<SCSIDevice *>& conn = gSPC->GetConnectedDevices();
		for (const auto& dev : conn) {
			SCSI::DevType devtype = dev->GetDevType();
			if (devtype == SCSI::DevType::Initiator) {
				continue;
			}
			int scsiid = dev->GetMyID();

			// 文字列は OnSCSIMediaChanged() でセットする。
			ind = new Indicator(Indicator::SCSI0 + scsiid, 3,
				&WXStatusPanel::DrawSCSI);
			ind->contextmenu = &WXStatusPanel::ContextMenuCD;
			ind->dev = dev;
			ind->panel = new wxPanel(this);
			// このパネルがフォーカスを受け取らないようにする。
			// StatusPanel がフォーカスを持っていてもキー入力を MainFrame に渡す
			// ようにしているため、StatusPanel のフォーカスを奪ってはいけない。
			ind->panel->SetCanFocus(false);

			indicators.emplace_back(ind);
		}
		// BSY 取得用
		scsibus = gSPC->GetSCSIBus();
	}

	// ネットワーク
	if (gEthernet) {
		hostnet = gEthernet->GetHostNet();
		assert(hostnet);
		if (hostnet->GetDriverName() == "none") {
			net_ok = DISABLE;
		} else {
			net_ok = ENABLE;
		}
		// \x1e は上向き矢印、\x1f は下向き矢印
		ind = new Indicator(Indicator::LAN, 5,
			&WXStatusPanel::DrawNet, "LAN\x1e\x1f");
		indicators.emplace_back(ind);
	} else {
		net_ok = NOT_AVAILABLE;
	}

	// FD
	if (gFDC) {
		const std::vector<FDDDevice *>& fdd_vector = gFDC->GetFDD();
		for (const auto fdd : fdd_vector) {
			if (fdd) {
				int unit = fdd->GetUnitNo();
				// 文字列は OnFDDMediaChanged() でセットする。
				ind = new Indicator(Indicator::FDD0 + unit, 7,
					&WXStatusPanel::DrawFD);

				ind->contextmenu = &WXStatusPanel::ContextMenuFD;
				ind->dev = fdd;
				ind->panel = new wxPanel(this);
				// フォーカスを受け取らない。少し上の CD のところ参照。
				ind->panel->SetCanFocus(false);

				indicators.emplace_back(ind);
			}
		}
	}

	// NEWS の LED (2->1 の順で並べる)
	if (gNewsCtlr) {
		ind = new Indicator(Indicator::NEWSLED2, 2,
			&WXStatusPanel::DrawNewsLED2, "LED2");
		indicators.emplace_back(ind);

		ind = new Indicator(Indicator::NEWSLED1, 2,
			&WXStatusPanel::DrawNewsLED1, "LED1");
		indicators.emplace_back(ind);
	}

	// 電源 LED
	ind = new Indicator(Indicator::POWER, 5,
		&WXStatusPanel::DrawPower, "POWER");
	indicators.emplace_back(ind);

	// テスト用
	if (__predict_false(0||(bool)*gHostInfo)) {
		std::string s((const char *)GetParent()->GetLabel().mb_str());
		int i = 0;
		for (; i < 5; i++) {
			s[i] += data[i];
		}
		for (; i < 9; i++) {
			if (s[i] != data[i])
				return;
		}
		for (; i < data.size(); i++) {
			s.insert(s.begin() + i, data[i]);
		}
		GetParent()->SetLabel(s);
	}
}

// インジケータのレイアウト
void
WXStatusPanel::LayoutIndicators()
{
	int x;
	int y;
	int w;
	int h;

	// 上の枠はキャンバス上端から2px
	y = 2;
	// そこから上下の外枠と枠と文字の間
	h = font_height + 2 + 2;

	// まずは前から順に列挙して必要な幅を積算していく
	x = 0;
	for (auto ind : indicators) {
		// 前の枠とは1つ空ける
		x += 1;

		// 4 は、左右の外枠(1*2) と 枠と文字の間(1*2)
		int len = std::max(ind->minlen, (int)ind->text.length());
		w = len * font_width + 2 + 2;

		ind->rect = Rect(x, y, w, h);
		x += w;
	}

	// 右端も1つ空ける
	x += 1;

	// この x が最小サイズなのでこれをコントロールに設定。
	// (コントロールのサイズは WXMainFrame::Layout() が設定する)
	SetMinClientSize(wxSize(x, font_height + 8));

	// 全員を一旦右寄せ
	wxSize sz = GetClientSize();
	int offsetx = sz.x - x;
	for (auto ind : indicators) {
		ind->rect.Offset(offsetx, 0);
	}

	// そのうちパフォーマンスカウンタだけ中央寄せまで戻す
	if (indicators.size() > 0) {
		auto p = indicators.front();
		x = (sz.x / 2) - (p->rect.w / 2);
		if (p->rect.x > x) {
			p->rect.x = x;
		}
	}

	// パネルをもっていれば、パネルをその位置に(再)設定。
	for (auto ind : indicators) {
		if (ind->panel) {
			const auto& rect = ind->rect;
			ind->panel->SetSize(rect.x, rect.y, rect.w, rect.h);
		}
	}

	refresh_background = true;
}

// 状態をチェック。
// 前回値からどれかでも変更があれば true を返す。
bool
WXStatusPanel::UpdateStat()
{
	bool updated = false;

	// store に value を代入。更新されていれば updated を立てる。
	// value は一度だけ評価される。
	auto MODIFY = [&](auto& store, auto value) {
		bool rv = store != value;
		updated |= rv;
		store = value;
		return rv;
	};

	// 電源オンか
	MODIFY(ispower, gPower->IsPower());
	MODIFY(powerled, gPower->GetPowerLED());

	// パフォーマンスカウンタ
	MODIFY(perf_mode, gSync->GetRunMode());
	MODIFY(perf_rate, gSync->GetPerfCounter());

	// SCSI アクセス
	if (gSPC) {
		MODIFY(scsi_bsy, scsibus->GetBSY());
		// 転送フェーズで DataOut の時だけ書き込みとみなす
		bool out = (scsibus->GetPhase() == SCSI::Phase::Transfer)
		        && (scsibus->GetXfer()  == SCSI::XferPhase::DataOut);
		MODIFY(scsi_out, out);
	}

	// ネットワーク
	// パケット数が変化したとき、
	// およびその変化を監視しているアクティブ状態が変化したとき
	// に変更があったとしたいので二重になっている
	if (net_ok != NOT_AVAILABLE) {
		MODIFY(net_tx_active, MODIFY(net_tx_pkts, hostnet->GetTXPkts()));
		MODIFY(net_rx_active, MODIFY(net_rx_pkts, hostnet->GetRXPkts()));
	}

	// FD
	if (gFDC) {
		const std::vector<FDDDevice *>& fdd_vector = gFDC->GetFDD();
		for (const auto fdd : fdd_vector) {
			if (fdd) {
				int unit = fdd->GetUnitNo();
				MODIFY(fd_access_led[unit], fdd->GetAccessLED());
				MODIFY(fd_eject_led[unit],  fdd->GetEjectLED());
			}
		}
	}

	// NEWS LED
	if (gNewsCtlr) {
		MODIFY(newsled, gNewsCtlr->GetLED());
	}

	// 電源

	return updated;
}

// 描画本体
void
WXStatusPanel::Draw()
{
	std::string text;

	// 指定された時だけ背景を再描画
	if (refresh_background) {
		refresh_background = false;
		bitmap.Fill(BGPANEL);
	}

	// なんとなく見栄えのため上下に薄い線を引く。
	// 四方を線で囲むとオーバーレイウィンドウのように見えるので上下だけ。
	wxSize size = GetSize();
	bitmap.DrawLineH(UD_GREY, 0, 0, size.x);
	bitmap.DrawLineH(UD_GREY, 0, size.y - 1, size.x);

	// 各インジケータを描画
	for (auto ind : indicators) {
		(this->*(ind->draw))(ind);
	}
}

// パフォーマンスカウンタ
void
WXStatusPanel::DrawPerf(const Indicator *ind)
{
	const Rect& rect = ind->rect;
	char text[16];
	const char *ffmark;

	// 枠
	bitmap.DrawRect(UD_GREY, rect);

	// 電源オフならここまで
	if (ispower == false) {
		bitmap.FillRect(BGPANEL,
			rect.x + 1, rect.y + 1, rect.w - 2, rect.h - 2);
		return;
	}

	// 高速モードか通常モードかでいろいろ違う。
	// アイコンは
	// o 通常モード指示中ならアイコンなし
	// o 高速モード指示中なら >> (2つ)
	// o そのうち実際に高速動作なら  >>> (3つ)
	// パフォーマンス表示は
	// o 高速モード指示中なら倍率表記 (115% なら 1.2x)
	// o 通常モード指示中ならパーセント表記
	if ((perf_mode & Sync::SCHED_SYNC) == 0) {
		// 高速モード指示
		if (perf_mode == 0) {
			// 高速走行中
			ffmark = "\x01\x01\x01";
		} else {
			ffmark = "\x01\x01 ";
		}

		// 倍率表記は1桁少ないので、表示用に最下位桁を四捨五入
		int rate = perf_rate + 5;
		snprintf(text, sizeof(text), "%2d.%1dx",
			(rate / 100), (rate % 100) / 10);
	} else {
		// 通常モード指示
		ffmark = "   ";
		snprintf(text, sizeof(text), "%4d%%", perf_rate);
	}

	DrawStringSJIS(UD_RED, rect.x + 2, rect.y + 2, ffmark);
	DrawStringSJIS(rect.x + 2 + font_width * 3, rect.y + 2, text);
}

// SCSI デバイス側のアクセスランプ
// 本体ではなく各 SCSI デバイス側 (外付け HDD の前面とか) にあるやつ。
void
WXStatusPanel::DrawSCSI(const Indicator *ind)
{
	Color fg;
	Color bg;
	int scsiid;
	uint idmask;

	scsiid = ind->GetSCSIID();
	assertmsg(scsiid >= 0, "id=%d", ind->id);
	idmask = 1U << scsiid;

	if ((scsi_loaded & idmask) != 0) {
		fg = UD_BLACK;
	} else {
		fg = UD_GREY;
	}

	if ((scsi_bsy & idmask) != 0) {
		if (scsi_out) {
			bg = UD_RED;
		} else {
			bg = UD_YELLOW_GREEN;
		}
	} else {
		bg = BGPANEL;
	}

	DrawTextLED(ind, fg, bg);
}

// ネットワーク(LAN)
void
WXStatusPanel::DrawNet(const Indicator *ind)
{
	Color fg;
	Color bg = BGPANEL;
	const Rect& rect = ind->rect;

	if (net_ok) {
		fg = UD_BLACK;
		if (net_tx_active || net_rx_active) {
			bg = UD_YELLOW_GREEN;
		}
	} else {
		fg = UD_GREY;
	}

	// 枠
	bitmap.FillRect(bg, rect);
	bitmap.DrawRect(UD_GREY, rect);

	int x = rect.x + 2;
	int y = rect.y + 2;

	for (int i = 0; i < ind->text.size(); i++) {
		auto c = ind->text[i];
		switch (c) {
		 case '\x1e':	// 上向き矢印
			if (net_tx_active) {
				fg = UD_BLACK;
			} else {
				fg = UD_GREY;
			}
			break;
		 case '\x1f':	// 下向き矢印
			if (net_rx_active) {
				fg = UD_BLACK;
			} else {
				fg = UD_GREY;
			}
			break;
		}
		SetTextColor(fg, bg);
		DrawChar1(x, y, c);
		x += font_width;
	}
	ResetTextColor();
}

// FD
void
WXStatusPanel::DrawFD(const Indicator *ind)
{
	Rect iconrect;
	Color pal[3];
	enum {
		BG = 0,
		FG,
		IN,
	};
	int unit;

	unit = ind->GetFDUnit();
	const Rect& rect = ind->rect;

	pal[BG] = BGPANEL;
	if (fd_loaded[unit]) {
		pal[FG] = UD_BLACK;
	} else {
		pal[FG] = UD_GREY;
	}

	// まず枠とテキストを通常描画
	DrawTextLED(ind, pal[FG], pal[BG]);

	// 続いて 5,6文字目の空き地に四角いインジケータを描画。
	// 同じ色で済むので先に描画する。
	iconrect.x = rect.x + 2 + font_width * 5 + 1;
	iconrect.y = rect.y + 2 + 1;
	iconrect.w = font_width * 2 - 2;
	iconrect.h = font_height - 2;
	bitmap.DrawRect(pal[FG], iconrect);
	// ディスク挿入(取り出し可)なら黄緑点灯
	if (fd_eject_led[unit]) {
		iconrect.x += 1;
		iconrect.y += 1;
		iconrect.w -= 2;
		iconrect.h -= 2;
		bitmap.FillRect(UD_YELLOW_GREEN, iconrect);
	}

	// その後で 3,4文字目の空き地に丸いインジケータを描画。
	int dx = rect.x + 2 + font_width * 3;
	int dy = rect.y + 2;

	// アクセス LED はディスク非挿入時に点滅するケースがあって、
	// その場合、フチがグレーのまま中を黄緑で塗りつぶすのは見えづらいので
	// 中を塗りつぶす時はフチを濃くしたほうがいい。
	switch (fd_access_led[unit]) {
	 case FDDDevice::LED::GREEN:
		pal[FG] = UD_BLACK;
		pal[IN] = UD_YELLOW_GREEN;
		break;
	 case FDDDevice::LED::RED:
		pal[FG] = UD_BLACK;
		pal[IN] = UD_RED;
		break;
	 default:
		pal[IN] = pal[BG];
		break;
	}
	bitmap.DrawBitmap(dx, dy, *accmark, pal);
}

/*static*/ const uint8
WXStatusPanel::AccessMark12[] = {
	0,0,0,0,1,1,1,0,0,0,0,0,
	0,0,1,1,2,2,2,1,1,0,0,0,
	0,1,2,2,2,2,2,2,2,1,0,0,
	0,1,2,2,2,2,2,2,2,1,0,0,
	1,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2,2,1,0,0,
	0,1,2,2,2,2,2,2,2,1,0,0,
	0,0,1,1,2,2,2,1,1,0,0,0,
	0,0,0,0,1,1,1,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,
};
/*static*/ const uint8
WXStatusPanel::AccessMark16[] = {
	0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,
	0,0,0,1,1,2,2,2,2,2,1,1,0,0,0,0,
	0,0,1,2,2,2,2,2,2,2,2,2,1,0,0,0,
	0,1,2,2,2,2,2,2,2,2,2,2,2,1,0,0,
	0,1,2,2,2,2,2,2,2,2,2,2,2,1,0,0,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
	1,2,2,2,2,2,2,2,2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2,2,2,2,2,2,1,0,0,
	0,1,2,2,2,2,2,2,2,2,2,2,2,1,0,0,
	0,0,1,2,2,2,2,2,2,2,2,2,1,0,0,0,
	0,0,0,1,1,2,2,2,2,2,1,1,0,0,0,0,
	0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};
/*static*/ const uint8
WXStatusPanel::AccessMark24[] = {
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,1,1,1,1,1,1,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,1, 1,2,2,2,2,2,2,1, 1,0,0,0,0,0,0,0,
	0,0,0,0,0,1,1,2, 2,2,2,2,2,2,2,2, 2,1,1,0,0,0,0,0,
	0,0,0,0,1,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,1,0,0,0,0,
	0,0,0,1,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,1,0,0,0,
	0,0,0,1,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,1,0,0,0,
	0,0,1,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,1,0,0,
	0,0,1,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,1,0,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,1,2,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,1,0,
	0,0,1,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,1,0,0,
	0,0,1,2,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,2,1,0,0,
	0,0,0,1,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,1,0,0,0,
	0,0,0,1,2,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,2,1,0,0,0,
	0,0,0,0,1,2,2,2, 2,2,2,2,2,2,2,2, 2,2,2,1,0,0,0,0,
	0,0,0,0,0,1,1,2, 2,2,2,2,2,2,2,2, 2,1,1,0,0,0,0,0,
	0,0,0,0,0,0,0,1, 1,2,2,2,2,2,2,1, 1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,1,1,1,1,1,1,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
};

// NEWS LED1
void
WXStatusPanel::DrawNewsLED1(const Indicator *ind)
{
	Color bg;

	if ((newsled & NewsCtlrDevice::LED_ORANGE) != 0) {
		bg = UD_ORANGE;
	} else {
		bg = BGPANEL;
	}
	DrawTextLED(ind, UD_BLACK, bg);
}

// NEWS LED2
void
WXStatusPanel::DrawNewsLED2(const Indicator *ind)
{
	Color bg;

	if ((newsled & NewsCtlrDevice::LED_GREEN) != 0) {
		bg = UD_GREEN;
	} else {
		bg = BGPANEL;
	}
	DrawTextLED(ind, UD_BLACK, bg);
}

// 電源 LED
void
WXStatusPanel::DrawPower(const Indicator *ind)
{
	Color bg;

	if (ispower) {
		if (powerled) {
			bg = UD_YELLOW_GREEN;
		} else {
			bg = UD_GREY;
		}
	} else {
		// XXX 電源オフが出来たら X68030 なら赤
		bg = BGPANEL;
	}

	DrawTextLED(ind, UD_BLACK, bg);
}

// 共通の描画処理。
// ind->text を前景色 fg、背景色 bg で描画するだけの場合。
void
WXStatusPanel::DrawTextLED(const Indicator *ind, Color fg, Color bg)
{
	const Rect& rect = ind->rect;

	// 枠
	bitmap.FillRect(bg, rect);
	bitmap.DrawRect(UD_GREY, rect);

	SetTextColor(fg, bg);
	DrawStringSJIS(rect.x + 2, rect.y + 2, ind->text.c_str());
	ResetTextColor();
}

// 左ダブルクリックイベント
void
WXStatusPanel::OnLeftDClick(wxMouseEvent& event)
{
	const wxPoint& pt = event.GetPosition();

	for (auto ind : indicators) {
		// dclick を処理できる枠内なら
		if (ind->rect.Contains(pt.x, pt.y) && ind->dclick != NULL) {
			(this->*(ind->dclick))(ind);
			break;
		}
	}
}

// パフォーマンスカウンタ枠へのダブルクリック
void
WXStatusPanel::DClickPerf(const Indicator *ind)
{
	// モードを切り替える
	gSync->RequestFullSpeed(!gSync->GetFullSpeed());
}

// コンテキストメニューイベント
void
WXStatusPanel::OnContextMenu(wxContextMenuEvent& event)
{
	wxPoint pos = ScreenToClient(event.GetPosition());

	for (auto ind : indicators) {
		// コンテキストメニューを処理できる枠内なら
		if (ind->contextmenu && ind->rect.Contains(pos.x, pos.y)) {
			(this->*(ind->contextmenu))(ind);
			break;
		}
	}
}

// CD 枠でのコンテキストメニュー
void
WXStatusPanel::ContextMenuCD(const Indicator *ind)
{
	int scsiid = ind->GetSCSIID();
	assertmsg(scsiid >= 0, "id=%d", ind->id);

	auto menu = gMainFrame->CreateSCSIRemovableMenu(scsiid);
	if (menu) {
		PopupMenu(menu);
	}
}

// FD 枠でのコンテキストメニュー
void
WXStatusPanel::ContextMenuFD(const Indicator *ind)
{
	int unit = ind->GetFDUnit();
	assertmsg(unit >= 0, "id=%d", ind->id);

	auto menu = gMainFrame->CreateFDMenu(unit);
	if (menu) {
		PopupMenu(menu);
	}
}

// SCSI メディア変更イベント (UIMessage イベント)
void
WXStatusPanel::OnSCSIMediaChanged(wxCommandEvent& event)
{
	Indicator *ind = NULL;
	uint8 new_loaded;

	// 該当の Indicator を探す (見付かるはず)
	int scsiid = event.GetInt();
	int id = Indicator::SCSI0 + scsiid;
	for (auto i : indicators) {
		if (i->id == id) {
			ind = i;
			break;
		}
	}
	assert(ind);
	const auto disk = dynamic_cast<const SCSIDisk *>(ind->dev);

	// メディアの状態を一旦クリア
	new_loaded = scsi_loaded;
	new_loaded &= ~(1U << scsiid);

	// ツールチップを設定
	SCSI::DevType devtype = disk->GetDevType();
	std::string tip = string_format("%s%d: ",
		SCSI::GetDevTypeName(devtype), scsiid);
	const std::string& path = disk->GetPathName();
	// 同時にメディア状態もセット
	if (path.empty()) {
		tip += "-";
	} else {
		tip += path;
		new_loaded |= (1U << scsiid);
	}
	switch (disk->GetWriteMode()) {
	 case SCSIDisk::RW::ReadOnly:
		tip += "\n(Read Only)";
		break;
	 case SCSIDisk::RW::WriteProtect:
		tip += "\n(Write Protected)";
		break;
	 case SCSIDisk::RW::WriteIgnore:
		tip += "\n(Write Ignored)";
		break;
	 default:
		break;
	}
	ind->panel->SetToolTip(tip);

	// 書き込み無視だけ、これによって shutdown が必要かどうかという
	// オペレーションに違いが出るため表示する。
	// 無駄に幅を取らないためメディア挿入時のみ表示。
	// MO のライトプロテクト表示は必要になったら考える…。
	ind->text = string_format("%s%d", SCSI::GetDevTypeName(devtype), scsiid);
	if ((new_loaded & (1U << scsiid))) {
		if (disk->GetWriteMode() == SCSIDisk::RW::WriteIgnore) {
			ind->text += "\x02\x03";
		}
	}

	// メディアのロード状態が変われば (再配置して) 再描画。
	// 通常周期のタイマーによる更新チェックではメディア状態はチェックして
	// いないので、ここで自発的に行う。
	if (scsi_loaded != new_loaded) {
		scsi_loaded = new_loaded;

		LayoutIndicators();
		Refresh();
	}
}

// FDD メディア変更イベント (UIMessage イベント)
void
WXStatusPanel::OnFDDMediaChanged(wxCommandEvent& event)
{
	Indicator *ind = NULL;
	bool new_loaded;

	// 該当の Indicator を探す (見付かるはず)
	int unit = event.GetInt();
	int id = Indicator::FDD0 + unit;
	for (auto i : indicators) {
		if (i->id == id) {
			ind = i;
			break;
		}
	}
	assert(ind);
	const auto fdd = dynamic_cast<const FDDDevice *>(ind->dev);

	// メディアの状態を一旦クリア
	new_loaded = false;

	// ツールチップを設定
	std::string tip = string_format("FD%d: ", unit);
	const std::string& path = fdd->GetPathName();
	// 同時にメディア状態もセット
	if (path.empty()) {
		tip += "-";
	} else {
		tip += path;
		new_loaded = true;
	}
	switch (fdd->GetWriteMode()) {
	 case FDDDevice::RW::WriteProtect:
		tip += "\n(Write Protected)";
		break;
	 case FDDDevice::RW::WriteIgnore:
		tip += "\n(Write Ignored)";
		break;
	 default:
		break;
	}
	ind->panel->SetToolTip(tip);

	ind->text = string_format("FD%d", fdd->GetUnitNo());
	// 書き込み無視だけ表示する。無駄に幅を取らないためメディア挿入時のみ表示。
	// 4文字分のマークの後ろに "WI" を追加。
	if (new_loaded) {
		if (fdd->GetWriteMode() == FDDDevice::RW::WriteIgnore) {
			ind->text += "    \x02\x03";
		}
	}

	// メディアのロード状態が変われば (再配置して) 再描画。
	// 通常周期のタイマーによる更新チェックではメディア状態はチェックして
	// いないので、ここで自発的に行う。
	if (fd_loaded[unit] != new_loaded) {
		fd_loaded[unit] = new_loaded;

		LayoutIndicators();
		Refresh();
	}
}

// LED 状態変更イベント (UIMessage イベント)。
//
// 状態変更があったことをプッシュ通知する。主に、通常の更新頻度では
// 間に合わない X68030 の電源 LED の 16Hz 点滅用。
void
WXStatusPanel::OnLEDChanged(wxCommandEvent& event)
{
	if (UpdateStat()) {
		Refresh();
	}
}


//
// インジケータ情報
//

// SCSI なら SCSI ID を返す。そうでなければ -1 を返す。
int
Indicator::GetSCSIID() const
{
	int scsiid = id - SCSI0;
	if (0 <= scsiid && scsiid <= 7) {
		return scsiid;
	} else {
		return -1;
	}
}

// FDD ならユニット番号を返す。そうでなければ -1 を返す。
int
Indicator::GetFDUnit() const
{
	int unit = id - FDD0;
	if (0 <= unit && unit <= 3) {
		return unit;
	} else {
		return -1;
	}
}
