//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// ステータスパネル
//

#pragma once

#include "wxtextpanel.h"
#include "fdd.h"
#include "scsibus.h"
#include <array>

class Indicator;
class HostNetDevice;
class WXStatusPanel;

// インジケータ情報 (ほぼ構造体)
class Indicator
{
 public:
	// インジケータの識別子
	enum {
		NONE = 0,
		PERF,
		POWER,
		SCSI0,
		SCSI1,
		SCSI2,
		SCSI3,
		SCSI4,
		SCSI5,
		SCSI6,
		SCSI7,
		LAN,
		FDD0,
		FDD1,
		FDD2,
		FDD3,
		NEWSLED1,
		NEWSLED2,
	};
 protected:
	using dispfunc_t = void (WXStatusPanel::*)(const Indicator *);
	using eventfunc_t = void (WXStatusPanel::*)(const Indicator *);
 public:
	Indicator() { }
	Indicator(int id_, int minlen_, dispfunc_t draw_) {
		id = id_;
		minlen = minlen_;
		draw = draw_;
	}
	Indicator(int id_, int minlen_, dispfunc_t draw_, const std::string& text_)
		: Indicator(id_, minlen_, draw_)
	{
		text = text_;
	}
	virtual ~Indicator() = default;

	// インジケータ識別子
	int id {};

	// 枠内の最小文字列長。
	// 枠の大きさはこの文字列長と text.length() の長いほうで描画される。
	int minlen {};

	// 表示する文字列。
	std::string text {};

	// 描画関数 (必須)
	dispfunc_t draw {};

	// ダブルクリックの処理関数 (必要なら)
	eventfunc_t dclick {};

	// コンテキストメニューの処理関数 (必要なら)
	eventfunc_t contextmenu {};

	// 対応するデバイス (必要なら)
	Device *dev {};

	// ToolTip を表示するためのダミーパネル
	wxPanel *panel {};

	// 枠の位置と大きさは panel.GetRect() で取得できる (実行時に計算する)
	Rect rect {};

 public:
	// SCSI なら SCSI ID を返す。そうでなければ -1 を返す。
	int GetSCSIID() const;

	// FDD ならユニット番号を返す。そうでなければ -1 を返す。
	int GetFDUnit() const;
};

// ステータスパネル
class WXStatusPanel : public WXTextPanel
{
	using inherited = WXTextPanel;
 public:
	WXStatusPanel(wxWindow *parent);
	virtual ~WXStatusPanel() override;

	void FontChanged() override;

 private:
	void OnClose(wxCloseEvent&);
	void OnShow(wxShowEvent&);
	void OnSize(wxSizeEvent&);
	void OnTimer(wxTimerEvent&);
	void OnLeftDClick(wxMouseEvent&);
	void OnContextMenu(wxContextMenuEvent&);
	void OnSCSIMediaChanged(wxCommandEvent&);
	void OnFDDMediaChanged(wxCommandEvent&);
	void OnLEDChanged(wxCommandEvent&);

	bool UpdateStat();
	void Draw() override;
	void InitIndicators();
	void LayoutIndicators();
	void DrawPerf(const Indicator *);
	void DrawSCSI(const Indicator *);
	void DrawNet(const Indicator *);
	void DrawFD(const Indicator *);
	void DrawNewsLED1(const Indicator *);
	void DrawNewsLED2(const Indicator *);
	void DrawPower(const Indicator *);
	void DrawTextLED(const Indicator *, Color fg, Color bg);
	void DClickPerf(const Indicator *);
	void ContextMenuCD(const Indicator *);
	void ContextMenuFD(const Indicator *);

	// インジケータ情報
	std::vector<Indicator*> indicators {};

	// インジケータ表示用
	bool ispower {};
	bool powerled {};
	int perf_mode {};
	int perf_rate {};
	std::shared_ptr<SCSIBus> scsibus {};
	uint8 scsi_bsy {};
	bool scsi_out {};
	uint8 scsi_loaded {};
	enum {
		NOT_AVAILABLE = -1,	// 存在しない (Nereid なし X68k とか)
		DISABLE = 0,		// 存在してホスト側が無効 (driver = none)
		ENABLE = 1,			// 存在してホスト側が有効
	} net_ok {};
	HostNetDevice *hostnet {};
	uint64 net_tx_pkts {};
	uint64 net_rx_pkts {};
	bool net_tx_active {};
	bool net_rx_active {};
	std::array<bool, 4> fd_loaded {};
	std::array<FDDDevice::LED, 4> fd_access_led {};
	std::array<bool, 4> fd_eject_led {};
	int newsled {};

	// 丸いアクセスマーク (X68k FD 用)
	std::unique_ptr<BitmapI8> accmark /*{}*/;

	// 背景再描画
	bool refresh_background {};

	// タイマー
	wxTimer timer {};

	// イベントテーブル
	wxDECLARE_EVENT_TABLE();

	static const uint8 AccessMark12[];
	static const uint8 AccessMark16[];
	static const uint8 AccessMark24[];
};

extern WXStatusPanel *gStatusPanel;
